/*
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@
@@ PROGRAM:  integrate.h 
@@
@@ PURPOSE:  Shared header file for pixel integration calculations
@@
@@   Routines with "_2" in the name also process sigmas
@@   Routines with "_two_" in the name process two arrays for extra speed,
@@      as requested by David Makovoz
@@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

#ifndef _INTEGRATE_H_
#define _INTEGRATE_H_

#ifndef FALSE
#define FALSE	0
#define TRUE	!FALSE
#endif

/*============================== CODE VERSION ==============================*/
/*
**define INT_VERSION	"1.02"   Version for release 6.2 
**define INT_VERSION	"1.10"   Added three routines for D. Makovoz: 
**                                integrate_two_quad()         CR3030 
**                                distribute_two_quad()        CR3031 
**                                unweighted_distribute_quad() CR3032 
**define INT_VERSION	"1.11"   Changed for D. Makovoz:                  
**                                in unweighted_distribute_quad()         
**                                REPLACED image[j] += pseudopixel_value; 
**                                WITH     image[j]  = pseudopixel_value; 
**define INT_VERSION	"1.20"   Changed for D. Makovoz:  CR4270          
**define INT_VERSION	"1.21"   Added integrate_quad_3() for mask combining 
**define INT_VERSION	"1.22"   Updated integrate_quad_3()                  
*/
#define INT_VERSION	"1.30"  /* Added two new routines for David Makovoz:  */
                                /*   mask_distribute_quad_short()   CR5496    */
                                /*   mask_distribute_quad_int()               */
#define INT_NAME  	"integrate (library)"

/*========================== CONSTANT DEFINITIONS ==========================*/
#define MAX_POLY_LEN	100  /* Max number of pts in a clipped polygon */
#define MAX_WORKING_PTS 256  /* Max number of pts to store intermediate clipping results */
#define TOL	       1E-6  /* Set maximum error tolerance percent to 0.0001% */
#define N_QUAD		  5  /* Number of points in a closed quadrilateral */


/*============================ TYPE DEFINITIONS ============================*/

typedef unsigned char boolean;

/* A pixel consists of the INTEGER x and y coordinates of a vertex */
typedef struct {
  int x;
  int y;
} pixel_type;

/* A point consists of the REAL x and y coordinates of a vertex */
typedef struct {
  double x;
  double y;
} point_type;

/* A poly consists of up to MAX_POLY_LEN vertices.  Maintain as ALWAYS CLOSED 
** which means it will always have an additional copy of the first vertex added 
** to the end of its list, yielding up to MAX_POLY_LEN+1 points in the poly.
*/
typedef struct {	 
   int 		len;		 	/* How many vertices are stored  */
   					/* (including repeated vertex)   */
   point_type 	p[MAX_POLY_LEN+1];	/* The vertices, in CCW order, ending with */
 					/* a repetition of the first vertex        */
} poly_type;


/*=============================== PROTOTYPES ===============================*/

/************* FROM integrate_area.c *************/

void  bounding_rect( poly_type pp, int *ixmin, int *ixmax, 
                     int *iymin, int *iymax , int ixdim, int iydim );
                     
double area_poly(poly_type *poly);
double area_poly1(point_type *pts, int npts);

poly_type stuff_four( double rx1, double ry1,  double rx2, double ry2, 
                      double rx3, double ry3,  double rx4, double ry4  );
                       
void print_poly(poly_type *poly);

int rmax4ceil( double r1, double r2, double r3, double r4);
int rmin4floor(double r1, double r2, double r3, double r4);

/*===== PRIMARY EXPORTED ROUTINES =====*/
char* integrate_name( void );
char* integrate_version( void );

double integrate_quad(    double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double image[], int ixdim, int iydim, double *area );

double integrate_quad_2(  double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double image[], double sigmas[],
                          int    ixdim,   int iydim,
                          double *area,   double *area_sq, double *sigma_result );
 
double integrate_quad_3(  double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double image[],            double sigmas[],
                          short int    mask[],       short int mask_fatal,
                          int    ixdim,              int iydim,
                          double *area,              double *area_sq,
                          double *sigma_result,      short int *mask_result );
 
void integrate_two_quad(  double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double image1[],           double image2[], 
                          double *integrated_value1, double *integrated_value2,
                          int ixdim, int iydim,      
                          double *area1, double *area2 );

void distribute_quad(     double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double pseudopixel_value,  double image[],
                          int xdim, int ydim                                 );

void distribute_quad_2(   double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double pseudopixel_value,  double pseudopixel_sigma,
                          double image[],            double sigmas[],
                          int xdim, int ydim                                 );

void distribute_quad_n(   double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double *pseudopixel_vals,  double **main_image,
                          int    n_images,
                          int    xdim, int ydim                              );
                          

void distribute_quad_n_1D(double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double *pseudopixel_vals,  double *main_image,
                          int    n_images,
                          int    xdim, int ydim                              );
                          

void distribute_two_quad( double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double pseudopixel_value1, double image1[],
                          double pseudopixel_value2, double image2[],
                          int xdim, int ydim                                 );

void unweighted_distribute_quad(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double pseudopixel_value,  double image[],
                          int xdim, int ydim                                 );

void distribute_quad_BB(  double rx1, double ry1,   double rx2, double ry2,
                          double rx3, double ry3,   double rx4, double ry4,
                          double pseudopixel_value, double pseudopixel_sigma,
                          double image[], double sigmas[],
                          int xdim, int ydim,
                          double mask_image[], double order_image[],
                          int order_number                                   );

void mask_distribute_quad_short(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          short pseudopixel_value,   short image[],
                          int xdim, int ydim                                 );
                        
void mask_distribute_quad_int(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          int    pseudopixel_value,  int image[],
                          int    xdim, int ydim                              );
 
                       
/* DEBUGGING ROUTINES */                   
int verify_cyclical( poly_type poly );

                   
/************* FROM poly_clip.c *************/

void clip_polygon( pixel_type  wMin,    pixel_type  wMax, 
                   poly_type  *polyIn,  poly_type  *polyOut  );

            
#endif
