/* "C" function definitions for uncertainty propagation.
   Intended for both C and C++ programs; calling conventions are always C.
       (IMPORTANT linking note - requires the C++ libraries, so
        use CC or g++ to link; this means the SIRTF linker name is CCC

   Used with the C++ to C wrapper uprop.cpp. 
   
   This code calls (using a C calling interface) C++ routines associated
   with the global class instance defined in uprop.cpp. 
   
   Note that the C routines are all using a single class instance and 
   share a common definition for bit masks, NAN values and 
   Cauchy/Illdef limit thresholds.

   The following uncertainty propagation calculation functions are provided:
            uprop_add         
	    uprop_sub
	    uprop_mul
	    uprop_div
	    uprop_interp

   Typical call:
            uprop_mul( &result, &result_sigma, &result_mask, x, x_sigma, x_mask, y, y_sigma, y_mask);

   Algebraic operations:
            result = x * y;
            result_sigma =sqrt (x * x * sigma_y * sigma_y + y * y * sigma_x * sigma_x + 
                                sigma_x * sigma_x * sigma_y * sigma_y);
            result_mask = x_mask | y_ mask;

 --------------------------------------------------------------------------------------

   The following utility routines provided to return a pointer
   to a C++ object that holds the parameters for the object that
   owns the C methods doing the work:

	uprop_globptr

 --------------------------------------------------------------------------------------

   Utility Code added to manipulate parallel data, sigma and mask FITS files:
 
               For both float's and doubles:
                   given a data array, a mask array and a sigma array, handle fetching a value, sigma and mask 
               ffetch ( - )
               dfetch ( - )

   Rationale - its complicated because the initial implementation has the sigma's optional.



*/
/* $Id: uprop.h,v 1.12 2006/06/21 19:40:52 davidh Exp $ */
#ifndef uprop_hh
#define uprop_hh 

/* enum type, possible return status from uncertainty propagation */
typedef  enum {uprop_normal, uprop_missing, uprop_warning, uprop_cauchy} uprop_state;


#include <math.h>

/* Public data inside the upropclass is intended to be inside this data structure.
   This allows the data structure to be visible outside C++ should the need arise.
*/
typedef struct {
  int dmask_missing; /* bit to set when error occurs */
  int  dmask_illdef; /* masks, when set equiv to NaN */
  int  dmask_warning; /* warning bit - by default its 0  */
  int  dmask_cauchy;  /* warning bit - by default its same as warning */
  int  state_2_mask[3];

  float CauchyThreshold,ApproxThreshold;

  int div_exceptions;
} uprop_data;

#ifdef	__cplusplus
extern "C" {
#endif

  uprop_state uprop_add(float*u, float*usigma, int*umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask);
  uprop_state uprop_sub(float*u, float*usigma, int*umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask);
  uprop_state uprop_mul(float*u, float*usigma, int*umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask);
  uprop_state uprop_div(float*u, float*usigma, int*umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask);
  int uprop_interp(float*u, float*usigma, int*umask, 
                   float x, int order, int ny, float *xtab, float *ytab, float *ysigma);

  /* change the mask values from the default (for the global used by C routines only) */
  void uprop_mask_setup(int missing_mask_input, int illdef_mask_input);

  /* double precision version of the binary operation and interplation routines */

  uprop_state uprop_dadd(double*u, double*usigma, int*umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask);
  uprop_state uprop_dsub(double*u, double*usigma, int*umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask);
  uprop_state uprop_dmul(double*u, double*usigma, int*umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask);
  uprop_state uprop_ddiv(double*u, double*usigma, int*umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask);
  int uprop_dinterp(double*u, double*usigma, int*umask, 
		    double x, int order, int ny, double *xtab, double *ytab, double *ysigma);

  /* return a pointer to the data area C wrapper routines */
  uprop_data * uprop_globpubptr();

  /* given a data array, a mask array and a sigma array, handle fetching a value, sigma and mask */
  int dfetch(double *val, double *vsigma, unsigned short *vmask, int ndx, int missingmask,
	     short exist_val, double *array_val, 
	     short exist_vsigma, double *array_vsigma,
	     short exist_vmask,unsigned short* array_vmask)
       ;
       int ffetch(float *val, float *vsigma, unsigned short *vmask, int ndx, int missingmask,
		  short exist_val, float *array_val, 
		  short exist_vsigma, float *array_vsigma,
		  short exist_vmask,unsigned short* array_vmask)
       ;

#ifdef __cplusplus
       class upropclass;
       /* return a pointer to the C++ object used for the C wrapper routines */
       typedef upropclass* upropclass_ptr ;
  upropclass_ptr uprop_globptr();
}
#endif
#endif
