/* Purpose: class library for uncertainty propagation
              1. standardize interfaces
              2. centralize methods and techniques
              3. central part of a "C" wrapper for general usage
   Conventions:
              1. NaN's may appear in any calculation and must be propagated.
              2. Random variables are zero mean with a width parameter, sigma
              3. Positive sigmas indicate a standard deviation,
                 for those distributions that have a variance.
              4. Negative sigma's indicate the FWHM of a Cauchy or other distribution
                 without a variance.  (Result from a division operation ONLY)
   Implementation notes:
              1. This is a C++ class to allow various threshold parameters to be specified.
              2. I envision a "C" wrapper to have a fixed threshold for Cauchy and divide-by-zero ratio's.
              3. The divide routine generates a positive value for usigma;
                 Cauchy distributions return a sigma corresponding to a gaussian FWHM
              4. No execptions are thrown unless the member variable div_exceptions is true
	      5. creating NAN's is not very system independent, so I have defined a value that
                 will work for various machines.
	      6. A table lookup/interpolation routine is provided. A present the functionality
	         is linear interpolation only with a planned cubic spline to be added at a future date.
                 The parameter "order" is examined to specify linear interpolation vs cubic splines.
  Mathematical notes:
              1. Random variables are zero mean, and have a zero third moment
              2. All random variables with a positive width parameter map the width
                 parameter to a variance/standard deviation corresponding to a Gaussian.
              3. A graceful degradation of accuracy is allowed by introducing a Cauchy
                 distribution as (x+X)/(y+Y) approaches indeterminate. Modeled as u=0 and U=X/Y
		 The "width" parameter is standard deviation of a Gaussian having the
		 same FWHM.

*/
/* $Id: upropclass.h,v 1.14 2006/06/21 19:38:08 davidh Exp $ */
#ifndef upropclass_hh
#define upropclass_hh

#include "uprop.h"

#include <math.h>

#include "nanvalue.h"

class upropclass {

 public:
  uprop_data  pubdata;

  /*  Try this change - TYB 5/1/09 */
 
  bool NotNumber(float x, float xsigma, int xmask){return iznand((double) x) || iznand((double) xsigma) || ((xmask&pubdata.dmask_missing)!=0) ;}

  //   bool NotNumber(float x, float xsigma, int xmask){return iznanf(x) || iznanf(xsigma) || ((xmask&pubdata.dmask_missing)!=0) ;}
  bool NotNumber(double x, double xsigma, int xmask){return iznand(x) || iznand(xsigma) || ((xmask&pubdata.dmask_missing)!=0) ;}

  void mask_setup(int missing_mask_input, int illdef_mask_input)
    {
      pubdata.dmask_missing=missing_mask_input;
      pubdata.dmask_illdef=illdef_mask_input;
      pubdata.dmask_warning=0;
      pubdata.state_2_mask[uprop_normal]=0;
      pubdata.state_2_mask[uprop_missing]=missing_mask_input;
      pubdata.state_2_mask[uprop_warning]=pubdata.dmask_warning;
      pubdata.state_2_mask[uprop_cauchy]=pubdata.dmask_warning;
    }
  upropclass();
  void mask_setup(int missing_mask_input, int illdef_mask_input,int warning_mask_input)
    {
      pubdata.dmask_missing=missing_mask_input;
      pubdata.dmask_illdef=illdef_mask_input;
      pubdata.dmask_warning=warning_mask_input;
      pubdata.state_2_mask[uprop_normal]=0;
      pubdata.state_2_mask[uprop_missing]=missing_mask_input;
      pubdata.state_2_mask[uprop_cauchy]=pubdata.dmask_warning;
    }
  upropclass(float CauchyRatio, float ApproxRatio );
   
  uprop_state add(float & u, float & usigma, int & umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask)
    {float tsigmasquare;
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_float.f;
      uval=uprop_missing;
      u=x+y;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_float.f;
      uval=uprop_missing;
      u=x+y;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      tsigmasquare=xsigma*xsigma+ysigma*ysigma;
      u=x+y;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }     
   
   
  uprop_state sub(float & u, float & usigma, int & umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask)
    {float tsigmasquare; 
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_float.f;
      u=x-y;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_float.f;
      u=x-y;
      uval=uprop_missing;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      tsigmasquare=xsigma*xsigma+ysigma*ysigma;
      u=x-y;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }     

  uprop_state mul(float & u, float & usigma, int & umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask)
    {float tsigmasquare;
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_float.f;
      u=x*y;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_float.f;
      u=x*y;
      uval=uprop_missing;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      u=x*y;
      tsigmasquare=x*x*ysigma*ysigma + y*y*xsigma*xsigma + ysigma*ysigma*xsigma*xsigma;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }

  uprop_state div(float & u, float & usigma, int & umask, float x, float xsigma, int xmask, float y, float ysigma, int ymask)
    {float tsigmasquare,xsq,ysq,xsigmasq,ysigmasq;
     uprop_state xval,yval,uval;
     int throwit=0;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_float.f;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_float.f;
      uval=uprop_missing;
      }
    else if(fabs(ysigma)>fabs(pubdata.ApproxThreshold*y))
      {if( (fabs(xsigma)>fabs(pubdata.CauchyThreshold*x)) && (fabs(ysigma)>fabs(pubdata.CauchyThreshold*y)) )
	{/* Cauchy result */
	  usigma=EquivSigma(2.0*fabs(xsigma/ysigma));
	  uval=uprop_cauchy;
	}
      else
	{/* illdef - illdefined series approximation */
	  xsq=x*x;
	  ysq=y*y;
	  xsigmasq=xsigma*xsigma;
	  ysigmasq=ysigma*ysigma;
	  tsigmasquare=xsigmasq/ysq + (3.0*xsigmasq*ysigmasq + xsq*ysigmasq)/(ysq*ysq);
	  u=x*(1.0+ysigmasq/ysq)/y;
	  usigma=sqrt(tsigmasquare);
	  uval=uprop_warning;
	  throwit=pubdata.div_exceptions;
	}
      }
    else
      { uval=mask_2_state(xmask|ymask);
      xsq=x*x;
      ysq=y*y;
      xsigmasq=xsigma*xsigma;
      ysigmasq=ysigma*ysigma;
      tsigmasquare=xsigmasq/ysq + (3.0*xsigmasq*ysigmasq + xsq*ysigmasq)/(ysq*ysq);
      usigma=sqrt(tsigmasquare);
      }
    if(y!=0)
      u=x/y;
    else
      u=const_nan_float.f;
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    if(throwit)
	    throw ("Un-approximate result in uprop.div() - divide by zero??");
    return uval;
    }
  uprop_state add(double & u, double & usigma, int & umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask)
    {double tsigmasquare;
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_double.f;
      uval=uprop_missing;
      u=x+y;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_double.f;
      uval=uprop_missing;
      u=x+y;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      tsigmasquare=xsigma*xsigma+ysigma*ysigma;
      u=x+y;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }     
   
   
  uprop_state sub(double & u, double & usigma, int & umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask)
    {double tsigmasquare; 
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_double.f;
      u=x-y;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_double.f;
      u=x-y;
      uval=uprop_missing;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      tsigmasquare=xsigma*xsigma+ysigma*ysigma;
      u=x-y;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }     

  uprop_state mul(double & u, double & usigma, int & umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask)
    {double tsigmasquare;
    uprop_state xval,yval,uval;
    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_double.f;
      u=x*y;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_double.f;
      u=x*y;
      uval=uprop_missing;
      }
    else
      { uval=mask_2_state(xmask|ymask);
      u=x*y;
      tsigmasquare=x*x*ysigma*ysigma + y*y*xsigma*xsigma + ysigma*ysigma*xsigma*xsigma;
      usigma=sqrt(tsigmasquare);
      }
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    return uval;
    }

  uprop_state div(double & u, double & usigma, int & umask, double x, double xsigma, int xmask, double y, double ysigma, int ymask)
    {double tsigmasquare,xsq,ysq,xsigmasq,ysigmasq;
    uprop_state xval,yval,uval;
    int throwit=0;

    if(NotNumber(x,xsigma,xmask) )
      { usigma=const_nan_double.f;
      uval=uprop_missing;
      }
    else if( NotNumber(y,ysigma,ymask) )
      { usigma=const_nan_double.f;
      uval=uprop_missing;
      }
    else if(fabs(ysigma)>fabs(pubdata.ApproxThreshold*y))
      {if( (fabs(xsigma)>fabs(pubdata.CauchyThreshold*x)) && (fabs(ysigma)>fabs(pubdata.CauchyThreshold*y)) )
	{/* Cauchy result */
	  usigma=EquivSigma(2.0*fabs(xsigma/ysigma));
	  uval=uprop_cauchy;
	}
      else
	{/* illdef - illdefined series approximation */
	  xsq=x*x;
	  ysq=y*y;
	  xsigmasq=xsigma*xsigma;
	  ysigmasq=ysigma*ysigma;
	  tsigmasquare=xsigmasq/ysq + (3.0*xsigmasq*ysigmasq + xsq*ysigmasq)/(ysq*ysq);
	  usigma=sqrt(tsigmasquare);
	  uval=uprop_warning;
	  throwit=(pubdata.div_exceptions);
	}
      }
    else
      { uval=mask_2_state(xmask|ymask);
      xsq=x*x;
      ysq=y*y;
      xsigmasq=xsigma*xsigma;
      ysigmasq=ysigma*ysigma;
      tsigmasquare=xsigmasq/ysq + (3.0*xsigmasq*ysigmasq + xsq*ysigmasq)/(ysq*ysq);
      usigma=sqrt(tsigmasquare);
      }
    if(y!=0)
      u=x/y;
    else
      u=const_nan_double.f;
    umask = (umask & ~(pubdata.dmask_missing|pubdata.dmask_illdef)) | pubdata.state_2_mask[uval];
    if(throwit)
	    throw ("Un-approximate result in uprop.div() - divide by zero??");
    return uval;
    }

  uprop_state mask_2_state(int m)
    {if (m& pubdata.dmask_missing)
      return uprop_missing;
    else if(m& pubdata.dmask_illdef )
      return uprop_missing;
    else
      return uprop_normal;
    }

  int interp(float & u, float & usigma, int & umask, float x, int order, int ny, float *xtab, float *ytab, float *ysigma);
  int interp(double & u,double & usigma, int & umask, double x, int order, int ny,double *xtab,double *ytab,double *ysigma);

 private:

  double EquivSigma(double Cauchy_fwhm);
  int linear_interp(float & u, float & usigma, int & umask, float x, int ny, float *xtab,float *ytab, float *ysigma);
  int spline_interp(float & u, float & usigma, int & umask, float x, int ny, float *xtab,float *ytab, float *ysigma);
  int linear_interp(double & u, double & usigma, int & umask, double x, int ny, double *xtab,double *ytab, double *ysigma);
  int spline_interp(double & u, double & usigma, int & umask, double x, int ny, double *xtab,double *ytab, double *ysigma);
   
};

#endif
