/* unit tests for uncertainty propagation 
#
# Makefile for unit tests
#
# 1. run make
#      compiles & links all .o files and unit test: upropmain
#      runs latex to create theory .dvi file
#      create postscript file of theory
#      creates a .tar file with all the source needed to remake on a generic platform

# 2. Run the unit tests
#     ./upropmain
#
all: upropmain upropmaind ProbModel.ps uprop.tar uprop.a
upropmain : upropmain.o uprop.a
	CC -o upropmain upropmain.o uprop.a
upropmaind : upropmaind.o uprop.a
	CC -o upropmaind upropmaind.o uprop.a
uprop.a:  uprop.o upropclass.o tridiagonal.o
	ar -r uprop.a uprop.o upropclass.o  tridiagonal.o
upropmain.o : upropmain.cpp upropclass.h uprop.h
	CC -o upropmain.o -c upropmain.cpp
upropmaind.o : upropmaind.cpp upropclass.h uprop.h
	CC -o upropmaind.o -c upropmaind.cpp
uprop.o : uprop.cpp uprop.h upropclass.h
	CC -o uprop.o -c uprop.cpp
upropclass.o : upropclass.cpp upropclass.h 
	CC -o upropclass.o -c upropclass.cpp 
ProbModel.ps: ProbModel.dvi
	dvips ProbModel.dvi -o ProbModel.ps
ProbModel.dvi: ProbModel.tex
	latex ProbModel.tex
uprop.tar: uprop.cpp uprop.h upropclass.h upropclass.cpp Makefile upropmain.cpp upropmaind.cpp ProbModel.tex tridiagonal.c tridiagonal.h 
	tar -c -f uprop.tar uprop.cpp uprop.h upropclass.h upropclass.cpp Makefile upropmain.cpp upropmaind.cpp ProbModel.tex tridiagonal.c tridiagonal.h

 */

/* $Id: upropmain.cpp,v 1.10 2005/03/10 19:49:50 davidh Exp $ */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "uprop.h"

#include "upropclass.h"

#include "nanvalue.h"

upropclass_ptr  uc_ptr =uprop_globptr();

float rtest,rtestsigma;
int rmask;
/* compare the relative error between x and y using
   (for a scale factor) the maximum of scalex and scaley
*/
float relerr(float x,float y, float scalex, float scaley) {
  float absdif,maxval;

  absdif=fabs(x-y);
  maxval=fabs(scalex);
  if (maxval<fabs(scaley))
    maxval=fabs(scaley);
  if(maxval==0)
    return 0;
  else
    return absdif/maxval;
}

/* evaluate the actual vs expected values of calculate and its sigma */
void eval(const char* fname,float act, float actsigma, int actmask, float x, float xsigma, int xmask, float y, float ysigma, int ymask,float u, float usigma,int umask)
{static float eps=1e-4;
 float udif,usigmadif;
 int umaskdif;
      
 udif=fabs(u-act);
 usigmadif=fabs(fabs(usigma)-actsigma);
 
 if(umask & uc_ptr->pubdata.dmask_missing)
   usigmadif=0;
    
 umaskdif=(actmask-umask);
     
 if((udif<eps) && (usigmadif<eps) && (umaskdif==0))
   printf("passed: (%f,%f,%4x) %s (%f,%f,%4x))\n",x,xsigma,xmask,fname,y,ysigma,ymask);
 else
   {printf("failed: (%f,%f,%4x) %s (%f,%f,%4x))",x,xsigma,xmask,fname,y,ysigma,ymask);
   printf(" u expected =%f, uobserved=%f, usigma expected=%f, usigma observed=%f, mask expected %4x, mask observed=%4x\n",
	  u,act,usigma,actsigma,umask,actmask);
   }
}
extern "C" {
  /* perform a binary operation and check the results */
  void binoptest(const char *func_name, uprop_state (*func)(float * fu, float * fusigma, int * fumask, float fx, float fxsigma, int fxmask, float fy, float fysigma, int fymask),
		 float x,float xsigma,int xmask,float y,float ysigma,int ymask, float u,float usigma,int umask) 
  {    static float eps=1e-6;
  float udif,usigmadif;
     
  (*func)(&rtest,&rtestsigma,&rmask,(float)x,(float)xsigma,xmask,(float)y,(float)ysigma,ymask);
     
  eval(func_name,rtest,rtestsigma,rmask,(float)x,(float)xsigma,xmask,(float)y,(float)ysigma,ymask,(float)u,(float)usigma,umask);}
}

void binoptest();

void interptest();

void nantest();


static float nan_value=NANVALUE;

int main(char** argv, int argc)
{
  nantest();
  binoptest();
  interptest();
}
int diagnose(float res, int fi, int si, char* fname,int dzero);
int diagnose(float res, int fi, int si, char* fname,int dzero){
  int result=0;
  if((fi==0) || (si==0)) {

    /*  Try these changes - TYB 5/1/09 */

    if(!iznand((double) res)) {
    //    if(!isnan(res)) {
      fprintf(stdout,"nantest for %s failed nan expected, number=%f , fi=%d, si=%d\n",fname,res,fi,si);
      result++;
    }
  }
  else {
    if(iznand((double) res) && !dzero) {
      //    if(isnan(res) && !dzero) {
      fprintf(stdout,"nantest for %s failed, nan calculated, number expected,  fi=%d, si=%d\n",fname,res,fi,si);
      result++;
    }
  }
  return (result);
}
void nantest() {
  float first_arg[4]= {0, 0, 0.1, 100};
  float second_arg[4]={0, 0, 0.2, 200};
  int n=sizeof(first_arg)/sizeof(first_arg[0]);
  float res, res_unc;
  int mask,dz;
  float z=0;
  int iz=0,errcount=0;

  first_arg[0]=nan_value;
  second_arg[0]=nan_value;

  for(int fi=0;fi<n;fi++) {
    for(int si=0;si<n;si++) {
      uprop_add(&res,&res_unc,&mask, first_arg[fi],z,iz,second_arg[fi],z,iz);
      errcount+=diagnose(res,fi,si,"uprop_add",0);
      uprop_sub(&res,&res_unc,&mask, first_arg[fi],z,iz,second_arg[fi],z,iz);
      errcount+=diagnose(res,fi,si,"uprop_sub",0);
      uprop_mul(&res,&res_unc,&mask, first_arg[fi],z,iz,second_arg[fi],z,iz);
      errcount+=diagnose(res,fi,si,"uprop_mul",0);
      uprop_div(&res,&res_unc,&mask, first_arg[fi],z,iz,second_arg[fi],z,iz);
      dz=0;
      if((si==1) && (fi!=0))
        dz=1;
      errcount+=diagnose(res,fi,si,"uprop_div",dz);
    }
  }
  fprintf(stdout,"error count from nan propagate test=%d\n",errcount);
}

void binoptest()
{
  float x,y,u,xsigma,ysigma,usigma;
  int xmask_ndx,ymask_ndx,umask_ndx,binopmask[4][4];
   
   
  /* deliberate failure */
  x=1;
  y=2;
  u=0;
  xsigma=3.0;
  ysigma=4.0;
  usigma=0;
  xmask_ndx=ymask_ndx=0;
  umask_ndx=uprop_cauchy;
  binoptest("uprop_add deliberate test failure",&uprop_add,x,xsigma,0,y,ysigma,0,u,usigma,uc_ptr->pubdata.dmask_missing);
    
   
  /* all forms of add */
   
  x=1;
  y=2;
  u=3;
  xsigma=3.0;
  ysigma=4.0;
  usigma=5.0;
   
  for(xmask_ndx=0;xmask_ndx<=2;xmask_ndx++)
    for(ymask_ndx=0;ymask_ndx<=2;ymask_ndx++)
      {umask_ndx=xmask_ndx;
      if(umask_ndx<ymask_ndx) 
	umask_ndx=ymask_ndx;
      binoptest("uprop_add synthetic",&uprop_add,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
      }
   
  /* all forms of sub */
  x=1;
  y=2;
  u=-1;
  xsigma=13.0;
  ysigma=84.0;
  usigma=85.0;
   
  for(xmask_ndx=0;xmask_ndx<=2;xmask_ndx++)
    for(ymask_ndx=0;ymask_ndx<=2;ymask_ndx++)
      {umask_ndx=xmask_ndx;
      if(umask_ndx<ymask_ndx) 
	umask_ndx=ymask_ndx;
      binoptest("uprop_sub synthetic",&uprop_sub,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
      }
   
  /* all forms of mul */
  x=2.5;
  y=3.5;
  u=x*y;
  xsigma=3.0;
  ysigma=4.0;
  usigma=sqrt(x*x*ysigma*ysigma + y*y*xsigma*xsigma + xsigma*xsigma * ysigma*ysigma);
   
  for(xmask_ndx=0;xmask_ndx<=2;xmask_ndx++)
    for(ymask_ndx=0;ymask_ndx<=2;ymask_ndx++)
      {umask_ndx=xmask_ndx;
      if(umask_ndx<ymask_ndx) 
	umask_ndx=ymask_ndx;
      binoptest("uprop_mul synthetic",&uprop_mul,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
      }
   
  /* forms of div that simply propagate uncertainty types, small value approximation holds */
  x=100000;
  y=1000;
  u=100;
  xsigma=0;
  ysigma=0;
  usigma=0;
   
  for(xmask_ndx=0;xmask_ndx<=2;xmask_ndx++)
    for(ymask_ndx=0;ymask_ndx<=2;ymask_ndx++)
      {umask_ndx=xmask_ndx;
      if(umask_ndx<ymask_ndx) 
	umask_ndx=ymask_ndx;
      binoptest("uprop_div synthetic",&uprop_div,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
      }
   
  /* deliberate generation of a cauchy from two normals */
  x=0;
  y=0;
  u=0;
  xsigma=3.0;
  ysigma=4.0;
  usigma=1.5;
   
  xmask_ndx=ymask_ndx=0;
  umask_ndx=uprop_cauchy;
  binoptest("uprop_div deliberate cauchy",&uprop_div,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
   
  /* deliberate generation of a illdef from two normals */
  x=10;
  y=2;
  xsigma=3.0;
  ysigma=1.0;
  u=x*(1.0+ysigma*ysigma/(y*y))/y;
  usigma=sqrt(xsigma*xsigma/(y*y) + x*x*(ysigma*ysigma)/(y*y*y*y) + 3.0*xsigma*xsigma*ysigma*ysigma);
   
  xmask_ndx=ymask_ndx=0;
  umask_ndx=uprop_globpubptr()->dmask_illdef;
  binoptest("uprop_div deliberate illdef",&uprop_div,x,xsigma,uc_ptr->pubdata.state_2_mask[xmask_ndx],y,ysigma,uc_ptr->pubdata.state_2_mask[ymask_ndx],u,usigma,uc_ptr->pubdata.state_2_mask[umask_ndx]);
}

#define maxtab 100

/* unit test for cubic splines with uncertainty estimates */
float fcubic(float x) {return (x-maxtab)*x*(x+maxtab);}
float fquad(float x)  {return (x-maxtab)*x*x*(x+maxtab);}

void interptest()
{float xtab[maxtab],ytab[maxtab],ysigma[maxtab];

 int n=maxtab,yhatmask,errcnt,stat,i;
 float xlo, xhi, dx, yhat, yhatsigma, yexpect, yexpectsigma, x, interr;

 errcnt=0;

 xlo=-maxtab-1;
 xhi=maxtab+1;
 dx=(xhi-xlo)/(n-1);

 for(i=0;i<n;i++)
   {xtab[i]= xlo + i*(dx);
   ytab[i]= fcubic(xtab[i]);
   ysigma[i]=1e-4*(i+1);
   }

 /* test at interpolation points for piecewise linear exact match */
 printf("beginning linear interpolation test exactly on data points, n=%i\n",n);
 for(i=0;i<n;i++)
   {x = xlo + i*(dx);
   yexpect = fcubic(xtab[i]);
   yexpectsigma=ysigma[i];
   yhatmask=0;
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 2, n, xtab, ytab, ysigma);
   if(stat!=0)
     {errcnt++;
     printf("linear interp failed:  unexpected nonzero status returned from data point during piecewise linear interpolation, stat=%d\n",stat);
     }
   interr=relerr(yhat,yexpect,yhat,yexpect);
   if(interr>1E-2)
     {errcnt++;
     printf("linear interp failed:  unexpected large error returned from data point during piecewise linear interpolation, interp err=%f\n",interr);
     }
   interr=relerr(yhatsigma,yexpectsigma,yhat,yexpect);
   if(interr>1E-5)
     {errcnt++;
     printf("linear interp failed:  unexpected large error returned from variance of data point during piecewise linear interpolation, interp err=%f\n",interr);
     }
   }

 /* test at 3/4 between interpolation points for piecewise linear exact match */
 printf("beginning linear interpolation test between data points, n=%i\n",n);

 for(i=1;i<n;i++)
   {x = (3.0*xtab[i-1]+xtab[i])*0.25;
   yexpect = (3.0*ytab[i-1]+ytab[i])*0.25;
   yexpectsigma = sqrt((9.0*ysigma[i-1]*ysigma[i-1]+ysigma[i]*ysigma[i])/16.0);
   yhatmask=0;
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 2, n, xtab, ytab, ysigma);
   if(stat!=0)
     {errcnt++;
     printf("linear interp failed:  unexpected nonzero status returned from between points during piecewise linear interpolation, stat=%d\n",stat);
     }
   interr=relerr(yhat,yexpect,yhat,yexpect);
   if(interr>1E-5)
     {errcnt++;
     printf("linear interp failed:  unexpected large error returned from data point during piecewise linear interpolation, interp err=%f\n",interr);
     }
   interr=relerr(yhatsigma,yexpectsigma,yhat,yexpect);
   if(interr>1E-5)
     {errcnt++;
     printf("linear interp failed:  unexpected large error returned from variance of data point during piecewise linear interpolation, interp err=%f\n",interr);
     }
   }
   
 /* test at interpolation points for spline fit exact match */
 printf("beginning spline interpolation test at data points, n=%i\n",n);

 for(i=0;i<n;i++)
   {x = xlo + i*(dx);
   yexpect = fcubic(xtab[i]);
   yexpectsigma=ysigma[i];
   yhatmask=0;
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 3, n, xtab, ytab, ysigma);
   if(stat!=0)
     {errcnt++;
     printf("spline interp failed:  unexpected nonzero status returned from exact cubic interpolation, stat=%d\n",stat);
     }
   interr=relerr(yhat,yexpect,yhat,yexpect);
   if(interr>1E-5)
     {errcnt++;
     printf("spline interp failed:  unexpected large error returned from exact cubic interpolation, interp err=%f\n",interr);
     }
   interr=relerr(yhatsigma,yexpectsigma,yhat,yexpect);
   if(interr>1E-5)
     {errcnt++;
     printf("spline interp failed:  unexpected large error returned from exact cubic interpolation, interp err=%f\n",interr);
     }
   }

 /* test at 3/4 between interpolation points for spline exact match */
 printf("beginning spline interpolation test between data points, n=%i\n",n);

 for(i=1;i<n;i++)
   {x = (3.0*xtab[i-1]+xtab[i])*0.25;
   yexpect =fcubic(x);
   yexpectsigma = sqrt((9.0*ysigma[i-1]*ysigma[i-1]+ysigma[i]*ysigma[i])/16.0);
   yhatmask=0;
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 3, n, xtab, ytab, ysigma);
   if(stat!=0)
     {errcnt++;
     printf("spline interp failed:  unexpected nonzero status returned from between points during spline interpolation, stat=%d\n",stat);
     }
   interr=relerr(yhat,yexpect,yhat,yexpect);
   if(interr>1E-3)
     {errcnt++;
     printf("spline interp failed:  unexpected large error returned from data point during spline interpolation, interp err=%f\n",interr);
     }
   interr=relerr(yhatsigma,yexpectsigma,yhat,yexpect);
   if(interr>1E-3)
     {errcnt++;
     printf("spline interp failed:  unexpected large error returned from variance of data point during spline interpolation, interp err=%f\n",interr);
     }
   }
 if(errcnt==0)
   printf("passed: piecewise linear interpolation and basic spline interpolation errcnt=0\n");
 else
   printf("failed: nonzero error count in interpolation routine unit tests, errcnt= %d\n",errcnt);
    
 /* timing test of 5000 interpolations, linear and cubic */
 time_t  tcpux, tcpuy, delta_linear, delta_spline;
 int iterations=5000;
   
 x=(xtab[0]+xtab[1])*0.5;
   
 time(&tcpux);
   
 for(i=1;i<iterations;i++)
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 2, n, xtab, ytab, ysigma);
 time(&tcpuy);
 delta_linear=tcpuy-tcpux;
      
 for(i=1;i<iterations;i++)
   stat = uprop_interp(&yhat , &yhatsigma, &yhatmask, 
		       x , 3, n, xtab, ytab, ysigma);
 time(&tcpux);
 delta_spline=tcpux-tcpuy;
   
 printf("for %i iterations, table size = %d, wall clock delta for linear=%i, wall clock delta for cubic=%i\n",iterations, n, delta_linear, delta_spline);
}
