/* $Id: cropfits_compute.c,v 1.3 2007/11/12 18:12:16 laher Exp $ */


/*********************************************************
 * cropfits_compute.c
 *********************************************************
 Purpose
 Main computational loop of program

 Overview
 0. Allocate heap memory
 1. Loop over input image
    if CurrentPixel is cropped out
       record it in DumpStruct
    else
       copy CurrentPixel to cropped image
 2. Create Image of original pixel mapping

 Definition of Variables
 External:
 See cropfits_defs.h and cropfits_exec.c

 Internal:
 index - loop increment variable
 status - checks return value of major function calls
 ImagePixels - number of pix in image
 GridPixels - num of grid squares
 DoModel - boolean to indicate if model calculation selected
 ********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "cropfits.h"
#include "cropfits_defs.h"
#include "cropfits_errcodes.h"

void cropfits_compute( CPF_Filenames     *CPFP_Fnames,
		       CPF_Computation   *CPFP_Comp,
		       CPF_ImageInfo     *CPFP_Image,
		       CPF_State         *CPFP_Stat)
{
  int i, j, k, index, startz, stopz;
  int cropdex = 0;
  int iCropStart, jCropStart, iCropStop, jCropStop;
  int LengthX, LengthY, LengthZ;
  int CropLengthX, CropLengthY;
  int PlanePixels, UncropPix, UncropPixPlane;

  /* boolean state indicators */
  int Cropped;
  int DoTable = FALSE;
  int DoMap = FALSE;
  
  time_t now;
  FILE *tablefp;

  double *  CroppedImage=NULL;
  int *     Map = NULL;

  if(CPFP_Stat->Status) return;

  /***************************************************************
   * set the edge conditions & copy information from 
   * structs
   */
  if(CPFP_Stat->Verbose)
    printf("cropfits_compute: Initializing parameters for compute routine\n");

  if(strcmp(CPFP_Fnames->Filename_Table,""))
    DoTable = TRUE;
  if(strcmp(CPFP_Fnames->Filename_Map,""))
    DoMap = TRUE;

  LengthX = CPFP_Image->LengthX;
  LengthY = CPFP_Image->LengthY;
  LengthZ = CPFP_Image->LengthZ;
  iCropStart = CPFP_Comp->istart-1;
  iCropStop = iCropStart + CPFP_Comp->width;
  jCropStart = CPFP_Comp->jstart-1;
  jCropStop = jCropStart + CPFP_Comp->height;
  PlanePixels = LengthX * LengthY;

  /* take care of possibility of going beyond edge */
  if(jCropStop>=LengthY) {
    jCropStop = LengthY;
    CPFP_Comp->height = LengthY - jCropStart;
  }
  if(iCropStop>=LengthX) {
    iCropStop = LengthX;
    CPFP_Comp->width = LengthX - iCropStart;
  }

  UncropPix = CPFP_Comp->height * CPFP_Comp->width;
  UncropPixPlane = UncropPix;
 
  /* set plane to process */
  if(CPFP_Image->Plane==2) {
    startz = 0;
    stopz = 1;
  } else if(CPFP_Image->Plane==3) {
    startz = LengthZ-1;
    stopz = LengthZ;
  } else {
    startz = 0;
    stopz = LengthZ;
    UncropPix *= LengthZ;
  }
    

  /*******************************************************************
   * heap memory gets allocated for:
   *     CroppedImage (local & external)
   *     Map
   */
  if(CPFP_Stat->Verbose)
    printf("cropfits_compute: Allocating heap memory\n");

  CroppedImage = (void *) calloc(UncropPix, sizeof(double));
  if(CroppedImage==NULL) {
    CPFP_Stat->Status = COMPUTE | MALLOC_FAILED;
    return;
  }

  CPFP_Image->CroppedImage = (double *) calloc(UncropPix, sizeof(double));
  if(CPFP_Image->CroppedImage==NULL)  {
    CPFP_Stat->Status = COMPUTE | MALLOC_FAILED;
    free(CroppedImage); 
    return;
  }

  if(CPFP_Stat->Debug)
    printf("debug - allocated image memory, moving on to map memory\n");

  if(DoMap) {
    Map = (int *) calloc(2*UncropPixPlane, sizeof(int));
    if(Map==NULL) {
      CPFP_Stat->Status = COMPUTE | MALLOC_FAILED;
      free(CroppedImage);
      return;
    }
    CPFP_Image->Map = (int *) calloc(2*UncropPixPlane, sizeof(int));
    if(CPFP_Image->Map == NULL) {
      CPFP_Stat->Status = COMPUTE | MALLOC_FAILED;
      free(CroppedImage);
      free(Map);
    }
  }

  /*******************************************************************
   * breaking the abstraction....i still haven't devised
   * a good, low-overhead way to do this....so output
   * the table within this compute routine. don't use a
   * secondary data structure at all.
   *
   * wishes & dreams - would like to have a separate program
   * which listens for input of this type. gathers it all
   * together through separate feedings, and then writes
   * to a file when it receives a terminate signal.
   */
  if(DoTable) {
    /* open the output table file */
    if(CPFP_Stat->Verbose)
      printf("cropfits_compute: Writing dump table header \n");

    tablefp = fopen(&(CPFP_Fnames->Filename_Table[1]), "w");
    if(tablefp==NULL) {
      CPFP_Stat->Status = COMPUTE | FOPEN_FAILED;
      return;
    }

    /* prime with comments & headers */
    fprintf(tablefp, "\\comment = CROPFITS pixel dump file\n");
    now = time(NULL);
    fprintf(tablefp, "\\date = %s", ctime(&now));

    fprintf(tablefp, "\\InpImage = %s\n", CPFP_Fnames->Filename_Input);
    fprintf(tablefp, "\\ImageWidth = %d\n", LengthX);
    fprintf(tablefp, "\\ImageHeight = %d\n", LengthY);
    fprintf(tablefp, "\\CroppedWidth = %d\n", CPFP_Comp->width);
    fprintf(tablefp, "\\CroppedHeight = %d\n", CPFP_Comp->height);
    fprintf(tablefp, "\\CropIStart = %d\n", CPFP_Comp->istart);
    fprintf(tablefp, "\\CropJStart = %d\n", CPFP_Comp->jstart);
    fprintf(tablefp, "\\TotalCroppedPix = %d\n", (PlanePixels-UncropPixPlane)*LengthZ);
    fprintf(tablefp, "\\rows = %d\n", (PlanePixels-UncropPixPlane));
    fprintf(tablefp, "\\cols = %d\n", LengthZ+2);

    fprintf(tablefp, "|i    |j    ");
    for(index=startz; index < stopz; index++) {
      if(((index+1)/10) < 1)
	fprintf(tablefp, "|P%d         ", (index+1));
      else if(((index+1)/10) < 10)
	fprintf(tablefp, "|P%d        ", (index+1));
      else
	fprintf(tablefp, "|P%d       ", (index+1));
    }
    fprintf(tablefp, "|\n|i    |i    ");
    for(index=startz; index < stopz; index++)
      fprintf(tablefp, "|r          ");
    fprintf(tablefp, "|\n");
  }
      
  /*****************************************************************
   * loop over input image - turn the loop inside out
   * from "regular" orientation...put the z leg innermost
   * can decide outside the z leg if a pixel is cropped
   * out or no.
   */
  if(CPFP_Stat->Verbose) 
    printf("cropfits_compute: Cropping image\n");

  for(i=0; i < LengthX; i++)
    for(j=0; j < LengthY; j++) {
      if( ((i>=iCropStart)&&(i<iCropStop)) &&
	  ((j>=jCropStart)&&(j<jCropStop)) ) 
	Cropped = FALSE;
      else {
	Cropped = TRUE;
	if(DoTable) {
	  fprintf(tablefp, " %4d  %4d ", (i+1), (j+1));
	}
      }

      for(k=startz; k < stopz; k++) {
	if(!Cropped) {
	  CroppedImage[cropdex+(k-startz)*UncropPixPlane] =
	    CPFP_Image->Image[(k*PlanePixels)+(i*LengthY)+j];
	  if(CPFP_Stat->Debug)
	    printf("debug - cropdex=%d croppedimage=%f\n", 
		   cropdex, CroppedImage[cropdex]); 
	  if(DoMap&&(k==startz)&&(cropdex<UncropPixPlane)) {
	    Map[cropdex] = j+1;  
	    Map[cropdex+UncropPixPlane] = i+1; 
	  }
/* 	  cropdex++; */
	} else 
	  if((DoTable) && (Cropped)) {
	    fprintf(tablefp, " %10.2f ", CPFP_Image->Image[(k*PlanePixels)+(i*LengthY)+j]);
	  }
      } /* end of the single pixel z run */
      
      if(!Cropped)
	cropdex++;

      if(DoTable&&Cropped)
	fprintf(tablefp, "\n");
    }

  
  /*****************************************************************
   * copy CroppedImage back to global struct
   */
   for(index=0; index < UncropPix; index++) {
    CPFP_Image->CroppedImage[index] = CroppedImage[index];
  }
    
  if(DoMap)
    for(index=0; index < UncropPixPlane; index++) {
      CPFP_Image->Map[index] = Map[index]; 
      CPFP_Image->Map[index+UncropPixPlane] = Map[index+UncropPixPlane]; 
    }
  
  free(CroppedImage);
  free(Map);
  if(DoTable)
    fclose(tablefp);

  if(CPFP_Stat->Verbose) {
    if(DoTable)
      printf("cropfits_compute: Table dump finished & file closed\n");
    printf("cropfits_compute: Successful, exiting compute\n");
  }

  return;
}


