/* $Id: cropfits_log.c,v 1.4 2007/11/12 18:12:17 laher Exp $ */


/************************************************************ 
 *  cropfits_log.c
 ************************************************************

 Purpose:
 Decode error messages and report.
 
 Overview:
 Append Error Messages to log file.
 
 Definitions of Variables

 External:
 See cropfits_defs.h and cropfits_exec.c
 
 Internal:
 Time = Running time of program.
 cropfitsMessage = Code of error that occurred.
 cropfitsRoot = Code of function in which error occurred.
 
 CP_Message = String interpretation of error code.
 CP_Function = String interpretation of function of error.
 
 fp_Logfile = pointer into logfile.
 
 ***********************************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include "cropfits_errcodes.h"
#include "cropfits.h"
#include "cropfits_defs.h"
#include "fitsio.h"

void cropfits_log( CPF_Filenames   *CPFP_Fnames,
	           CPF_Computation *CPFP_Comp,
	           CPF_ImageInfo   *CPFP_Image,
	           CPF_Pathnames   *CPFP_Paths,
	           CPF_State	   *CPFP_Stat)
{

  int	Time;
  int	cropfitsMessage;
  int	cropfitsRoot;
 
  char	CP_Message[MESSAGE_LENGTH];
  char	CP_Function[FCNNAME_LENGTH];

  FILE	*fp_Logfile;
  
  time_t  currenttime;
  char    *datetime;

  /* open the designated log file */
  if(strcmp(CPFP_Fnames->Filename_Log, "stdout") == 0) {
    fp_Logfile = stdout;
  }
  else	{
    fp_Logfile = fopen(CPFP_Fnames->Filename_Log, "a");
    if(fp_Logfile==NULL)	{
      if(CPFP_Stat->Verbose)
	fprintf(stdout, "cropfits_log: Could not open Log File, using stdout\n");
      fp_Logfile = stdout;
    }
  }
  

  fprintf(fp_Logfile,"\nProgram %s, Version %2.1f\n", 
	  PROGRAM, CPFVERSN);
  if(strcmp(CPFP_Fnames->Filename_Namelist, "")) 
    fprintf(fp_Logfile, "Namelist File  = %s\n", 
	    CPFP_Fnames->Filename_Namelist);
  if(strcmp(CPFP_Fnames->Filename_Input, "")) 
    fprintf(fp_Logfile, "Input FITS Image = %s\n", 
	    CPFP_Fnames->Filename_Input);
  if(strcmp(CPFP_Fnames->Filename_Output, "")) 
    fprintf(fp_Logfile, "Output Image = %s\n", 
	    &(CPFP_Fnames->Filename_Output[1]));
  if(strcmp(CPFP_Fnames->Filename_Table,""))
    fprintf(fp_Logfile, "Table dump = %s\n",
	    &(CPFP_Fnames->Filename_Table[1]));
  if(strcmp(CPFP_Fnames->Filename_Map, ""))
    fprintf(fp_Logfile, "Original Pixel Map = %s\n",
	    CPFP_Fnames->Filename_Map);

  fprintf(fp_Logfile, "i Start = %d\n", CPFP_Comp->jstart);
  fprintf(fp_Logfile, "j Start = %d\n", CPFP_Comp->istart);
  fprintf(fp_Logfile, "Width = %d\n", CPFP_Comp->height);
  fprintf(fp_Logfile, "Height = %d\n", CPFP_Comp->width);
  
  fprintf(fp_Logfile, "Data plane processed ");
  if(CPFP_Image->Plane==1)
    fprintf(fp_Logfile, "1 = all\n");
  else if(CPFP_Image->Plane==2)
    fprintf(fp_Logfile, "2 = first\n");
  else if(CPFP_Image->Plane==3)
    fprintf(fp_Logfile, "3 = last\n");

  fprintf(fp_Logfile, "Ancillary Data-File Path = %s\n", 
	  CPFP_Paths->CP_Ancillary_File_Path);
  fprintf(fp_Logfile, "Verbose flag = %d\n", 
	  CPFP_Stat->Verbose);
  fprintf(fp_Logfile, "Super-verbose flag = %d\n", 
	  CPFP_Stat->SuperVerbose);
  fprintf(fp_Logfile, "Debug flag = %d\n", 
	  CPFP_Stat->Debug);

  cropfitsMessage = RIGHT_MASK&(CPFP_Stat->Status);
  cropfitsRoot = (LEFT_MASK&(CPFP_Stat->Status))>>16;
  if(cropfitsMessage&FITS_MSG_MASK) {
    cropfitsMessage -= FITS_MSG_MASK;
    ffrprt(fp_Logfile, cropfitsMessage);
  }

  if (CPFP_Stat->Verbose)
    printf("cropfits_log: Decoding messages...\n");
  
  decode_message(cropfitsMessage, cropfitsRoot,
		 CP_Message, CP_Function, CPFP_Paths, CPFP_Stat);

  cropfitsMessage = RIGHT_MASK&(CPFP_Stat->Status);
  cropfitsRoot = (LEFT_MASK&(CPFP_Stat->Status))>>16;
  if(cropfitsMessage&FITS_MSG_MASK) {
    cropfitsMessage -= FITS_MSG_MASK;
  }

  if (cropfitsMessage != 0) 
    fprintf(fp_Logfile, 
	    "cropfits Status Message      0x%04x\n%s from Function 0x%04x: %s\n",
	    cropfitsMessage, CP_Message,  cropfitsRoot, CP_Function);
  else 
    fprintf(fp_Logfile, 
	    "cropfits Status Message      0x%04x\n%s from Function 0x%04x: %s\n",
	    cropfitsMessage, "Normal exit",  cropfitsRoot, CP_Function);

  Time = clock();
  fprintf(fp_Logfile, "Processing time: %f seconds\n", (float)Time/CLOCKS_PER_SEC);
  time(&currenttime);
  datetime = ctime(&currenttime);
  
  fprintf(fp_Logfile,"Current date/time: %s",datetime);
  fprintf(fp_Logfile,"Program %s, version %2.1f, terminated\n\n", 
	  PROGRAM, CPFVERSN);
  
  if(fp_Logfile != stdout) 
    fclose(fp_Logfile);
  
  if(CPFP_Stat->Verbose)
    printf("cropfits_log_writer: Log Writer finished\n");

  return;
}


void decode_message( int           I_message, 
                     int           I_root,
		     char	   *CP_Message,
		     char	   *CP_Function,
		     CPF_Pathnames *CPFP_Paths,
		     CPF_State     *CPFP_Stat)
{

  char           codefile[2*MAX_FILENAME_LENGTH],line[80],mess[80],numstr[80],*rem;
  int            ifun=0, imess=0;
  unsigned long  num, nums;

  FILE *GETCODEBOOK, *fopen();

  sprintf(CP_Function, "%s", "Unspecified function.");
  sprintf(CP_Message, "%s", "Unknown error");
  strcpy(codefile, CPFP_Paths->CP_Ancillary_File_Path);
  strcat(codefile, "/cropfits_errcodes.h");
  
  if(CPFP_Stat->Verbose) 
    printf("cropfits_log_writer: Message-code file = %s\n",codefile);
  
  if ((GETCODEBOOK = fopen(codefile, "r")) == NULL) {
    printf("cropfits_log_writer: Could not open cropfits_errcodes.h\n");
    CPFP_Stat->Status = LOG_WRITER|LOG_WRITER;
    sprintf(CP_Message,"%s","ERRCODE_FILE_NOT_FOUND");
    sprintf(CP_Function,"%s","LOG_WRITER");
    return;
  } else {
    while (!feof(GETCODEBOOK)) {
      fscanf(GETCODEBOOK, "%s",line);
      if (line[0]== '#') {
	fscanf(GETCODEBOOK, "%s",mess);
	fscanf(GETCODEBOOK, "%s",numstr);
	num = strtoul(numstr,&rem,16);
	nums = (LEFT_MASK & num) >> 16;
	if ((nums == I_root) && (ifun == 0)) {
	  /* printf("%s %X %s\n",mess,nums,rem); */
	  sprintf(CP_Function, "%s", mess);
	  ifun = 1;
	}
	if ((num == I_message) && (ifun == 1)) {
	  /* printf("%s %X %s\n",mess,num,rem); */
	  sprintf(CP_Message, "%s", mess);
	  imess = 1;
	  break;
	}
      }
    }
    fclose(GETCODEBOOK);
  }

  if ((ifun != 1) || (imess != 1)) 
    printf("cropfits_log_writer: Error in decoding error message\n");

  return;

}



