/* $Id: cropfits_parse_namelist.c,v 1.3 2007/11/12 18:12:17 laher Exp $ */


/************************************************************ 
 *  cropfits_parse_namelist.c
 ************************************************************
 Purpose
 Read and parse a parameter file to obtain input 
 and output file names, constants, and processing flags.
 
 Overview
 Use string functions to search for separator characters,
 then a series of conditionals to find what variables are
 being defined.
 
 Definitions of Variables
 
 Input:
 Expects a either a namelist file or command-line inputs.

 External:
 See cropfits_defs.h and cropfits_exec.c

 Internal:
 fp_namelist = Pointer to open namelist file.
 CP_buf = Holds contents of namelist file.
 CP_NextWord = First entry in parsed namelist data segment.
 Must be name of namelist variable.
 CP_Equals = Must be the "=" sign.
 CP_Value = Value of a namelist variable.
 CP_Comment = Holds comment in namelist file, if any.
 CP_Pointer = Keeps track of parsing namelist data.
 CP_Comma = Holds position of comment character token.
 I_Got_Namelist = Flags whether namelist was specified on 
 command line.
 I_SectionsFound = Counts how many namelist sections 
 have been found. Not used.
 
 I_fildes_nl = File descriptor for opening namelist file.
 I_str_length = Holds length of namelist file.
 I_past_end = Flags if end of namelist file reached.
 I_found = Flags if desired section has been found.
 STAT_nlbuf = Holds files status return information.
 STATP_nlbuf = Pointer to STAT_nlbuf.
 
 ************************************************************/

#include <stdio.h>
#include "cropfits_errcodes.h"
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include "cropfits.h"
#include "cropfits_defs.h"

#define	STRING_BUFFER_SIZE	 256

void cropfits_parse_namelist( int		argc,
			      char	        **argv,
			      CPF_Filenames	*CPFP_Fnames,
			      CPF_Pathnames	*CPFP_Paths,
			      CPF_ImageInfo     *CPFP_Image,
			      CPF_Computation   *CPFP_Comp,
			      CPF_State	        *CPFP_Stat)
{
  
  FILE	*fp_namelist = NULL;
  
  char	*CP_buf;
  char	CP_NextWord[STRING_BUFFER_SIZE];
  char  CP_Equals[STRING_BUFFER_SIZE];
  char  CP_Value[STRING_BUFFER_SIZE];
  char  CP_Comment[STRING_BUFFER_SIZE];
  char	*CP_Pointer, *CP_Comma;
  
  int	i=1, I_Got_Namelist=0, I_SectionsFound=0;
  int	I_fildes_nl=0, I_str_length=0;
  int	I_past_end=0, I_found=0;
  int   temp=0;

  struct stat  STAT_nlbuf, *STATP_nlbuf;
 
  /* if no command line args given, print mini-help to stdout */
  if(argc==1)	{
    fprintf(stdout, "\nUsage: cropfits\n\n");
    fprintf(stdout, "       Filenames:\n");
    fprintf(stdout, "       -n  Input namelist filename (Optional)\n");
    fprintf(stdout, "       -i  Input image filename (Required)\n");
    fprintf(stdout, "       -o  Output cropped image filename (Required) \n");
    fprintf(stdout, "       -m  Pixel orientation map (Optional) \n");
    fprintf(stdout, "       -t  IPAC table filename (Optional)\n");
    fprintf(stdout, "       -l  Log filename (Default is stdout)\n\n");
    fprintf(stdout, "       Computational Parameters & Switches:\n");
    fprintf(stdout, "       -i1 start i (from 1)\n");
    fprintf(stdout, "       -j1 start j (from 1)\n");
    fprintf(stdout, "       -w  width\n");
    fprintf(stdout, "       -h  height\n\n");
    fprintf(stdout, "       Standard Options: \n");
    fprintf(stdout, "       -a  Ancillary file path (Default is current directory)\n");
    fprintf(stdout, "       -z  Data plane to process (1=all, 2=first, 3=last; Default is all)\n");
    fprintf(stdout, "       -b  Print debug statements \n");
    fprintf(stdout, "       -v  Verbose output          \n");
    fprintf(stdout, "       -vv SuperVerbose output\n\n");

    CPFP_Stat->Status = NAMELIST|NO_ARGS; 
    return;
  }
  
  /* loop through command line arguements to find namelist */
  while( i < argc ) {
    if( argv[i][0] == '-' ) {
      switch( argv[i][1] ) {
      case 'n':      /* -n <input_namelist_fname>  */
	if( ++i >= argc )
	  printf("cropfits_parse_namelist: -n <input_namelist_fname> missing argument\n" );
	else {
	  if(CPFP_Fnames->Filename_Namelist) /* Filename can be stored */
	    I_Got_Namelist = sscanf(argv[i], "%s", CPFP_Fnames->Filename_Namelist); 
	}
	break;
      default: 
	break;
      }
    } 
    i++;
  } 
  
  if (!I_Got_Namelist) {
    printf("cropfits_parse_namelist: Status Information: %s\n",
	   "Namelist not specified, parsing command line");
    return;
  }

  /* Open the namelist file to get its length */  
  if ( (I_fildes_nl = open(CPFP_Fnames->Filename_Namelist,O_RDONLY)) < 0) {
    printf("cropfits_parse_namelist: Error opening file %s\n",
	   CPFP_Fnames->Filename_Namelist);
    CPFP_Stat->Status = NAMELIST|FOPEN_FAILED; 
    return;
  }

  /* Find number of characters in namelist file */
  STATP_nlbuf = &STAT_nlbuf;
  fstat(I_fildes_nl, STATP_nlbuf);
  I_str_length = STATP_nlbuf->st_size;
 
  
#if 0
  /* Open the namelist file */  
  /* check for STDIN        */
  if(strcmp(CPFP_Fnames->Filename_Namelist, "stdin") == 0)	{
    fprintf(stdout, "cropfits_parse_namelist: Taking namelist from stdin\n");
    fp_namelist = stdin;
  }
  else if ( (fp_namelist=fopen(CPFP_Fnames->Filename_Namelist,"r"))==NULL) {
    printf("cropfits_parse_namelist: Error opening file %s",
	   CPFP_Fnames->Filename_Namelist);
    CPFP_Stat->Status = NAMELIST|FOPEN_FAILED; 
    return;
    }
#endif

  /* Associate a stream with the file designator generated above */
  fp_namelist=fdopen(I_fildes_nl,"r");
  
  /* Allocate memory for namelist buffer, */
  CP_buf = (char *) calloc(I_str_length+2, sizeof(char));
  
  /* Read namelist data */
  if(fread(CP_buf, sizeof(char), I_str_length, fp_namelist) == 0) {
    printf("cropfits_parse_namelist: Could not read from namelist file %s",
	   CPFP_Fnames->Filename_Namelist);
    CPFP_Stat->Status = NAMELIST|COULD_NOT_READ; 
    return;
  }

  /* pad with EOL character */
  CP_buf[I_str_length] = '\0';
  
  /* Close the stream */
  fclose(fp_namelist);
  
#if 0
  /* Look for the cropfits section */
  for(CP_Pointer = CP_buf;
      strncmp("&CROPFITSIN", CP_Pointer, strlen("&CROPFITSIN")) != 0;
      CP_Pointer = strchr(CP_Pointer+1, '&'))
    I_SectionsFound++;
#endif

#if 1
  CP_Pointer = CP_buf;
  while(CP_Pointer <= CP_buf + I_str_length && !I_past_end && !I_found) {
    CP_Pointer = strchr(CP_Pointer, '&');
    if (CP_Pointer + strlen("&CROPFITSIN") >= CP_buf + I_str_length)
      I_past_end = 1;
    else {
      if (strncmp("&CROPFITSIN", CP_Pointer, strlen("&CROPFITSIN")))
	CP_Pointer += strlen("&");
      else
	I_found = 1;
    }
  }
  if (I_past_end) {
    CPFP_Stat->Status = NAMELIST|COULD_NOT_READ; 
    return;
  }
#endif
  
  /* parsing the namelist .... */
  
  CP_Pointer += strlen("&CROPFITSIN");

  /* Match the string values read to the variable definitions and
     write the  values into the appropriate variables. */
  for(CP_Comma = strtok(CP_Pointer, ","); 
      CP_Comma != NULL && strchr(CP_Comma, '&') == 0;
      CP_Comma = strtok(NULL, ","))	{
    
    sscanf(CP_Comma, "%s%s%s", CP_NextWord, CP_Equals, CP_Value);
    if(strcmp(CP_NextWord, "Comment")==0) 
      strcpy(CP_Comment, CP_Value);
    if(strcmp(CP_NextWord, "Input_FITS_Image")==0)
      sscanf(&CP_Value[1], "%[^']", CPFP_Fnames->Filename_Input);
    if(strcmp(CP_NextWord, "Output_FITS_Image")==0) {
      sscanf("!", "%c", CPFP_Fnames->Filename_Output);  /* clobber */ 
      sscanf(&CP_Value[1], "%[^']", &(CPFP_Fnames->Filename_Output[1]));
    }
    if(strcmp(CP_NextWord, "Map_FITS_Image")==0) {
      sscanf("!", "%c", CPFP_Fnames->Filename_Map);
      sscanf(&CP_Value[1], "%[^']", &(CPFP_Fnames->Filename_Map[1]));
    }
    if(strcmp(CP_NextWord, "TableDump")==0) {
      sscanf("!", "%c", CPFP_Fnames->Filename_Table);
      sscanf(&CP_Value[1], "%[^']", &(CPFP_Fnames->Filename_Table[1]));
    }
    if(strcmp(CP_NextWord, "Log_Filename")==0) 
      sscanf(&CP_Value[1], "%[^']", CPFP_Fnames->Filename_Log);
    if(strcmp(CP_NextWord, "Ancillary_File_Path")==0) 
      sscanf(&CP_Value[1], "%[^']", CPFP_Paths->CP_Ancillary_File_Path);
 
    if(strcmp(CP_NextWord, "i1")==0)
      CPFP_Comp->jstart = atoi(CP_Value);
    if(strcmp(CP_NextWord, "j1")==0)
      CPFP_Comp->istart = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Width")==0)
      CPFP_Comp->height = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Height")==0)
      CPFP_Comp->width = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Data_Plane")==0)
      CPFP_Image->Plane = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Verbose")==0)
      CPFP_Stat->Verbose = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Superverbose")==0)
      CPFP_Stat->SuperVerbose = atoi(CP_Value);
    if(strcmp(CP_NextWord, "Debug")==0)
      CPFP_Stat->Debug = atoi(CP_Value);
  }
  
  free(CP_buf);
  
  if(CPFP_Stat->Verbose)	
    printf("cropfits_parse_namelist: Namelist File %s processed\n",
	   CPFP_Fnames->Filename_Namelist);
  
  return;
  
}



