/* $ID */

/***************************************************************
 *  cropfits_read_image.c
 ***************************************************************

 Purpose:
 Read the allocate memory for the image & its masks,
 read & store the dimensions, and read the image &
 masks (create blank masks if none specified).

 Overview:
 Use malloc to allocate the memory, CFITSIO routines to
 read the fits headers and image data 
 
 Variable Definitions
 External:
 See cropfits_defs.h and cropfits_exec.c

 Internal:
 *ffp_FITS_Input = Pointer to input file
 I_fits_return_status = hold return status of any fitsio call
 I_Anynull = flag for null pixels in images
 NumDims = dimensionality of image
 Axes = array dimensional lengths
 ImagePixels = Number of pixels contained in image
 TempImage = temporary holding spot for the image data
             
 As always,
 CP_Keyname = holds keyname for fitsio call
 CP_Comment = holds a comment for/from fits header

 ***************************************************************/

#include <time.h>
#include <stdio.h>
#include "cropfits_errcodes.h"
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "cropfits.h"
#include "cropfits_defs.h"

void cropfits_read_image( CPF_Filenames   *CPFP_Fnames,
		          CPF_ImageInfo   *CPFP_Image,
		          CPF_State       *CPFP_Stat)
{
  fitsfile *ffp_FITS_Input = NULL;
  double*   TempImage = NULL;

  int      index;
  int      NumDims;
  int      I_fits_return_status=0;
  int      I_dims_found;
  int      I_Anynull;
  double   D_nullval = 0;
  long     Axes[3] = {0};
  long     ImagePixels = 1;
  double   temp_storage;
  
  char CP_Keyname[FLEN_KEYWORD];
  char CP_Comment[FLEN_COMMENT];
  
  if(CPFP_Stat->Verbose)	
    printf("cropfits_read_image: Opening image file\n");

  fits_open_file(&ffp_FITS_Input,
		 CPFP_Fnames->Filename_Input,
		 READONLY,
		 &I_fits_return_status);
  if(I_fits_return_status){
    CPFP_Stat->Status = READ_IMAGE | FITS_MSG_MASK | I_fits_return_status;
    return;
  }

  /***********************************************
   * read data dimension keys 
   */  
  sprintf(CP_Keyname, "%s", "NAXIS");
  fits_read_keys_lng(ffp_FITS_Input,
		     CP_Keyname,
		     1, 3,
		     Axes,
		     &I_dims_found,
		     &I_fits_return_status);

  if(I_dims_found < 3)
    Axes[2] = 1;

  if(I_fits_return_status){
    CPFP_Stat->Status = READ_IMAGE | FITS_MSG_MASK | I_fits_return_status;
    return;
  }

  /* 
     Because of the way in which image data are stored in the 2-D array,
     the first image dimension (i) corresponds to NAXIS2, and the second
     dimension (j) corresponds to NAXIS1.  Hence, Axes[1] is associated
     with LengthX and Axes[0] is associated with LengthY.
  */
  CPFP_Image->LengthX = Axes[1]; 
  CPFP_Image->LengthY = Axes[0];
  CPFP_Image->LengthZ = Axes[2];
  ImagePixels = Axes[0] * Axes[1] * Axes[2];

  /***********************************************
   * search for any assigned world coordinates
   */
  
  /* using CRVAL1 to search - if WCS have been assigned, right */
  /* ascension value must be present */
  fits_read_key(   ffp_FITS_Input,
                   TDOUBLE,
                   "CRVAL1",
                   &temp_storage,
                   CP_Comment,
                   &I_fits_return_status);
  
  if(I_fits_return_status==0) {
    printf("imfliprot_read_image: Image contains world coordinates\n");
    printf("imfliprot_read_image: World coordinates cannot be processed by this module\n");
    printf("imfliprot_read_image: Terminating execution\n\n");
    CPFP_Stat->Status = READ_IMAGE| IMAGE_HAS_WCS;
    return;
  } else
    I_fits_return_status = 0;


  /***********************************************
   * allocate memory & read images
   */
  CPFP_Image->Image = (double *) calloc(ImagePixels, sizeof(double));
  TempImage = (double *) calloc(ImagePixels, sizeof(double));
  if((CPFP_Image->Image==NULL)||(TempImage==NULL)) {
    CPFP_Stat->Status = READ_IMAGE | MALLOC_FAILED;
    return;
  }

  if(CPFP_Stat->Verbose)	
    printf("cropfits_read_image: Reading image\n");  
  
  fits_read_img(ffp_FITS_Input,
		TDOUBLE,
		1, ImagePixels,
		&D_nullval,
	        TempImage,
		&I_Anynull,
		&I_fits_return_status);
  if(I_fits_return_status) {
    CPFP_Stat->Status = READ_IMAGE | FITS_MSG_MASK | I_fits_return_status;
    free(TempImage);
    return;
  }
 
  for(index=0; index < ImagePixels; index++) {
    CPFP_Image->Image[index] = TempImage[index];
  }


  /**********************************************
   * cleanup & return to caller
   */
  fits_close_file(ffp_FITS_Input, &I_fits_return_status);

  if(CPFP_Stat->SuperVerbose)  {
    printf("cropfits_read_image: X Length = %d\n", CPFP_Image->LengthX);
    printf("cropfits_read_image: Y Length = %d\n", CPFP_Image->LengthY);
  }

  free(TempImage);

  return;
}
