/* $Id: cubesub_read_data.c,v 1.13 2004/07/19 22:40:20 bob Exp $ */


/*********************************************** 
cubesub_read_data.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

Open the FITS file, allocate memory, read the data into an array, 

Definitions of Variables

Input:
CP_Filename_FITS_In = In Data File Name.

Output:
fpp_FITS_in = Pointer to fits file handle.
DPP_Data_In = Pointer to image data in array form.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "cubesub.h"
#include "cubesub_defs.h"


cubesub_read_data(	
            Filenames   *Fnames,
	    Constants   *Const,
            Dims        *Dim_In,
            float	**FPP_Data_In,
            float       **FPP_Vars_In,
            Status	*Stat,
            short       **SPP_pmask,
            short       **SPP_dmask,
            short       **SPP_bmask,
            short       **SPP_cmask,
            float       **FPP_Dark,
            float       **FPP_Dark_Vars)
{
char    I_bunit[FLEN_VALUE], D_bunit[FLEN_VALUE];
char    IU_bunit[FLEN_VALUE], DU_bunit[FLEN_VALUE];
int	i, j, k;
int	I_fits_return_status=0, I_ndims_found, I_anynull;
int     I_naxis, D_naxis, I_bitpix, D_bitpix;
long	LP_naxes[3];
long	LP_naxes1[3];
long    dimhigh;
float   I_gain, I_tint;
float	D_nullval=0;
fitsfile 	*ffp_FITS_In;
fitsfile        *ffp_Vars_In;
fitsfile 	*ffp_FITS_pmask;
fitsfile 	*ffp_FITS_dmask;
fitsfile 	*ffp_FITS_bmask;
fitsfile        *ffp_FITS_cmask;
fitsfile 	*ffp_FITS_Dark;
fitsfile        *ffp_FITS_Dark_Vars;
char    fits_comment[FLEN_COMMENT];

if(Stat->I_status) return;

/*  Open and get array data from input fits file */

fits_open_file(	&ffp_FITS_In,
		Fnames->CP_Filename_FITS_In,
		READONLY,
		&I_fits_return_status);

fits_read_key( ffp_FITS_In,
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
if(I_bitpix != -32) {
   printf("cubesub_read_data: input data file has wrong data type\n");
   Stat->I_status = READ_DATA|BIT_ERROR;
   return;
}

fits_read_key( ffp_FITS_In,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_naxis == 2) LP_naxes[2] = 1;

Dim_In->I_NX = LP_naxes[0];
Dim_In->I_NY = LP_naxes[1];
Dim_In->I_NZ = LP_naxes[2];

fits_read_key( ffp_FITS_In,
               TSTRING,
               "BUNIT",
               I_bunit,
               fits_comment,
               &I_fits_return_status);
if ( !(strcmp(I_bunit, "")) ) {
   printf("cubesub_read_data: value of BUNIT key missing in input data file\n");
   Stat->I_status = READ_DATA|MISSING_KEY;
   return;
}

/* Allocate memory for the data */
/* Don't try if problem with fits file */

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

*FPP_Data_In = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2], 
          sizeof(float));

/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In==NULL) I_fits_return_status = MALLOC_FAILED;

/* Read the data */

fits_read_img(ffp_FITS_In,
              TFLOAT,
              1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
              &D_nullval,
              *FPP_Data_In,
              &I_anynull,
              &I_fits_return_status);

fits_close_file(ffp_FITS_In,
		&I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and get array data from data uncertainty file.  Optional input.
    Data default at zero for all pixels if input file not specified.
*/

*FPP_Vars_In = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if (*FPP_Vars_In == NULL) I_fits_return_status = MALLOC_FAILED;


if(strcmp(Fnames->CP_Filename_Vars_In, "")) {
  
   fits_open_file( &ffp_Vars_In,
                Fnames->CP_Filename_Vars_In,
                READONLY,
                &I_fits_return_status);

   fits_read_key( ffp_Vars_In,
                  TINT,
                  "NAXIS",
                  &I_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_Vars_In,
                       "NAXIS",
                       1, I_naxis,
                       LP_naxes,
                       &I_ndims_found,
                       &I_fits_return_status);

   if (I_naxis == 2) LP_naxes[2] = 1;

   if (LP_naxes[0] != Dim_In->I_NX || LP_naxes[1] != Dim_In->I_NY || LP_naxes[2] != Dim_In->I_NZ) {
      printf("cubesub_read_data: dimension inconsistent between input data and uncertainty images\n");
      Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
   }

   fits_read_key( ffp_Vars_In,
                  TSTRING,
                  "BUNIT",
                  IU_bunit,
                  fits_comment,
                  &I_fits_return_status);
   if ( !(strcmp(IU_bunit, "")) ) {
      printf("cubesub_read_data: value of BUNIT key missing in input uncertainty file\n");
      Stat->I_status = READ_DATA|MISSING_KEY;
      return;
   }  

   fits_read_img(ffp_Vars_In,
                 TFLOAT,
                 1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                 &D_nullval,
                 *FPP_Vars_In,
                 &I_anynull,
                 &I_fits_return_status);

   fits_close_file(ffp_Vars_In,
                &I_fits_return_status);

}
else {
   for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
      *(*FPP_Vars_In +i) = 0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read data from dark current file */

fits_open_file( &ffp_FITS_Dark,
                Fnames->CP_Filename_Dark,
                READONLY,
                &I_fits_return_status);

fits_read_key( ffp_FITS_Dark,
               TINT,
               "BITPIX",
               &D_bitpix,
               fits_comment,
               &I_fits_return_status);
if(D_bitpix != -32) {
   printf("cubesub_read_data: input dark file has wrong data type\n");
   Stat->I_status = READ_DATA|BIT_ERROR;
   return;
}

fits_read_key( ffp_FITS_Dark,
               TINT,
               "NAXIS",
               &D_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_Dark,
                    "NAXIS",
                    1, D_naxis,
                    LP_naxes1,
                    &I_ndims_found,
                    &I_fits_return_status);

if (D_naxis == 2) LP_naxes1[2] = 1;

Dim_In->D_NX = LP_naxes1[0];
Dim_In->D_NY = LP_naxes1[1];
Dim_In->D_NZ = LP_naxes1[2];

if (Dim_In->I_NX != Dim_In->D_NX || Dim_In->I_NY != Dim_In->D_NY)
   {
   printf("Error when reading input data:\n\n");
   printf("Data and dark-current images have different pixels in at least one of the two dimensions\n");
   Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
   }

dimhigh = Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ;
if (Dim_In->I_NZ > Dim_In->D_NZ)
   {
   printf("\nInput data cube has more planes than input dark cube\n");
   printf("Dark subtraction applied only up to the highest plane corresponding to that of the input dark cube\n\n");
   }
else if (Dim_In->I_NZ < Dim_In->D_NZ)
   {
   dimhigh = Dim_In->D_NX*Dim_In->D_NY*Dim_In->D_NZ;
   printf("\nInput data cube has fewer planes than input dark cube\n");
   printf("Using dark cube up to the highest plane corresponding to that of the input data cube\n\n");
   }
   
fits_read_key( ffp_FITS_Dark,
               TSTRING,
               "BUNIT",
               D_bunit,
               fits_comment,
               &I_fits_return_status);
if ( !(strcmp(D_bunit, "")) ) {
   printf("cubesub_read_data: value of BUNIT key missing in input dark file\n");
   Stat->I_status = READ_DATA|MISSING_KEY;
   return;
}

if (strcmp(D_bunit, I_bunit)) {
   printf("cubesub_read_data: value of BUNIT key inconsistent in input files\n");
   Stat->I_status = READ_DATA|INCONSIS_KEY;
   return;
}

if (Const->Operation != 1)
{
   fits_read_key( ffp_FITS_Dark,
                  TFLOAT,
                  "GAIN",
                  &I_gain,
                  fits_comment,
                  &I_fits_return_status);

   Const->Gain = I_gain;
   if (Stat->I_Verbose) {
      printf("cubesub_read_data: FITS keyword value read from file: %s\n", Fnames->CP_Filename_Dark);
      printf("cubesub_read_data: GAIN = %f\n", Const->Gain);
   }

   if (Const->Operation == 3)
   {
       fits_read_key( ffp_FITS_Dark,
                      TFLOAT,
                      "T_INT",
                      &I_tint,
                      fits_comment,
                      &I_fits_return_status);

       Const->T_Int = I_tint;
       if (Stat->I_Verbose) {
          printf("cubesub_read_data: FITS keyword value read from file: %s\n", Fnames->CP_Filename_Dark);
          printf("cubesub_read_data: T_INT = %f\n", Const->T_Int);
       }

   }

}

*FPP_Dark = (float *) calloc(dimhigh, sizeof(float));
if(*FPP_Dark==NULL) I_fits_return_status = MALLOC_FAILED;

fits_read_img(  ffp_FITS_Dark,
                TFLOAT,
                1, Dim_In->D_NX*Dim_In->D_NY*Dim_In->D_NZ,
                &D_nullval,
                *FPP_Dark,
                &I_anynull,
                &I_fits_return_status);

fits_close_file(ffp_FITS_Dark,
                &I_fits_return_status);

if (I_fits_return_status)
   {
   Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; 
   return;
   }

/*  Open and read dark uncertainty image file.  Optional input.
    Data default at zero for all pixels if input file not specified.
*/

*FPP_Dark_Vars = (float *) calloc( dimhigh, sizeof(float));
if(*FPP_Dark_Vars == NULL) I_fits_return_status = MALLOC_FAILED;

if (strcmp(Fnames->CP_Filename_Dark_Vars, "")) {

   fits_open_file( &ffp_FITS_Dark_Vars,
                   Fnames->CP_Filename_Dark_Vars,
                   READONLY,
                   &I_fits_return_status);

   fits_read_key( ffp_FITS_Dark_Vars,
                  TINT,
                  "NAXIS",
                  &D_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_FITS_Dark_Vars,
                       "NAXIS",
                       1, D_naxis,
                       LP_naxes1,
                       &I_ndims_found,
                       &I_fits_return_status);

   if (D_naxis == 2) LP_naxes1[2] = 1;

   if (LP_naxes1[0] != Dim_In->D_NX || LP_naxes1[1] != Dim_In->D_NY || LP_naxes1[2] != Dim_In->D_NZ) {
      printf("cubesub_read_data: dimension inconsistent between input dark and uncertainty images\n");
      Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
   }
  
   fits_read_key( ffp_FITS_Dark_Vars,
                  TSTRING,
                  "BUNIT",
                  DU_bunit,
                  fits_comment,
                  &I_fits_return_status);
   if ( !(strcmp(DU_bunit, "")) ) {
      printf("cubesub_read_data: value of BUNIT key missing in input dark uncertainty file\n");
      Stat->I_status = READ_DATA|MISSING_KEY;
      return;
   }

   if (strcmp(DU_bunit, IU_bunit)) {
      printf("cubesub_read_data: value of BUNIT key inconsistent in input uncertainty files\n");
      Stat->I_status = READ_DATA|INCONSIS_KEY;
      return;
   }

   fits_read_img(ffp_FITS_Dark_Vars,
                 TFLOAT,
                 1, Dim_In->D_NX*Dim_In->D_NY*Dim_In->D_NZ,
                 &D_nullval,
                 *FPP_Dark_Vars,
                 &I_anynull,
                 &I_fits_return_status);

   fits_close_file(ffp_FITS_Dark_Vars,
                &I_fits_return_status);

}
else {
   for (i=0; i<dimhigh; ++i) {
      *(*FPP_Dark_Vars +i) = 0;
   }
}


/*  Open and read pmask file.  Optional input. 
    If file can't be opened create pmask assuming all pixes are good.
    Otherwise read in from file.
*/

*SPP_pmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Pmask, ""))
   {
   fits_open_file( &ffp_FITS_pmask,
                   Fnames->CP_Filename_Pmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_pmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_pmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_pmask, 
                   &I_fits_return_status);
   }
else
   {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i)
       {
       *(*SPP_pmask +i) = 0;
       }
   }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read dmask file.  Optional input.  */

*SPP_dmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Dmask, ""))
   {
   fits_open_file( &ffp_FITS_dmask,
                   Fnames->CP_Filename_Dmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_dmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  &D_nullval,
                  *SPP_dmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_dmask, 
                   &I_fits_return_status);
   }
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i)
       {
       *(*SPP_dmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read Bmask file.  Optional input.  */

*SPP_bmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Bmask, ""))
   {
   fits_open_file( &ffp_FITS_bmask,
                   Fnames->CP_Filename_Bmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_bmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_bmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_bmask, 
                   &I_fits_return_status);
   }
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i)
       {
       *(*SPP_bmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read Cmask file.  Optional input.  */

*SPP_cmask = (short *) calloc(dimhigh, sizeof(short));

if (strcmp(Fnames->CP_Filename_Cmask, ""))
   {
   fits_open_file( &ffp_FITS_cmask,
                   Fnames->CP_Filename_Cmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_cmask,
                  TSHORT,
                  1, Dim_In->D_NX*Dim_In->D_NY*Dim_In->D_NZ,
                  &D_nullval,
                  *SPP_cmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_cmask,
                   &I_fits_return_status);
   }
else
   for (i=0; i<dimhigh; ++i)
       {
       *(*SPP_cmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

printf("fits %d\n",I_fits_return_status);

/* Return the fits status as cubesub status, with mask to say so */
if(I_fits_return_status)
Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status;

if(Stat->I_Verbose)	{
printf("cubesub_read_data: Data read from file:\n%s\n",
Fnames->CP_Filename_FITS_In);
printf("cubesub_read_data: I_NX = %d\n", Dim_In->I_NX);
printf("cubesub_read_data: I_NY = %d\n", Dim_In->I_NY);
printf("cubesub_read_data: I_NZ = %d\n", Dim_In->I_NZ);
}
}
