/* $Id: darkest_read_data.c,v 1.9 2003/07/25 23:41:16 fan Exp $ */


/*********************************************** 
darkest_read_data.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

call subroutines to read data and dmask FITS files.
Open the FITS file, allocate memory, read the data into an array 
for other mask files.

Definitions of Variables

Input:
CP_Filename_FITSlist_In = In Data List File Name.
CP_Filename_Dmasklist = In Dmask List file Name.
SPP_pmask = In Pmask File Name.

Output:
FPP_Data_In = Pointer to hold all input FITS data.
SPP_dmask = Pointer to hold all input dmask FITS data.
DP_Data_In = Pointer to hold individual image data.
DD_Data_In = Pointer to hold individual dmask FITS data.
SPP_cmask = pointer to hold cmask array.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "darkest.h"
#include "darkest_errcodes.h"
#include "darkest_defs.h"


darkest_read_data(	
            Filenames   *Fnames,
            Dims        *Dim_In,
	    Constants   *Const,
            float	**FPP_Data_In,
            float       **FPP_Vars_In,
            float       **DP_Data_In,
	    float       **DP_Vars_In,
            short       **DD_Data_In,
            Status	*Stat,
            short       **SPP_pmask,
            short       **SPP_dmask, 
            short       **SPP_cmask) {
char    line[MAX_FILENAME_LENGTH];
int	i, j, k;
int     goodfile_index;
int     cnt_FITS, cnt_Vars, cnt_Dmask;
int	I_fits_return_status=0, I_ndims_found, I_anynull;
int     readflag;
long	LP_naxes[3];
long	LP_naxes1[3];
float	D_nullval=0;
fitsfile 	*ffp_FITS_pmask;
fitsfile 	*ffp_FITS_cmask;
char    fits_comment[FLEN_CARD];
FILE    *inFITS, *inVars, *indmask;

if(Stat->I_status) return;

/*  Open input file containing a list of FITS files */

if ((inFITS = fopen(Fnames->CP_Filename_FITSlist_In, "r")) == NULL) {
     printf("Cannot open input file %s\n", Fnames->CP_Filename_FITSlist_In);
     Stat->I_status = READ_DATA|FOPEN_FAILED;
     return;
}

cnt_FITS = 0;
while(fgets(line, MAX_FILENAME_LENGTH, inFITS) != NULL) {
     sscanf(line, "%s", Fnames->CP_Filename_FITS_Image[cnt_FITS]);
     if (Stat->I_Verbose)
           printf("input file %d = %s\n", cnt_FITS+1, Fnames->CP_Filename_FITS_Image[cnt_FITS]);
     cnt_FITS++;
}

fclose(inFITS);

/*  Open input file containing a list of Uncertainty files */

if(strcmp(Fnames->CP_Filename_Varslist_In, "")) {

   if ((inVars = fopen(Fnames->CP_Filename_Varslist_In, "r")) == NULL) {
        printf("Cannot open input file %s\n", Fnames->CP_Filename_Varslist_In);
        Stat->I_status = READ_DATA|FOPEN_FAILED;
        return;
   }

   cnt_Vars = 0;
   while(fgets(line, MAX_FILENAME_LENGTH, inVars) != NULL) {
        sscanf(line, "%s", Fnames->CP_Filename_Vars_Image[cnt_Vars]);
        if (Stat->I_Verbose)
              printf("input file %d = %s\n", cnt_Vars+1, Fnames->CP_Filename_Vars_Image[cnt_Vars]);
        cnt_Vars++;
   }

   if (cnt_Vars != cnt_FITS) {
      printf("darkest_read_data: Input FITS and uncertainty file lists contain different number of files\n");
      Stat->I_status = READ_DATA|NUM_FILE_INCONS;
   }

   fclose(inVars);

}

/*  Open and read dmask file.  Optional input.  */

if(strcmp(Fnames->CP_Filename_Dmasklist, "")) {

   if ((indmask = fopen(Fnames->CP_Filename_Dmasklist, "r")) == NULL) {
        printf("Cannot open input file %s\n", Fnames->CP_Filename_Dmasklist);
        Stat->I_status = READ_DATA|FOPEN_FAILED;
        return;
   }

   cnt_Dmask = 0;
   while(fgets(line, MAX_FILENAME_LENGTH, indmask) != NULL) {
        sscanf(line, "%s", Fnames->CP_Filename_FITS_Dmask[cnt_Dmask]);
        if (Stat->I_Verbose)
            printf("input file %d = %s\n", cnt_Dmask+1, Fnames->CP_Filename_FITS_Dmask[cnt_Dmask]);
        cnt_Dmask++;
   }

   if (cnt_Dmask != cnt_FITS) {
      printf("darkest_read_data: Input and dmask file lists contain different number of files\n");
      Stat->I_status = READ_DATA|NUM_FILE_INCONS;
   }

   fclose(indmask);

}

Const->Num_FITSfile = cnt_FITS;

if (Const->Num_FITSfile > MAX_INPUT_IMAGES) {
   printf("darkest_read_data: too many files in input FITS list\n");
   Stat->I_status = READ_DATA|TOO_MANY_FITS;
   return;
}

/* find dimensions of data based on the first DCE with full planes */

if (Stat->I_Verbose)
   printf("\ndarkest_read_data: reading and setting dimension from file %s\n\n", Fnames->CP_Filename_FITS_Image[0]);
goodfile_index = 0;
if (Const->Num_Cube_Abn != 0) {
   for (j=0; j<Const->Num_Cube_Abn; ++j) {
       if (Const->Cube_Abn[j] == j+1) goodfile_index++;  /* input file counter start from 1 */
   }
   if (goodfile_index == cnt_FITS) {
       goodfile_index = 0;
       printf("darkest_read_data: no full DCE indicated by input\n");
       Stat->I_status = READ_DATA|NO_USEFUL_DATA;
   }
   Const->Goodfile_Index = goodfile_index;
}
if (Stat->I_Verbose)
   printf("\ndarkest_read_data: data dimension based on FITS file %d in list\n", goodfile_index+1);

readflag = 1;
darkest_read_image(Fnames->CP_Filename_FITS_Image[goodfile_index],
                   readflag,
                   Dim_In,
                   Const,
                   DP_Data_In,
                   Stat);

if(Stat->I_Verbose) {
printf("darkest_read_data: I_NX = %d\n", Dim_In->I_NX);
printf("darkest_read_data: I_NY = %d\n", Dim_In->I_NY);
printf("darkest_read_data: I_NZ = %d\n\n", Dim_In->I_NZ);
}
 
/* Allocate memory for the data */
*FPP_Data_In = (float *) calloc(Const->Num_FITSfile*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                sizeof(float));
*DP_Data_In = (float *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  sizeof(float));
/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In==NULL || *DP_Data_In==NULL) I_fits_return_status = MALLOC_FAILED;
if(I_fits_return_status) return;

/* read image files */
for (k=0; k<Const->Num_FITSfile; ++k) {

    if(I_fits_return_status) return;

    readflag = 0;
    if (Const->Num_Cube_Abn > 0) {
       for (j=0; j<Const->Num_Cube_Abn; ++j) {
          if ((k+1) == Const->Cube_Abn[j]) readflag = 2;
       }
    }

    darkest_read_image(Fnames->CP_Filename_FITS_Image[k],
                       readflag,
                       Dim_In,
		       Const,
                       DP_Data_In,
                       Stat);

    for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
        *(*FPP_Data_In + i + k*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ) = *(*DP_Data_In + i);
    }

}

/*  read uncertainty files.  Optional input.  */
/*  Allocate memory for uncertainty data */
*FPP_Vars_In = (float *) calloc(Const->Num_FITSfile*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                sizeof(float));
*DP_Vars_In = (float *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, sizeof(float));
if(*FPP_Vars_In==NULL || *DP_Vars_In == NULL) I_fits_return_status = MALLOC_FAILED;
if(I_fits_return_status) return;

/* read image files if available */

if (strcmp(Fnames->CP_Filename_Vars_Image[0], "") && strcmp(Fnames->CP_Filename_Varslist_In, "")) {

   for (k=0; k<Const->Num_FITSfile; ++k) {

       if(I_fits_return_status) return;

       readflag = 0;
       if (Const->Num_Cube_Abn > 0) {
          for (j=0; j<Const->Num_Cube_Abn; ++j) {
             if ((k+1) == Const->Cube_Abn[j]) readflag = 2;
          }
       }

       darkest_read_image(Fnames->CP_Filename_Vars_Image[k],
                          readflag,
                          Dim_In,
                          Const,
                          DP_Vars_In,
                          Stat);

       for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
           *(*FPP_Vars_In + i + k*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ) = *(*DP_Vars_In + i);
       }

   }  

}
else {
   for (i=0; i<=Const->Num_FITSfile*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i) {
       *(*FPP_Vars_In +i) = 0;
   }
}

/*  read dmask files.  Optional input.  */
/*  Allocate memory for dmask data */
*SPP_dmask = (short *) calloc(Const->Num_FITSfile*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
              sizeof(short));
*DD_Data_In = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                     sizeof(short));

if(*SPP_dmask==NULL || *DD_Data_In == NULL) I_fits_return_status = MALLOC_FAILED;
if(I_fits_return_status) return;

if (strcmp(Fnames->CP_Filename_FITS_Dmask[0], "") && strcmp(Fnames->CP_Filename_Dmasklist, "")) {

   for (k=0; k<Const->Num_FITSfile; ++k) {

       darkest_read_dmask(Fnames->CP_Filename_FITS_Dmask[k],
                          Dim_In,
                          Const,
                          DD_Data_In,
                          Stat);

       for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
           *(*SPP_dmask + i + k*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ) = *(*DD_Data_In + i);
       }

   }

}
else {
   for (i=0; i<=Const->Num_FITSfile*Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i) {
       *(*SPP_dmask +i) = 0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read pmask file.  Optional input. 
    If file can't be opened create dmask assuming all pixes are good.
    Otherwise read in from file.
*/
*SPP_pmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Pmask, "")) {
   fits_open_file( &ffp_FITS_pmask,
                   Fnames->CP_Filename_Pmask,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_pmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_pmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_pmask, 
                   &I_fits_return_status);
}
else {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_pmask +i) = 0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}
/*  Open and read cmask file.  Optional input.  */

*SPP_cmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Cmask_In, "")) {
   fits_open_file( &ffp_FITS_cmask,
                   Fnames->CP_Filename_Cmask_In,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_cmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  &D_nullval,
                  *SPP_cmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_cmask, 
                   &I_fits_return_status);
}
else {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i) {
       *(*SPP_cmask +i) = 0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

printf("fits %d\n",I_fits_return_status);

if(Stat->I_Verbose)
   printf("darkest_read_data: Data read from file: %s\n\n", Fnames->CP_Filename_FITSlist_In);

}
