/* $Id: droopop_read_data.c,v 1.12 2004/07/19 22:40:20 bob Exp $ */


/*********************************************** 
droopop_read_data.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

Open the FITS file, allocate memory, read the data into an array, 

Definitions of Variables

Input:
CP_Filename_FITS_In = In Data File Name.

Output:
fpp_FITS_in = Pointer to fits file handle.
FPP_Data_In = Pointer to image data in array form.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "droopop.h"
#include "droopop_errcodes.h"
#include "droopop_defs.h"


droopop_read_data(	
            Filenames   *Fnames,
            Dims        *Dim_In,
            float	**FPP_Data_In,
            float       **FPP_Vars_In,
            float       **FPP_Refs_In,
            float       **FPP_Vars_Refs_In,
            Status	*Stat,
            short       **SPP_pmask,
            short       **SPP_dmask, 
            short       **SPP_bmask,
            short       **SPP_cmask) {
int	i, j, k;
int	I_fits_return_status=0, I_ndims_found, I_anynull;
int     I_naxis, I_bitpix;
long	LP_naxes[3];
long	LP_naxes1[3];
float	D_nullval=0;
fitsfile 	*ffp_FITS_In;
fitsfile        *ffp_Vars_In;
fitsfile        *ffp_Refs_In;
fitsfile        *ffp_Vars_Refs_In;
fitsfile 	*ffp_FITS_pmask;
fitsfile 	*ffp_FITS_dmask;
fitsfile 	*ffp_FITS_bmask;
fitsfile        *ffp_FITS_cmask;
char    fits_comment[FLEN_CARD];

if(Stat->I_status) return;

/*  Open and get array data from input fits file */

fits_open_file(	&ffp_FITS_In,
		Fnames->CP_Filename_FITS_In,
		READONLY,
		&I_fits_return_status);

fits_read_key( ffp_FITS_In,
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
if(I_bitpix != -32)
{Stat->I_status = READ_DATA|BIT_ERROR; return;}

fits_read_key( ffp_FITS_In,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_naxis == 2) LP_naxes[2] = 1;

Dim_In->I_NX = LP_naxes[0];
Dim_In->I_NY = LP_naxes[1];
Dim_In->I_NZ = LP_naxes[2];

/* Allocate memory for the data */
/* Don't try if problem with fits file */

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

*FPP_Data_In = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2], 
          sizeof(float));

/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In==NULL) I_fits_return_status = MALLOC_FAILED;

/* Read the data */

fits_read_img(ffp_FITS_In,
              TFLOAT,
              1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
              &D_nullval,
              *FPP_Data_In,
              &I_anynull,
              &I_fits_return_status);

fits_close_file(ffp_FITS_In,
		&I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/* read the input uncertainty data.  Optional input.
   Data default at zero for all pixels if input not specified.
*/

*FPP_Vars_In = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if(*FPP_Vars_In==NULL) I_fits_return_status = MALLOC_FAILED;

if (strcmp(Fnames->CP_Filename_Vars_In, "")) {

   fits_open_file( &ffp_Vars_In,
                   Fnames->CP_Filename_Vars_In,
                   READONLY,
                   &I_fits_return_status);

   fits_read_key( ffp_Vars_In,
                  TINT,
                  "NAXIS",
                  &I_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_Vars_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if (I_naxis == 2) LP_naxes[2] = 1;
   if (LP_naxes[0] != Dim_In->I_NX || LP_naxes[1] !=Dim_In->I_NY || LP_naxes[2] != Dim_In->I_NZ) {
      printf("droopop_read_data: input data uncertainty image dimension inconsistent\n");
      Stat->I_status = READ_DATA|DIM_INCONSIS;
   }

   /* Read the data */

   fits_read_img(ffp_Vars_In,
                 TFLOAT,
                 1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
                 &D_nullval,
                 *FPP_Vars_In,
                 &I_anynull,
                 &I_fits_return_status);

   fits_close_file(ffp_Vars_In,
                 &I_fits_return_status);

}
else {
   for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
       *(*FPP_Vars_In +i) = 0.0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read reference data file.  Optional Input.
    If file is not specified create reference array with all zero value
*/
*FPP_Refs_In = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if(*FPP_Refs_In==NULL) I_fits_return_status = MALLOC_FAILED;

if (strcmp(Fnames->CP_Filename_Refs_In, "")) {

   fits_open_file( &ffp_Refs_In,
                   Fnames->CP_Filename_Refs_In,
                   READONLY,
                   &I_fits_return_status);

   fits_read_key( ffp_Refs_In,
                  TINT,
                  "NAXIS",
                  &I_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_Refs_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if (I_naxis == 2) LP_naxes[2] = 1;
   if (LP_naxes[0] != Dim_In->I_NX || LP_naxes[1] !=Dim_In->I_NY || LP_naxes[2] != Dim_In->I_NZ) {
      printf("droopop_read_data: input reference image dimension inconsistent\n");
      Stat->I_status = READ_DATA|DIM_INCONSIS;
   }

   /* Read the data */

   fits_read_img(ffp_Refs_In,
                 TFLOAT,
                 1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
                 &D_nullval,
                 *FPP_Refs_In,
                 &I_anynull,
                 &I_fits_return_status);

   fits_close_file(ffp_Refs_In,
                 &I_fits_return_status);

}
else {
   for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
       *(*FPP_Refs_In +i) = 0.0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read reference data uncertainty file.  Optional Input.
    If file is not specified create reference array with all zero value
*/
*FPP_Vars_Refs_In = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if(*FPP_Vars_Refs_In==NULL) I_fits_return_status = MALLOC_FAILED;

if (strcmp(Fnames->CP_Filename_Vars_Refs_In, "")) {

   fits_open_file( &ffp_Vars_Refs_In,
                   Fnames->CP_Filename_Vars_Refs_In,
                   READONLY,
                   &I_fits_return_status);

   fits_read_key( ffp_Vars_Refs_In,
                  TINT,
                  "NAXIS",
                  &I_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_Vars_Refs_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if (I_naxis == 2) LP_naxes[2] = 1;
   if (LP_naxes[0] != Dim_In->I_NX || LP_naxes[1] !=Dim_In->I_NY || LP_naxes[2] != Dim_In->I_NZ) {
      printf("droopop_read_data: input reference uncertainty image dimension inconsistent\n");
      Stat->I_status = READ_DATA|DIM_INCONSIS;
   }

   /* Read the data */

   fits_read_img(ffp_Vars_Refs_In,
                 TFLOAT,
                 1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
                 &D_nullval,
                 *FPP_Vars_Refs_In,
                 &I_anynull,
                 &I_fits_return_status);

   fits_close_file(ffp_Vars_Refs_In,
                 &I_fits_return_status);

}
else {
   for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
       *(*FPP_Vars_Refs_In +i) = 0.0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}
/*  Open and read pmask file.  Optional input. 
    If file can't be opened create dmask assuming all pixes are good.
    Otherwise read in from file.
*/

*SPP_pmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Pmask, "")) {
   fits_open_file( &ffp_FITS_pmask,
                   Fnames->CP_Filename_Pmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_pmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_pmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_pmask, 
                   &I_fits_return_status);
}
else {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_pmask +i) = 0;
   }
}

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read dmask file.  Optional input.  */

*SPP_dmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Dmask, "")) {
   fits_open_file( &ffp_FITS_dmask,
                   Fnames->CP_Filename_Dmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_dmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  &D_nullval,
                  *SPP_dmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_dmask, 
                   &I_fits_return_status);
}
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i) {
       *(*SPP_dmask +i) = 0;
   }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read bmask file.  Optional input.  */

*SPP_bmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Bmask, "")) {
   fits_open_file( &ffp_FITS_bmask,
                   Fnames->CP_Filename_Bmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_bmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_bmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_bmask, 
                   &I_fits_return_status);
}
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_bmask +i) = 0;
   }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read cmask file.  Optional input.  */

*SPP_cmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, sizeof(short));

if (strcmp(Fnames->CP_Filename_Cmask, "")) {
   fits_open_file( &ffp_FITS_cmask,
                   Fnames->CP_Filename_Cmask,
                   READONLY,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_cmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_cmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_cmask,
                   &I_fits_return_status);
}
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_cmask +i) = 0;
   }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

printf("fits %d\n",I_fits_return_status);

if(Stat->I_Verbose)	{
printf("droopop_read_data: Data read from file:\n%s\n",
Fnames->CP_Filename_FITS_In);
printf("droopop_read_data: I_NX = %d\n", Dim_In->I_NX);
printf("droopop_read_data: I_NY = %d\n", Dim_In->I_NY);
printf("droopop_read_data: I_NZ = %d\n", Dim_In->I_NZ);
}

}
