/*********************************************** 
flatap_output.c

Purpose

Save the flatap outputs to files.

Overview
Write the results of the computations as a FITS file.

Definitions of Variables

Input:
I_NX = Number of pixels in a row of the image
I_NY = Number of pixels in a column of the image
CP_Filename_FITS_In = Input Data File Name. 

Output:


Internal:
ffp_FITS = Handle to output FITS file.
I_fits_return_status = Resturn value of fits function.
CP_Keyname = name of Key to be written.
CP_String = String value to be written to header.
CP_Comment = String value to be written in comment field of header.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "flatap.h"
#include "flatap_defs.h"

void writeDLhistoryline (fitsfile *ffp_FITS, char *programname, 
                         char *versionstring, int *I_fits_return_status) 
{
  time_t  timer;
  char    CP_History[81], program[129], version[32], timestring[32];
  
  time(&timer);
  sprintf(timestring, "%s", asctime(localtime(&timer)));
  sprintf(program, "%s", programname);
  sprintf(version, "%s", versionstring);
  program[25] = '\0';
  version[12] = '\0';
  timestring[24] = '\0';
  sprintf(CP_History, "%-25s %2s %5s   %24s", program, "v.", 
          version, timestring);
  fits_write_history(ffp_FITS,
                     CP_History,
                     I_fits_return_status);
  
}

int flatap_output(
            Filenames	  *Fnames,
            Dims          *Dim_In,
            Status	  *Stat,
            struct tm     *tmp_time,
            short         *SPP_dmask,
            float        *FPP_Data_In,
	    float        *sigma)
{
  int	I_fits_return_status=0;
  int	I_Num_In;
  int   string_length;
  int   i;
  
  char    CP_Keyname[FLEN_KEYWORD], CP_Comment[FLEN_COMMENT], CP_Filename_In[FLEN_FILENAME];
  char    CP_String[22];
  char    format[10];
  char    progname [16];
  char    verstr[32];

  fitsfile 	*ffp_FITS_Out;
  fitsfile 	*ffp_FITS_In;
  fitsfile 	*ffp_FITS_Dmask;
  
  printf(asctime(tmp_time));
  
  if(Stat->I_status) return(-1);
  
  /* Open dmask file. If found replace data.  */
  
  if(strcmp(Fnames->CP_Filename_Dmask, "")!=0) {
    fits_open_file( &ffp_FITS_Dmask,
		    Fnames->CP_Filename_Dmask,
		    READWRITE,
		    &I_fits_return_status);
    
    /* Add module specific keywords */
    
/*     sprintf(CP_Comment, "%s", "Begin FLATAP module additions"); */
/*     fits_write_comment(ffp_FITS_Dmask, */
/* 		       CP_Comment, */
/* 		       &I_fits_return_status); */
    
/*     sprintf(CP_Comment, "%s", asctime(tmp_time)); */
/*     string_length = strlen(CP_Comment)-1; */
/*     sprintf(format, "%%.%ds", string_length); */
/*     sprintf(CP_Comment, format, asctime(tmp_time)); */
/*     fits_write_comment(ffp_FITS_Dmask, */
/* 		       CP_Comment, */
/* 		       &I_fits_return_status); */

    /* write HISTORY line for flatap processing */
    sprintf(progname, "%s", "FLATAP");
    sprintf(verstr, "%.3f", VERSN);
    writeDLhistoryline(ffp_FITS_Dmask, progname, verstr, &I_fits_return_status);
    if (I_fits_return_status) {
      Stat->I_status = DATE|I_fits_return_status; 
      return;
    }              

    fits_write_img(ffp_FITS_Dmask,
             TSHORT,
             1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
             SPP_dmask,
             &I_fits_return_status);		
    
    fits_flush_file(ffp_FITS_Dmask,
		    &I_fits_return_status);
    
    fits_close_file(ffp_FITS_Dmask,
		    &I_fits_return_status);
  }
  
  if (I_fits_return_status) {
    Stat->I_status = OUTPUT_DATA|FITS_MSG_MASK|I_fits_return_status; 
    return(-1);
  }

  /* write out the image data */
  fits_open_file( &ffp_FITS_In,
		  Fnames->CP_Filename_FITS_In,
		  READWRITE,
		  &I_fits_return_status);
  
  fits_create_file( &ffp_FITS_Out,
		    Fnames->CP_Filename_FITS_Out,
		    &I_fits_return_status);
  
  if (I_fits_return_status) {
    Stat->I_status = OUTPUT_DATA|FITS_MSG_MASK|I_fits_return_status; 
    return(-1);
  }
  
  fits_copy_header(ffp_FITS_In,
		   ffp_FITS_Out,
		   &I_fits_return_status);
  
  /* Add module specific keywords */
  /* write HISTORY line for flatap processing */
  sprintf(progname, "%s", "FLATAP");
  sprintf(verstr, "%.3f", VERSN);
  writeDLhistoryline(ffp_FITS_Out, progname, verstr, &I_fits_return_status);
  if (I_fits_return_status) {
    Stat->I_status = DATE|I_fits_return_status; 
    return;
  }              
  
/*   strcpy(CP_Comment, "Begin FLATAP module additions"); */
/*   fits_write_comment(ffp_FITS_Out, */
/* 		     CP_Comment, */
/* 		     &I_fits_return_status); */
  
/*   sprintf(CP_Comment, "%s", asctime(tmp_time)); */
/*   string_length = strlen(CP_Comment)-1; */
/*   sprintf(format, "%%.%ds", string_length); */
/*   sprintf(CP_Comment, format, asctime(tmp_time)); */
/*   fits_write_comment(ffp_FITS_Out, */
/* 		     CP_Comment, */
/* 		     &I_fits_return_status); */
  
  fits_write_img( ffp_FITS_Out,
		  TFLOAT,
		  1, (Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ),
		  FPP_Data_In,
		  &I_fits_return_status);
  
  fits_flush_file(ffp_FITS_Out,
		  &I_fits_return_status);
  
  fits_close_file(ffp_FITS_Out,
		  &I_fits_return_status);

  fits_close_file(ffp_FITS_In,
		  &I_fits_return_status);
  
  if(I_fits_return_status) 
    Stat->I_status = OUTPUT_DATA|FITS_MSG_MASK|I_fits_return_status;

  /* write out uncertainty data, if uncertainty image provided */
  if(((strcmp(Fnames->insigma, "")!=0))&&
     ((strcmp(Fnames->outsigma, "")!=0)))  {
    fits_open_file(&ffp_FITS_In,
		   Fnames->insigma,
		   READONLY,
		   &I_fits_return_status);
    fits_create_file(&ffp_FITS_Out,
		     Fnames->outsigma,
		     &I_fits_return_status);
    if(I_fits_return_status) {
      Stat->I_status = OUTPUT_DATA|FITS_MSG_MASK|I_fits_return_status;
      return(-1);
    }
    fits_copy_header(ffp_FITS_In,
		     ffp_FITS_Out,
		     &I_fits_return_status);
/*     strcpy(CP_Comment, "Begin FLATAP module additions"); */
/*     fits_write_comment(ffp_FITS_Out, */
/* 		       CP_Comment, */
/* 		       &I_fits_return_status); */
    writeDLhistoryline(ffp_FITS_Out, progname, verstr, &I_fits_return_status);
    if (I_fits_return_status) {
      Stat->I_status = DATE|I_fits_return_status; 
      return;
    }              

    fits_write_date(ffp_FITS_Out,
		    &I_fits_return_status);
    fits_flush_file(ffp_FITS_Out,
		    &I_fits_return_status);
    fits_write_img(ffp_FITS_Out,
		   TFLOAT,
		   1, (Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ),
		   sigma,
		   &I_fits_return_status);
    if(I_fits_return_status)
      Stat->I_status = OUTPUT_DATA|FITS_MSG_MASK|I_fits_return_status;
    fits_close_file(ffp_FITS_In,
		    &I_fits_return_status);
    fits_close_file(ffp_FITS_Out,
		    &I_fits_return_status);
  }

  if(Stat->I_Verbose)
    printf("flatap_output: Results written to file(s).\n");
  
  return(0);
}
