/*************************************************
flatap_parse_namelist.c

Purpose

Read and parse a namelist file to get the coefficients for 
the magnitude of the bandwidth effect.

Overview

Use string functions to search for separator characters,
then a series of conditionals to find what variables are
being defined.

Definitions of Variables

Input:

CP_Filename_Namelist = Namelist Filename.
 
Local:
fp_namelist = Pointer to open namelist file.
CP_buf = Holds contents of namelist file.
CP_NextWord = First entry in parsed namelist data segment.
Must be name of namelist variable.
CP_Equals = Must be the "=" sign.
CP_Value = Value given to namelist variable.
CP_Comment = Holds comment in namelist file, if any.
CP_Pointer = Keeps track of parsing of namelist data.


CHANGES - March 2001 (heidi@ipac.caltech.edu)
please see my comments in flatap_exec.c

***********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include "flatap.h"
#include "flatap_defs.h"

#define	STRING_BUFFER_SIZE	 256
#define	FILE_BUFFER_SIZE	8192

int flatap_parse_namelist(int         argc,
		      char        **argv,
		      Filenames   *Fnames,
		      Status	  *Stat)
{
  FILE	*fp_namelist;

  char	*CP_buf;
  char	CP_NextWord[STRING_BUFFER_SIZE];
  char  CP_Equals[STRING_BUFFER_SIZE];
  char  CP_Value[STRING_BUFFER_SIZE];
  char  CP_Comment[STRING_BUFFER_SIZE];
  char	*CP_Pointer, *CP_Comma;
  
  int   Got_Namelist = 0;
  int   I_fildes_nl=0;
  int	i=1, I_Got_Namelist=0, I_nlchk=1;				
  int   past_end = 0, found = 0, str_length = 0;

  struct stat  nlbuf, *p_nlbuf;

  p_nlbuf = &nlbuf;
  
  if(Stat->I_status) return(-1);
  
  if(argc==1) {
    printf("flatap:\n");
    printf("        Required arguments\n");
    printf("        -i  <file>    Input DCE\n");
    printf("        -o  <file>    Output DCE\n");
    printf("        -f  <file>    Flatfield\n");
    printf("        Optional arguments\n");
    printf("        -ui <file>    Input Uncertainty image\n");
    printf("        -uo <file>    Output Uncertainty image\n");
    printf("        -n  <file>    Namelist\n");
    printf("        -d  <file>    Dmask or IRS Bmask\n");
    printf("        -dm <integer> Dmask fatality level (default 16384)\n");
    printf("        -p  <file>    Pmask\n");
    printf("        -pm <integer> Pmask fatality level (default 128)\n");
    printf("        -c  <file>    Cmask\n");
    printf("        -cm <integer> Cmask fatality level (default 256)\n");
    printf("        -t  <file>    Transmission Calibrator Flatfield\n");
    printf("        -ct <file>    Cmask for Transmission Calibrator Flatfield\n");
    printf("        -l  <file>    Processing Log File (optional)\n");
    printf("        -v            Verbose\n");
    printf("Note: If a Cmask is given for the transmission calibrator flat (-tc),\n");
    printf("the Cmask fatality level provided in the -cm option will be used for \n");
    printf("masking the flat.\n\n");
	  
    Stat->I_status = PARSE_ARGS|CL_ARGS_INCRT;
    return(-1);
  }

  /* get namelist filename from command line */
  while( i < argc )     {
    if( argv[i][0] == '-' )     {
      switch( argv[i][1] ) {
      case 'n':      /* -n <input_namelist_fname>  */
        if( ++i >= argc )
          printf("flatap_parse_namelist: -n <input_namelist_fname> missing argument\n" );
        else
          if(Fnames->CP_Filename_Namelist) 
            Got_Namelist = sscanf(argv[i], "%s", Fnames->CP_Filename_Namelist); 
        break;
      default:
        break;
      }
    }
    i++;
  }

  if(!Got_Namelist){
    printf("flatap_parse_namelist: Information only: %s\n",
	   "Namelist not specified.");
    return(0);
  }

  /* Open the namelist file */  
  if ((I_fildes_nl = open(Fnames->CP_Filename_Namelist,O_RDONLY)) == -1) {
    printf("flatap_parse_namelist: Error opening file %s\n",
	   Fnames->CP_Filename_Namelist);
    Stat->I_status = NAMELIST|FOPEN_FAILED; 
    return(-1);
  }
  
  /* Find number of characters in namelist file */
  fstat(I_fildes_nl,p_nlbuf);

  /* associate a stream with the fildes generated above */
  fp_namelist=fdopen(I_fildes_nl,"r");

  /* Allocate memory for filename buffer */
  /* add 1 to file length so string can be null terminated */
  str_length = p_nlbuf->st_size + 1;
  CP_buf = (char *) calloc(str_length, sizeof(char));
  if (CP_buf == NULL) {
    Stat->I_status = NAMELIST|MALLOC_FAILED;
    return(-1);
  }
  
  /* Read the namelist data */
  if(fread(CP_buf, sizeof(char), p_nlbuf->st_size, fp_namelist) == 0) {
    printf("flatap_parse_namelist: Could not read from file %s",
	   Fnames->CP_Filename_Namelist);
    Stat->I_status = NAMELIST|COULDNOTREAD;
    return(-1);
  }
  
  CP_buf[str_length-1] = '\0';
  
  /* Close the stream */
  fclose(fp_namelist);
  
  /* Look for the flatap section */
  CP_Pointer = CP_buf;
  
  while(CP_Pointer <= CP_buf + str_length && !past_end && !found) {
    CP_Pointer = strchr(CP_Pointer, '&');
    if (CP_Pointer + strlen("&FLATAPIN") >= CP_buf + str_length)
      past_end = 1;
    else {
      if (strncmp("&FLATAPIN", CP_Pointer, strlen("&FLATAPIN")))
	CP_Pointer += strlen("&");
      else
	found = 1;
    }
  }
  
  if (past_end) {
    Stat->I_status = NAMELIST|NL_INCRT; 
    return(-1);
  }
  
  CP_Pointer += strlen("&FLATAPIN");

  /* Match the string values read to the variable definitions and
     write the  values into the appropriate variables. */
  for(CP_Comma = strtok(CP_Pointer, ",\n"); 
      CP_Comma != NULL && strchr(CP_Comma, '&') == 0;
      CP_Comma = strtok(NULL, ",\n")) {
    sscanf(CP_Comma, "%s%s%s", CP_NextWord, CP_Equals, CP_Value);
    
    if(strcmp(CP_NextWord, "Comment")==0) 
      strcpy(CP_Comment, CP_Value);
    
    if(strcmp(CP_NextWord, "FITS_Input_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_FITS_In);
    
    if(strcmp(CP_NextWord, "Input_Uncertainty_File") == 0)
      sscanf(&CP_Value[1], "%[^']", Fnames->insigma);
    
    if(strcmp(CP_NextWord, "Output_Uncertainty_File") == 0) {
      sscanf("!", "%c", Fnames->outsigma);
      sscanf(&CP_Value[1], "%[^']", &(Fnames->outsigma[1]));
    }

    if(strcmp(CP_NextWord, "FITS_Output_File") == 0) {
      sscanf("!", "%c", Fnames->CP_Filename_FITS_Out);
      sscanf(&CP_Value[1], "%[^']", &(Fnames->CP_Filename_FITS_Out[1]));
    }
    
    if(strcmp(CP_NextWord, "FITS_Flatfield_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Flat);
    
    if(strcmp(CP_NextWord, "Cmask_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Cmask);
    
    if(strcmp(CP_NextWord, "Dmask_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Dmask);
    
    if(strcmp(CP_NextWord, "Pmask_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Pmask);

    if(strcmp(CP_NextWord, "DmaskMask") == 0) 
      Stat->DmaskMask = atoi(CP_Value);
    
    if(strcmp(CP_NextWord, "PmaskMask") == 0) 
      Stat->PmaskMask = atoi(CP_Value);
    
    if(strcmp(CP_NextWord, "CmaskMask") == 0) 
      Stat->CmaskMask = atoi(CP_Value);
    
    if(strcmp(CP_NextWord, "Transmission_Cal_Flatfile") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Transcal);

    if(strcmp(CP_NextWord, "Processing_Log_File") == 0) 
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Log);
    
    if(strcmp(CP_NextWord, "Transmission_Cal_Mask_File") == 0)
      sscanf(&CP_Value[1], "%[^']", Fnames->CP_Filename_Cmasktc);
    
    if(strcmp(CP_NextWord, "Verbose") == 0) 
      Stat->I_Verbose = atoi(CP_Value);

  }
  
  free(CP_buf);
  
  if(Stat->I_Verbose)	
    printf("flatap_parse_namelist: Namelist File = \t%s\n",
	   Fnames->CP_Filename_Namelist);
  
  return(0);
}

