/*********************************************** 
flatap_read_data.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

Open the FITS file, allocate memory, read the data into an array, 

Definitions of Variables

Input:
CP_Filename_FITS_In = In Data File Name.

Output:
fpp_FITS_in = Pointer to fits file handle.
DPP_Data_In = Pointer to image data in array form.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "flatap.h"
#include "flatap_defs.h"

int flatap_read_data(	
            Filenames   *Fnames,
            Dims        *Dim_In,
            float	**data_in,
	    float      **sigma,
            Status	*Stat,
            short       **pmask,
            short       **dmask,
            short       **cmask,
            float       **flat,
            float       **transcal,
            short       **cmasktc)
{
  int	 i, status=0, numdims=0, nulls=0, naxis=0;
  int    pixels, planes;
  long	 naxes[3] = {0,0,0};
  float  nullval=0;
  
  fitsfile  *handle;
  char      fits_comment[FLEN_COMMENT];

  if(Stat->I_status) return(-1);

  /*  Open and get array data from input fits file */
  fits_open_file(&handle, Fnames->CP_Filename_FITS_In, READONLY, &status);
  fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
  fits_read_keys_lng(handle,"NAXIS", 1, naxis, naxes, &numdims, &status);
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    fits_close_file(handle, &status);
    return(-1);
  }

  if(naxis == 2) 
    naxes[2] = 1;

  Dim_In->I_NY = naxes[0];
  Dim_In->I_NX = naxes[1];
  Dim_In->I_NZ = naxes[2];
  pixels = Dim_In->I_NY *Dim_In->I_NX; 
  planes = Dim_In->I_NZ;

  *data_in = (float *) malloc(pixels*planes* sizeof(float));
  if(*data_in==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    fits_close_file(handle, &status);
    return(-1);
  }

  fits_read_img(handle, TFLOAT, 1, pixels*planes, &nullval, *data_in, &nulls, &status);
  fits_close_file(handle, &status);
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    return(-1);
  }

  /*  Open and get array data from flat field file */
  nulls=0; naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *flat = (float *) malloc(2*pixels*sizeof(float));
  if(*flat==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  fits_open_file(&handle, Fnames->CP_Filename_Flat, READONLY, &status);
  fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
  fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    fits_close_file(handle, &status);
    return(-1);
  }
  if((Dim_In->I_NY != naxes[0])||(Dim_In->I_NX != naxes[1])){
    Stat->I_status = READ_DATA|DIMS_INCRT; 
    fits_close_file(handle, &status);
    return(-1);
  }
  if((naxis<3)||(naxes[2]<2)) {
    printf("flatap_read_data: WARNING - Flatfield does not contain 2nd plane for uncertainty!\n");
    printf("flatap_read_data: No uncertainty calculations will be made.\n");
    Stat->uncertainty = FALSE;
    strcpy(Fnames->insigma, "");
    strcpy(Fnames->outsigma, "");
  }
  if(Stat->uncertainty == TRUE) 
    fits_read_img(handle, TFLOAT, 1, 2*pixels, &nullval, *flat, &nulls, &status);
  else
    fits_read_img(handle, TFLOAT, 1, pixels, &nullval, *flat, &nulls, &status);
  fits_close_file(handle, &status);
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    return(-1);
  }
 
  /* load data from DCE uncertainty file */
  nulls=0; naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *sigma = (float *) calloc(pixels*planes,sizeof(float));
  if(*sigma==NULL){
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  if(strcmp(Fnames->insigma, "")!=0) {
    fits_open_file(&handle, Fnames->insigma, READONLY, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status;
      fits_close_file(handle, &status);
      return(-1);
    }
    if(numdims==2)
      naxes[2] = 1;

    if((Dim_In->I_NY != naxes[0])||
       (Dim_In->I_NX != naxes[1])||
       (Dim_In->I_NZ != naxes[2])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    fits_read_img(handle, TFLOAT, 1, pixels*planes, &nullval, *sigma, &nulls, &status);
    fits_close_file(handle, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      return(-1);
    }
  } else {
    printf("flatap_read_data: WARNING - no input DCE uncertainty image!\n");
    printf("flatap_read_data: No uncertainty calculations will be made.\n");
    Stat->uncertainty = FALSE;
  }

  /* PMASK - optional input. 1 plane pixel mask only, default values of 0 */
  nulls = 0; naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *pmask = (short *) calloc(pixels, sizeof(short));
  if(*pmask==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  if(strcmp(Fnames->CP_Filename_Pmask, "")) {
    fits_open_file(&handle, Fnames->CP_Filename_Pmask, READONLY, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if((Dim_In->I_NY != naxes[0])||(Dim_In->I_NX != naxes[1])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    fits_read_img(handle,TSHORT, 1, pixels, &nullval, *pmask, &nulls, &status);
    fits_close_file(handle, &status);
  } 
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    return(-1);
  }


  /* DMASK - optional input. cube same size as data, default values of 0  */
  nulls = 0;  naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *dmask = (short *) calloc(pixels*planes, sizeof(short));
  if(*dmask==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  if(strcmp(Fnames->CP_Filename_Dmask, "")) {
    fits_open_file(&handle, Fnames->CP_Filename_Dmask, READWRITE, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if(numdims==2) 
      naxes[2]=1;
    if((Dim_In->I_NY != naxes[0])||
       (Dim_In->I_NX != naxes[1])||
       (Dim_In->I_NZ != naxes[2])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    fits_read_img(handle, TSHORT, 1, pixels*planes, &nullval, *dmask, &nulls, &status);
    fits_close_file(handle, &status);
  } 
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    return(-1);
  }

  /* CMASK - optional input. 1 plane calibration mask only, default values of 0 */
  nulls = 0;  naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *cmask = (short *) calloc(pixels, sizeof(short));
  if(*cmask==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  if(strcmp(Fnames->CP_Filename_Cmask, "")) {
    fits_open_file(&handle, Fnames->CP_Filename_Cmask, READONLY, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if((Dim_In->I_NY != naxes[0])||(Dim_In->I_NX != naxes[1])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    fits_read_img(handle, TSHORT, 1, pixels,&nullval, *cmask, &nulls, &status);
    fits_close_file(handle, &status);
  }
  if(status) {
    Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
    return(-1);
  }



  /* Transcal flat field - optional input. 1 plane flat only,defaults to 1 */
  if(strcmp(Fnames->CP_Filename_Transcal, "")) {
    nulls = 0; naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
    *transcal = (float *) malloc(2*pixels*sizeof(float));
    if(*transcal==NULL) {
      Stat->I_status = READ_DATA | MALLOC_FAILED;
      return(-1);
    }
    fits_open_file(&handle, Fnames->CP_Filename_Transcal, READONLY, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if((Dim_In->I_NY != naxes[0])||(Dim_In->I_NX != naxes[1])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if((naxis<3)||(naxes[2]<2)) {
      printf("flatap_read_data: WARNING - Trans Cal Flat does not contain 2nd plane for uncertainty!\n");
      printf("flatap_read_data: No uncertainty calculations will be made.\n");
      Stat->uncertainty = FALSE;
      strcpy(Fnames->insigma, "");
      strcpy(Fnames->outsigma, "");
    }
    if(Stat->uncertainty == TRUE) 
      fits_read_img(handle, TFLOAT, 1, 2*pixels, &nullval, *transcal, &nulls, &status);
    else
      fits_read_img(handle, TFLOAT, 1, pixels, &nullval, *transcal, &nulls, &status);
    fits_close_file(handle, &status);
    if(status){
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status;
      return(-1);
    }
  } else {
    if(Stat->uncertainty==FALSE) {
      *transcal = (float *) malloc(2*pixels*sizeof(float));
      if(*transcal==NULL) {
	Stat->I_status = READ_DATA | MALLOC_FAILED;
	return(-1);
      }
      for(i=0; i<pixels; i++) 
	*(*transcal+i) = 1.0;
    }
  }

  /* CMASKTC - optional. Calibration mask for transcal image. default pix value 0*/
  nulls = 0; naxes[0] = 0; naxes[1] = 0; naxes[2] = 0;
  *cmasktc = (short *) calloc(pixels, sizeof(short));
  if(*cmasktc==NULL) {
    Stat->I_status = READ_DATA | MALLOC_FAILED;
    return(-1);
  }
  if(strcmp(Fnames->CP_Filename_Cmasktc, "")) {
    fits_open_file(&handle, Fnames->CP_Filename_Cmasktc, READONLY, &status);
    fits_read_key(handle, TINT, "NAXIS", &naxis, fits_comment, &status);
    fits_read_keys_lng(handle, "NAXIS", 1, naxis, naxes, &numdims, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status; 
      fits_close_file(handle, &status);
      return(-1);
    }
    if((Dim_In->I_NY != naxes[0])||(Dim_In->I_NX != naxes[1])){
      Stat->I_status = READ_DATA|DIMS_INCRT; 
      fits_close_file(handle, &status);
      return(-1);
    }
    fits_read_img(handle, TSHORT, 1, pixels, &nullval, *cmasktc, &nulls, &status);
    fits_close_file(handle, &status);
    if(status) {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|status;
      return(-1);
    }
  }

  if(Stat->I_Verbose)	{
    printf("flatap_read_data: Data read from file:\n%s\n", Fnames->CP_Filename_FITS_In);
    printf("flatap_read_data: I_NX = %d\n", Dim_In->I_NX);
    printf("flatap_read_data: I_NY = %d\n", Dim_In->I_NY);
    printf("flatap_read_data: I_NZ = %d\n", Dim_In->I_NZ);
  }
  return(0);
}
