#include  "fpgen.h"

/*
 *  Declare a global variable fpg_version that will
 *  contain the program version number.
 *  Declare a global variable temp_log that will be
 *  a pointer to a character string that temporarily
 *  will contain output information until a log file
 *  has been opened (assuming this option is chosen
 *  at run time).
 *  Declare a global variable verbose_mode that is
 *  initialized to zero (indicating verbose mode
 *  is turned OFF).  This variable will be set to
 *  a positive value in the function parse_cmdln,
 *  if the command line option '-v' is included
 *  (indicating verbose mode is turned ON).  The
 *  '-v' option is followed by a number (currently
 *  1 or 2), indicating what level of verbosity is
 *  desired (a higher value yields more information).
 *  Alternatively, this option can also be set via
 *  the namelist file.
 *  Declare a global variable log_stream that will
 *  be a pointer to the output log stream (i.e.,
 *  either a log file if one is specified or
 *  stdout).
 */
char
      fpg_version[PROG_VER_STR_LEN],
      quiet_mode = 0,
     *temp_log,
      verbose_mode = 0;

FILE
      *log_stream = NULL;


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  main (fpgen)                                      *** *
 *                                                                       *
 *                                                                       *
 * ***  This function (or program) - FPGen or Final Product          *** *
 * ***  Generator - reads an input DCE (FITS) file and a template    *** *
 * ***  file, using the information in the input files to produce    *** *
 * ***  an output DCE (FITS), final product, file.                   *** *
 *                                                                       *
 * ***  Input:  argc                      - number of command line   *** *
 * ***                                      arguments.               *** *
 * ***          argv                      - array containing the     *** *
 * ***                                      command line arguments.  *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS_EXIT      - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  WARNING_EXIT      - successful processing    *** *
 * ***                                      completed, but a         *** *
 * ***                                      WARNING message has      *** *
 * ***                                      been issued (indicating  *** *
 * ***                                      minor - but not fatal -  *** *
 * ***                                      error or problem).       *** *
 * ***                  FAILURE_EXIT      - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  main(int     argc,
          char**  argv)

{

   char
         input_file_name[MAX_FNAM_LEN],
         log_file_name[MAX_FNAM_LEN],
         namelist_file_name[MAX_FNAM_LEN],
         output_file_name[MAX_FNAM_LEN],
         sodb_entries_file_name[MAX_FNAM_LEN],
         template_file_name[MAX_FNAM_LEN];

   int
        blank_blocks = NOT_SET,
        exit_code    = SUCCESS_EXIT,
        hist_copy    = NOT_SET,
        line_count   = NOT_SET,
        mask         = NO_MASK,
        rcode;

   struct  sodb_tbl
                     *sodb_table;


   /*
    *  First allocate memory for the temporary
    *  log that will hold output information
    *  until the log file has been opened,
    *  assuming this option has been chosen at
    *  run time (which we won't know until
    *  after parsing of the command line and
    *  the namelist file).
    */
   temp_log = malloc(TEMP_LOG_SIZE * sizeof(char) );

   if (temp_log == NULL)  {
      printf("\n* * * E R R O R:  Allocation of memory for");
      printf(" temporary log failed - ABORTING.\n");

      exit(FAILURE_EXIT);
   }

   /*** DEBUG ***/  strcpy(fpg_version, "1.25");

   /*
    *  Print program name & version number + date & time info.
    */
   printf("\nFPGen:  Final Product Generator, version %s, executed on ",
          fpg_version);

   strcpy(temp_log, "\nFPGen:  Final Product Generator, version ");
   strcat(temp_log, fpg_version);
   strcat(temp_log, ", executed on ");

   rcode = print_current_date_and_time();

   if (rcode != SUCCESS)  {
      printf("\n* * * W A R N I N G:  System time unknown -");
      printf(" date and time set to zero.\n");

      strcat(temp_log, "\n* * * W A R N I N G:  System time unknown -");
      strcat(temp_log, " date and time set to zero.\n");

      /*
       *  Set exit code to WARNING_EXIT value.
       */
      exit_code = WARNING_EXIT;
   }

   /*
    *  Initialize all file names to "NONE"
    *  indicating that no file names have
    *  been found yet.
    */
   strcpy(input_file_name, "NONE");
   strcpy(log_file_name, "NONE");
   strcpy(namelist_file_name, "NONE");
   strcpy(output_file_name, "NONE");
   strcpy(sodb_entries_file_name, "NONE");
   strcpy(template_file_name, "NONE");

   /*
    *  Check the number of arguments on the command
    *  line.  If the number is between 7 and 19,
    *  parse the command line.  If not, print
    *  command line format information and exit.
    */
   if ( (argc > 6) && (argc < 20) )  {
      rcode = parse_cmdln(argc,
                          argv,
                          &input_file_name[0],
                          &log_file_name[0],
                          &namelist_file_name[0],
                          &output_file_name[0],
                          &sodb_entries_file_name[0],
                          &template_file_name[0],
                          &hist_copy,
                          &blank_blocks,
                          &mask);

      if (rcode == FAILURE)  {
         printf("\n* * * E R R O R:  Parsing of command line failed");
         printf(" - ABORTING.\n");

         /*
          *  Print command line format information
          *  since parsing of command line failed.
          */
         print_cmdln_info();

         exit(FAILURE_EXIT);
      }

      else  if (rcode == WARNING)  {
         /*
          *  Set exit code to WARNING_EXIT value.
          */
         exit_code = WARNING_EXIT;
      }

   }  /*  End of 'if ( (argc > 6) && (argc < 18) )' segment.  */

   else  {
      print_cmdln_info();

      exit(FAILURE_EXIT);
   }

   /*
    *  Read namelist information (if the 'n' option
    *  was included on the command line).
    */
   if (strncmp(namelist_file_name, "NONE", 4) != 0)  {
      rcode = read_namelist(&namelist_file_name[0],
                            &input_file_name[0],
                            &log_file_name[0],
                            &output_file_name[0],
                            &sodb_entries_file_name[0],
                            &template_file_name[0],
                            &hist_copy,
                            &blank_blocks,
                            &mask);

      if (rcode == FAILURE)  {
         printf("\n* * * E R R O R:  Reading of namelist");
         printf(" information failed - ABORTING.\n");

         exit(FAILURE_EXIT);
      }
      else  if (rcode == WARNING)  {
         /*
          *  Set exit code to WARNING_EXIT value.
          */
         exit_code = WARNING_EXIT;
      }

   }  /*  End of 'if (strncmp(namelist_file_name, "NONE", 4) != 0)' segment.  */

   /*
    *  Check if the log file option was included
    *  either on the command line or in the
    *  namelist file.  If so, open the log file
    *  and copy the information from temp_log to
    *  this file.
    */
   if (strcmp(log_file_name, "NONE") != 0)  {
      log_stream = freopen(log_file_name, "ab", stdout);

      if (log_stream == NULL)  {
         printf("\n* * * E R R O R:  Failed to open log file %s\n",
                log_file_name);

         exit(FAILURE_EXIT);
      }

      fprintf(log_stream, temp_log);
   }

   /*
    *  Make sure all the required information
    *  was given either on the command line
    *  or in the namelist file.
    */
   if (strncmp(input_file_name, "NONE", 4) == 0)  {
      printf("\n* * * E R R O R:  Name of input DCE file");
      printf(" not found (required).\n");

      exit_code = FAILURE_EXIT;
   }

   if (strncmp(output_file_name, "NONE", 4) == 0)  {
      printf("\n* * * E R R O R:  Name of output DCE file");
      printf(" not found (required).\n");

      exit_code = FAILURE_EXIT;
   }

   if (strncmp(template_file_name, "NONE", 4) == 0)  {
      printf("\n* * * E R R O R:  Name of template file");
      printf(" not found (required).\n");

      exit_code = FAILURE_EXIT;
   }

   if (exit_code == FAILURE_EXIT)  {
      printf("\n                  ABORTING.\n");

      exit(exit_code);
   }

   /*
    *  Check if a SODB entries file has
    *  been specified on the command
    *  line or in the namelist file.
    */
   if (strncmp(sodb_entries_file_name, "NONE", 4) == 0)  {
      printf("\n* * * W A R N I N G:  SODB entries file not found");
      printf(" - DATABASE keywords");
      printf("\n                      will NOT be output.\n");

      exit_code = WARNING_EXIT;
   }
   else  {
      /*
       *  A SODB entries files has been
       *  given, so we read it.
       */
      sodb_table = read_sodb_entries_file(&sodb_entries_file_name[0],
                                          &line_count);

      if (sodb_table == NULL)  {
         printf("\n* * * E R R O R:  Denied access to");
         printf(" SODB entries file - ABORTING.\n");

         exit(FAILURE_EXIT);
      }
   }

   /*
    *  Check if the history copy option was
    *  included on either the command line
    *  or in the namelist file.  If not,
    *  set hist_copy to FALSE.
    */
   if (hist_copy == NOT_SET)  {
      hist_copy = FALSE;
   }

   /*
    *  Check if the remove blank blocks option
    *  was included on either the command line
    *  or in the namelist file.  If not, set
    *  blank_blocks to FALSE.
    */
   if (blank_blocks == NOT_SET)  {
      blank_blocks = FALSE;
   }

   /*
    *  Generate final product.
    */
   rcode = generate_final_product(&input_file_name[0],
                                  &output_file_name[0],
                                  &template_file_name[0],
                                  sodb_table,
                                  line_count,
                                  hist_copy,
                                  blank_blocks);

   if (rcode == FAILURE)  {
      printf("\n* * * E R R O R:  FPGen failed - ABORTING.\n");

      exit(FAILURE_EXIT);
   }
   else  if (rcode == WARNING)  {
      exit_code = WARNING_EXIT;
   }
   else  if ( (rcode == NO_ACTION) || (rcode == NO_VALUE) )  {
      exit_code = FAILURE_EXIT;
   }

   /*
    *  Processing completed, so write sign-
    *  off message (incl. completion date &
    *  time info.) and terminate.
    */
   printf("\nFPGen processing completed on ");

   rcode = print_current_date_and_time();

   if (rcode != SUCCESS)  {
      printf("\n* * * W A R N I N G:  System time unknown -");
      printf(" date and time set to zero.\n");

      /*
       *  Set exit code to WARNING_EXIT value.
       */
      exit_code = WARNING_EXIT;
   }
   else  {
      printf("\n");
   }

   /*
    *  Free the temp_log memory.
    */
   free(temp_log);

   exit(exit_code);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  print_current_date_and_time                       *** *
 *                                                                       *
 *                                                                       *
 * ***  This function gets the current system time, converts it to   *** *
 * ***  local date and time, and prints this information.  If the    *** *
 * ***  system time can't be acquired, zeros are printed.            *** *
 *                                                                       *
 * ***  Input:  None.                                                *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - system time has been     *** *
 * ***                                      successfully acquired,   *** *
 * ***                                      converted, and printed.  *** *
 * ***                  FAILURE           - no system time was       *** *
 * ***                                      available (zeros         *** *
 * ***                                      printed).                *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  print_current_date_and_time()

{

   char
         help_str[SHORT_TEMP_STR_LEN];

   struct  tm
               *sys_time;

   time_t
           t_time;


   /*
    *  Get the current system time.
    */
   t_time = time(NULL);

   if (t_time != -1)  {
      /*
       *  Convert the current system time into local time.
       */
      sys_time = localtime(&t_time);

      if (log_stream == NULL)  {
         sprintf(help_str, "%.2d/%.2d/%.2d at %.2d:%.2d:%.2d.\n",
                 (sys_time->tm_mon + 1), sys_time->tm_mday,
                 (sys_time->tm_year - 100), sys_time->tm_hour,
                 sys_time->tm_min, sys_time->tm_sec);

         strcat(temp_log, help_str);
      }
      printf("%.2d/%.2d/%.2d at %.2d:%.2d:%.2d.\n",
             (sys_time->tm_mon + 1), sys_time->tm_mday,
             (sys_time->tm_year - 100), sys_time->tm_hour,
             sys_time->tm_min, sys_time->tm_sec);

      return(SUCCESS);
   }
   else  {
      /*
       *  No system time available - print
       *  zeros for date & time.
       */
      if (log_stream == NULL)  {
         strcat(temp_log, "00/00/00 at 00:00:00.\n");
      }
      printf("00/00/00 at 00:00:00.\n");

      return(FAILURE);
   }

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  parse_cmdln                                       *** *
 *                                                                       *
 *                                                                       *
 * ***  This function parses the command line and in the process     *** *
 * ***  performs certain verifications of the parsed information's   *** *
 * ***  format and contents.                                         *** *
 *                                                                       *
 * ***  Input:  argc                      - number of command line   *** *
 * ***                                      arguments.               *** *
 * ***          argv                      - array containing the     *** *
 * ***                                      command line arguments.  *** *
 *                                                                       *
 * ***  Output:  input_file_name          - name of the input DCE    *** *
 * ***                                      file.                    *** *
 * ***           log_file_name            - name of the output log   *** *
 * ***                                      file.                    *** *
 * ***           namelist_file_name       - name of the namelist     *** *
 * ***                                      file.                    *** *
 * ***           output_file_name         - name of the output DCE   *** *
 * ***                                      file.                    *** *
 * ***           sodb_entries_file_name   - name of the SODB         *** *
 * ***                                      entries file.            *** *
 * ***           template_file_name       - name of the template     *** *
 * ***                                      file.                    *** *
 * ***           hist_copy                - flag indicating if       *** *
 * ***                                      HISTORY records should   *** *
 * ***                                      be copied from input to  *** *
 * ***                                      output DCE file.         *** *
 * ***           blank_blocks             - flag indicating if       *** *
 * ***                                      blank blocks should be   *** *
 * ***                                      removed from the output  *** *
 * ***                                      DCE file.                *** *
 * ***           mask                     - mask (NOT IMPLEMENTED).  *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  WARNING           - successful processing    *** *
 * ***                                      completed (but a         *** *
 * ***                                      WARNING message has      *** *
 * ***                                      been issued,indicating   *** *
 * ***                                      minor - but not fatal -  *** *
 * ***                                      error or problem).       *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  parse_cmdln(int     argc,
                 char**  argv,
                 char*   input_file_name,
                 char*   log_file_name,
                 char*   namelist_file_name,
                 char*   output_file_name,
                 char*   sodb_entries_file_name,
                 char*   template_file_name,
                 int*    hist_copy,
                 int*    blank_blocks,
                 int*    mask)

{

   char
         argv_char,
       **argv_help,
         help_str[SHORT_TEMP_STR_LEN];

   int
        arg_count   = 1,
        bb_help,
        hist_help,
        i,
        opt         = 0,
        qm_opt      = 0,
        return_code = SUCCESS,
        vm_opt      = 0;


   /*
    *  First print out the contents of
    *  the command line for this run.
    */
   argv_help = argv;

   printf("\nContents of the command line (for this run):\n");

   strcat(temp_log, "\nContents of the command line (for this run):\n");

   for (i = 0; i < argc; i++)  {
      strcat(temp_log, *argv_help);
      strcat(temp_log, " ");

      printf("%s ", *argv_help++);
   }

   printf("\n");
   strcat(temp_log, "\n");

   /*
    *  Make sure the number of arguments on
    *  the command line is an odd number.
    */
   if ( (2 * (argc/2) ) == argc)  {
      printf("\n* * * E R R O R:  Unexpected number of command line");
      printf(" arguments:  %d", argc);
      printf("\n                  (should be an odd number between 7-19)\n");

      return(FAILURE);
   }

   /*
    *  Parse the command line.
    */
   while (arg_count < argc)  {
      /*
       *  Skip to the next command line argument.
       */
      *argv++;
      arg_count++;

      /*
       *  Make sure the option begins with a hyphen.  If
       *  so, copy the option letter to 'argv_char' (since
       *  the switch statement only operates on integers
       *  or single characters).  If not, print error
       *  message and return.
       */
      if (!strncmp(*argv, "-", 1) )  {
         argv_char = *(*argv + 1);
      }
      else  {
         printf("\n* * * E R R O R:  Undefined option (missing hyphen): %s\n",
                *argv);

         return(FAILURE);
      }

      /*
       *  Branch to the proper code for each option.
       */
      switch (argv_char)  {

         /*
          *  Option '-b' indicates whether or not
          *  blank header blocks should be
          *  removed from the output DCE file.
          */
         case 'b':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (*blank_blocks == NOT_SET)  {
               bb_help = atoi(*argv);

               /*
                *  Make sure the value is legal.
                */
               if ( (bb_help == TRUE) || (bb_help == FALSE) )  {
                  *blank_blocks = bb_help;
               }
               else  {
                  printf("\n* * * W A R N I N G:  Undefined remove blank");
                  printf(" blocks value: %d (should", bb_help);
                  printf("\n                      be 0 or 1; ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Undefined");
                  strcat(temp_log, " remove blank blocks value: ");
                  sprintf(help_str, "%d", bb_help);
                  strcat(temp_log, help_str);
                  strcat(temp_log, " (should");
                  strcat(temp_log, "\n                      be 0 or 1;");
                  strcat(temp_log, " ignored).\n");

                  return_code = WARNING;
               }
            }
            else  {
               printf("\n* * * W A R N I N G:  Remove blank blocks option");
               printf(" specified multiple");
               printf("\n                      times on command line");
               printf(" (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Remove blank blocks");
               strcat(temp_log, " option specified multiple");
               strcat(temp_log, "\n                      times on");
               strcat(temp_log, " command line (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-d' is followed by the SODB
          *  entries file name.
          */
         case 'd':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(sodb_entries_file_name, "NONE", 4) == 0)  {
               strcpy(sodb_entries_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  SODB entries file name");
               printf(" specified multiple times");
               printf("\n                      on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  SODB entries file");
               strcat(temp_log, " name specified multiple times");
               strcat(temp_log, "\n                      on command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-h' indicates whether or not
          *  HISTORY records should be copied
          *  from input to output DCE file.
          */
         case 'h':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (*hist_copy == NOT_SET)  {
               hist_help = atoi(*argv);

               /*
                *  Make sure the value is legal.
                */
               if ( (hist_help == TRUE) || (hist_help == FALSE) )  {
                  *hist_copy = hist_help;
               }
               else  {
                  printf("\n* * * W A R N I N G:  Undefined HISTORY copy");
                  printf(" value: %d (should be", hist_help);
                  printf("\n                      0 or 1; ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Undefined");
                  strcat(temp_log, " HISTORY copy value: ");
                  sprintf(help_str, "%d", hist_help);
                  strcat(temp_log, help_str);
                  strcat(temp_log, " (should be");
                  strcat(temp_log, "\n                      0 or 1;");
                  strcat(temp_log, " ignored).\n");

                  return_code = WARNING;
               }
            }
            else  {
               printf("\n* * * W A R N I N G:  HISTORY copy option specified");
               printf(" multiple times");
               printf("\n                      on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  HISTORY copy");
               strcat(temp_log, " option specified multiple times");
               strcat(temp_log, "\n                      on command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-i' is followed by the input file name.
          */
         case 'i':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(input_file_name, "NONE", 4) == 0)  {
               strcpy(input_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  Input file name specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Input file name");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-l' is followed by the log file name.
          */
         case 'l':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(log_file_name, "NONE", 4) == 0)  {
               strcpy(log_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  Log file name specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Log file name");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-m' is followed by a mask.
          *  N.B.  This option has not been implemented
          *        yet, so the following is basically a
          *        place holder.
          */
          case 'm':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if ( (opt%10) == 0)  {
               *mask = atoi(*argv);

               /*
                *  Increment opt by 1 in case we
                *  need to know whether or not
                *  this option was included.
                */
               opt += 1;

               printf("\n* * * E R R O R:  Mask option not implemented.\n");

               return(FAILURE);
            }
            else  {
               printf("\n* * * W A R N I N G:  Too many mask");
               printf(" values specified on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Too many mask");
               strcat(temp_log, " values specified");
               strcat(temp_log, " on command line (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-n' is followed by the namelist file name.
          */
         case 'n':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(namelist_file_name, "NONE", 4) == 0)  {
               strcpy(namelist_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  Namelist file name specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Namelist file name");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-o' is followed by the output file name.
          */
         case 'o':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(output_file_name, "NONE", 4) == 0)  {
               strcpy(output_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  Output file name specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Output file name");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-q' indicates quiet mode and should be
          *  followed by either a 1 (indicating quiet mode
          *  ON) or a 0 (indicating quiet mode OFF).
          */
         case 'q':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (qm_opt == 0)  {
               quiet_mode = atoi(*argv);

               /*
                *  Make sure the quiet mode value
                *  is equal to either 1 or 0.
                */
               if ( (quiet_mode != 1) && (quiet_mode != 0) )  {
                  printf("\n* * * W A R N I N G:  Unknown quiet mode level:");
                  printf(" %d (should be 1 or 0; ignored).\n", quiet_mode);

                  strcat(temp_log, "\n* * * W A R N I N G:  Unknown");
                  strcat(temp_log, " quiet mode level:");
                  strcat(temp_log, *argv);
                  strcat(temp_log, " (should be 1 or 0; ignored).\n");

                  quiet_mode = 0;

                  return_code = WARNING;
               }

               /*
                *  Set qm_opt to one, indicating
                *  quiet mode has been set.
                */
               qm_opt = 1;

            }  /*  End of 'if (qm_opt == 0)' segment.  */

            else  {
               printf("\n* * * W A R N I N G:  Quiet mode specified");
               printf(" multiple times on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Quiet mode");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, " command line (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-t' is followed by the template file name.
          */
         case 't':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (strncmp(template_file_name, "NONE", 4) == 0)  {
               strcpy(template_file_name, *argv);
            }
            else  {
               printf("\n* * * W A R N I N G:  Template file name specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Template file name");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  Option '-v' indicates verbose mode and
          *  should be followed by the level value,
          *  which should be either 1 (for major
          *  steps) or 2 (for more detailed output).
          */
         case 'v':
            *argv++;
            arg_count++;

            /*
             *  Make sure this isn't a multiple entry.
             */
            if (vm_opt == 0)  {
               verbose_mode = atoi(*argv);

               /*
                *  Make sure the verbosity level
                *  value is equal to either 1 or 2.
                */
               if ( (verbose_mode != 1) && (verbose_mode != 2) )  {
                  printf("\n* * * W A R N I N G:  Unknown verbosity level:");
                  printf(" %d (should be 1 or 2; ignored).\n", verbose_mode);

                  strcat(temp_log, "\n* * * W A R N I N G:  Unknown");
                  strcat(temp_log, " verbosity level:");
                  strcat(temp_log, *argv);
                  strcat(temp_log, " (should be 1 or 2; ignored).\n");

                  verbose_mode = 0;

                  return_code = WARNING;
               }

               /*
                *  Set vm_opt to one, indicating
                *  verbose mode has been set.
                */
               vm_opt = 1;

            }  /*  End of 'if (vm_opt == 0)' segment.  */

            else  {
               printf("\n* * * W A R N I N G:  Verbose mode specified");
               printf(" multiple times on");
               printf("\n                      command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Verbose mode");
               strcat(temp_log, " specified multiple times on");
               strcat(temp_log, "\n                      command line");
               strcat(temp_log, " (ignored).\n");

               return_code = WARNING;
            }

            break;

         /*
          *  We should only end up here for undefined options.
          */
         default:
            printf("\n* * * E R R O R:  Unknown option: %s\n", *argv);

            return(FAILURE);

      }  /*  End of 'switch (argv_char)' segment.  */

   }  /*  End of 'while (arg_count < argc)' loop.  */

   /*
    *  Parsing of the command line has been completed.
    *  Check if the global variable verbose_mode has
    *  been set, in which case we print what we found
    *  while parsing the command line.
    */
   if (verbose_mode > 0)  {
      printf("\n\nThe following options were found:\n");

      strcat(temp_log, "\n\nThe following options were found:\n");

      if (*blank_blocks != NOT_SET)  {
         printf("\n   -b: ");

         strcat(temp_log, "\n   -b: ");
        
         if (*blank_blocks == FALSE)  {
            printf(" Do NOT remove");

            strcat(temp_log, " Do NOT remove");
         }
         else  {
            printf(" Remove");

            strcat(temp_log, " Remove");
         }

         printf(" blank blocks from output file\n");

         strcat(temp_log, " blank blocks from output file\n");
      }

      if (strncmp(sodb_entries_file_name, "NONE", 4) != 0)  {
         printf("\n   -d:  Read the following SODB entries file:\n\t   %s\n",
                sodb_entries_file_name);

         strcat(temp_log, "\n   -d:  Read the following SODB");
         strcat(temp_log, " entries file:\n\t   ");
         strcat(temp_log, sodb_entries_file_name);
         strcat(temp_log, "\n");
      }

      if (*hist_copy != NOT_SET)  {
         printf("\n   -h: ");

         strcat(temp_log, "\n   -h: ");
        
         if (*hist_copy == FALSE)  {
            printf(" Do NOT copy");

            strcat(temp_log, " Do NOT copy");
         }
         else  {
            printf(" Copy");

            strcat(temp_log, " Copy");
         }

         printf(" HISTORY lines from input to output file\n");

         strcat(temp_log, " HISTORY lines from input to output file\n");
      }

      if (strncmp(input_file_name, "NONE", 4) != 0)  {
         printf("\n   -i:  Read the following input DCE file:\n\t   %s\n",
                input_file_name);

         strcat(temp_log, "\n   -i:  Read the following");
         strcat(temp_log, " input DCE file:\n\t   ");
         strcat(temp_log, input_file_name);
         strcat(temp_log, "\n");
      }

      if (strcmp(log_file_name, "NONE") != 0)  {
         printf("\n   -l:  Write to the following log file:\n\t   %s\n",
                log_file_name);

         strcat(temp_log, "\n   -l:  Write to the following");
         strcat(temp_log, " log file:\n\t   ");
         strcat(temp_log, log_file_name);
         strcat(temp_log, "\n");
      }

      if (opt%10 != 0)  {
         printf("\n   -m:  Set mask to %d\n", *mask);

         strcat(temp_log, "\n   -m:  Set mask to ");
         sprintf(help_str, "%d\n", *mask);
         strcat(temp_log, help_str);
      }

      if (strncmp(namelist_file_name, "NONE", 4) != 0)  {
         printf("\n   -n:  Read the following namelist file:\n\t   %s\n",
                namelist_file_name);

         strcat(temp_log, "\n   -n:  Read the following");
         strcat(temp_log, " namelist file:\n\t   ");
         strcat(temp_log, namelist_file_name);
         strcat(temp_log, "\n");
      }

      if (strncmp(output_file_name, "NONE", 4) != 0)  {
         printf("\n   -o:  Write the following output DCE file:\n\t   %s\n",
                output_file_name);

         strcat(temp_log, "\n   -o:  Write the following");
         strcat(temp_log, " output DCE file:\n\t   ");
         strcat(temp_log, output_file_name);
         strcat(temp_log, "\n");
      }

      if (qm_opt == 1)  {
         printf("\n   -q:  Quiet mode turned ");
         strcat(temp_log, "\n   -q:  Quiet mode turned ");

         if (quiet_mode == 1)  {
            printf("ON\n");
            strcat(temp_log, "ON\n");
         }
         else  {
            printf("OFF\n");
            strcat(temp_log, "OFF\n");
         }
      }

      if (strncmp(template_file_name, "NONE", 4) != 0)  {
         printf("\n   -t:  Read the following template file:\n\t   %s\n",
                template_file_name);

         strcat(temp_log, "\n   -t:  Read the following");
         strcat(temp_log, " template file:\n\t   ");
         strcat(temp_log, template_file_name);
         strcat(temp_log, "\n");
      }

      printf("\n   -v:  Verbose mode set to ON (level %d)\n\n", verbose_mode);

      strcat(temp_log, "\n   -v:  Verbose mode set to ON (level ");
      sprintf(help_str, "%d)\n\n", verbose_mode);
      strcat(temp_log, help_str);

   }  /*  End of 'if (verbose_mode > 0)' segment.  */

   return(return_code);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  print_cmdln_info                                  *** *
 *                                                                       *
 *                                                                       *
 * ***  This function prints the command line input format.          *** *
 *                                                                       *
 * ***  Input:  None.                                                *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  None.                                        *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

void  print_cmdln_info()

{


   /*
    *  Print command line format and contents information.
    */
   printf("\n\nusage:  fpgen <option flags> <specifications>\n");

   printf("\n        where <option flags> may be:\n");
   printf("\n           -b, -d, -h, -i, -l, -m, -n, -o, -q, -t, or -v,\n");

   printf("\n        any one of which, if specified, must be followed by a");
   printf("\n        blank and a corresponding specification, as follows.\n");

   printf("\n\t-b bb_remove       remove blank blocks from output DCE file");
   printf("\n\t                   (if set to 1, do not remove if set to 0);");
   printf(" optional,");
   printf("\n\t                   (DEFAULT: do NOT remove blank blocks),\n");

   printf("\n\t-d SODB_file       name of SODB entries input file;");
   printf(" optional,");
   printf("\n\t                   (DEFAULT: no SODB entries file used).\n");

   printf("\n\t-h history_copy    copy HISTORY lines from input");
   printf(" to output DCE file");
   printf("\n\t                   (if set to 1, do not copy if");
   printf(" set to 0); optional,");
   printf("\n\t                   (DEFAULT: do NOT copy),\n");

   printf("\n\t-i input_DCE       name of input DCE file;");
   printf(" required (on command");
   printf("\n\t                   line or in namelist file),\n");

   printf("\n\t-l log_file        name of output log file; optional,");
   printf("\n\t                   (DEFAULT: no log file used),\n");

   printf("\n\t-m mask            treat input as a mask file");
   printf(" and strip header;");
   printf("\n\t                   CURRENTLY NOT SUPPORTED,\n");

   printf("\n\t-n namelist        name of namelist file; optional,");
   printf("\n\t                   (DEFAULT: no namelist file used),\n");

   printf("\n\t-o output_DCE      name of output DCE file;");
   printf(" required (on command");
   printf("\n\t                   line or in namelist file),\n");

   printf("\n\t-q value           quiet mode ON (value = 1) or");
   printf(" OFF (value = 0;");
   printf("\n\t                   DEFAULT: quiet mode turned OFF),\n");

   printf("\n\t-t template_file   name of template file;");
   printf(" required (on command");
   printf("\n\t                   line or in namelist file),\n");

   printf("\n\t-v verbose_level   verbose mode - verbose_level = 1");
   printf(" for major steps,");
   printf("\n\t                   verbose_level = 2 for more detailed output;");
   printf("\n\t                   optional, (DEFAULT: verbose mode");
   printf(" turned OFF).\n");

   printf("\n        The option flags (with corresponding specifications)");
   printf(" may be given\n        in arbitrary order.\n\n");

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  read_namelist                                     *** *
 *                                                                       *
 *                                                                       *
 * ***  This function reads namelist input data.  This function is   *** *
 * ***  adapted from Ted Hesselroth's fitsop_parse_namelist          *** *
 * ***  function.                                                    *** *
 *                                                                       *
 * ***  Input:  namelist_file_name        - name of the namelist     *** *
 * ***                                      file.                    *** *
 *                                                                       *
 * ***  Output:  input_file_name          - name of the input DCE    *** *
 * ***                                      file.                    *** *
 * ***           log_file_name            - name of the output log   *** *
 * ***                                      file.                    *** *
 * ***           output_file_name         - name of the output DCE   *** *
 * ***                                      file.                    *** *
 * ***           sodb_entries_file_name   - name of the SODB         *** *
 * ***                                      entries file.            *** *
 * ***           template_file_name       - name of the template     *** *
 * ***                                      file.                    *** *
 * ***           hist_copy                - flag indicating if       *** *
 * ***                                      HISTORY records should   *** *
 * ***                                      be copied from input to  *** *
 * ***                                      output DCE file.         *** *
 * ***           blank_blocks             - flag indicating if       *** *
 * ***                                      blank blocks should be   *** *
 * ***                                      removed from the output  *** *
 * ***                                      DCE file.                *** *
 * ***           mask                     - mask (NOT IMPLEMENTED).  *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  WARNING           - successful processing    *** *
 * ***                                      completed (but a         *** *
 * ***                                      WARNING message has      *** *
 * ***                                      been issued,indicating   *** *
 * ***                                      minor - but not fatal -  *** *
 * ***                                      error or problem).       *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Ted Hesselroth, modified for SIRTF  *** *
 * ***  Final Product Generator application by Mikael Aronsson.      *** *
 *                                                                       *
 * ********************************************************************* */

int  read_namelist(char*  namelist_file_name,
                   char*  input_file_name,
                   char*  log_file_name,
                   char*  output_file_name,
                   char*  sodb_entries_file_name,
                   char*  template_file_name,
                   int*   hist_copy,
                   int*   blank_blocks,
                   int*   mask)

{

   char
         *buffer_ptr,
         *comma_token,
          comment[MAX_CMNT_LEN],
          equal_sign[SHORT_TEMP_STR_LEN],
          help_str[SHORT_TEMP_STR_LEN],
          keyword[LONG_TEMP_STR_LEN],
         *nmlst_buffer,
         *nmlst_out_buf,
          value[LONG_TEMP_STR_LEN];

   FILE
         *nmlst_file;

   int
        nmlst_file_descr         = 0,
        nmlst_infile_count       = 0,
        nmlst_logfile_count      = 0,
        nmlst_outfile_count      = 0,
        nmlst_past_end           = 0,
        nmlst_print              = 0,
        nmlst_section_found      = 0,
        nmlst_strlen             = 0,
        nmlst_sodbfile_count     = 0,
        nmlst_templatefile_count = 0,
        rcode,
        return_code              = SUCCESS,
        token_length;

   struct  stat
                 stat_buffer,
                *stat_buffer_ptr;


   if (verbose_mode > 0)  {
      printf("\nRead namelist information.\n");

      strcat(temp_log, "\nRead namelist information.\n");
   }

   /*
    *  Open the namelist file in order to get
    *  its length (and eventually read it).
    */
   nmlst_file_descr = open(namelist_file_name, O_RDONLY);

   if (nmlst_file_descr == -1)  {
      printf("\n* * * E R R O R:  Opening of namelist file %s failed.\n",
             namelist_file_name);

      return(FAILURE);
   }

   /*
    *  Determine the number of characters
    *  in the namelist file.
    */
   stat_buffer_ptr = &stat_buffer;

   fstat(nmlst_file_descr, stat_buffer_ptr);

   nmlst_strlen = stat_buffer_ptr->st_size;

   /*
    *  Associate a stream with the namelist
    *  file descriptor generated above.
    */
   nmlst_file = fdopen(nmlst_file_descr, "r");

   /*
    *  Allocate memory for the namelist buffer.
    */
   nmlst_buffer  = (char*) calloc( (nmlst_strlen + 2), sizeof(char) );

   if (nmlst_buffer)  {
      /*
       *  Read the namelist data.
       */
      rcode = fread(nmlst_buffer, sizeof(char), nmlst_strlen, nmlst_file);

      if (rcode == 0)  {
         printf("\n* * * E R R O R:  Reading of namelist file %s failed.\n",
                namelist_file_name);

         return(FAILURE);
      }

      /*
       *  Close the stream and the file descriptor.
       */
      fclose(nmlst_file);
      close(nmlst_file_descr);

      /*
       *  Save a copy of the namelist, in
       *  case user wants it printed out.
       */
      nmlst_out_buf = (char*) calloc( (nmlst_strlen + 2), sizeof(char) );

      if (nmlst_out_buf)  {
         strcpy(nmlst_out_buf, nmlst_buffer);
      }

      /*
       *  Look for the "&FINALPRODGENIN" section.
       */
      buffer_ptr = nmlst_buffer;

      token_length = strlen("&FINALPRODGENIN");

      while ( (buffer_ptr <= (nmlst_buffer + nmlst_strlen) ) &&
              !(nmlst_past_end) && !(nmlst_section_found) )  {

         buffer_ptr = strchr(buffer_ptr, '&');

         /*
          *  Check that we don't move past the
          *  end of the namelist buffer.
          */
         if ( (buffer_ptr + token_length) >= (buffer_ptr + nmlst_strlen) )  {
            /*
             *  We end up here if we're past the
             *  end, so we set the past end flag.
             */
            nmlst_past_end = 1;
         }
         else  {
            /*
             *  We end up here if there is still more to go
             *  of the namelist buffer, so let's check if
             *  we have found the desired section.
             */
            if (strncmp("&FINALPRODGENIN", buffer_ptr, token_length) )  {
               /*
                *  Desired section NOT found,
                *  so we move forward.
                */
               buffer_ptr += strlen("&");
            }
            else  {
               /*
                *  Desired section was found, so
                *  we set section found flag.
                */
               nmlst_section_found = 1;
            }

         }

      }  /*  End of 'while ( (buffer_ptr <= (nmlst_buffer + nmlst_strlen) ) &&*/
         /*          !(nmlst_past_end) && !(nmlst_section_found) )' loop.     */

      /*
       *  Make sure we've found what
       *  we were looking for.
       */
      if (!nmlst_section_found)  {
         printf("\n* * * E R R O R:  Failed to find desired section");
         printf(" of namelist %s\n", namelist_file_name);

         return(FAILURE);
      }

      buffer_ptr += token_length;

      /*
       *  Find and parse each "keyword = value"
       *  part of the namelist.
       */
      for (comma_token = strtok(buffer_ptr, ",");
           (comma_token != NULL) && (strchr(comma_token, '&') == 0);
           comma_token = strtok(NULL, ",") )  {

         sscanf(comma_token, "%s%s%s", keyword, equal_sign, value);

         /*
          *  If this is a comment part, we
          *  place the comment in the local
          *  parameter of the same name.
          */
         if (strcmp(keyword, "Comment") == 0)  {
            strcpy(comment, value);
         }

         /*
          *  Parse the value field into the
          *  appropriate namelist parameter.
          */
         else  if (strcmp(keyword, "Input_File") == 0)  {
            /*
             *  Make sure we haven't already found an
             *  input file name on the command line
             *  or multiple input file name entries
             *  in the namelist file.
             */
            if (strncmp(input_file_name, "NONE", 4) == 0)  {
               sscanf(&value[1], "%[^']", input_file_name);

               nmlst_infile_count++;
            }
            else  {
               if (nmlst_infile_count == 0)  {
                  printf("\n* * * W A R N I N G:  Input file");
                  printf(" name specified both on command line");
                  printf("\n                      and in");
                  printf(" namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Input file");
                  strcat(temp_log, " name specified both on command line");
                  strcat(temp_log, "\n                      and in");
                  strcat(temp_log, " namelist file (ignored).\n");
               }
               else  {
                  printf("\n* * * W A R N I N G:  Input file");
                  printf(" name specified multiple times in");
                  printf("\n                      namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Input file");
                  strcat(temp_log, " name specified multiple times in");
                  strcat(temp_log, "\n                      namelist");
                  strcat(temp_log, " file (ignored).\n");
               }

               return_code = WARNING;
            }

         }  /*  End of                                                    */
            /*  'else  if (strcmp(keyword, "Input_File") == 0)' segment.  */

         else  if (strcmp(keyword, "Output_File") == 0)  {
            /*
             *  Make sure we haven't already found an
             *  output file name on the command line
             *  or multiple output file name entries
             *  in the namelist file.
             */
            if (strncmp(output_file_name, "NONE", 4) == 0)  {
               sscanf(&value[1], "%[^']", output_file_name);

               nmlst_outfile_count++;
            }
            else  {
               if (nmlst_outfile_count == 0)  {
                  printf("\n* * * W A R N I N G:  Output file");
                  printf(" name specified both on command line");
                  printf("\n                      and in");
                  printf(" namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Output file");
                  strcat(temp_log, " name specified both on command line");
                  strcat(temp_log, "\n                      and in");
                  strcat(temp_log, " namelist file (ignored).\n");
               }
               else  {
                  printf("\n* * * W A R N I N G:  Output file");
                  printf(" name specified multiple times in");
                  printf("\n                      namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Output file");
                  strcat(temp_log, " name specified multiple times in");
                  strcat(temp_log, "\n                      namelist");
                  strcat(temp_log, " file (ignored).\n");
               }

               return_code = WARNING;
            }

         }  /*  End of                                                     */
            /*  'else  if (strcmp(keyword, "Output_File") == 0)' segment.  */

         else  if (strcmp(keyword, "SODB_Entries_File") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  SODB entries file name on the command
             *  line or multiple SODB entries file
             *  name entries in the namelist file.
             */
            if (strncmp(sodb_entries_file_name, "NONE", 4) == 0)  {
               sscanf(&value[1], "%[^']", sodb_entries_file_name);

               nmlst_sodbfile_count++;
            }
            else  {
               if (nmlst_sodbfile_count == 0)  {
                  printf("\n* * * W A R N I N G:  SODB entries file");
                  printf(" name specified both on command line");
                  printf("\n                      and in");
                  printf(" namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  SODB entries file");
                  strcat(temp_log, " name specified both on command line");
                  strcat(temp_log, "\n                      and in");
                  strcat(temp_log, " namelist file (ignored).\n");
               }
               else  {
                  printf("\n* * * W A R N I N G:  SODB entries file");
                  printf(" name specified multiple times in");
                  printf("\n                      namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  SODB entries file");
                  strcat(temp_log, " name specified multiple times in");
                  strcat(temp_log, "\n                      namelist");
                  strcat(temp_log, " file (ignored).\n");
               }

               return_code = WARNING;
            }

         }  /*  End of                                                  */
            /*  'else  if (strcmp(keyword, "SODB_Entries_File") == 0)'  */
            /*  segment.                                                */

         else  if (strcmp(keyword, "Template_File") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  template file name on the command
             *  line or multiple template file name
             *  entries in the namelist file.
             */
            if (strncmp(template_file_name, "NONE", 4) == 0)  {
               sscanf(&value[1], "%[^']", template_file_name);

               nmlst_templatefile_count++;
            }
            else  {
               if (nmlst_templatefile_count == 0)  {
                  printf("\n* * * W A R N I N G:  Template file");
                  printf(" name specified both on command line");
                  printf("\n                      and in");
                  printf(" namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Template file");
                  strcat(temp_log, " name specified both on command line");
                  strcat(temp_log, "\n                      and in");
                  strcat(temp_log, " namelist file (ignored).\n");
               }
               else  {
                  printf("\n* * * W A R N I N G:  Template file");
                  printf(" name specified multiple times in");
                  printf("\n                      namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Template file");
                  strcat(temp_log, " name specified multiple times in");
                  strcat(temp_log, "\n                      namelist");
                  strcat(temp_log, " file (ignored).\n");
               }

               return_code = WARNING;
            }

         }  /*  End of                                                       */
            /*  'else  if (strcmp(keyword, "Template_File") == 0)' segment.  */

         else  if (strcmp(keyword, "Log_File") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  log file name on the command line or
             *  multiple log file name entries in
             *  the namelist file.
             */
            if (strncmp(log_file_name, "NONE", 4) == 0)  {
               sscanf(&value[1], "%[^']", log_file_name);

               nmlst_logfile_count++;
            }
            else  {
               if (nmlst_logfile_count == 0)  {
                  printf("\n* * * W A R N I N G:  Log file");
                  printf(" name specified both on command line");
                  printf("\n                      and in");
                  printf(" namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Log file");
                  strcat(temp_log, " name specified both on command line");
                  strcat(temp_log, "\n                      and in");
                  strcat(temp_log, " namelist file (ignored).\n");
               }
               else  {
                  printf("\n* * * W A R N I N G:  Log file");
                  printf(" name specified multiple times in");
                  printf("\n                      namelist file (ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Log file");
                  strcat(temp_log, " name specified multiple times in");
                  strcat(temp_log, "\n                      namelist");
                  strcat(temp_log, " file (ignored).\n");
               }

               return_code = WARNING;
            }

         }  /*  End of                                                  */
            /*  'else  if (strcmp(keyword, "Log_File") == 0)' segment.  */

         else  if (strcmp(keyword, "Verbose") == 0)  {
            /*
             *  Make sure we haven't already found the
             *  verbose mode option on the command
             *  line or multiple verbose mode options
             *  in the namelist file.
             */
            if (verbose_mode == 0)  {
               verbose_mode = atoi(value);

               /*
                *  Check that the verbose level is
                *  either 0 (= OFF), 1 (= ON, give
                *  major steps), or 2 (= ON, give
                *  more detailed output).
                */
               if ( (verbose_mode < 0) || (verbose_mode > 2) )  {
                  printf("\n* * * W A R N I N G:  Unexpected verbosity");
                  printf(" level specified: %d (ignored).\n", verbose_mode);

                  strcat(temp_log, "\n* * * W A R N I N G:  Unexpected");
                  strcat(temp_log, " verbosity level specified: ");
                  sprintf(help_str, "%d", verbose_mode);
                  strcat(temp_log, help_str);
                  strcat(temp_log, " (ignored).\n");

                  verbose_mode = 0;

                  return_code = WARNING;
               }

            }
            else  {
               printf("\n* * * W A R N I N G:  Verbose");
               printf(" mode specified both on command line");
               printf("\n                      and in");
               printf(" namelist file (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Verbose");
               strcat(temp_log, " mode specified both on command line");
               strcat(temp_log, "\n                      and in");
               strcat(temp_log, " namelist file (ignored).\n");

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword, "Verbose") == 0)' segment.  */

         else  if (strcmp(keyword, "Quiet_Mode") == 0)  {
            /*
             *  Make sure we haven't already found the
             *  quiet mode option on the command line
             *  or multiple verbose mode options in
             *  the namelist file.
             */
            if (quiet_mode == 0)  {
               quiet_mode = atoi(value);

               /*
                *  Check that the quiet mode value is
                *  either 0 (= OFF) or 1 (= ON).
                */
               if ( (quiet_mode < 0) || (quiet_mode > 1) )  {
                  printf("\n* * * W A R N I N G:  Unexpected quiet mode");
                  printf(" value specified: %d (ignored).\n", quiet_mode);

                  strcat(temp_log, "\n* * * W A R N I N G:  Unexpected");
                  strcat(temp_log, " quiet mode value specified: ");
                  sprintf(help_str, "%d", quiet_mode);
                  strcat(temp_log, help_str);
                  strcat(temp_log, " (ignored).\n");

                  quiet_mode = 0;

                  return_code = WARNING;
               }

            }
            else  {
               printf("\n* * * W A R N I N G:  Quiet");
               printf(" mode specified both on command line");
               printf("\n                      and in");
               printf(" namelist file (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Quiet");
               strcat(temp_log, " mode specified both on command line");
               strcat(temp_log, "\n                      and in");
               strcat(temp_log, " namelist file (ignored).\n");

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword, "Quiet_Mode") == 0)'  */
            /*  segment.                                                */

         else  if (strcmp(keyword, "Copy_Input_History_to_Output") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  history copy option on the command
             *  line or multiple history copy
             *  options in the namelist file.
             */
            if (*hist_copy == NOT_SET)  {
               *hist_copy = atoi(value);

               /*
                *  Make sure the value is legal.
                */
               if ( (*hist_copy != TRUE) && (*hist_copy != FALSE) )  {
                  printf("\n* * * W A R N I N G:  Undefined HISTORY");
                  printf(" copy option value: %d", *hist_copy);
                  printf("\n                      (should be 1");
                  printf(" or 0; ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Undefined");
                  strcat(temp_log, " HISTORY copy option value: ");
                  sprintf(help_str, "%d", *hist_copy);
                  strcat(temp_log, help_str);
                  strcat(temp_log, "\n                      (should be 1");
                  strcat(temp_log, " or 0; ignored).\n");

                  *hist_copy = FALSE;

                  return_code = WARNING;
               }
            }
            else  {
               printf("\n* * * W A R N I N G:  HISTORY copy option");
               printf(" specified multiple times in");
               printf("\nnamelist file and/or on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  HISTORY copy option");
               strcat(temp_log, " specified multiple times in");
               strcat(temp_log, "\nnamelist file and/or on");
               strcat(temp_log, " command line (ignored).\n");

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword,                        */
            /*          "Copy_Input_History_to_Output") == 0)' segment.  */

         else  if (strcmp(keyword, "Remove_Blank_Blocks") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  remove blank blocks option on the
             *  command line or multiple remove blank
             *  blocks options in the namelist file.
             */
            if (*blank_blocks == NOT_SET)  {
               *blank_blocks = atoi(value);

               /*
                *  Make sure the value is legal.
                */
               if ( (*blank_blocks != TRUE) && (*blank_blocks != FALSE) )  {
                  printf("\n* * * W A R N I N G:  Undefined remove blank");
                  printf(" blocks option value: %d", *blank_blocks);
                  printf("\n                      (should be 1");
                  printf(" or 0; ignored).\n");

                  strcat(temp_log, "\n* * * W A R N I N G:  Undefined");
                  strcat(temp_log, " remove blank blocks option value: ");
                  sprintf(help_str, "%d", *blank_blocks);
                  strcat(temp_log, help_str);
                  strcat(temp_log, "\n                      (should be 1");
                  strcat(temp_log, " or 0; ignored).\n");

                  *blank_blocks = FALSE;

                  return_code = WARNING;
               }
            }
            else  {
               printf("\n* * * W A R N I N G:  Remove blank blocks option");
               printf(" specified multiple times in");
               printf("\nnamelist file and/or on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Remove blank blocks");
               strcat(temp_log, " option specified multiple times in");
               strcat(temp_log, "\nnamelist file and/or on");
               strcat(temp_log, " command line (ignored).\n");

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword,                         */
            /*                    "Remove_Blank_Blocks") == 0)' segment.  */

         else  if (strcmp(keyword, "Mask") == 0)  {
            /*
             *  Make sure we haven't already found a
             *  mask value on the command line or
             *  multiple mask value entries in the
             *  namelist file.
             */
            if (*mask == NO_MASK)  {
               *mask = atoi(value);

               printf("\n* * * E R R O R:  Mask option not implemented.\n");

               return(FAILURE);
            }
            else  {
               printf("\n* * * W A R N I N G:  Mask value");
               printf(" specified multiple times");
               printf("\n                      in namelist file and/or");
               printf(" on command line (ignored).\n");

               strcat(temp_log, "\n* * * W A R N I N G:  Mask value");
               strcat(temp_log, " specified multiple times");
               strcat(temp_log, "\n                      in namelist file");
               strcat(temp_log, " and/or on command line (ignored).\n");

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword, "Mask") == 0)' segment.  */

         else  if (strcmp(keyword, "NL_Print") == 0)  {
            nmlst_print = atoi(value);

            /*
             *  Check that the specified
             *  value is either 1 or 0.
             */
            if ( (nmlst_print != 0) && (nmlst_print != 1) )  {
               printf("\n* * * W A R N I N G:  Undefined NL_Print");
               printf(" value: %d (should be 1 or 0; ignored).\n",
                      nmlst_print);

               strcat(temp_log, "\n* * * W A R N I N G:  Undefined");
               strcat(temp_log, " NL_Print value: ");
               sprintf(help_str, "%d", nmlst_print);
               strcat(temp_log, help_str);
               strcat(temp_log, " (should be 1 or 0; ignored).\n");

               nmlst_print = 0;

               return_code = WARNING;
            }

         }  /*  End of 'else  if (strcmp(keyword, "NL_Print") == 0)'  */
            /*  segment.                                              */

         else  {
            printf("\n* * * W A R N I N G:  Unknown namelist keyword:");
            printf(" %s - ignored.\n", keyword);

            strcat(temp_log, "\n* * * W A R N I N G:  Unknown");
            strcat(temp_log, " namelist keyword:");
            strcat(temp_log, keyword);
            strcat(temp_log, " - ignored.\n");

            return_code = WARNING;
         }

      }  /*  End of 'for (comma_token = strtok(buffer_ptr, ",");    */
         /*               (comma_token != NULL) &&                  */
         /*               (strchr(comma_token, '&') == 0);          */
         /*               comma_token = strtok(NULL, ",") )' loop.  */

   }  /*  End of 'if (nmlst_buffer)' segment.  */

   else  {
      printf("\n* * * E R R O R:  Allocation of nmlst_buffer");
      printf(" memory space failed.\n");

      return(FAILURE);
   }

   /*
    *  Check if the contents of the namelist
    *  should be printed to stdout.
    */
   if (nmlst_print == 1)  {
      if (nmlst_out_buf)  {
         printf("\n%s\n", nmlst_out_buf);

         strcat(temp_log, "\n");
         strcat(temp_log, nmlst_out_buf);
         strcat(temp_log, "\n");
      }
      else  {
         printf("\n* * * W A R N I N G:  Unknown contents");
         printf(" of namelist file - data not saved");
         printf("\n                      due to lack of memory space.\n");

         strcat(temp_log, "\n* * * W A R N I N G:  Unknown contents");
         strcat(temp_log, " of namelist file - data not saved");
         strcat(temp_log, "\n                      due to lack of");
         strcat(temp_log, " memory space.\n");

         return_code = WARNING;
      }
   }

   free(nmlst_out_buf);
   free(nmlst_buffer);

   return(return_code);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  read_sodb_entries_file                            *** *
 *                                                                       *
 *                                                                       *
 * ***  This function reads the SODB entries file (into the array    *** *
 * ***  sodb_table).                                                 *** *
 *                                                                       *
 * ***  Input:  sodb_entries_file_name    - name of the SODB         *** *
 * ***                                      entries file.            *** *
 *                                                                       *
 * ***  Output:  line_count               - number of lines in the   *** *
 * ***                                      sodb_table array.        *** *
 *                                                                       *
 * ***  Return values:  sodb_table        - pointer to the SODB      *** *
 * ***                                      table array.             *** *
 * ***                  NULL              - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

struct  sodb_tbl*  read_sodb_entries_file(char*  sodb_entries_file_name,
                                          int*   line_count)

{

   char
         buffer[LONG_TEMP_STR_LEN],
         help_iline[MAX_SODB_LINE_LEN],
         input_line[MAX_SODB_LINE_LEN],
        *line_part;

   FILE
         *pipe_ptr,
         *sodb_fp;

   int
        i,
        imax,
        ix = 0,
        jx,
        part_count,
        quote_flag = 0;

   struct  sodb_tbl
                     *sodb_table;


   /*
    *  Initiate a pipe that will count
    *  the number of lines in the SODB
    *  entries file.
    */
   strcpy(buffer, "wc -l ");
   strcat(buffer, sodb_entries_file_name);

   if ( (pipe_ptr = popen(buffer, "r") ) != NULL)  {
      fgets(buffer, LONG_TEMP_STR_LEN, pipe_ptr);
      pclose(pipe_ptr);

      sscanf(buffer, "%d", line_count);
   }
   else  {
      printf("\n* * * E R R O R:  Failed to access SODB entries file.\n");

      return(NULL);
   }

   /*
    *  Allocate memory space for
    *  the sodb_table array.
    */
   sodb_table = malloc(*line_count * sizeof(struct  sodb_tbl) );

   if (sodb_table == NULL)  {
      printf("\n* * * E R R O R:  Allocation of memory for");
      printf(" SODB entries array failed.\n");

      return(NULL);
   }

   /*
    *  Open and read the SODB entries file.
    */
   sodb_fp = fopen(sodb_entries_file_name, "r");

   if (sodb_fp)  {
      while (fgets(input_line, MAX_SODB_LINE_LEN, sodb_fp) )  {
         if (ix >= *line_count)  {
            printf("\n* * * E R R O R:  Unexpected number of");
            printf(" lines in SODB entries file.\n");

            return(NULL);
         }

         /*
          *  Save a copy of the input line in case
          *  the value field is surrounded by
          *  single or double quotes.
          */
         strcpy(help_iline, input_line);

         /*
          *  Parse the parts of the input line.
          */
         part_count = 0;

         line_part = strtok(input_line, ", ");

         strcpy(sodb_table[ix].keyword, line_part);
         if (strlen(line_part) < MAX_KEYWORD_LEN)  {
            imax = MAX_KEYWORD_LEN - strlen(line_part);

            for (i = 0; i < imax; i++)  {
               strcat(sodb_table[ix].keyword, " ");
            }
         }

         while (line_part = strtok(NULL, ", ") )  {
            part_count++;

            if (part_count == 1)  {
               strcpy(sodb_table[ix].table, line_part);
            }
            else  if (part_count == 2)  {
               strcpy(sodb_table[ix].field, line_part);
            }
            else  if (part_count == 3)  {
               strcpy(sodb_table[ix].type, line_part);
            }
            else  if (part_count == 4)  {
               if (strncmp(line_part, "\"", 1) == 0)  {
                  quote_flag        = 1;
                  line_part         = strtok(help_iline, "\"");
                  line_part         = strtok(NULL, "\n");
                  jx                = strlen(line_part);
                  line_part[jx - 1] = '\0';
               }
               else  if (strncmp(line_part, "\'", 1) == 0)  {
                  quote_flag        = 1;
                  line_part         = strtok(help_iline, "\'");
                  line_part         = strtok(NULL, "\n");
                  jx                = strlen(line_part);
                  line_part[jx - 1] = '\0';
               }

               strcpy(sodb_table[ix].value, line_part);

               if (quote_flag != 0)  {
                  line_part = strtok(NULL, ", ");
                  quote_flag = 0;
               }
               else  {
                  jx = strlen(sodb_table[ix].value);
                  sodb_table[ix].value[jx - 1] = '\0';
               }

            }  /*  End of 'else  if (part_count == 4)' segment.  */

            else  {
               printf("\n* * * E R R O R:  Undefined SODB");
               printf(" entries file format.\n");

               return(NULL);
            }

         }  /*  End of 'while (line_part = strtok(NULL, ", ") )' loop.  */

         ix++;

      }  /*  End of 'while (fgets(input_line, MAX_SODB_LINE_LEN, sodb_fp) )'  */
         /*  loop.                                                            */

   }  /*  End of 'if (sodb_fp)' segment.  */

   else  {
      printf("\n* * * E R R O R:  Failed to open SODB entries file.\n");

      return(NULL);
   }

   fclose(sodb_fp);

   return(sodb_table);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  generate_final_product                            *** *
 *                                                                       *
 *                                                                       *
 * ***  This function generates the final product, i.e., open an     *** *
 * ***  input DCE (FITS) file, an output DCE (FITS) file, and a      *** *
 * ***  template file.  Read the input and template files and        *** *
 * ***  generate the output file according to the template file.     *** *
 *                                                                       *
 * ***  Input:  input_file_name           - name of the input DCE    *** *
 * ***                                      file.                    *** *
 * ***          output_file_name          - name of the output DCE   *** *
 * ***                                      file.                    *** *
 * ***          template_file_name        - name of the template     *** *
 * ***                                      file.                    *** *
 * ***          sodb_table                - SODB table array.        *** *
 * ***          line_count                - number of lines in the   *** *
 * ***                                      sodb_table array.        *** *
 * ***          hist_copy                 - flag indicating if       *** *
 * ***                                      HISTORY records should   *** *
 * ***                                      be copied from input to  *** *
 * ***                                      output DCE file.         *** *
 * ***          blank_blocks              - flag indicating if       *** *
 * ***                                      blank blocks should be   *** *
 * ***                                      removed from the output  *** *
 * ***                                      DCE file.                *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  WARNING           - successful processing    *** *
 * ***                                      completed (but a         *** *
 * ***                                      WARNING message has      *** *
 * ***                                      been issued,indicating   *** *
 * ***                                      minor - but not fatal -  *** *
 * ***                                      error or problem).       *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  generate_final_product(char*              input_file_name,
                            char*              output_file_name,
                            char*              template_file_name,
                            struct  sodb_tbl*  sodb_table,
                            int                line_count,
                            int                hist_copy,
                            int                blank_blocks)

{

   char
         actn_str[NUM_ACTIONS][MAX_KEYWORD_LEN + 1] = {"NOACTION",
                                                       "PASSTHRU",
                                                       "RENAME",
                                                       "DERIVED",
                                                       "DATABASE"},
         cmnt_help[MAX_KEYWD_CMNT + 1],
         comment[MAX_KEYWD_CMNT + 1],
         keywd_cmnt[MAX_KEYWD_CMNT + 1],
         keyword[MAX_KEYWORD_LEN + 1],
         naxis_comment[MAX_KEYWD_CMNT + 1],
         naxis3_comment[MAX_KEYWD_CMNT + 1],
         naxis_keywd[4][MAX_KEYWORD_LEN + 1]        = {"NAXIS   ",
                                                       "NAXIS1  ",
                                                       "NAXIS2  ",
                                                       "NAXIS3  "},
         prog_name[PROG_VER_STR_LEN],
         record[MAX_KEYWD_CMNT + 1],
         repl_keywd[MAX_KEYWORD_LEN + 1],
         str_value[MAX_KEYWD_CMNT],
         temp_file_name[MAX_FNAM_LEN],
         template_line[MAX_TEMPL_LINE_LEN + 1];

	 /*  Added TYB - 7/1/09 */
	 char S_value[FLEN_VALUE];

   double
           value;

   FILE
         *tmplt_fp;

   fitsfile
             *infp   = NULL,
             *outfp  = NULL,
             *tempfp = NULL;

   float
          float_value;

   int
        action,
        bitpix                = NOT_SET,
        blank_comment_history,
        i,
        int_val,
        last,
        missing_keyword_count = 0,
        naxis                 = 0,
        rcode,
        return_code           = SUCCESS,
        status                = 0,
        template_line_count   = 0,
        template_naxis[4]     = {0, 0, 0, 0},
        type;

   long
         pixels;

   union  convert  {
      double  dbl_value;
      int     int_value[2];
   }  conv_value;


   if (verbose_mode > 0)  {
      printf("\nPerform FPGen processing.\n");
      printf("\n   Open input DCE file.\n");
   }

   strcpy(naxis_comment, "");
   strcpy(naxis3_comment, "");

   /*
    *  Open the input DCE (FITS) file.
    */
   rcode = fits_open_file(&infp,
                          input_file_name,
                          READONLY,
                          &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to open input");
      printf(" file %s - status = %d\n", input_file_name, status);

      return(FAILURE);
   }

   /*
    *  Open the output DCE (FITS) file.
    */
   if (verbose_mode > 0)  {
      printf("\n   Open output DCE file.\n");
   }

   rcode = fits_create_file(&outfp,
                            output_file_name,
                            &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to create output");
      printf(" file %s - status = %d\n", output_file_name, status);

      return(FAILURE);
   }

   /*
    *  Check if the remove blank blocks
    *  option has been set, in which
    *  case we must open a temporary
    *  output file.  If this option has
    *  not been set, we set the tempfp
    *  pointer to outfp.
    */
   if (blank_blocks == TRUE)  {
      strcpy(temp_file_name, "!FPGen_temp_file.hdr");

      rcode = fits_create_file(&tempfp,
                               temp_file_name,
                               &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to create temporary header");
         printf(" file %s - status = %d\n", &temp_file_name[1], status);

         return(FAILURE);
      }
   }
   else  {
      tempfp = outfp;
   }

   /*
    *  Open the template file.
    */
   if (verbose_mode > 0)  {
      printf("\n   Open template file.\n");
   }

   tmplt_fp = fopen(template_file_name, "r");

   if (tmplt_fp)  {
      /*
       *  Read one template line at a time.
       */
      if (verbose_mode > 0)  {
         printf("\n   Parse and process template entries.\n");
      }

      while (fgets(template_line, MAX_TEMPL_LINE_LEN, tmplt_fp) )  {
         /*
          *  Skip lines that begin with a "#" character,
          *  since they are comment lines.
          */
         if (strncmp(template_line, "#", 1) != 0)  {
            /*
             *  Parse the template line.
             */
            rcode = parse_template_line(&template_line[0],
                                        &keyword[0],
                                        &type,
                                        &action,
                                        &repl_keywd[0],
                                        &comment[0],
                                        &blank_comment_history);

            if (rcode != SUCCESS)  {
               return(FAILURE);
            }

            /*
             *  If this template line is either blank
             *  or a COMMENT/HISTORY line, make sure
             *  it is 80 characters long.
             */
            if (blank_comment_history != 0)  {
               if (strncmp(keyword, "\n", 1) == 0)  {
                  strcpy(template_line, "                    ");
                  strcat(template_line, "                    ");
                  strcat(template_line, "                    ");
                  strcat(template_line, "                    ");
               }
               else  if (strlen(template_line) < 80)  {
                  last = strlen(template_line) - 1;
                  strncpy(&template_line[last], "\0", 1);

                  for (i = 0; i < (80 - strlen(template_line) ); i++)  {
                     strcat(template_line, " ");
                  }
               }
            }

            if (verbose_mode > 1)  {
               if (blank_comment_history != 0)  {
                  printf("\n      blank KEYIN = %s\n", template_line);
               }
               else  {
                  if ( (action == DERIVED) || (action == DATABASE) )  {
                     printf("\n      keyword = %s   action = %s",
                            repl_keywd, actn_str[action]);
                  }
                  else  {
                     printf("\n      keyword = %s   action = %s",
                            keyword, actn_str[action]);
                  }

                  if (action == RENAME)  {
                     printf("   to keyword = %s\n", repl_keywd);
                  }
                  else  {
                     printf("\n");
                  }
               }
            }

            /*
             *  Check if the blank_comment_history flag
             *  is set.  If it is, the current template
             *  line is either a blank (comment) line,
             *  a comment line, or a history line, that
             *  should be written verbatim to the
             *  output DCE file.
             */
            if (blank_comment_history != 0)  {
               rcode = fits_write_record(tempfp,
                                         &template_line[0],
                                         &status);

               if (status != 0)  {
                  printf("\n* * * E R R O R:  Failed to write");
                  printf(" blank/COMMENT/HISTORY record - status = %d\n",
                         status);

                  return(FAILURE);
               }

               continue;
            }

            /*
             *  Check through the comment string for new-
             *  line characters and strip off if found.
             */
            for (i = 0; i < strlen(comment); i++)  {
               if (comment[i] == '\n')  {
                  strcpy(cmnt_help, comment);
                  LOCAL_COPY_OF_strlcpy(comment, cmnt_help, (i + 1) );

                  break;
               }
            }

            /*
             *  Increment (valid) template line counter and
             *  save information if this is a NAXIS line
             *  (which, according to the FITS format, must
             *  be among the first six lines of the header).
             */
            template_line_count++;

            if (template_line_count < 7)  {
               for (i = 0; i < 4; i++)  {
                  if (strcmp(keyword, naxis_keywd[i]) == 0)  {
                     template_naxis[i] = 1;

                     if (i == 0)  {
                        strcpy(naxis_comment, comment);
                     }
                     else  if (i == 3)  {
                        strcpy(naxis3_comment, comment);
                     }

                     break;
                  }
               }
            }

            /*
             *  Check if the current template keyword is equal
             *  to either "ORIGIN  " or "CREATOR " and, if so,
             *  make sure action is equal to RENAME.
             */
            if ( (strcmp(keyword, "ORIGIN  ") == 0) ||
                 (strcmp(keyword, "CREATOR ") == 0) )  {
               if (action != RENAME)  {
                  printf("\n* * * E R R O R:  Unexpected action for");
                  printf(" template keyword %s: %s\n",
                         keyword, actn_str[action]);

                  return(FAILURE);
               }
            }

            /*
             *  Take appropriate action.  We begin
             *  with PASSTHRU and RENAME (these
             *  two actions are very similar).
             */
            if ( (action == PASSTHRU) || (action == RENAME) )  {
               /*
                *  This means we will read this keyword
                *  from the input file, write it
                *  unchanged to the output file if
                *  action is PASSTHRU or rename the
                *  keyword and write it to the output
                *  file if action is RENAME.
                *  The reason for zeroing the parameter
                *  value before calling the function
                *  fits_read_key is to avoid internal
                *  error when processing the keyword
                *  NAXIS (for which the value is saved
                *  in the parameter naxis - this step
                *  is what causes the internal error,
                *  "improper use of Symbol::offset", if
                *  the parameter value isn't zeroed).
                */
               value = 0.0;

	       /*  There was some problem with strings on Mac ppc - TYB 7/1/09 */

               if(type == TSTRING){
                 rcode = fits_read_key(infp,
                                       type,
                                       &keyword[0],
                                       &S_value,
                                       &keywd_cmnt[0],
                                       &status);
	       }
	       else{
                 rcode = fits_read_key(infp,
                                       type,
                                       &keyword[0],
                                       &value,
                                       &keywd_cmnt[0],
                                       &status);
	       }

               if (status != 0)  {
                  /*
                   *  Check is status is equal to 202,
                   *  which means the keyword was not
                   *  found in the input file header.
                   *  If so, write WARNING message,
                   *  increment missing keyword
                   *  counter, and continue.
                   */
                  if (status == KEY_NO_EXIST)  {
                     status = 0;

                     if (strncmp(keyword, "NAXIS", 5) == 0)  {
                        if (strcmp(keyword, "NAXIS3  ") != 0)  {
                           printf("\n* * * E R R O R:  Mandatory keyword not");
                           printf(" found in input file: %s\n", keyword);

                           return(FAILURE);
                        }
                     }

                     if (!quiet_mode)  {
                        printf("\n* * * W A R N I N G:  %s", actn_str[action]);
                        printf(" keyword not found in input file: %s\n",
                               keyword);
                     }

                     missing_keyword_count++;

                     continue;
                  }
                  else  {
                     printf("\n* * * E R R O R:  Failed to read");
                     printf(" %s keyword %s - status = %d\n",
                            actn_str[action], keyword, status);

                     /*
                      *  Check if status is equal to 412, which
                      *  means there was an overflow during
                      *  datatype conversion.  If so, write an
                      *  output record with a NULL value.
                      */
                     if (status == NUM_OVERFLOW)  {
                        strcpy(comment, "ERROR: Could not determine");
                        strcat(comment, " keyword value");
                        status = 0;

                        rcode = fits_update_key_null(tempfp,
                                                     &repl_keywd[0],
                                                     &comment[0],
                                                     &status);

                        if (status != 0)  {
                           printf("\n* * * E R R O R:  Failed to write %s",
                                  actn_str[action]);
                           printf(" keyword %s with NULL value - status = %d\n",
                                  keyword, status);

                           return(FAILURE);
                        }

                        return_code = NO_VALUE;

                        continue;
                     }

                     return(FAILURE);
                  }

               }  /*  End of 'if (status != 0)' segment.  */

               /*
                *  We must copy the input keyword
                *  to the repl_keywd string if
                *  this is a PASSTHRU keyword.
                */
               if (action == PASSTHRU)  {
                  strcpy(repl_keywd, keyword);
               }

	       /*  There was some problem with strings on Mac ppc - TYB 7/1/09 */

               if(type == TSTRING){
                 rcode = fits_write_key(tempfp,
                                        type,
                                        &repl_keywd[0],
                                        &S_value,
                                        &comment[0],
                                        &status);
	       }
               else{
                 rcode = fits_write_key(tempfp,
                                        type,
                                        &repl_keywd[0],
                                        &value,
                                        &comment[0],
                                        &status);
	       }
               if (status != 0)  {
                  printf("\n* * * E R R O R:  Failed to write");
                  printf(" %s keyword %s - status = %d\n",
                         actn_str[action], keyword, status);

                  /*
                   *  Write keyword with NULL value field.
                   */
                  strcpy(comment, "ERROR: Could not determine keyword value");
                  status = 0;

                  rcode = fits_update_key_null(tempfp,
                                               &repl_keywd[0],
                                               &comment[0],
                                               &status);

                  if (status != 0)  {
                     printf("\n* * * E R R O R:  Failed to write");
                     printf(" %s keyword %s with NULL value - status = %d\n",
                            actn_str[action], keyword, status);

                     return(FAILURE);
                  }

                  return_code = NO_VALUE;
               }

               /*
                *  Check if this is one of the first six
                *  header keyword lines, since we must
                *  do some special processing for some
                *  of these first six lines.
                */
               if (template_line_count < 7)  {
                  /*
                   *  Save the value if the keyword
                   *  is equal to BITPIX or NAXIS.
                   */
                  if (strcmp(keyword, "BITPIX  ") == 0)  {
                     if (type == TINT)  {
                        conv_value.dbl_value = value;
                        bitpix               = conv_value.int_value[0];
                     }
                     else  {
                        printf("\n* * * E R R O R:  Unexpected data type");
                        printf(" for BITPIX keyword: %d\n", type);

                        return(FAILURE);
                     }

                     if ( (bitpix != FLOAT_IMG) && (bitpix != SHORT_IMG) )  {
                        printf("\n* * * E R R O R:  Unexpected BITPIX value:");
                        printf(" %d (must be %d or %d).\n",
                               bitpix, SHORT_IMG, FLOAT_IMG);

                        return(FAILURE);
                     }
                  }
                  else if (strcmp(keyword, "NAXIS   ") == 0)  {
                     if (type == TINT)  {
                        conv_value.dbl_value = value;
                        naxis                = conv_value.int_value[0];
                     }
                     else  {
                        printf("\n* * * E R R O R:  Unexpected data type");
                        printf(" for NAXIS keyword: %d\n", type);

                        return(FAILURE);
                     }
                  }

                  /*
                   *  Check if the current keyword is
                   *  the final NAXIS<i> keyword, in
                   *  which case it is time to add
                   *  the ORIGIN and CREATOR keywords.
                   */
                  if ( ( (strcmp(keyword, "NAXIS2  ") == 0) &&
                         (naxis == 2) ) ||
                       ( (strcmp(keyword, "NAXIS3  ") == 0) &&
                         (naxis == 3) ) )  {
                     rcode = add_origin_and_creator_keywds(tempfp);

                     if (rcode != SUCCESS)  {
                        return(FAILURE);
                     }
                  }

               }  /*  End of 'if (template_line_count < 7)' segment.  */

            }  /*  End of 'if ( (action == PASSTHRU) ||         */
               /*               (action == RENAME) )' segment.  */

            /*
             *  The third action is DERIVED.  Each
             *  keyword with this action has its
             *  own unique processing.
             */
            else  if (action == DERIVED)  {
               /*
                *  The first DERIVED keyword is PROGCAT.
                */
               if (strcmp(repl_keywd, "PROGCAT ") == 0)  {
                  /*
                   *  PROGCAT is based on a database quantity,
                   *  which hasn't been implemented yet.
                   */
                  printf("\n# # # FPGen:  Action DERIVED not yet");
                  printf(" implemented for keyword PROGCAT.\n");
               }

               /*
                *  The second DERIVED keyword is BOOSTFRM.
                */
               else  if (strcmp(repl_keywd, "BOOSTFRM") == 0)  {
                  /*
                   *  Set BOOSTFRM to T(rue) if the input
                   *  header keyword DCENUM is equal to 0.
                   *  If not, set BOOSTFRM to F(alse).
                   */
                  strcpy(keyword, "DCENUM  ");

                  rcode = fits_read_key(infp,
                                        TINT,
                                        &keyword[0],
                                        &value,
                                        &keywd_cmnt[0],
                                        &status);

                  if (status != 0)  {
                     status = 0;

                     if (!quiet_mode)  {
                        printf("\n* * * W A R N I N G:  Failed to generate");
                        printf(" DERIVED keyword BOOSTFRM - ");
                        printf("\n                      keyword DCENUM not");
                        printf(" found in input file.\n");
                     }

                     missing_keyword_count++;

                     continue;
                  }

                  if (value == 0)  {
                     value = TRUE;
                  }
                  else  {
                     value = FALSE;
                  }

                  rcode = fits_write_key(tempfp,
                                         TLOGICAL,
                                         &repl_keywd[0],
                                         &value,
                                         &comment[0],
                                         &status);

                  if (status != 0)  {
                     printf("\n* * * E R R O R:  Failed to write");
                     printf(" DERIVED keyword BOOSTFRM - status = %d\n",
                            status);

                     return(FAILURE);
                  }

               }  /*  End of 'else  if                                */
                  /*          (strcmp(repl_keywd, "BOOSTFRM") == 0)'  */
                  /*  segment.                                        */

               /*
                *  We end up here if the DERIVED
                *  keyword is undefined.
                */
               else  {
                  if (!quiet_mode)  {
                     printf("\n* * * W A R N I N G:  Unknown DERIVED");
                     printf(" keyword: %s\n", repl_keywd);
                  }

                  missing_keyword_count++;

                  continue;
               }

            }  /*  End of 'else  if (action == DERIVED)' segment.  */

            /*
             *  The fourth action is DATABASE.
             */
            else  if (action == DATABASE)  {
               /*
                *  Get the value for this keyword
                *  from the sodb_table array.
                */
               if (line_count > 0)  {
                  rcode = get_sodb_entry(sodb_table,
                                         line_count,
                                         type,
                                         &repl_keywd[0],
                                         &str_value[0]);

                  if (rcode != SUCCESS)  {
                     if (rcode == FAILURE)  {
                        return(FAILURE);
                     }

                     missing_keyword_count++;

                     continue;
                  }

                  /*
                   *  Write complete line to the FITS output file.
                   */
                  if ( (type == TSTRING) || (type == TBYTE) )  {
                     rcode = fits_write_key(tempfp,
                                            type,
                                            &repl_keywd[0],
                                            &str_value[0],
                                            &comment[0],
                                            &status);
                  }
                  else  {
                     if (type == TLOGICAL)  {
                        str_value[0] = toupper(str_value[0]);

                        if (strncmp(str_value, "T", 1) == 0)  {
                           value = TRUE;
                        }
                        else  if (strncmp(str_value, "F", 1) == 0)  {
                           value = FALSE;
                        }
                        else  {
                           printf("\n* * * E R R O R:  Undefined logical");
                           printf(" value in SODB entries file: %s\n",
                                  str_value);

                           return(FAILURE);
                        }

                        rcode = fits_write_key(tempfp,
                                               type,
                                               &repl_keywd[0],
                                               &value,
                                               &comment[0],
                                               &status);
                     }
                     else  if (type == TINT)  {
                        int_val = atoi(str_value);

                        rcode = fits_write_key(tempfp,
                                               type,
                                               &repl_keywd[0],
                                               &int_val,
                                               &comment[0],
                                               &status);
                     }
                     else  if (type == TDOUBLE)  {
                        value = atof(str_value);

                        rcode = fits_write_key(tempfp,
                                               type,
                                               &repl_keywd[0],
                                               &value,
                                               &comment[0],
                                               &status);
                     }
                     else  if (type == TFLOAT)  {
                        value       = atof(str_value);
                        float_value = (float) value;

                        rcode = fits_write_key(tempfp,
                                               type,
                                               &repl_keywd[0],
                                               &float_value,
                                               &comment[0],
                                               &status);
                     }
                     else  {
                        printf("\n* * * E R R O R:  Unknown type value in");
                        printf(" template for action DATABASE: %d\n", type);

                        return(FAILURE);
                     }

                  }  /*  End of 'else' (i.e., (type != TSTRING)  */
                     /*  && (type != TBYTE) segment.             */
                  
                  if (status != 0)  {
                     printf("\n* * * E R R O R:  Failed to write DATABASE");
                     printf(" keyword %s - status = %d\n",
                            repl_keywd, status);

                     /*
                      *  Write keyword with NULL value field.
                      */
                     strcpy(comment, "ERROR: Could not determine");
                     strcat(comment, " keyword value");
                     status = 0;

                     rcode = fits_update_key_null(tempfp,
                                                  &repl_keywd[0],
                                                  &comment[0],
                                                  &status);

                     if (status != 0)  {
                        printf("\n* * * E R R O R:  Failed to write");
                        printf(" %s keyword %s with NULL value - status = %d\n",
                               actn_str[action], keyword, status);

                        return(FAILURE);
                     }

                     return_code = NO_VALUE;
                  }

               }  /*  End of 'if (line_count > 0)' segment.  */

               else  {
                  missing_keyword_count++;
               }

            }  /*  End of 'else  if (action == DATABASE)' segment.  */

            /*
             *  Undefined action.
             */
            else  {
               printf("\n* * * E R R O R:  Undefined action (keyword =");
               printf(" %s): %d\n", keyword, action);

               return_code = NO_ACTION;
            }

         }  /*  End of 'if (strncmp(template_line, "#", 1) != 0)' segment.  */

      }  /*  End of                                                         */
         /*  'while (fgets(template_line, MAX_TEMPL_LINE_LEN, tmplt_fp) )'  */
         /*  loop.                                                          */

   }  /*  End of 'if (tmplt_fp)' segment.  */

   else  {
      printf("\n* * * E R R O R:  Failed to open template file %s\n",
             template_file_name);

      return(FAILURE);
   }

   /*
    *  Make sure the mandatory keywords
    *  NAXIS, NAXIS1, and NAXIS2 were
    *  included in the template file.
    */
   for (i = 0; i < 3; i++)  {
      if (template_naxis[i] == 0)  {
         printf("\n* * * E R R O R:  Mandatory keyword not found");
         printf(" in template file: %s\n", naxis_keywd[i]);

         return(FAILURE);
      }
   }

   /*
    *  Check if the HISTORY records should be
    *  copied from input to output DCE file.
    */
   if (hist_copy == TRUE)  {
      rcode = copy_history(infp,
                           tempfp);

      if (rcode != SUCCESS)  {
         return(FAILURE);
      }
   }

   /*
    *  Generate a HISTORY record for this program.
    */
   strcpy(prog_name, "FPGen");

   writeDLhistoryline(tempfp,
                      prog_name,
                      fpg_version,
                      &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to write FPGen");
      printf(" HISTORY record - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Check (and, if necessary, adjust) the NAXIS
    *  and NAXISi keywords in the output file.
    */
   if (verbose_mode > 1)  {
      printf("\n   Check/adjust NAXIS keywords.\n");
   }

   rcode = check_naxis_keywords(infp,
                                tempfp,
                                &template_naxis[0],
                                &naxis_comment[0],
                                &naxis3_comment[0],
                                &pixels);

   if (rcode != SUCCESS)  {
      return(FAILURE);
   }

   /*
    *  Check if the remove blank blocks option
    *  has been set, in which case we must
    *  copy the completed header from the
    *  temporary to the output DCE file.
    */
   if (blank_blocks == TRUE)  {
      rcode = copy_temp_to_out_hdr(tempfp,
                                   outfp,
                                   temp_file_name);

      if (rcode != SUCCESS)  {
         return(FAILURE);
      }
   }

   /*
    *  Copy image data from input to output file.
    */
   if (verbose_mode > 0)  {
      printf("\n   Copy image data.\n");
   }

   rcode = copy_image_data(infp,
                           outfp,
                           bitpix,
                           pixels);

   if (rcode != SUCCESS)  {
      return(FAILURE);
   }

   /*
    *  Close the input and output files,
    *  as well as the template file.
    */
   if (verbose_mode > 0)  {
      printf("\n   Close in- & output files.\n");
   }

   rcode = fits_close_file(infp,
                           &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to close input");
      printf(" file %s - status = %d\n", input_file_name, status);

      return(FAILURE);
   }

   rcode = fits_close_file(outfp,
                           &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to close output");
      printf(" file %s - status = %d\n", output_file_name, status);

      return(FAILURE);
   }

   fclose(tmplt_fp);

   /*
    *  Check if any (PASSTHRU or RENAME) keywords
    *  were missing in the input file header and,
    *  if so, write a summary.
    */
   if (missing_keyword_count > 0)  {
      if (!quiet_mode)  {
         printf("\n* * * W A R N I N G:  Total number of keywords not found");
         printf(" in input file = %d\n", missing_keyword_count);
      }

      if (return_code > 0)  {
         return_code = WARNING;
      }
   }

   return(return_code);

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  parse_template_line                               *** *
 *                                                                       *
 *                                                                       *
 * ***  This function parses a template line, taking one template    *** *
 * ***  line string as input and returning a keyword, a data type,   *** *
 * ***  an action, a replacement keyword (that may be blank), and a  *** *
 * ***  comment string.                                              *** *
 *                                                                       *
 * ***  Input:  template_line             - template line string.    *** *
 *                                                                       *
 * ***  Output:  keyword                  - FITS keyword.            *** *
 * ***           type                     - FITS data type.          *** *
 * ***           action                   - action type.             *** *
 * ***           repl_keywd               - replacement FITS         *** *
 * ***                                      keyword.                 *** *
 * ***           comment                  - comment string.          *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  parse_template_line(char*  template_line,
                         char*  keyword,
                         int*   type,
                         int*   action,
                         char*  repl_keywd,
                         char*  comment,
                         int*   blank_comment_history)

{

   char
         chr_action[9],
         chr_type[12],
        *str_ptr;


   /*
    *  Set the blank_comment_history flag to
    *  zero, indicating that the current
    *  template line is NOT a blank
    *  (comment) line, a COMMENT line, or a
    *  HISTORY line.
    */
   *blank_comment_history = 0;

   /*
    *  Check if the template line consists of
    *  only blanks or blanks and a new-line
    *  character and, if so, return with
    *  blank_comment_history set to one.
    */
   if (strlen(template_line) == strspn(template_line, " \n") )  {
      *blank_comment_history = BLNK_LINE;

      return(SUCCESS);
   }

   /*
    *  Set str_ptr to the beginning of
    *  the template line and copy the
    *  first 9 characters into keyword.
    */
   str_ptr = template_line;

   LOCAL_COPY_OF_strlcpy(keyword, str_ptr, 9);

   /*
    *  Check if the keyword is COMMENT or
    *  HISTORY, in which case we set the
    *  blank_comment_history flag to the
    *  corresponding value and return.
    *  If keyword is neither of these,
    *  check if it is blank, in which
    *  case we set the flag in order to
    *  facilitate further testing of
    *  blank (comment) lines later.
    */
   if (strncmp(keyword, "COMMENT ", 8) == 0)  {
      *blank_comment_history = CMNT_LINE;

      return(SUCCESS);
   }
   else  if (strncmp(keyword, "HISTORY ", 8) == 0)  {
      *blank_comment_history = HIST_LINE;

      return(SUCCESS);
   }
   else  if (strncmp(keyword, "        ", 8) == 0)  {
      *blank_comment_history = BLNK_LINE;
   }
   else  if (strncmp(keyword, "\n", 1) == 0)  {
      *blank_comment_history = BLNK_LINE;

      return(SUCCESS);
   }

   /*
    *  Move past the keyword part of the
    *  template line and copy the next
    *  12 characters into chr_type.
    */
   str_ptr += 10;

   LOCAL_COPY_OF_strlcpy(chr_type, str_ptr, 12);

   /*
    *  Check if this is a block comment
    *  line and, if so, return.
    */
   if (strncmp(chr_type, "/", 1) == 0)  {
      if (*blank_comment_history == BLNK_LINE)  {
         return(SUCCESS);
      }
   }

   /*
    *  Move past the data type part of the
    *  template line and copy the next
    *  9 characters into chr_action.
    */
   str_ptr += 13;

   LOCAL_COPY_OF_strlcpy(chr_action, str_ptr, 9);

   /*
    *  Move past the action type part of
    *  the template line and copy the
    *  next 9 characters into repl_keywd.
    */
   str_ptr += 11;

   LOCAL_COPY_OF_strlcpy(repl_keywd, str_ptr, 9);

   /*
    *  Move past the replacement keyword
    *  part of the template line and copy
    *  the remainder into comment.
    */
   str_ptr += 12;

   strcpy(comment, str_ptr);

   /*
    *  Identify action type.
    */
   if (strncmp(chr_action, "PASSTHRU", 8) == 0)  {
      *action = PASSTHRU;
   }
   else  if (strncmp(chr_action, "RENAME", 6) == 0)  {
      *action = RENAME;
   }
   else  if (strncmp(chr_action, "DERIVED", 7) == 0)  {
      *action = DERIVED;

      if (*blank_comment_history == BLNK_LINE)  {
         *blank_comment_history = 0;
      }
   }
   else  if (strncmp(chr_action, "DATABASE", 8) == 0)  {
      *action = DATABASE;

      if (*blank_comment_history == BLNK_LINE)  {
         *blank_comment_history = 0;
      }
   }
   else  {
      *action = NOACTION;

      printf("\n* * * E R R O R:  Undefined action: %s\n", chr_action);

      return(FAILURE);
   }

   /*
    *  Check if the blank_comment_history flag
    *  is set, in which case there is no FITS
    *  data type to identify.
    */
   if (*blank_comment_history != 0)  {
      return(SUCCESS);
   }

   /*
    *  Identify FITS data type.
    */
   if (strncmp(chr_type, "TBIT", 4) == 0)  {
      *type = TBIT;
   }
   else  if (strncmp(chr_type, "TBYTE", 5) == 0)  {
      *type = TBYTE;
   }
   else  if (strncmp(chr_type, "TLOGICAL", 8) == 0)  {
      *type = TLOGICAL;
   }
   else  if (strncmp(chr_type, "TSTRING", 7) == 0)  {
      *type = TSTRING;
   }
   else  if (strncmp(chr_type, "TUSHORT", 7) == 0)  {
      *type = TUSHORT;
   }
   else  if (strncmp(chr_type, "TSHORT", 6) == 0)  {
      *type = TSHORT;
   }
   else  if (strncmp(chr_type, "TINT", 4) == 0)  {
      *type = TINT;
   }
   else  if (strncmp(chr_type, "TULONG", 6) == 0)  {
      *type = TULONG;
   }
   else  if (strncmp(chr_type, "TLONG", 5) == 0)  {
      *type = TLONG;
   }
   else  if (strncmp(chr_type, "TFLOAT", 6) == 0)  {
      *type = TFLOAT;
   }
   else  if (strncmp(chr_type, "TDOUBLE", 7) == 0)  {
      *type = TDOUBLE;
   }
   else  if (strncmp(chr_type, "TCOMPLEX", 8) == 0)  {
      *type = TCOMPLEX;
   }
   else  {
      printf("\n* * * E R R O R:  Undefined data type: %s", chr_type);
      printf("  [keyword = %s]\n", keyword);

      return(FAILURE);
   }

   return(SUCCESS);

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  add_origin_and_creator_keywds                     *** *
 *                                                                       *
 *                                                                       *
 * ***  This function adds the keywords ORIGIN and CREATOR to the    *** *
 * ***  oputput DCE (FITS) file header.                              *** *
 *                                                                       *
 * ***  Input:  tempfp                    - temporary header file    *** *
 * ***                                      pointer.                 *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  add_origin_and_creator_keywds(fitsfile*  tempfp)

{

   char
        *env_value,
         hdr_line[LONG_TEMP_STR_LEN];

   int
        i,
        line_length,
        rcode,
        status = 0;


   /*
    *  Add the ORIGIN keyword to the header.
    */
   strcpy(hdr_line, "ORIGIN  = \'Spitzer Science Center\'");
   strcat(hdr_line, " / Organization generating this FITS file     ");

   rcode = fits_write_record(tempfp,
                             &hdr_line[0],
                             &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to add ORIGIN");
      printf(" keyword - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Add the CREATOR keyword to the header.
    *  This keyword gets its value from the
    *  environmental variables.
    */
   strcpy(hdr_line, "CREATOR = \'");

   env_value = getenv("SOS_REL");

   if (env_value != NULL)  {
      strcat(hdr_line, env_value);

      line_length = strlen(hdr_line);

      if (line_length > 79)  {
         hdr_line[79] = '\'';
      }
      else  {
         if (line_length < 19)  {
            for (i = 0; i < (19 - line_length); i++)  {
               strcat(hdr_line, " ");
            }
         }

         strcat(hdr_line, "\'");

         line_length = strlen(hdr_line);

         if (line_length <= 30)  {
            for (i = 0; i < (30 - line_length); i++)  {
               strcat(hdr_line, " ");
            }

            strcat(hdr_line, " / SW version used to create");
            strcat(hdr_line, " this FITS file       ");
         }
         else  {
            strcat(hdr_line, " / SW version used to create this FITS file");

            if (line_length < 37)  {
               for (i = 0; i < (37 - line_length); i++)  {
                  strcat(hdr_line, " ");
               }
            }

         }

      }  /*  End of 'else' (i.e., line_length <= 79) segment.  */

   }  /*  End of 'if (env_value != NULL)' segment.  */

   else  {
      strcat(hdr_line, "        \'           / SW version used");
      strcat(hdr_line, " to create this FITS file       ");
   }

   rcode = fits_write_record(tempfp,
                             &hdr_line[0],
                             &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to add CREATOR");
      printf(" keyword - status = %d\n", status);

      return(FAILURE);
   }

   return(SUCCESS);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  get_sodb_entry                                    *** *
 *                                                                       *
 *                                                                       *
 * ***  This function finds the value for a given keyword from the   *** *
 * ***  SODB entries file.                                           *** *
 *                                                                       *
 * ***  Input:  sodb_table                - pointer to the SODB      *** *
 * ***                                      table array.             *** *
 * ***          line_count                - number of lines in the   *** *
 * ***                                      sodb_table array.        *** *
 * ***          type                      - FITS data type.          *** *
 * ***          repl_keywd                - replacement FITS         *** *
 * ***                                      keyword.                 *** *
 *                                                                       *
 * ***  Output:  str_value                - value of the given       *** *
 * ***                                      keyword (as a character  *** *
 * ***                                      string).                 *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  WARNING           - successful processing    *** *
 * ***                                      completed (but a         *** *
 * ***                                      WARNING message has      *** *
 * ***                                      been issued,indicating   *** *
 * ***                                      minor - but not fatal -  *** *
 * ***                                      error or problem).       *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  get_sodb_entry(struct  sodb_tbl*  sodb_table,
                    int                line_count,
                    int                type,
                    char*              repl_keywd,
                    char*              str_value)

{

   char
         data_type[NUM_DATA_TYPES][MAX_KEYWORD_LEN] = {"char",
                                                       "logical",
                                                       "string",
                                                       "integer",
                                                       "float",
                                                       "double"},
         str_val_help[MAX_KEYWD_CMNT],
        *str_val_ptr;

   int
        i,
        ix,
        type_val[NUM_DATA_TYPES] = {TBYTE,
                                    TLOGICAL,
                                    TSTRING,
                                    TINT,
                                    TFLOAT,
                                    TDOUBLE};


   /*
    *  Find the given keyword in the sodb_table array.
    */
   ix = NOT_SET;

   for (i = 0; i < line_count; i++)  {
      if (strcasecmp(repl_keywd, sodb_table[i].keyword) == 0)  {
         ix = i;

         break;
      }
   }

   /*
    *  Write a WARNING message if the
    *  keyword was not found.
    */
   if (ix == NOT_SET)  {
      if (!quiet_mode)  {
         printf("\n* * * W A R N I N G:  DATABASE type keyword not found");
         printf(" in SODB entries file:\n                      %s\n",
                repl_keywd);
      }

      strcpy(str_value, "NONE");

      return(WARNING);
   }

   /*
    *  The keyword WAS found, so copy
    *  the value to str_value.  Also,
    *  make sure the type from the
    *  template matches the type from
    *  the sodb_table array.
    */
   if (strcmp(sodb_table[ix].value,"\n") != 0)  {
      strcpy(str_value, sodb_table[ix].value);

      /*
       *  Strip off new line character if found.
       */
      str_val_ptr = &str_value[0];
      str_val_ptr += strlen(str_value) - 1;

      if (strcmp(str_val_ptr, "\n") == 0)  {
         strcpy(str_val_help, str_value);
         LOCAL_COPY_OF_strlcpy(str_value, str_val_help, strlen(str_value));
      }

      for (i = 0; i < NUM_DATA_TYPES; i++)  {
         if (strcmp(data_type[i], sodb_table[ix].type) == 0)  {
            if (type_val[i] != type)  {
               printf("\n* * * E R R O R:  Unexpected data type in SODB");
               printf(" entries file: %s [%d]\n", sodb_table[ix].type, type);

               return(FAILURE);
            }

            return(SUCCESS);
         }
      }
   }
   else  {
      if (!quiet_mode)  {
         printf("\n* * * W A R N I N G:  Blank value field in SODB entries");
         printf(" file for keyword\n                      %s\n", repl_keywd);
      }

      strcpy(str_value, "NONE");

      return(WARNING);
   }

   printf("\n* * * E R R O R:  Unexpected data type in SODB");
   printf(" entries file: %s\n", sodb_table[ix].type);

   return(FAILURE);

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  copy_history                                      *** *
 *                                                                       *
 *                                                                       *
 * ***  This function copies the HISTORY records from the input DCE  *** *
 * ***  (FITS) file to the output DCE (FITS) file.                   *** *
 *                                                                       *
 * ***  Input:  infp                      - input DCE (FITS) file    *** *
 * ***                                      pointer.                 *** *
 * ***          outfp                     - output DCE (FITS) file   *** *
 * ***                                      pointer.                 *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  copy_history(fitsfile*  infp,
                  fitsfile*  outfp)

{

   char
         record[MAX_KEYWD_CMNT + 1];

   int
        i,
        keywd_count,
        keywd_pos,
        rcode,
        status = 0;


   if (verbose_mode > 0)  {
      printf("\n   Copy history lines.\n");
   }

   /*
    *  Find out how many keywords there
    *  are in the input DCE file.
    */
   rcode = fits_get_hdrpos(infp,
                           &keywd_count,
                           &keywd_pos,
                           &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to get input file");
      printf(" header info. - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Loop through the keyword records
    *  in the input DCE file and copy
    *  all the HISTORY records to the
    *  output DCE file.
    */
   for (i = 0; i < keywd_count; i++)  {
      rcode = fits_read_record(infp,
                               (i + 1),
                               &record[0],
                               &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to read HISTORY record");
         printf(" from input file - status = %d\n", status);

         return(FAILURE);
      }

      if (strncmp(record, "HISTORY ", 8) == 0)  {
         rcode = fits_write_record(outfp,
                                   &record[0],
                                   &status);

         if (status != 0)  {
            printf("\n* * * E R R O R:  Failed to write HISTORY record");
            printf(" to output file - status = %d\n", status);

            return(FAILURE);
         }
      }

   }  /*  End of 'for (i = 0; i < keywd_count; i++)' loop.  */

   return(SUCCESS);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  check_naxis_keywords                              *** *
 *                                                                       *
 *                                                                       *
 * ***  This function checks the NAXIS and NAXISi keywords in the    *** *
 * ***  output DCE (FITS) file (or the temporary FITS header file    *** *
 * ***  if one is present) in order to ensure that the NAXIS value   *** *
 * ***  matches the number of NAXISi keywords that will be output.   *** *
 *                                                                       *
 * ***  Input:  infp                      - input DCE (FITS) file    *** *
 * ***                                      pointer.                 *** *
 * ***          outfp                     - output or temporary DCE  *** *
 * ***                                      (FITS) file pointer.     *** *
 * ***          template_naxis            - four-element array that  *** *
 * ***                                      indicates whether (= 1)  *** *
 * ***                                      or not (= 0) the         *** *
 * ***                                      keywords NAXIS, NAXIS1,  *** *
 * ***                                      NAXIS2, and NAXIS3 were  *** *
 * ***                                      included in the          *** *
 * ***                                      template file.           *** *
 * ***          naxis_comment             - string that contains     *** *
 * ***                                      the FITS header comment  *** *
 * ***                                      field value for the      *** *
 * ***                                      keyword NAXIS from the   *** *
 * ***                                      template file.           *** *
 * ***          naxis3_comment            - string that contains     *** *
 * ***                                      the FITS header comment  *** *
 * ***                                      field value for the      *** *
 * ***                                      keyword NAXIS3 from the  *** *
 * ***                                      template file.           *** *
 *                                                                       *
 * ***  Output:  pixels                   - number of pixels in the  *** *
 * ***                                      (binary) image data.     *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  check_naxis_keywords(fitsfile*  infp,
                          fitsfile*  outfp,
                          int*       template_naxis,
                          char*      naxis_comment,
                          char*      naxis3_comment,
                          long*      pixels)

{

   char
         hdr_line[LONG_TEMP_STR_LEN];

   int
        flush_flag = 0,
        nfound,
        rcode,
        status     = 0,
        value;

   long
         naxis[3] = {1, 1, 1};


   /*
    *  Read the NAXIS and NAXISi keywords
    *  from the input file header.
    */
   rcode = fits_read_keys_lng(infp,
                              "NAXIS",
                              1,
                              3,
                              &naxis[0],
                              &nfound,
                              &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to read NAXIS data");
      printf(" from input file - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Calculate the size of the image data.
    */
   if (nfound == 2)  {
      *pixels = naxis[0] * naxis[1];

      /*
       *  CR-3860:  Do _NOT_ modify NAXIS (from 2
       *            to 3), i.e., the following
       *            code segment has been
       *            commented out (but left in
       *            place, since this CR may be
       *            reversed in the future).
       *
       *  Check if the keyword NAXIS3 was included
       *  in the template file.  If so, modify the
       *  value field of the keyword NAXIS (from 2
       *  to 3) and insert the keyword NAXIS3 in
       *  the output FITS header.
       *  <====  This is where one of the changes was made.
      if (template_naxis[3] == 1)  {
         value = 3;

         rcode = fits_update_key(outfp,
                                 TINT,
                                 "NAXIS   ",
                                 &value,
                                 naxis_comment,
                                 &status);

         if (status != 0)  {
            printf("\n* * * E R R O R:  Failed to modify NAXIS");
            printf(" keyword - status = %d\n", status);

            return(FAILURE);
         }

         strcpy(hdr_line, "NAXIS3  =                    1 / ");
         strcat(hdr_line, naxis3_comment);

         rcode = fits_insert_record(outfp,
                                    6,
                                    &hdr_line[0],
                                    &status);

         if (status != 0)  {
            printf("\n* * * E R R O R:  Failed to add NAXIS3");
            printf(" keyword - status = %d\n", status);

            return(FAILURE);
         }

         flush_flag = 1;

      }   *  End of 'if (template_naxis[3] == 1)' segment.  */

      /*
       *  End of CR-3860 commented out code, i.e.,
       *  the above comment would need a leading
       *  slant in order to reverse this CR.
       */

   }  /*  End of 'if (nfound == 2)' segment.  */

   else  if (nfound == 3)  {
      *pixels = naxis[0] * naxis[1] * naxis[2];

      /*
       *  Check if the keyword NAXIS3 was not
       *  included in the template file.  If
       *  so, modify the value field of the
       *  keyword NAXIS (from 3 to 2).
       */
      if (template_naxis[3] == 0)  {
         /*
          *  First make sure NAXIS3 is equal 
          *  to 1 in the input DCE file.
          */
         if (naxis[2] != 1)  {
            printf("\n* * * E R R O R:  NAXIS3 = %d in input", naxis[2]);
            printf(" in input DCE file, but missing in template.\n");
         }

         value = 2;

         rcode = fits_update_key(outfp,
                                 TINT,
                                 "NAXIS   ",
                                 &value,
                                 naxis_comment,
                                 &status);

         if (status != 0)  {
            printf("\n* * * E R R O R:  Failed to modify NAXIS");
            printf(" keyword - status = %d\n", status);

            return(FAILURE);
         }

         flush_flag = 1;

      }  /*  End of 'if (template_naxis[3] == 0)' segment.  */

   }  /*  End of 'else  if (nfound == 3)' segment.  */

   else  {
      printf("\n* * * E R R O R:  Unexpected dimension: %d\n",
              nfound);

      return(FAILURE);
   }

   /*
    *  Check if flush_flag is set to 1,
    *  which indicates we have updated
    *  the FITS header of the output
    *  file and we must flush the
    *  internal FITS buffer.
    */
   if (flush_flag == 1)  {
      rcode = fits_flush_file(outfp,
                              &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to flush");
         printf(" output file - status = %d\n", status);

         return(FAILURE);
      }
   }

   return(SUCCESS);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  copy_temp_to_out_hdr                              *** *
 *                                                                       *
 *                                                                       *
 * ***  This function copies the completed output FITS header from   *** *
 * ***  the temporary file to the output DCE (FITS) file.            *** *
 *                                                                       *
 * ***  Input:  tempfp                    - temporary header file    *** *
 * ***                                      pointer.                 *** *
 * ***          outfp                     - output DCE (FITS) file   *** *
 * ***                                      pointer.                 *** *
 * ***          temp_file_name            - name of the temporary    *** *
 * ***                                      file.                    *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  copy_temp_to_out_hdr(fitsfile*  tempfp,
                          fitsfile*  outfp,
                          char*      temp_file_name)

{

   char
         blank_record[MAX_KEYWD_CMNT + 1],
         record[MAX_KEYWD_CMNT + 1];

   int
        blank_count = 0,
        i,
        j,
        keywd_count,
        keywd_pos,
        prev_keywd  = 1,
        rcode,
        status      = 0;


   if (verbose_mode > 1)  {
      printf("\n   Copy completed output header from temp. to output file.\n");
   }

   /*
    *  Flush the internal FITS buffer.
    */
   rcode = fits_flush_file(tempfp,
                           &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to flush");
      printf(" temp. file - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Find out how many keywords there
    *  are in the temporary header file.
    */
   rcode = fits_get_hdrpos(tempfp,
                           &keywd_count,
                           &keywd_pos,
                           &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to get temp. file");
      printf(" header info. - status = %d\n", status);

      return(FAILURE);
   }

   /*
    *  Loop through the keyword records
    *  in the temporary header file and
    *  copy the records to the output
    *  DCE file unless they are part
    *  of a blank header block.
    */
   for (i = 0; i < keywd_count; i++)  {
      rcode = fits_read_record(tempfp,
                               (i + 1),
                               &record[0],
                               &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to read keyword record");
         printf(" from temp. file - status = %d\n", status);

         return(FAILURE);
      }

      if ( (strlen(record) != 0) &&
           (strncmp(record, "          ", 10) != 0) )  {
         /*
          *  The curent keyword record is neither
          *  totally blank nor does it begin with
          *  ten blank characters, i.e., this is
          *  a real keyword record that should be
          *  copied to the output header.  First
          *  check if we have a keyword block
          *  comment (with accompanying blank
          *  records) to output.
          */
         if (prev_keywd == 0)  {
            rcode = fits_write_record(outfp,
                                      &blank_record[0],
                                      &status);

            if (status != 0)  {
               printf("\n* * * E R R O R:  Failed to write keyword record");
               printf(" to output DCE file - status = %d\n", status);

               return(FAILURE);
            }

            strcpy(blank_record, "                    ");
            strcat(blank_record, "                    ");
            strcat(blank_record, "                    ");
            strcat(blank_record, "                    ");

            for (j = 0; j < blank_count; j++)  {
               rcode = fits_write_record(outfp,
                                         &blank_record[0],
                                         &status);

               if (status != 0)  {
                  printf("\n* * * E R R O R:  Failed to write keyword record");
                  printf(" to output DCE file - status = %d\n", status);

                  return(FAILURE);
               }
            }

            prev_keywd = 1;

         }  /*  End of 'if (prev_keywd == 0)' segment.  */

         rcode = fits_write_record(outfp,
                                   &record[0],
                                   &status);

         if (status != 0)  {
            printf("\n* * * E R R O R:  Failed to write keyword record");
            printf(" to output DCE file - status = %d\n", status);

            return(FAILURE);
         }

      }  /*  End of 'if ( (strlen(record) != 0) &&                     */
         /*               (strncmp(record, "          ", 10) != 0) )'  */
         /*  segment.                                                  */

      else  {
         /*
          *  The curent keyword record is either a
          *  totally blank line, a regular comment
          *  line (beginning with ten blanks), or
          *  a keyword block comment (beginning
          *  with ten blanks followed by a forward
          *  slant).
          */  
         if ( (strlen(record) == 0) ||
              (strncmp(record, "          /", 11) != 0) )  {
            /*
             *  We have one of the first two types.
             *  Check if we should output this
             *  record or if we should count blank
             *  lines.
             */
            if (prev_keywd == 1)  {
               rcode = fits_write_record(outfp,
                                         &record[0],
                                         &status);

               if (status != 0)  {
                  printf("\n* * * E R R O R:  Failed to write keyword record");
                  printf(" to output DCE file - status = %d\n", status);

                  return(FAILURE);
               }
            }
            else  {
               blank_count++;
            }
         }  /*  End of 'if ( (strlen(record) == 0) ||                      */
            /*               (strncmp(record, "          /", 11) != 0) )'  */
            /*  segment.                                                   */

         else  {
            /*
             *  We have a keyword block comment.
             *  Save this record in case we
             *  need to output it later.
             */
            prev_keywd  = 0;
            blank_count = 0;

            strcpy(blank_record, record);
         }

      }  /*  End of 'else' (i.e., if ( (strlen(record) == 0) ||         */
         /*                (strncmp(record, "          ", 10) == 0) )'  */
         /*  segment.                                                   */

   }  /*  End of 'for (i = 0; i < keywd_count; i++)' loop.  */

   /*
    *  Delete the temporary header file.
    */
   rcode = fits_delete_file(tempfp,
                            &status);

   if (status != 0)  {
      printf("\n* * * E R R O R:  Failed to delete temporary");
      printf(" header file %s - status = %d\n", &temp_file_name[1], status);

      return(FAILURE);
   }

   return(SUCCESS);

}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  copy_image_data                                   *** *
 *                                                                       *
 *                                                                       *
 * ***  This function copies the image (binary) data from the input  *** *
 * ***  DCE (FITS) file to the output DCE (FITS) file.               *** *
 *                                                                       *
 * ***  Input:  infp                      - input DCE (FITS) file    *** *
 * ***                                      pointer.                 *** *
 * ***          outfp                     - output DCE (FITS) file   *** *
 * ***                                      pointer.                 *** *
 * ***          bitpix                    - bits per pixel in the    *** *
 * ***                                      image data.              *** *
 * ***          pixels                    - number of pixels in the  *** *
 * ***                                      (binary) image data.     *** *
 *                                                                       *
 * ***  Output:  None.                                               *** *
 *                                                                       *
 * ***  Return values:  SUCCESS           - successful processing    *** *
 * ***                                      completed.               *** *
 * ***                  FAILURE           - processing was halted    *** *
 * ***                                      due to error (specified  *** *
 * ***                                      in stdout log).          *** *
 *                                                                       *
 * ***  Original code written by Mikael Aronsson.                    *** *
 *                                                                       *
 * ********************************************************************* */

int  copy_image_data(fitsfile*  infp,
                     fitsfile*  outfp,
                     int        bitpix,
                     long       pixels)

{

   float
          *img_data_float;

   int
        anynul,
        rcode,
        status = 0;

   short
          *img_data_short;


   /*
    *  We have separate branches for
    *  32-bit floating point and
    *  16-bit (short) integer data.
    */
   if (bitpix == FLOAT_IMG)  {
      /*
       *  Allocate an image data array.
       */
      img_data_float = (float*) malloc(pixels * sizeof(float) );

      if (img_data_float == NULL)  {
         printf("\n* * * E R R O R:  Failed to allocate image data array.\n");

         return(FAILURE);
      }

      /*
       *  Read the image data from the input file.
       */
      rcode = fits_read_img(infp,
                            TFLOAT,
                            1,
                            pixels,
                            0,
                            img_data_float,
                            &anynul,
                            &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to read image");
         printf(" data - status = %d\n", status);

         return(FAILURE);
      }

      /*
       *  Write the image data to the output file.
       */
      rcode = fits_write_img(outfp,
                             TFLOAT,
                             1,
                             pixels,
                             img_data_float,
                             &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to write image");
         printf(" data - status = %d\n", status);

         return(FAILURE);
      }

   }  /*  End of 'if (bitpix == FLOAT_IMG)' segment.  */

   else  {
      /*
       *  Allocate an image data array.
       */
      img_data_short = (short*) malloc(pixels * sizeof(short) );

      if (img_data_short == NULL)  {
         printf("\n* * * E R R O R:  Failed to allocate image data array.\n");

         return(FAILURE);
      }

      /*
       *  Read the image data from the input file.
       */
      rcode = fits_read_img(infp,
                            TSHORT,
                            1,
                            pixels,
                            0,
                            img_data_short,
                            &anynul,
                            &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to read image");
         printf(" data - status = %d\n", status);

         return(FAILURE);
      }

      /*
       *  Write the image data to the output file.
       */
      rcode = fits_write_img(outfp,
                             TSHORT,
                             1,
                             pixels,
                             img_data_short,
                             &status);

      if (status != 0)  {
         printf("\n* * * E R R O R:  Failed to write image");
         printf(" data - status = %d\n", status);

         return(FAILURE);
      }

   }  /*  End of 'else' (i.e., 'if (bitpix != FLOAT_IMG)') segment.  */

   return(SUCCESS);
}


/* ********************************************************************* *
 *                                                                       *
 * ***  Function:  LOCAL_COPY_OF_strlcpy                             *** *
 *                                                                       *
 *                                                                       *
 * ***  This is a LOCAL COPY of the function strlcpy.  The reason    *** *
 * ***  for including this function here, is that while strlcpy is   *** *
 * ***  recognized under Solaris, it is not defined under Linux, so  *** *
 * ***  by including the function here, this program will run under  *** *
 * ***  BOTH Solaris and Linux.                                      *** *
 *                                                                       *
 * ********************************************************************* */

/*      $OpenBSD: strlcpy.c,v 1.8 2003/06/17 21:56:24 millert Exp $     */

/*
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char *rcsid = "$OpenBSD: strlcpy.c,v 1.8 2003/06/17 21:56:24 millert Exp 
$";
#endif /* LIBC_SCCS and not lint */

#include <sys/types.h>
#include <string.h>

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t
LOCAL_COPY_OF_strlcpy(char *dst, const char *src, size_t siz)
{
        register char *d = dst;
        register const char *s = src;
        register size_t n = siz;

        /* Copy as many bytes as will fit */
        if (n != 0 && --n != 0) {
                do {
                        if ((*d++ = *s++) == 0)
                                break;
                } while (--n != 0);
        }

        /* Not enough room in dst, add NUL and traverse rest of src */
        if (n == 0) {
                if (siz != 0)
                        *d = '\0';              /* NUL-terminate dst */
                while (*s++)
                        ;
        }

        return(s - src - 1);    /* count does not include NUL */
}

