#define PROGRAM "imsubtract"
#define VERSION "1.2"

/**UPDATES
070822 1.2 Add input _unc's in quadriture to get out_unc
           Eliminate warning messages: (1) use im_interface.h and param_interface.h
            (2) specify return type for subroutines (3) get calling paramter types right
060830 1.1a In call to im_hdr, add a third parameter (=0) (CR-7348)
050726 1.1 For output _unc, just pass on in1_unc [Bob]
050413 1.0 Initial version [Bob]
This subtracts two images, including mask and uncertainty files

 NOTE:  uses program downlink/pipeline/include/tools/nl_and_params.c
**/

#include <math.h>
/* Added - TYB 6/1/09 */
#include "fitsio.h"

/* Termination codes as specified in:
**     SIRTF Downlink Software Development Guidelines  20010515
**     (file: /ssc/pipe/dlmtg/sw_dev_guidelines)
*/
#define TERMINATE_NORMAL         0
#define TERMINATE_WARNING       32
#define TERMINATE_ERROR	        64

#ifndef FALSE
#define FALSE		         0
#define TRUE              (!FALSE)
#endif

#define MAXFILENAMESIZE        256
#define MAXIMAGESIZE           20000

#define DATATYPE_D "double"
#define DATATYPE_S "short"

#define NL_START     "&IMSUBTRACT"
#define NL_END              "&END"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "im_interface.h"
#include "param_interface.h"


/* PARAMETERS (GLOBAL) */
  char filename_in1_image[MAXFILENAMESIZE];
  char filename_in2_image[MAXFILENAMESIZE];
  char filename_out_image[MAXFILENAMESIZE];
  char filename_in1_unc[MAXFILENAMESIZE];
  char filename_in2_unc[MAXFILENAMESIZE];
  char filename_out_unc[MAXFILENAMESIZE];
  char filename_in1_mask[MAXFILENAMESIZE];
  char filename_in2_mask[MAXFILENAMESIZE];
  char filename_out_mask[MAXFILENAMESIZE];
  int  verbose;

/* GLOBAL DATA */
  int n_unc_files;
  int n_mask_files;
  double array_in1[MAXIMAGESIZE];
  double array_in2[MAXIMAGESIZE];
  double array_out[MAXIMAGESIZE];
  int fid_in1, fid_in2, fid_out; /* im file i/o FID */
  int npixels = 0; /* number of pixels in array (0 if not set yet) */
  int naxis1, naxis2; /* dimensions of image */

/**************************************************************************/
void FATAL_char (char *message, char *param) {
  /* prints error msg, with printf parameter; then exits with error status */
  char msg_nl[300];
  strcpy(msg_nl, message);
  strcat(msg_nl, "\n");
  printf(msg_nl, param);
  exit(TERMINATE_ERROR);
}


/**************************************************************************/
int process_parameters(int argc, char **argv)
{
  int stat;

  param(filename_in1_image, "-i1 IN1_IMAGE name of first input image","infile");
  param(filename_in2_image,"-i2 IN2_IMAGE name of second input image","infile");
  param(filename_out_image, "-o  OUT_IMAGE name of output image", "outfile");

  param(filename_in1_unc, "-i1_unc IN1_UNC name of first input uncertainty","infile - NONE");
  param(filename_in2_unc,"-i2_unc IN2_UNC name of second input uncertainty","infile - NONE");
  param(filename_out_unc, "-o_unc  OUT_UNC name of output uncertainty", "outfile - NONE");

  param(filename_in1_mask,  "-i1_mask IN1_MASK name of first input mask file","infile - NONE");
  param(filename_in2_mask, "-i2_mask IN2_MASK name of second input mask file","infile - NONE");
  param(filename_out_mask,  "-o_mask  OUT_MASK name of output mask file", "outfile - NONE");


  param(&verbose, "-v VERBOSE level of printed messages", "int - 1");
  /**ENDPARAMS **/
  
  stat = nl_process(argc, argv, PROGRAM, VERSION, NL_START, NL_END);
  if (stat != 0) return stat;

  stat = param_begin(argc, argv, PROGRAM, VERSION);
  if (stat != 0) return stat;

  param_print();   /* Print the parameter settings to stdout. */

  /* check for high level consistency of parameters */
  n_unc_files = 0;
  if (strcmp(filename_in1_unc,"NONE")) n_unc_files++;
  if (strcmp(filename_in2_unc,"NONE")) n_unc_files++;
  if (strcmp(filename_out_unc,"NONE")) n_unc_files++;
  if (n_unc_files>0 && n_unc_files<3)
    FATAL_char("PARAM ERROR: For unc files, all or none must be given", "");
  n_mask_files = 0;
  if (strcmp(filename_in1_mask,"NONE")) n_mask_files++;
  if (strcmp(filename_in2_mask,"NONE")) n_mask_files++;
  if (strcmp(filename_out_mask,"NONE")) n_mask_files++;
  if (n_mask_files>0 && n_mask_files<3)
    FATAL_char("PARAM ERROR: For mask files, all or none must be given", "");
  
  return 0;
}


/**************************************************************************/
int read_image (char *filename, double *array, char *datatype)
{
  int fid;
  int nax1, nax2;
  int i;

  /* open image */
  fid = im_open(filename, "r");
  if (fid < 0) FATAL_char("Error opening or reading file %s", filename);

/* DEBUG */
/* 
printf ("DEBUG: setting read nans=0\n");
im_nans(fid, 0);
*/

  /* get size of axes */
  im_rkey_i(fid, &nax1, "NAXIS1");
  im_rkey_i(fid, &nax2, "NAXIS2");
  if (npixels <= 0) {
    naxis1 = nax1;
    naxis2 = nax2;
    npixels = naxis1 * naxis2;
    if (npixels > MAXIMAGESIZE)
      FATAL_char ("Image is too large", "");
  }
  else {
    if (nax1 != naxis1) 
      FATAL_char ("Image dimension mis-match NAXIS1 in file %s", filename);
    if (nax2 != naxis2) 
      FATAL_char ("Image dimension mis-match NAXIS2 in file %s", filename);
  }

  if (!strcmp(datatype, DATATYPE_D)) {
     i = im_rpix_d(fid, array, npixels);
     if (i <0) FATAL_char("Trouble reading pixels of file %s", filename); 

/* DEBUG */
/* if (!finite(array[i])) */
/*****
for (i=0; i<npixels; i++)
 printf ("DEBUG: array[%2d] = %14lf %#16lx\n", i, array[i], array[i]);
****/

  }

  else if (!strcmp(datatype, DATATYPE_S)) {
     i = im_rpix_s(fid, (short*) array, npixels);
     if (i <0) FATAL_char("Trouble reading pixels of file %s", filename); 
  }

  else FATAL_char("PROGRAMMING ERROR: read_image: bad datatype: %s", datatype);

  return fid;
}


/** DEBUG *** DEBUG ******* DEBUG *****************************************/
/*double get_one() {double valval; valval = 0.5; valval+=0.5; return valval;} */


/**************************************************************************/
int write_image (char *filename, double *array, char *datatype, int fid_i1)
{
  int i;
  int fid;
  int fid_in_array[3];
  char header_card[200];
  char datetime[200];
  char *timeStamp1();

/* DEBUG */
/****
double one, zero, inf, inf2;
union {double dbl; int nnn;} flex;
zero = 0.0;
one = get_one();
printf ("DEBUG: one=%8x\n", one);
flex.nnn = 0xbf800000;
printf ("DEBUG: flex.nnn =%8x\n", flex.nnn);
printf ("DEBUG: flex.dbl =%8x\n", flex.dbl);
printf ("DEBUG: flex.dbl =%lf\n", flex.dbl);
inf2 = flex.dbl;
printf ("DEBUG: inf2=%8x\n", inf2);
****/



  /* Open file and copy header from file in1 */
  fid = im_open(filename, "w");
  if (fid < 0) FATAL_char("Error opening file %s", filename);

/* DEBUG */
/*
printf ("DEBUG: setting write nans=0\n");
im_nans(fid, 0);
*/

  fid_in_array[0] = fid_i1;
  fid_in_array[1] = 0;
  i = im_hdr(fid, fid_in_array, 0);

  strcpy(datetime, timeStamp1());
  sprintf(header_card,"HISTORY IMSUBTRACT               v       %s,  ran %s", VERSION, datetime);
  im_wkey_t(fid, header_card, "end");

  if (!strcmp(datatype, DATATYPE_D)) {

/* DEBUG */
/* for (i=0; i<npixels; i++) */
/* if (!finite(array[i])) { */
/* printf ("DEBUG2: array[%d] = %16x\n", i, array[i]); */

    i = im_wpix_d(fid, array, npixels);
    if (i < npixels) FATAL_char("Error writing pixels to file %s", filename);
  }
  else if (!strcmp(datatype, DATATYPE_S)) {
    i = im_wpix_s(fid, (short*) array, npixels);
    if (i < npixels) FATAL_char("Error writing pixels to file %s", filename);
  }
  else FATAL_char("PROGRAMMING ERROR: write_image: bad datatype: %s", datatype);

  return fid;
}


/**************************************************************************/
void write_header_and_close (int f_in1, int f_in2, int f_out)
{
  im_close(f_in1);
  im_close(f_in2);
  im_close(f_out);
}


/**************************************************************************/
void compute_image(double *a, double *b, double *c)
{
  int i;
  for (i=0; i<npixels; i++) { 
    c[i] = a[i] - b[i];
/* printf ("%lf %lf %lf\n",c[i], a[i], b[i]); */
  }
}


/**************************************************************************/
void compute_unc(double *a, double *b, double *c)
{
  int i;
  for (i=0; i<npixels; i++) { c[i] = sqrt (a[i]*a[i] + b[i]*b[i]); }
}


/**************************************************************************/
void compute_mask(short *a, short *b, short *c)
{
  int i;
  for (i=0; i<npixels; i++) { c[i] = a[i] | b[i]; }
}


/**************************************************************************/
int extract(int argc, char **argv)
{
  /*  Add CFITSIO parameters - TYB 6/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int numofkeys = 0;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";
  char myoutuncfile[MAXFILENAMESIZE] = "";
  char header_card[200];
  char datetime[200];
  char *timeStamp1();

  process_parameters(argc, argv);

  fid_in1 = read_image (filename_in1_image, array_in1, DATATYPE_D);
  fid_in2 = read_image (filename_in2_image, array_in2, DATATYPE_D);
  compute_image(array_in1, array_in2, array_out);
  im_close(fid_in1);
  im_close(fid_in2);

  /* Replace with CFITSIO calls - TYB 6/1/09 */

  strcat(myoutfile, "!");
  strcat(myoutfile, filename_out_image);
  fits_create_file(&ffp_FITS,myoutfile,&I_fits_return_status);
  fits_open_file(&ffp_FITS_In,filename_in1_image,READONLY,&I_fits_return_status);
  fits_copy_hdu(ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);
  fits_close_file(ffp_FITS_In, &I_fits_return_status);
  fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
  strcpy(datetime, timeStamp1());
  sprintf(header_card,"HISTORY IMSUBTRACT               v       %s,  ran %s", VERSION, datetime);
  fits_write_record(ffp_FITS, header_card,&I_fits_return_status);
  fits_flush_file(ffp_FITS, &I_fits_return_status);
  fits_write_img(ffp_FITS,TDOUBLE,1, npixels, array_out, &I_fits_return_status);               
  fits_flush_file(ffp_FITS, &I_fits_return_status);
  fits_close_file(ffp_FITS, &I_fits_return_status);

  //  fid_out = write_image(filename_out_image, array_out, DATATYPE_D, fid_in1);
  //  write_header_and_close (fid_in1, fid_in2, fid_out);

  if (n_unc_files != 0) {
    fid_in1 = read_image (filename_in1_unc, array_in1, DATATYPE_D);
    fid_in2 = read_image (filename_in2_unc, array_in2, DATATYPE_D);
    compute_unc(array_in1, array_in2, array_out);
    im_close(fid_in1);
    im_close(fid_in2);

   /* Replace with CFITSIO calls - TYB 6/1/09 */

    strcat(myoutuncfile, "!");
    strcat(myoutuncfile, filename_out_unc);
    fits_create_file(&ffp_FITS,myoutuncfile,&I_fits_return_status);
    fits_open_file(&ffp_FITS_In,filename_in1_unc,READONLY,&I_fits_return_status);
    fits_copy_hdu(ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);
    fits_close_file(ffp_FITS_In, &I_fits_return_status);
    fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
    strcpy(datetime, timeStamp1());
    sprintf(header_card,"HISTORY IMSUBTRACT               v       %s,  ran %s", VERSION, datetime);
    fits_write_record(ffp_FITS, header_card, &I_fits_return_status);
    fits_flush_file(ffp_FITS, &I_fits_return_status);
    fits_write_img(ffp_FITS,TDOUBLE,1, npixels, array_out, &I_fits_return_status);               
    fits_flush_file(ffp_FITS, &I_fits_return_status);
    fits_close_file(ffp_FITS, &I_fits_return_status);

    //    fid_out = write_image(filename_out_unc, array_out, DATATYPE_D, fid_in1);
    //    write_header_and_close (fid_in1, fid_in2, fid_out);
  }

  if (n_mask_files != 0) {
    fid_in1 = read_image (filename_in1_mask, array_in1, DATATYPE_S);
    fid_in2 = read_image (filename_in2_mask, array_in2, DATATYPE_S);
    compute_mask((short *)array_in1, (short *)array_in2, (short *)array_out);
    fid_out = write_image(filename_out_mask, array_out, DATATYPE_S, fid_in1);
    write_header_and_close (fid_in1, fid_in2, fid_out);
  }

  return TERMINATE_NORMAL;
}


/**************************************************************************/
int main(argc, argv)
int argc;
char **argv;
{
  int status;
  status = extract(argc, argv);
  if (status < 0) return TERMINATE_ERROR;
  param_exit(status, PROGRAM);
}
