#define PROGRAM "irs_tune"
#define VERSION "3.3"

/*
 This module applies fluxcal factors and/or tuning coefficients to a spectrum.
*/

/* NOTES
 - uses program downlink/pipeline/include/tools/nl_and_params.c
 - need to check for a3, a4 in tunes table
*/

/**UPDATES
060530 3.3 option to call irs_retune twice
060130 3.2 Enlarged syscall_string to 5000 characters
051116 3.1 Added option to make system call to irs_retune (for SPICE)
050304 3.0a Changed output precision to 6 - Bob
050110 3.0 Fixed to work with C lib tblio  (IK)
031126 9.0 Fixed to work correctly with tablio (was passing flux unchanged)
031126 9.0 Changed eames of output columns
031126 9.0 Added option to apply BOTH fluxcon and tuning
031126 9.0 Changed default up-down mode to "tune_up"
031126 9.0 Changed default fluxcon-tune-both mode to 3 (both)
030615 2.0 Added a 'tune table' as input, defining photometric tuning and flux
           conversion. 
030615 2.0 Propagated all headers from Extract output.
030115 1.3 Allow a polynomial of up to 5 degree (JL)
030115 1.3 Added absolute flux conversion mode (JL)
020926 1.2 Added namelist facility (CW)
020926 1.2 Named TERMINATE_* codes (CW)
010303 1.0 New program design
**/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <tbl.h>

/******* GLOBAL DATA *************************************/

/* Termination codes as specified in:
**     SIRTF Downlink Software Development Guidelines  20010515
**     (file: /ssc/pipe/dlmtg/sw_dev_guidelines)
*/
#define TERMINATE_NORMAL        0
#define TERMINATE_WARNING      32
#define TERMINATE_ERROR	       64

#ifndef FALSE
#define FALSE		        0
#define TRUE             (!FALSE)
#endif

#define MAXFILENAMESIZE       256
#define MAX_ORDERS             30
#define NL_START       "&IRS_TUNE"
#define NL_END              "&END"

#define COEF_MAX 5

int tio_in, tio_out; /* table io table id */
int tstat, row;
struct tbl_filinfo *ifinfo, ofinfo;

int Verbose =1;

struct tune_item {
  int order;
  double wave0;
  double a[COEF_MAX+1];
  double fluxcon;
};

/******* parse_params ROUTINE **********************/
/* parse command line prameters */
int parse_params(
     int argc, char** argv,
     char* in_table, char* tune_table, char* out_table,
     int* _mode, int* _applymode, float* _noise0, float* _noise1,
     char* _recom, char* _retab, char* _reout, char* _retab2, char* _reout2 )
{
  int status;
  char ascii_mode[24];
  char nl_file[MAXFILENAMESIZE];

  /**PARAMS **/
  param(in_table,  "-i INFILE name of input table",              "infile jy.tbl");
  param(tune_table,"-t INFILE name of tuning table",                "infile tune.tbl");
  param(out_table, "-o OUTFILE name of output table",            "outfile jy.tbl");
  param(ascii_mode,"-m UP_DOWN_MODE mode (tune_up OR tune_down)",              "string - tune_up");
  param(_noise0,   "-noise0 NOISE0 sigma of offset noise to add","float Jy 0.0");
  param(_noise1,   "-noise1 NOISE1 sigma of gain noise to add",  "float - 0.0");
  param(&Verbose,  "-v VERBOSE level of printed messages",       "int - 1");
  param(_applymode, "-a APPLY (1=fluxcon, 2=tune, 3=both)", "int - 3");
  param(_recom, "-recom RETUNE_COMMAND full path to executable", "string - -");
  param(_retab, "-retab RETUNE_TABLE retune table", "infile - -");
  param(_reout, "-reout RETUNE_OUTFILE retune output file", "outfile - -");
  param(_retab2, "-retab2 RETUNE_TABLE2 second retune table", "infile - -");
  param(_reout2, "-reout2 RETUNE_OUTFILE2 second retune output file", "outfile - -");
  param(nl_file ,  "-n NAMELIST_FILE name of Namelist file",     "infile - namelist.nl");
  /**ENDPARAMS **/
  
  status = nl_process(argc, argv, PROGRAM, VERSION, NL_START, NL_END);
  if (status != 0) return status;

  status = param_begin(argc, argv, PROGRAM, VERSION);
  if (status != 0) return 1;

  param_print();   /* Print the parameter settings to stdout. */
  
  if (strcmp(ascii_mode, "tune_up") ==0) { *_mode = 1; }
  else if (strcmp(ascii_mode, "tune_down") ==0) { *_mode = -1; }
  else {
    printf("ERROR: requested mode not supported: %s\n", ascii_mode);
    return 1;
  }

  if (*_applymode < 1 || *_applymode > 3) {
    printf("Illegal value for applymode: must be 1, 2, or 3\n");
    return 1;
  }

  return 0;
}


/******* tio_getind SUB-ROUTINE ************************************/

int tio_getind(int iu, char* col_name, int *ind)
/* get index of the named column */
/* return 0=OK  non-zero = trouble */
{
  *ind = tbl_column(iu, col_name);
  if (*ind < 0) {
    printf("ERROR: bad in table file: missing column: %s\n", col_name);
    return -1;
  }
  return 0;
}


/******** read_tune_file SUB-ROUTINE *******************************/
int read_tune_file(char* file_name, struct tune_item tunes[], double* _fcon_unc, double* _tune_unc, int* _max_coef)
{

  int nparms;
  int ind_order, ind_wave0, ind_fluxcon; /* index to columns in table */
  int ind_fcon_unc, ind_tune_unc;
  int ind_a[COEF_MAX+1];
  char tmp_str[12];

  int status;
  int i, j;

  if (Verbose >=2) printf("reading %s ...\n", file_name);

  /* open file and get column numbers */
  tio_in =  tbl_open(file_name, &ifinfo);
  tstat = tbl_fileinfo(tio_in, &ifinfo);
  if (tstat !=  0)
  {
    printf("ERROR: Unable to read file: %s\n", file_name);
    return 1;
  }
  if( tio_getind(tio_in, "order", &ind_order) ) return 1;
  if( tio_getind(tio_in, "key_wavelength", &ind_wave0) ) return 1;
  if( tio_getind(tio_in, "fluxcon", &ind_fluxcon) ) return 1;


  *_max_coef = -1;
  for (j=0; j<= COEF_MAX; j++)
  {
    sprintf(tmp_str, "a%d", j);
    ind_a[j] = tbl_column(tio_in, tmp_str);
    if (ind_a[j] >=0) *_max_coef = j;
    else break;
  }
  if (*_max_coef < 1) 
  {
    printf("ERROR: There is no 'a0' column in tune file %s\n", file_name);
    return 1;
  }

  /* read in all of the data */
  row = 0;
  for(i=0; i<MAX_ORDERS; i++) {
    tstat = tbl_colval(tio_in, row, ind_order, TBL_INTEGER, &tunes[i].order);
    tstat = tbl_colval(tio_in, row, ind_fluxcon,TBL_DOUBLE,  &tunes[i].fluxcon);
    tstat = tbl_colval(tio_in, row, ind_wave0, TBL_DOUBLE, &tunes[i].wave0);

    if (tstat == -4) break;
    else if (tstat !=  0)
    {
      printf("ERROR: trouble while reading file: %s\n", file_name);
      return 1;
    }
    for (j=0; j<= *_max_coef; j++)
    {
      if (ind_a[j] >=0) tbl_colval(tio_in, row, ind_a[j], TBL_DOUBLE, &tunes[i].a[j]);
      else tunes[i].a[j] = 0.0;
    }

    /* get uncertainty values if first order */
    if (i == 0) {
      *_fcon_unc = -1.0;
      *_tune_unc = -1.0;
      if(tio_getind(tio_in, "fluxcon_sigma", &ind_fcon_unc) ==0) {
        tstat = tbl_colval(tio_in, row, ind_fcon_unc, TBL_DOUBLE, _fcon_unc);
      }
      if(tio_getind(tio_in, "a0_sigma",      &ind_tune_unc) ==0) {
        tstat = tbl_colval(tio_in, row, ind_tune_unc, TBL_DOUBLE, _tune_unc);
      }
    }

    row++;
  }

  tunes[i].order = -1; /* end of tune data */

  if (Verbose >= 4) {
    printf("\nTUNES\n");
    for(i=0; tunes[i].order >=0; i++)
    {
      printf ("order=%2d\n", tunes[i].order);
      printf ("key_wave=%-g\n", tunes[i].wave0);
      printf ("fluxcon=%-g\n", tunes[i].fluxcon);
      for (j=0; j<= *_max_coef; j++)
         printf(" a[%d]=%12.5e", j, tunes[i].a[j]);
      printf ("\n\n");
    }
    printf ("\n");
  }

  return 0;
}

/******** find_tunes_index SUB-ROUTINE *******************************/
int find_tunes_index(int order, struct tune_item tunes[], int* _tunes_index)
{
  static int last_index = 0; /* look here first for efficiency */
  int i;

  /* check if it is the same as the previous one */
  if (tunes[last_index].order == order) {
    *_tunes_index = last_index;
    return 0;
  
  }

  for (i=0; i<MAX_ORDERS; i++)
  {
    if (tunes[i].order < 0) {
      printf("ERROR: order %i not in tune table\n", order);
      return 1;
    }
    if (tunes[i].order == order) {
      *_tunes_index = i;
      last_index = i;
      return 0;
    }
  }

  printf("ERROR: order not found in tunes table: %d\n", order);
  return 1;
}


/******** compute_factor SUB-ROUTINE *******************************/
double compute_factor(double  wavelength, int applymode, struct tune_item *_t, int max_coef)
/* returns 0.0 if error */
{
  double factor;
  int j;
  double w;
  double wavetothej;

  if (applymode != 1) {
    factor = _t->a[0];
    w = wavelength - _t->wave0;
    wavetothej = w;
    for (j=1; j<=max_coef; j++)
    {
      factor += _t->a[j] * wavetothej;
      wavetothej *= w;
    }
  }
  else {
    factor = 1.0;
  }

  if (applymode != 2) {
    factor *= _t->fluxcon;
  }

  return(factor);
}


/******** do_update ROUTINE *******************************/
int do_update(char* in_file, char* out_file, struct tune_item tunes[], 
            int max_coef, int mode, int applymode, double noise0, double noise1,
            double fcon_unc, double tune_unc)
{
  int i;
  int status;
  int nparms;
  int tio_in; /* tio file ID*/
  int tio_out;
  int nrows;
  int ind_order, ind_wavelength, ind_flux, 
      ind_error, ind_bitflag;                 /* index to columns in table */
  int ind_out_flux, ind_out_error, ind_out_bitflag;  /* index to columns out table */
  char** hist_cards;
  int order;
  double wavelength;
  double flux;
  double error;
  int bitflag;
  int tunes_index;
  double factor;
  FILE *outfile;

  int *bitflag_out = 0;
  float *flux_out = 0;
  float *error_out = 0;


  if (Verbose >=2) printf("doing irs_tune %s -> %s ...\n", in_file, out_file);

  /* Open input and output and link them */
  tio_in = tbl_open(in_file, &ifinfo); 
  tstat   = tbl_fileinfo(tio_in, &ifinfo);
  if (tstat != 0) {
     printf("Cannot open table for reading: [%s]\n",in_file);
     exit(64); 
  }
  nparms = ifinfo->ncols;
  
  /* find column indexes */
  if(tio_getind(tio_in, "order",      &ind_order) ) return 1;
  if(tio_getind(tio_in, "wavelength", &ind_wavelength) ) return 1;
  if(tio_getind(tio_in, "flux",       &ind_flux) ) ind_flux = -999;
  if(tio_getind(tio_in, "error",      &ind_error)) return 1;
  if(tio_getind(tio_in, "bit-flag",   &ind_bitflag) ) return 1;

  /* Kludge to support (incorrect) column name "dn" until EXTRACT is fixed */
  if (ind_flux == -999) {
    printf("... using column 'dn' instead ...\n");
    if(tio_getind(tio_in, "dn",         &ind_flux) ) return 1;
  }

  /* open the output file */
  tio_out = tbl_openw(out_file, &ofinfo);
  tstat   = tbl_fileinfo(tio_out, &ofinfo);
  if (tstat != 0) {
     printf("Cannot open table for writing: [%s]\n",out_file);
     exit(64); 
  }

  tstat = tbl_link(tio_in, tio_out);

  tbl_delcol(tio_out, ind_flux);
  tbl_delcol(tio_out, ind_error);
  tbl_delcol(tio_out, ind_bitflag);

  tbl_setcol(tio_out, "flux", "real", " ", "Jy", 0., 0.,20,6, &ind_out_flux);
  tbl_setcol(tio_out, "error", "real", " ", "Jy", 0., 0.,20,6, &ind_out_error);
  tbl_setcol(tio_out, "bit-flag", "int", " ", " ", 0., 0.,20,0, &ind_out_bitflag);

  /* Add uncertainty cards */
printf ("fcon_unc = %lf\n", fcon_unc);
  if (fcon_unc > 0)
     tbl_setkey(tio_out, "FCON_UNC", "real", fcon_unc, "");       
  if (tune_unc > 0)
     tbl_setkey(tio_out, "TUNE_UNC", "real", tune_unc, "");       

  /* Add HISTORY cards */
  param_hcards(&hist_cards);
  for (i=0; hist_cards[i]; i++)
  {
    tbl_setkey(tio_out, "HISTORY", "char", hist_cards[i], "");       
  }

  tstat = tbl_numcheck(tio_in, &nrows);
  bitflag_out = (int *) malloc(nrows* sizeof(int));
  flux_out = (float *) malloc(nrows* sizeof(float));
  error_out = (float *) malloc(nrows* sizeof(float));
  if (bitflag_out == NULL || flux_out == NULL
      || error_out == NULL)
 {
   printf("Error in allocating memory for output table params\n");
   return TERMINATE_ERROR;
 }

  /* read, tune, write */
  row = 0;
  for (i=0; ; i++)
  {
    tstat = tbl_colval(tio_in, row, ind_order, TBL_INTEGER, &order);
    tstat = tbl_colval(tio_in, row, ind_flux,TBL_DOUBLE,  &flux);
    tstat = tbl_colval(tio_in, row, ind_wavelength, TBL_DOUBLE, &wavelength);
    tstat = tbl_colval(tio_in, row, ind_error, TBL_DOUBLE, &error);
    tstat = tbl_colval(tio_in, row, ind_bitflag, TBL_INTEGER, &bitflag);
    if (tstat == -4) break;
    else if (tstat !=  0) {
      printf("ERROR: Trouble reading %s near data card %d\n", in_file, i+1);
      return(1);
    }

    if (find_tunes_index(order, tunes, &tunes_index)) return 1;

    factor = compute_factor(wavelength,applymode,&tunes[tunes_index],max_coef);
    if (factor == 0.0) return 1;

    if (mode>0) { /* tune_up */
      flux  *= factor;
      error *= factor;
    }
    else { /* tune_down */
      flux  /= factor;
      error /= factor;
      if (noise1 != 0.0) flux *= (1.0 + noise1 * srang()); /* add gain noise */
      if (noise0 != 0.0) flux += noise0 * srang(); /* add offset noise */
    }

/*** save output arrays ***/
    error_out[row] = error;
    flux_out[row] = flux;
    bitflag_out[row] = bitflag;
    row++;
  }
    tbl_setcolval_f(tio_out, ind_out_flux,"real", row, flux_out);
    tbl_setcolval_f(tio_out, ind_out_error, "real", row, error_out);
    tbl_setcolval_i(tio_out, ind_out_bitflag, "int", row, bitflag_out);
    tbl_write(tio_out);

  /* close files and return */
  tbl_close(tio_in); /* close table file */
  tbl_close(tio_out);
  return 0;

}


/******** call_retune ROUTINE *******************************/
int call_retune(char *command, char *infile, char *table, char *outfile)
{
  int stat;
  char syscall_string[5000];

  /* check for required parameters */
  if (strcmp(command, "-") == 0) {
    printf("ERROR: -recom required if -retab specified\n");
    return TERMINATE_ERROR;
  }
  if (strcmp(outfile, "-") == 0) {
    printf("ERROR: -reout required if -retab specified\n");
    return TERMINATE_ERROR;
  }

  /* call retune */
  sprintf(syscall_string, "%s -i %s -t %s -o %s", 
           command, infile, table, outfile);
  printf("\nSYSTEM CALL: %s\n", syscall_string);
  stat = system(syscall_string);
  if (stat) {
    printf("ERROR: System call returned status=%d\n", stat);
    return TERMINATE_ERROR;
  }

}


/******* irs_tune HIGH LEVEL ROUTINE **********************/
irs_tune(argc, argv)
int argc;
char **argv;
{
  /* command line parameter values */
  char nl_file[MAXFILENAMESIZE];
  char in_file[MAXFILENAMESIZE];
  char tune_file[MAXFILENAMESIZE];
  char out_file[MAXFILENAMESIZE];
  int mode;  /* 1=tune,  -1=untune */
  int apply_mode; /* 1=fluxcon, 2=tune, 3=both */
  float noise0;
  float noise1;
  char retune_command[MAXFILENAMESIZE];
  char retune_table[MAXFILENAMESIZE];
  char retune_outfile[MAXFILENAMESIZE];
  char retune_table2[MAXFILENAMESIZE];
  char retune_outfile2[MAXFILENAMESIZE];

  /* data */
  struct tune_item tunes[MAX_ORDERS];
  double fcon_unc;
  double tune_unc;
  int max_coef;

  /* executable code */
  if( parse_params(argc, argv, in_file, tune_file, out_file,
       &mode, &apply_mode, &noise0, &noise1,
       retune_command, retune_table, retune_outfile,
                      retune_table2, retune_outfile2) ) return 1;

  if(read_tune_file(tune_file, tunes, &fcon_unc, &tune_unc, &max_coef) ) return 1;

  if(do_update(in_file, out_file, tunes, max_coef, mode, apply_mode, noise0, noise1, fcon_unc, tune_unc)) return 1;

if (strcmp(retune_table, "-"))
   call_retune(retune_command, out_file, retune_table, retune_outfile);
if (strcmp(retune_table2, "-"))
   call_retune(retune_command, retune_outfile, retune_table2, retune_outfile2);

  return 0;
}

/******* main MAIN **********************/
main(argc, argv)
int argc;
char **argv;
{
  int status;
  status = irs_tune(argc, argv);
  if (status < 0) return TERMINATE_ERROR;
  param_exit(status);
}
