#define PROGRAM "profile"
#define VERSION "3.6"

/*************************************************************************
 This module computes the profile in the cross-dispersion direction.

 NOTE:  uses program downlink/pipeline/include/tools/nl_and_params.c 
**************************************************************************/

/**UPDATES
061129 3.6 Output 0's instead of NaNs in profile.tbl
060929 3.5 Increase max cuts , removed subregions
060814 3.4 Replace atof("NaN") with NANVALUE (Bob)
050720 3.3 Rename global variable stat to tstat. (Ted H)
050316 3.3 computed with sliding median AR-6483 (Bob)
050111 3.2 fixed to use tblio instead of tableio (IK)
040427 3.0 fixed -ord parameter (Bob)
030617 2.0 Added FOV lookup table for SL and LL as input.           (JL)
030617 2.0 Added an option allowing the user to specify up to four pairs
           wavelengths that split a given order into subregions in 
           wavelength.                                              (JL)
030617 2.0 Added column subregion into the output table             (JL) 
030617 2.0 Propagated all keywords taken from the input BCD FITS    
           header into the header of the output profile table       (JL)
030617 2.0 Replaced pmask file with bmask file                      (JL)
020926 1.5 Modified to use integrate library instead of local copy of code (CW)
020926 1.5 Added namelist facility (CW)
020926 1.5 Named TERMINATE_* codes (CW)
011018 1.4 IRK  added 'stdev' column in output ipac table file      
011018 1.4 IRK  generated "QA" output file                          
010801 1.2d added back includes of stdlib.h and stdio.h
000920 1.2c reversed sense of 0% to 100% (now 0% is on the left)
000914 1.2b removed 0.5 pixel coordinate kludge
000828 1.2 update to new param subroutines
000814 1.2 added 0.5 pixel coordinate kludge (wavsamp vs. integrate_quad)
000810 1.2 Bob re-organized and plugged in Clare's integrate_quad subroutine
000301 1.2 Bob re-organized processing of argv/argc
000301 1.2 Bob added outputFilename (vs. just printing to stdout)
000302 1.2 Bob changed im_open from "u" to "r"
000307 1.2 Bob changed error messages from stderr to stdout (pipeline std)
**/


/* Termination codes as specified in:
**     SIRTF Downlink Software Development Guidelines  20010515
**     (file: /ssc/pipe/dlmtg/sw_dev_guidelines)
*/
#define TERMINATE_NORMAL        0
#define TERMINATE_WARNING      32
#define TERMINATE_ERROR	       64

#ifndef FALSE
#define FALSE		        0
#define TRUE             (!FALSE)
#endif

#define MAXFILENAMESIZE      256
#define MAX_CUTS            2000 /* rememeber to update param check */
#define MAX_MEDIAN_SIZE       21 /* rememeber to update param check */
#define MAX_ORDERS            15
#define NL_START       "&PROFILE"
#define NL_END             "&END"
#define PSEUDO_NAN -9999.999

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "integrate.h"
#include "tbl.h"
#include "nanvalue.h"

int iu ; /* for tableio */

char text_max_orders[40]; /* for passing to params routine */
int previous_order=-1; /* just used for verbose printout */

/* stuff for sliding_median subroutine */
int sliding_median_size=0; /* n samples for sliding_median (odd integer) */
struct median_struct {
  int sequence; /* number of values entered array so far */
  double val[MAX_MEDIAN_SIZE]; /* values to median */
  int    seq[MAX_MEDIAN_SIZE]; /* this val was Nth entered */
};
struct median_struct median_structures[MAX_CUTS+1];


/*==== sliding_median SUBROUTINE ==========================*/
double sliding_median (
 double value, /* next value to add into median */
 int index /* which sliding median working with now */

/**********************************************************
 returns median of last MEDIAN_SIZE values if there are enough values 
 otherwise, returns PSEUDO_NAN
 inputs one value to be median'd, and an index
 "index" indicates which of the many medians is being accessed.
 allows computation of many running medians at once
************************************************************/

) {

/*****
The idea is to maintain an array of sliding_median_size samples,
sorted in increasing order.  When a new value comes,
 (1) the oldest value in the array is deleted
 (2) the gap is closed (values are scooted down)
 (3) a gap is opened at the place where the new one needs to go
 (4) the new value is inserted 
Then the central value is returned as the median
******/

  int i,k;
  int old_seq=0; /* seq of old value for retirement  */
  int last_value_i; /* index of last actual value in array */
  double median;
  struct median_struct *m;

  /* access the right median struc and bump sequence */
  m = &median_structures[index];
  m->sequence ++;


  /* remove oldest value (unless array not full yet) */
  if (m->sequence > sliding_median_size) {

    /* find oldest value */
    last_value_i = sliding_median_size-1;
    old_seq =  m->sequence - sliding_median_size;
    for (k=0; k<=last_value_i; k++) {
      if (m->seq[k] ==  old_seq) break;
    }

    /* remove oldest value (scoot others down) */
    for (i=k+1; i<=last_value_i; i++) {
      m->val[i-1] = m->val[i];
      m->seq[i-1] = m->seq[i];
    }
    last_value_i--;
  }
  else {
    last_value_i = m->sequence - 2;
  }

  
  /* find where new value goes */
  for (k=0; k<=last_value_i; k++) {
    /* (find first one that is bigger) */
    if (m->val[k] >= value) break;
  }

  /* make room for new value (scoot others up) */
  last_value_i++;
  for (i=last_value_i; i>k;  i--) {
    m->val[i] = m->val[i-1];
    m->seq[i] = m->seq[i-1];
  }

  /* insert new value */
  m->val[k] = value;
  m->seq[k] = m->sequence;

  /* get median and return */
  if (m->sequence >= sliding_median_size) median = m->val[sliding_median_size/2];
  else                                   median = PSEUDO_NAN;

/*******  start DEBUG code
if (index == 0) {
 printf ("\nvalue = %lf sequence=%i old_seq=%i\n", value, m->sequence, old_seq);
 for (k=0; k<sliding_median_size; k++) {
  printf ("%2i %2i = %9.3lf\n", k, m->seq[k], m->val[k]);
 }
 printf ("median = %lf\n", median);
}
****** end DEBUG code  */

  return(median);

}


/*==== float2pct SUBROUTINE ==========================*/
float cut2pct(cut, n_cuts)
int cut;
int n_cuts;
{
 return (cut+0.5) * (100.0 / n_cuts);
}


/*==== tio_getind SUBROUTINE ==========================*/
int tio_getind(char* col_name, int *ind)
/* get index of the named column */
/* return 0=OK  non-zero = trouble */
{
  *ind = tbl_column(iu, col_name);
  if (*ind < 0) {
    printf("ERROR: bad in table file: missing column: %s\n", col_name);
    return -1;
  }
/* This feature not required now. The lib takes care of it */
/*  (*ind)--;  "feature" in tableio, requires subtract 1 */
  return TERMINATE_NORMAL;
}


/*==== profile ========================================*/
profile(argc, argv)
int argc;
char *argv[];
{

  /* file names */
  char fitsFilename[MAXFILENAMESIZE];
  char wavsampFilename[MAXFILENAMESIZE];
  char fovFilename[MAXFILENAMESIZE];
  char namelistFilename[MAXFILENAMESIZE];
  char in_bmask[MAXFILENAMESIZE];
  /* char in_pmask[MAXFILENAMESIZE]; */
  char outputFilename[MAXFILENAMESIZE];
  char QA_outputFilename[MAXFILENAMESIZE];
  char line[2880];
  int verbose; /* 1=normal, 2=small usr info, 3=small dev info,
                   4=big user info  5=big dev info */

  int i, j, status;
  int nparms;
  int fatal_bmask/*,fatal_pmask8*/;
  int fd, fdb, imstat;
  int peak_index, std_index, ratio_index;
  int chnlnum, fovidf=0, fovidt, exptype, 
      expid, plvid, aorkey, calid, paonum, campaign,
      dcenum, pipenum,plscrpid,aperture,proposid,airscmdn,
      ncycles,hp_cenq,epidflat;

  int order_list[MAX_ORDERS]; /* must be one more than value in call to param subroutine, below */
  int n_cuts; 
  FILE* outputFid; /* file ID for output file */

  char** hist_cards;	
  char tmp_string[16];

  /*keyword of image FITS header */
  char rawname[80], rawdate[80], date_obs[80], title[80], object[80],
       fovname[80], aot_type[80], ra_cmd[80], dec_cmd[80],filename[80], 
       date[80],   sis_sver[80],  obsrvr[80],expstrtc[80],
       pkupmode[80], ispupos[80], saturdet[80];

  float sos_ver,ramptime,ra_rqst,dec_rqst,pa_rqst,ra_slt,dec_slt,pa_slt;
  int nx, ny;

  /*image data*/
  double *ptr_pixels;
  int pixels_npixels;
  unsigned pixels_nbytes;

  /* bmask data */
  short *ptr_bmask;   
  int bmask_npixels;
  unsigned bmask_nbytes;

  /* pmask data             */
  /* short *ptr_pmask;      */   
  /* int pmask_npixels;     */
  /* unsigned pmask_nbytes; */

  char *xstr;
  int current_order;
  double x0, y0, x1, y1, x2, y2, x3, y3;
  double  wavelength;
  int ind_x0, ind_y0, ind_x1, ind_y1, ind_x2, ind_y2, ind_x3, ind_y3, 
      ind_wavelength;
  int ind_fov, ind_order;
  float xtop[MAX_CUTS+1],  ytop[MAX_CUTS+1];
  float xbottom[MAX_CUTS+1], ybottom[MAX_CUTS+1];
  float lambda_bgn, lambda_end;
  void param_print( void );
  
  double slit_dn, slit_area;  /* total in slit and area of slit */
  double slit_flux, slit_flux_sqr;  /* total of flux and square of flux in slit */


  double variance;
  double total_flux[MAX_CUTS];
  double total_flux_sqr[MAX_CUTS];
  int n_fluxes[MAX_CUTS];
  double mean_flux[MAX_CUTS];
  double stdv[MAX_CUTS];
  float dn_peak=-1, std_peak=-1, ratio, ratio_peak=-1;

  /* tblio params */
  struct tbl_filinfo  *rinfo;
  int tstat, row;

  /* Inintialization*/
  for (i=1;i<MAX_ORDERS;i++) order_list[i] = -1; 
  
  /**PARAMS **/
  param(fitsFilename,      "-i INFILE name of input Echelle image",  "infile echelle.fits");

  param(wavsampFilename,   "-w WAVSAMP_FILE name of Wavsamp file",   "infile wavsamp.tbl");

  param(fovFilename,       "-t FOVTABLE_FILE name of Fovtable file",   "infile - none");

  param(&n_cuts,           "-c N_CUTS number of cuts to sample",
       "int:>=1:<=2000 - 100"); /* remember to update MAX_CUTS */

  param(&sliding_median_size, "-m MEDIAN_SIZE size of sliding median (odd integer)",     "int:>=1:<=21 - 7"); /* remember to update MAX_MEDIAN_SIZE */

  sprintf (text_max_orders, "int[%d]:>=0 - -1 all", MAX_ORDERS-1);
  param(&order_list,       "-ord ORDERS which orders to include", text_max_orders);

  param(in_bmask,          "-b BMASKFILE name of bmask file", "infile bmask.fits");

  param(&fatal_bmask,      "-fb BMASKFILE_FATAL value of fatal", "int - 18432");

  /*param(in_pmask,          "-p PMASKFILE name of pmask file", "infile pmask.fits");

  param(&fatal_pmask,      "-fp PMASKFILE_FATAL value of fatal", "int - 32767");
  */

  param(outputFilename,    "-o OUTFILE name of output profile file", "outfile profile.tbl");

  param(QA_outputFilename, "-q QA_OUTFILE name of QA output Ascii file", "outfile QA");

  param(&verbose,          "-v VERBOSE level of printed messages",   "int - 1");

  param(namelistFilename,  "-n NAMELIST_FILE name of Namelist file", "infile - namelist.nl");

  /**ENDPARAMS **/

  status = nl_process(argc, argv, PROGRAM, VERSION, NL_START, NL_END);
  if (status != 0) return status;

  status = param_begin(argc, argv, PROGRAM, VERSION);
  if (status != 0) return status;

  param_print();   /* Print the parameter settings to stdout. */
  printf("%s, %s \n", integrate_name(), integrate_version() );
  param_subvers(integrate_name(), integrate_version() );

  /* check that parameters are valid */
  if (n_cuts > MAX_CUTS) {
    printf("ERROR: Value entered for -c (N_CUTS) is too large\n");
    return TERMINATE_ERROR;
  }
  if (sliding_median_size > MAX_MEDIAN_SIZE) {
    printf("ERROR: Value entered for -m (MEDIAN_SIZE) is too large\n");
    return TERMINATE_ERROR;
  }
  if (sliding_median_size %2 ==0) {
    printf("ERROR: Value entered for -m (MEDIAN_SIZE) must be an odd number\n");
    return TERMINATE_ERROR;
  }

  /* Read the fitsfile */
  fd = im_open(fitsFilename, "r");
  if (fd == -1)
  {
    printf("ERROR: %s -- not a proper image file.\n", fitsFilename);
    return TERMINATE_ERROR;
  }
  if (fd < 0)
  {
    printf("%s -- ERROR opening or reading file.\n", fitsFilename);
    return TERMINATE_ERROR;
  }

  *rawname = '\0';
  *rawdate = '\0';
  *date_obs = '\0';
  *title = '\0';
  *object = '\0';
  *fovname = '\0';
  *ra_cmd = '\0';
  *dec_cmd= '\0';
  *aot_type = '\0';
  *filename = '\0';
  *date = '\0';
  *sis_sver = '\0';
  *obsrvr = '\0';
  *expstrtc = '\0';
  *pkupmode = '\0';
  *ispupos = '\0';
  *saturdet= '\0';

  if ((imstat = im_rkey_i(fd, &nx,       "NAXIS1")) == -1)
     printf("Keyword NAXIS1 is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &ny,       "NAXIS2")) == -1)
     printf("Keyword NAXIS2 is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &chnlnum,  "CHNLNUM")) == -1)
     printf("Keyword CHNLNUM is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, rawname,   "RAWNAME")) == -1)
     printf("Keyword RAWNAME is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, rawdate,   "RAWDATE")) == -1)
     printf("Keyword RAWDATE is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, date_obs,  "DATE_OBS")) == -1)
     printf("Keyword DATE_OBS is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, title,     "TITLE")) == -1)
     printf("Keyword TITLE is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, object,    "OBJECT")) == -1)
     printf("Keyword OBJECT is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, ra_cmd,    "RA_RQST")) == -1)
     printf("Keyword RA_CMD is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, dec_cmd,   "DEC_RQST")) == -1)
     printf("Keyword DEC_CMD is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &fovidf,   "FOVID")) == -1)
     printf("Keyword FOVID is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, fovname,   "FOVNAME")) == -1)
     printf("Keyword FOVNAME is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &exptype,  "EXPTYPE")) == -1)
     printf("Keyword EXPTYPE is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, aot_type,  "AOT_TYPE")) == -1)
     printf("Keyword AOT_TYPE is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &aorkey,   "AORKEY")) == -1)
     printf("Keyword AORKEY is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &expid,    "EXPID")) == -1)
     printf("Keyword EXPID is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &plvid,    "PLVID")) == -1)
     printf("Keyword PLVID is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &calid,    "CALID")) == -1)
     printf("Keyword CALID is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &paonum,   "PAONUM")) == -1)
     printf("Keyword PAONUM is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &campaign, "CAMPAIGN")) == -1)
     printf("Keyword CAMPAIGN is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, filename,  "FILENAME")) == -1)
     printf("Keyword FILENAME is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, date,       "DATE")) == -1)
     printf("Keyword DATE is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &dcenum,  "DCENUM")) == -1)
     printf("Keyword DCENUM is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, sis_sver,   "SIS_SVER")) == -1)
     printf("Keyword SIS_SVER is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &pipenum,   "PIPENUM")) == -1)
     printf("Keyword PIPENUM is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &plscrpid,  "PLSCRPID")) == -1)
     printf("Keyword PLSCRPID is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &sos_ver,   "SOS_VER")) == -1)
     printf("Keyword SOS_VER is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &aperture,  "APERTURE")) == -1)
     printf("Keyword APERTURE is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, obsrvr,    "OBSRVR")) == -1)
     printf("Keyword OBSRVR is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &proposid, "PROPOSID")) == -1)
     printf("Keyword PROPOSID is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, expstrtc,   "EXPSTRTC")) == -1)
     printf("Keyword EXPSTRTC is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &airscmdn,   "AIRSCMDN")) == -1)
     printf("Keyword AIRSCMDN is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &ramptime,  "RAMPTIME")) == -1)
     printf("Keyword RAMPTIME is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &ra_rqst,  "RA_RQST")) == -1)
     printf("Keyword RA_RQST is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &dec_rqst,  "DEC_RQST")) == -1)
     printf("Keyword DEC_RQST is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &pa_rqst,   "PA_RQST")) == -1)
     printf("Keyword PA_RQST is missing from FITS header.\n"); 
  if ((imstat = im_rkey_r(fd, &ra_slt,  "RA_SLT")) == -1)
     printf("Keyword RA_SLT is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &dec_slt,  "DEC_SLT")) == -1)
     printf("Keyword DEC_SLT is missing from FITS header.\n");
  if ((imstat = im_rkey_r(fd, &pa_slt,   "PA_SLT")) == -1)
     printf("Keyword PA_SLT is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &ncycles,  "NCYCLES")) == -1)
     printf("Keyword NCYCLES is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, pkupmode,  "PKUPMODE")) == -1)
     printf("Keyword PKUPMODE is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, ispupos, "ISPUPOS")) == -1)
     printf("Keyword ISPUPOS is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &hp_cenq,   "HP_CENQ")) == -1)
     printf("Keyword HP_CENQ is missing from FITS header.\n");
  if ((imstat = im_rkey_c(fd, saturdet,   "SATURDET")) == -1)
     printf("Keyword SATURDET is missing from FITS header.\n");
  if ((imstat = im_rkey_i(fd, &epidflat,  "EPIDFLAT")) == -1)
     printf("Keyword EPIDFLAT is missing from FITS header.\n");

  pixels_npixels = nx*ny;
  pixels_nbytes = pixels_npixels * sizeof(double);
  ptr_pixels = (double *) malloc(pixels_nbytes);
  if (ptr_pixels == NULL)
  {
     printf("ERROR: Can't allocate required memory\n");
     return TERMINATE_ERROR;
  }

  for(i=0; i<ny*nx; i+=nx)
     im_rpix_d(fd, ptr_pixels+i, nx);

  /*im_close(fd);*/

  /* Read in the Bmask File */
  fdb = im_open(in_bmask, "r");
  if (fdb == -1)
  {
    printf("ERROR: %s -- not a proper Bmask file.\n", in_bmask);
    return TERMINATE_ERROR;
  }
  if (fdb < 0)
  {
    printf("%s -- ERROR opening or reading file.\n", in_bmask);
    return TERMINATE_ERROR;
  }

  imstat = im_rkey_i(fdb, &nx, "NAXIS1");
  imstat = im_rkey_i(fdb, &ny, "NAXIS2");

  bmask_npixels = nx*ny;
  bmask_nbytes = bmask_npixels * sizeof(short);
  ptr_bmask = (short *) malloc(bmask_nbytes);
  if (ptr_bmask == NULL)
  {
    printf("ERROR: Can't allocate required memory\n");
    return TERMINATE_ERROR;
  }
  status = im_rpix_s(fdb, ptr_bmask, bmask_npixels);
  im_close(fdb);

  /* Read in the Pmask File */
  /*fd = im_open(in_pmask, "r");
  if (fd == -1)
  {
    printf("ERROR: %s -- not a proper Pmask file.\n", in_pmask);
    return TERMINATE_ERROR;
  }
  if (fd < 0)
  {
    printf("%s -- ERROR opening or reading file.\n", in_pmask);
    return TERMINATE_ERROR;
  }

  imstat = im_rkey_i(fd, &nx, "NAXIS1");
  imstat = im_rkey_i(fd, &ny, "NAXIS2");

  pmask_npixels = nx*ny;
  pmask_nbytes = pmask_npixels * sizeof(short);

  ptr_pmask = (short *) malloc(pmask_nbytes);
  if (ptr_pmask == NULL)
  {
    printf("ERROR: Can't allocate required memory\n");
    return TERMINATE_ERROR;
  }
  status = im_rpix_s(fd, ptr_pmask, pmask_npixels);
  im_close(fd);
  */

  /* Set the detected fatal data in the image to be NAN */
  for (j=0; j<ny; ++j) 
      for (i=0; i<nx; ++i) {
         if ((ptr_bmask[j*nx+i] & fatal_bmask) != 0 ) 
            ptr_pixels[j*nx+i] = NANVALUE;
      }
 

  /* Read in the FOV table for SL and LL 
     to find out the ORDER number that corresponds to FOVID
     obtained from the input FITS header */
  
  if (order_list[1]<0 && (chnlnum==0 || chnlnum==2) ) {
     iu = tbl_open(fovFilename, &rinfo); 
     tstat   = tbl_fileinfo(iu, &rinfo);
     if (tstat != 0)
     {
        printf("Error open file '%s'\n", fovFilename);
        exit(64);
     }
     if (tio_getind("FOV", &ind_fov)) return TERMINATE_ERROR;
     if (tio_getind("order", &ind_order)) return TERMINATE_ERROR; 

     i = 0;
     row = 0;
     while (1) {
        status = tbl_colval(iu, row, ind_fov, TBL_INTEGER, &fovidt);
        if (status < 0) break;
        if (status > 0)
        {
           printf("Table I/O ERROR when reading file: [%s]\n",fovFilename);
           return TERMINATE_ERROR;
        }
        if (fovidt == fovidf) {
          i++;
          status = tbl_colval(iu, row, ind_order, TBL_INTEGER, &order_list[i]);
        }
        row++;
     }
     tbl_close(iu);
  }

  if (verbose >=3) { /* print included orders */
    printf ("order_list[1]=%d\n", order_list[1]);
    printf ("order_list[2]=%d\n", order_list[2]);
    printf ("order_list[3]=%d\n", order_list[3]);
  }

  /* initialize summation data */
  for (i=0; i<=n_cuts; i++) {
    n_fluxes[i] = 0;
    total_flux[i]  = 0.0;
    total_flux_sqr[i]  = 0.0;
    mean_flux[i] = 0.0;
    variance = 0.0;
    stdv[i] = 0.0; 
  }

  /**************************/
  /* Read the wavsamp table */
  /**************************/

  /* open table file and get column indices */
  iu = tbl_open( wavsampFilename, &rinfo); 
  tstat   = tbl_fileinfo(iu, &rinfo);
  if (tstat != 0)
  {
     printf("ERROR openning wavesamp file [%s]\n", wavsampFilename);
     exit(64);
  }

  if (tio_getind("x0", &ind_x0)) return TERMINATE_ERROR;
  if (tio_getind("y0", &ind_y0)) return TERMINATE_ERROR;
  if (tio_getind("x1", &ind_x1)) return TERMINATE_ERROR;
  if (tio_getind("y1", &ind_y1)) return TERMINATE_ERROR;
  if (tio_getind("x2", &ind_x2)) return TERMINATE_ERROR;
  if (tio_getind("y2", &ind_y2)) return TERMINATE_ERROR;
  if (tio_getind("x3", &ind_x3)) return TERMINATE_ERROR;
  if (tio_getind("y3", &ind_y3)) return TERMINATE_ERROR;
  if (tio_getind("wavelength", &ind_wavelength)) return TERMINATE_ERROR;

  /* read all the rows */
  row = 0;
  while (1)
  {

    tstat = tbl_colval(iu, row, 0, TBL_INTEGER, &current_order); 
    if (tstat == -4) break;
    else if (tstat < 0) {
       printf("Table I/O ERROR when reading file [%s]\n", wavsampFilename);
       return TERMINATE_ERROR;
    }
    if (order_list[1] >= 0) { /* only want selected orders added in */
        for (i=1; order_list[i]>=0; i++)
           if (order_list[i] == current_order) break;
        if (order_list[i]<0)  {row++; continue; }/* not found in list */

    }

    if (verbose >= 2) {
      if (current_order != previous_order) {
        printf ("Added in order %d\n", current_order);
        previous_order = current_order;
      }
    }

    xstr = tbl_colstr(iu, row,ind_x0, &tstat);
    status = tbl_colval(iu, row, ind_x0, TBL_DOUBLE, &x0);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_y0, TBL_DOUBLE, &y0);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_x1, TBL_DOUBLE, &x1);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_y1, TBL_DOUBLE, &y1);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_x2, TBL_DOUBLE, &x2);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_y2, TBL_DOUBLE, &y2);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_x3, TBL_DOUBLE, &x3);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_y3, TBL_DOUBLE, &y3);         
    if (status < 0) break;
    status = tbl_colval(iu, row, ind_wavelength, TBL_DOUBLE, &wavelength); 

    /*****/
    if (verbose >=5)
    {
      printf("1 %6.2f %6.2f\n", x0, y0);
      printf("2 %6.2f %6.2f\n", x1, y1);
      printf("3 %6.2f %6.2f\n", x2, y2);
      printf("4 %6.2f %6.2f\n", x3, y3);
    }
    /*****/

    /* Divide the slit into n_cuts sub-slits */
    for (i=0; i<=n_cuts; i++)
    {
      xtop[i]    = x1 + (x0-x1)/n_cuts * i;
      ytop[i]    = y1 + (y0-y1)/n_cuts * i;
      xbottom[i] = x2 + (x3-x2)/n_cuts * i;
      ybottom[i] = y2 + (y3-y2)/n_cuts * i;

      if (verbose >=5)  printf("X %6.2f %6.2f %6.2f %6.2f\n",
                   xtop[i], ytop[i], xbottom[i], ybottom[i]);

    }
  
    /* Call routine integrate_quad for each of n_cuts sub-slits */
    
    for (i=0; i<n_cuts; i++) {
      slit_dn = integrate_quad(
               xtop[i],      ytop[i],       xtop[i+1],  ytop[i+1],  
               xbottom[i+1], ybottom[i+1],  xbottom[i], ybottom[i],  
               ptr_pixels, nx, ny, &slit_area);

      /* Changed - TYB 6/1/09 */
      if (slit_area > 0.0 && iznand(slit_dn) == 0 ) {
	//      if (slit_area > 0.0 && isnan(slit_dn) == 0 ) {

        slit_flux = slit_dn / slit_area;
        slit_flux = sliding_median(slit_flux, i);
        if (slit_flux == PSEUDO_NAN) continue; /* no median available yet */
        slit_flux_sqr = pow(slit_flux, 2.0);
        total_flux[i] += slit_flux;
        total_flux_sqr[i] += slit_flux_sqr;
        n_fluxes[i]++;
        if (verbose >=5)  printf(" %d i n %d F=%12.5E fq=%12.5E \n",
               i,n_fluxes[i],  slit_flux, slit_flux_sqr);
      }

    }
  row++;
  }
  tbl_close(iu);

  /* normalize */
  for (i=0; i<n_cuts; i++) {
    if (n_fluxes[i] > 0) {
      mean_flux[i] = total_flux[i] / n_fluxes[i];
      variance = total_flux_sqr[i] - n_fluxes[i] * pow(mean_flux[i], 2.0);
      stdv[i] = sqrt(variance/n_fluxes[i]); 
    }
    else {
      mean_flux[i] = 0.0;
      stdv[i]      = 0.0;
    }
  }


  /* compute out "QA" params */
  for (i=0; i<n_cuts; i++) {       /* peak and its location */
    if (mean_flux[i] > dn_peak) {
       dn_peak = mean_flux[i];
       peak_index = i;
    }
  }

  for (i=0; i<n_cuts; i++) {       /* abs. std. dev.  and its location */
    if (stdv[i] > std_peak) {
       std_peak = stdv[i];
       std_index = i;
    }
  }

  for (i=0; i<n_cuts; i++) {       /* rel. std. dev.  and its location */
    ratio = stdv[i] / mean_flux[i];
    if (ratio > ratio_peak) {
       ratio_peak = ratio ;
       ratio_index = i;
    }
   }

  if (verbose >=2) {  
     printf("Maximum dn value is %g", dn_peak);
     printf(" at the %5.1f percent cut\n", cut2pct(peak_index, n_cuts)); 
     printf("Maximum std value is %g", std_peak);
     printf(" at the %5.1f percent cut\n", cut2pct(std_index, n_cuts)); 
     printf("Maximum ratio_peak value is %g", ratio_peak);
     printf(" at the %5.1f percent cut\n", cut2pct(ratio_index, n_cuts)); 
   }


  /* write the output file */
  outputFid = fopen(outputFilename, "w");
  if (outputFid == NULL) {
    printf("ERROR opening file %s\n", outputFilename);
    return TERMINATE_ERROR;
  }

  /* Copy the FITS header over */

  imstat = im_rkey_t(fd,line,"1");
  imstat = im_rkey_t(fd,line,"+"); /* (skip BITPIX card) */
  while (imstat >= 0) {
     if (strncmp(line, "HISTORY", 7) != 0 &&
         strncmp(line, "COMMENT", 7) != 0 &&
         strncmp(line, "FOVID", 5)   != 0)
         fprintf(outputFid,"\\char %s\n", line);
     imstat = im_rkey_t(fd,line,"+");
  }

  /* Add HISTORY cards */
  param_hcards(&hist_cards);
  for (i=0; hist_cards[i]; i++)
  {
    fprintf(outputFid,"\\HISTORY %s\n", hist_cards[i]);
  }

  fprintf(outputFid,"\\int CHNLNUM = %d\n", chnlnum);
  fprintf(outputFid,"\\int FOVID   = %d\n", fovidf);
  fflush(outputFid);
  for (i=1; order_list[i]>=0; i++) {
    printf("added in order %d\n", order_list[i]);
    fprintf(outputFid,"\\int ORDER%d = %d\n", i, order_list[i]);
  }


  fprintf(outputFid,"|  pct |     dn    |  stdev    |  n   | region |\n");
  fprintf(outputFid,"| real |    real   |   real    | int  | int    |\n");
  /* SAMPLE LINE:   "  50.00 2.08656e+06    87" */

  for (i=0; i<n_cuts; i++) {
    fprintf(outputFid,"%#7.2f%12.4e%12.4e%6d%8d\n",
    cut2pct(i, n_cuts), mean_flux[i], stdv[i], n_fluxes[i], 0);
  }

  fclose(outputFid);
  im_close(fd);

  /* write the QA output file */
  outputFid = fopen(QA_outputFilename, "a");
  if (outputFid == NULL)
  {
    printf("ERROR opening QA file %s\n", QA_outputFilename);
    return TERMINATE_ERROR;
  }


  strcpy(tmp_string,"profMaxFlux");
  fprintf(outputFid,"%-16s %12.5e \n", tmp_string, dn_peak);
  strcpy(tmp_string,"profLocMaxFlux");
  fprintf(outputFid,"%-16s %5.1f \n", tmp_string, cut2pct(peak_index, n_cuts) );

  strcpy(tmp_string,"profMaxStdev");
  fprintf(outputFid,"%-16s %12.5e \n", tmp_string, std_peak);
  strcpy(tmp_string,"profLocMaxStdev");
  fprintf(outputFid,"%-16s %5.1f \n", tmp_string, cut2pct(std_index, n_cuts) );

  strcpy(tmp_string,"profMaxRStdev");
  fprintf(outputFid,"%-16s %12.5e \n", tmp_string, ratio_peak);
  strcpy(tmp_string,"profLocMaxRStdev");
  fprintf(outputFid,"%-16s %5.1f \n", tmp_string, cut2pct(ratio_index, n_cuts) );


  fclose(outputFid);

  /* exit */
  return TERMINATE_NORMAL;
}


/********************/
main(argc, argv)
int argc;
char *argv[];
{
  int status;
  status = profile(argc, argv);
  if (status < 0) return TERMINATE_ERROR;
  param_exit(status);
}
