/*********************************************** 
find_max_ll_samp.cpp

Purpose

Find the sample number which has the highest log-likelyhood for containing
a possible radiation hit.

Overview
Call a series of functions which do a segmented fit the the data, finding
the sample number at which a breakpoint yields the best fit. Endpoints are
treated as special cases.

Definitions of Variables

External:
See radhit.h

Internal:
m, I_Posited_Radhit_Position_LL = Posited Radiation Hit Position.
FP_Mat_GPseudo = Pseudomatrix associated with I_Posited_Radhit_Position_LL.
DP_Fit_Vec_Charge_LL = Fit of data Charge vector.
D_Data_Fit_Inner_Prod = Inner Product of Data and Fit Vectors.
D_Data_Data_Inner_Prod = Inner Product of Data Vector and itself.

***********************************************/

#include <iostream>
#include <string.h>
#include <stdlib.h>

#include "more_includes.h"


 
int M_Computation::find_max_ll_samp(	int		I_SampStart,
					int		I_SampStop,
					M_Constants	*GP_ConstIn,
					M_Initialize	*GP_Pseudo,
					M_Status	*GP_Stat)
{
int	m;


/* Construct the pseudomatrix and determinants if not done already */
if(GP_Pseudo[I_NumSamples-1].FP_Mat_G == NULL)
GP_Pseudo[I_NumSamples-1].initialize(I_NumSamples, GP_ConstIn, GP_Stat);


if(GP_Stat->I_Debug)
printf("find_max_ll_samp: initialize done.\tI_NumSamples = %d\n", I_NumSamples);

/* Compute the log-likelyhood of the data
   for each possible radiation hit location. 
   Will write log likelyhood results into 
   the DP_LLikelyhood array */
//memset((void *) DP_LLikelyhood, 0, I_Num_Frames*sizeof(double));
for(m=0;m<I_Num_Frames;m++)
DP_LLikelyhood[m] = atof("-Inf");

for(m=I_SampStart;m<I_SampStop;m++)		
log_likelyhood(m, &GP_Pseudo[I_NumSamples-1], GP_Stat);

if(GP_Stat->I_Debug)
printf("find_max_ll_samp: log_likelyhood done 0x%x\n", GP_Stat->I_status);	

/* Set a default value in case maximum likelyhood cannot be found */
I_Max_LLikely_Position = 2;

/* Find the index of the maximum likelyhood */
maxll_findmax( &GP_Pseudo[I_NumSamples-1], GP_Stat);

if(GP_Stat->I_Debug)
printf("find_max_ll_samp: maxll_findmax done 0x%x\n", GP_Stat->I_status);

/* Find the sample number which has the highest
   log-likelyhood of having a radiation hit. */
//max_llikelyhood( &GP_Pseudo[I_NumSamples-1], GP_Stat);

//if(GP_Stat->I_Debug)
//printf("find_max_ll_samp: max_llikelyhood done.\n");


/* Estimate the charge added by the posited radiation hit. */
charge_addition( &GP_Pseudo[I_NumSamples-1], GP_Stat);


if(GP_Stat->I_Debug)
printf("find_max_ll_samp: charge_addition done 0x%x\n", GP_Stat->I_status);

/* For ambiguous cases, examine the endpoints for radiation hits. */		
endpoints( GP_Stat);

/* Free the stored matrices for longer lengths
   otherwise could run out of memory */
if (I_NumSamples > GP_ConstIn->I_MaximumStoredMatrixLen)  {
free(GP_Pseudo[I_NumSamples-1].FP_Mat_G);
GP_Pseudo[I_NumSamples-1].FP_Mat_G=NULL;
free(GP_Pseudo[I_NumSamples-1].DP_Det_G);
GP_Pseudo[I_NumSamples-1].DP_Det_G=NULL;
}


if(GP_Stat->I_Debug)
printf("find_max_ll_samp: I_Max_LLikely_Position = %d\n", I_Max_LLikely_Position);


  return(GP_Stat->I_status);
}
