/*********************************************** 
prob_radhit.c

Purpose

Compute the a posteriori probability that a radhit occurred. 			


Overview
Use Bayes rule based on a gaussian model with prior probabilities. 

Definitions of Variables

External:
See radhit.h

Internal:
D_Meas_Var = Estimated variance of typical charge measurement.
D_Meas_SDev = Estimated standard deviation of typical charge measurement.
D_NominalRadhitMag = Expected least detectable Radhit magnitude.
D_Normalization = Normalization of probability density function of charge measurement.
D_Exponent_NoRadhit = Exponent in probability density function when no radhit is assumed.
D_Exponent_Radhit = Exponent in probability density function when radhit is assumed.
D_ProbDens_NoRadhit = Probability density function when no radhit is assumed.
D_ProbDens_Radhit = Probability density function when radhit is assumed.
D_RadHitTerm = Term in a posterior probability of radhit due to radhit occurrence.
D_NoRadHitTerm = Term in a posterior probability of radhit due to no radhit occurrence.

***********************************************/

#include <stdio.h>
#include <math.h>
#include <float.h>

/*  Added TYB 5/1/09 */
#include "nanvalue.h"

#include "more_includes.h"


int M_Computation::prob_radhit(	M_Constants	*GP_Const,
				M_Initialize	*GP_Init,
				M_Status	*GP_Stat)
{
double	D_Meas_Var;
double	D_Meas_SDev;
double	D_NominalRadhitMag;
double	D_Normalization;
double	D_Exponent_NoRadhit;
double	D_Exponent_Radhit;


if(GP_Stat->I_status) return(GP_Stat->I_status);

D_Meas_Var  = fabs(D_Meas_Flux) + D_Readnoise*D_Readnoise;
if(D_Meas_Var < 0)
{GP_Stat->I_status = PROB_RADHIT|ROOT_NEG; return(GP_Stat->I_status);}
D_Meas_SDev = sqrt(D_Meas_Var);

D_NominalRadhitMag = GP_Const->D_NomRHMag*D_Meas_SDev;

if(D_Meas_SDev == 0)
{GP_Stat->I_status = PROB_RADHIT|DIVIDE_ZERO;  return(GP_Stat->I_status);}
D_Normalization = INV_SQRT_2PI/D_Meas_SDev;


/* Compute arguments for Gaussian Model. */
if(D_Meas_Var == 0)
{GP_Stat->I_status = PROB_RADHIT|DIVIDE_ZERO; return(GP_Stat->I_status);}
D_Exponent_NoRadhit = -( D_Meas_Charge_Add*D_Meas_Charge_Add)/(2*D_Meas_Var);
D_Exponent_Radhit   = -((fabs(D_Meas_Charge_Add)-D_NominalRadhitMag)*
			(fabs(D_Meas_Charge_Add)-D_NominalRadhitMag))/(2*D_Meas_Var);


/* Take care of noncomputable values */
if(D_Exponent_NoRadhit-D_Exponent_Radhit < -100)
{D_Prob_Radhit = 1; return(GP_Stat->I_status);}
if(D_Exponent_NoRadhit-D_Exponent_Radhit >  100)
{D_Prob_Radhit = 0; return(GP_Stat->I_status);}

/* Apply Bayes Rule. Avoid exponentiating large numbers. */
D_Prob_Radhit = exp(D_Exponent_NoRadhit - D_Exponent_Radhit);
D_Prob_Radhit = ((1-GP_Const->D_RPPriorProb)/GP_Const->D_RPPriorProb)*D_Prob_Radhit;
if(D_Prob_Radhit == -1)
{GP_Stat->I_status = PROB_RADHIT|DIVIDE_ZERO; return(GP_Stat->I_status);}
D_Prob_Radhit = 1/(1 + D_Prob_Radhit);

/* Check that DP_Prob_Radhit is a valid number */

/* Just use the one at bottom - TYB 5/1/09 */

//#if 0
//#ifdef HAVEISNAN
//if(iznand(D_Prob_Radhit))
//GP_Stat->I_status = PROB_RADHIT|NOT_A_NUMBER;
//#else
//if(D_Prob_Radhit != 0 && dnan(*(short *) &D_Prob_Radhit))
//GP_Stat->I_status = PROB_RADHIT|NOT_A_NUMBER;
//#endif
//#endif

if(iznand(D_Prob_Radhit))
  GP_Stat->I_status = PROB_RADHIT|NOT_A_NUMBER;

//if(GP_Stat->I_Verbose)
//printf("prob: D_Prob_Radhit = %f\n", D_Prob_Radhit);

  return(GP_Stat->I_status);
}	




