/*********************************************** 
initialize.cpp

Purpose

Initialization: Check input dimensions. Allocate output space.

Overview
Check that FITS X, Y, and Frames are the same. Allocate with calloc.

Definitions of Variables

External:
See radhit.h

Internal:


***********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Added TYB 5/1/09 */
#include "nanvalue.h"

#include "more_includes.h"

extern "C" void dlutl_vecmult(int, int, double*, double*, double*, int*);
extern "C" void dlutl_innerprod(int, double*, double*, double*, int*);
extern "C" void dlutl_transpmult(int, int, double*, double*, int*);
extern "C" void dlutl_transp(double*, int, int, double*, int*);
extern "C" void dlutl_pseudocomp(int, int, double*, double*, double*, int*);		
extern "C" void dlutl_projcomp(int, int, double*, double*, double*, int*);
extern "C" void dlutl_matmult(double*, int, int, double*, int, int, double*, int*);
extern "C" void dlutl_invertmat(int, double*, double*, int*);
extern "C" void dlutl_detcomp(int, double*, double*, int*);


M_Initialize::M_Initialize()
{
	/* Constructor */
	I_NumColumns = 0;
/* Basis matrices and their determinants */
	FP_Mat_G = NULL;
	DP_Det_G = NULL;
}


M_Initialize::~M_Initialize()
{
//printf("initialize: %d\t0x%x.\n", 1, DP_Det_G);
if(DP_Det_G!=NULL) free(DP_Det_G);
//printf("initialize: %d\t0x%x.\n", 2, FP_Mat_G);
if(FP_Mat_G!=NULL) free(FP_Mat_G);
}


int M_Initialize::initialize(	int 		I_Num_Frames,
				M_Constants	*GP_Const,
				M_Status	*GP_Stat)
{
int	k, l, m, I_Posited_Radhit_Position;
double	*DP_GMatrix;
double	*DP_GTransProd;

if(GP_Stat->I_status) return(GP_Stat->I_status);

double	D_DetGTransProd;
double	*DP_Mat_GPseudo;
double	*DP_Mat_GProj;
float	*FP_Mat_GProj;

/* Allocate space for init arrays */
allocate( I_Num_Frames, GP_Stat);
if(GP_Stat->I_status) return(GP_Stat->I_status);

I_NumColumns = GP_Const->I_NumColumns;

/* Allocate space for temporary arrays */
DP_Mat_GProj = (double *) calloc(I_Num_Frames*I_Num_Frames, sizeof(double));
DP_GMatrix = (double *) calloc(I_Num_Frames*I_NumColumns, sizeof(double));
DP_GTransProd = (double *) calloc(I_NumColumns*I_NumColumns, sizeof(double));
DP_Mat_GPseudo = (double *) calloc(I_Num_Frames*I_NumColumns, sizeof(double));


/* Set status if malloc failed */
if(DP_GMatrix==NULL) GP_Stat->I_status = INIT|MALLOC_FAILED;
if(DP_GTransProd==NULL) GP_Stat->I_status = INIT|MALLOC_FAILED;
if(DP_Mat_GPseudo==NULL) GP_Stat->I_status = INIT|MALLOC_FAILED;
if(DP_Mat_GProj==NULL) GP_Stat->I_status = INIT|MALLOC_FAILED;

/* Don't continue if memory not allocated */
if(GP_Stat->I_status) return(GP_Stat->I_status);

/* Loop over all possible changepoints for
   which the method can work */
for(m=2;m<I_Num_Frames-1;m++)	{
I_Posited_Radhit_Position = m;

//if(GP_Stat->I_Verbose)
//printf("init:  Computing pseudomatrices and determinants for I_Posited_Radhit_Position = %d.\n", m);

/* Create a matrix whose columns are the basis vectors */
mkbasis(	I_Posited_Radhit_Position,
		I_Num_Frames,
		I_NumColumns,
		DP_GMatrix,
		GP_Stat);
		
/* Compute the product of the transpose of the basis
   vector matrix and itself */
dlutl_transpmult(	I_Num_Frames,
			I_NumColumns,
			DP_GMatrix,
			DP_GTransProd,
			&(GP_Stat->I_status));
	
/* Compute the determinant of the product computed in
   radhit_init_transp */
dlutl_detcomp(		I_NumColumns,
			DP_GTransProd,
			&D_DetGTransProd,
			&(GP_Stat->I_status));

/* If Determinant Computation was OK, take square root */
if(D_DetGTransProd < 0) GP_Stat->I_status = INIT|ROOT_NEG;

if(GP_Stat->I_status) return(GP_Stat->I_status);

DP_Det_G[I_Posited_Radhit_Position] = sqrt(D_DetGTransProd);

/* Check that square root returned a valid number */
if(iznand(DP_Det_G[m]))
GP_Stat->I_status = INIT|NOT_A_NUMBER;

/* Compute the pseudoinverse of DP_GMatrix */
dlutl_pseudocomp(	I_Num_Frames,
			I_NumColumns,
			DP_GMatrix,
			DP_GTransProd,
			DP_Mat_GPseudo,
			&(GP_Stat->I_status));		

/* Multiply DP_Mat_GPseudo on the left by G to get the projector matrix */
dlutl_matmult(	DP_GMatrix,
		I_Num_Frames,
		I_NumColumns,
		DP_Mat_GPseudo,
		I_NumColumns,
		I_Num_Frames,
		DP_Mat_GProj,
		&(GP_Stat->I_status));

FP_Mat_GProj =
&(FP_Mat_G[I_Num_Frames*I_Num_Frames*I_Posited_Radhit_Position]);
for(k=0;k<I_Num_Frames;k++)
for(l=0;l<I_Num_Frames;l++)	
FP_Mat_GProj[k*I_Num_Frames+l] = DP_Mat_GProj[k*I_Num_Frames+l];

//if(GP_Stat->I_Verbose)
//printf("init:  Computed for I_Posited_Radhit_Position = %d.\n", m);

}

/* There are cases in resolving the ambiguity of decimation in which
   the endpoint gets checked as a normal segmentation. The log likelyhood
   computation then fails unless we give these determinant dummy values.
   The large values cause the likelyhoods for these segmentations to
   be ignored. */ 
DP_Det_G[1] = (double) 3.40282346638528860e38;
DP_Det_G[I_Num_Frames-1] = (double) 3.40282346638528860e38;

/* Deallocate Memory */
if(DP_GMatrix!=NULL) free(DP_GMatrix);
if(DP_GTransProd!=NULL) free(DP_GTransProd);
if(DP_Mat_GPseudo!=NULL) free(DP_Mat_GPseudo);
if(DP_Mat_GProj!=NULL) free(DP_Mat_GProj);


if(GP_Stat->I_Verbose)
if(GP_Stat->I_status==0)
printf("init:  Computed pseudomatrices and determinants for I_Num_Frames = %d.\n", I_Num_Frames);
else
printf("init:  Error %d\n", GP_Stat->I_status);

  return(GP_Stat->I_status);
}

