

/*********************************************** 
rowdroop_output.c

Purpose

Save the rowdroop outputs to files.

Overview
Write the results of the computations as a FITS file.

Definitions of Variables

Input:
I_NX = Number of pixels in a row of the image
I_NY = Number of pixels in a column of the image
CP_Filename_FITS_In = Input Data File Name. 

Output:


Internal:
ffp_FITS = Handle to output FITS file.
I_fits_return_status = Resturn value of fits function.
CP_Keyname = name of Key to be written.
CP_Comment = String value to be written in comment field of header.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "rowdroop.h"
#include "rowdroop_errcodes.h"
#include "rowdroop_defs.h"


rowdroop_output(
            Filenames	  *Fnames,
            Constants     *Const,
            Dims          *Dim_In,
            Status	  *Stat,
            struct tm     *tmp_time,
            short         *SPP_dmask,
            float         *FPP_Data_In,
            float         *FPP_Data_Unc_In,
            float         *FPP_Data_Rd_Unc)   
{

int	I_fits_return_status=0;
int	I_Num_In;
int     string_length;
int     i, unc_out;

float   F_Num_In;

char    CP_Keyname[FLEN_KEYWORD], CP_Comment[FLEN_COMMENT];
char    format[10];

fitsfile 	*ffp_FITS_Out;
fitsfile 	*ffp_FITS_Unc_Out;
fitsfile 	*ffp_FITS_In;
fitsfile 	*ffp_FITS_Unc_In;
fitsfile 	*ffp_FITS_Dmask;

printf(asctime(tmp_time));

if(Stat->I_status) return;

/* Open dmask file. If found replace data.  */

if (strcmp(Fnames->CP_Filename_Dmask, ""))
   {
   fits_open_file( &ffp_FITS_Dmask,
                   Fnames->CP_Filename_Dmask,
                   READWRITE,
                   &I_fits_return_status);

/* Add module specific keywords */

   sprintf(CP_Comment, "%s", "Begin ROW DROOP module additions");
   fits_write_comment(ffp_FITS_Dmask,
                      CP_Comment,
                      &I_fits_return_status);

   sprintf(CP_Comment, "%s", asctime(tmp_time));
   string_length = strlen(CP_Comment)-1;
   sprintf(format, "%%.%ds", string_length);
   sprintf(CP_Comment, format, asctime(tmp_time));
   fits_write_comment(ffp_FITS_Dmask,
                      CP_Comment,
                      &I_fits_return_status);
   sprintf(CP_Comment, "%s", "Row Droop correction performed using dmask");
   fits_write_comment(ffp_FITS_Dmask,
                      CP_Comment,
                      &I_fits_return_status);
   
   
   fits_write_img(ffp_FITS_Dmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  SPP_dmask,
                  &I_fits_return_status);		

   fits_flush_file(ffp_FITS_Dmask,
                   &I_fits_return_status);
		
   fits_close_file(ffp_FITS_Dmask,
                   &I_fits_return_status);
   }

if (I_fits_return_status)
{Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status; return;}

/* write out the image data */
fits_open_file( &ffp_FITS_In,
                Fnames->CP_Filename_FITS_In,
                READONLY, 
                &I_fits_return_status);

fits_create_file( &ffp_FITS_Out,
                Fnames->CP_Filename_FITS_Out,
                &I_fits_return_status);

if (I_fits_return_status)
{Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status; return;}

fits_copy_header(ffp_FITS_In,
                 ffp_FITS_Out,
                 &I_fits_return_status);

/* Add module specific keywords */

sprintf(CP_Comment, "%s", "Begin ROWDROOP module additions");
fits_write_comment(ffp_FITS_Out,
                   CP_Comment,
                   &I_fits_return_status);

sprintf(CP_Comment, "%s", asctime(tmp_time));
string_length = strlen(CP_Comment)-1;
sprintf(format, "%%.%ds", string_length);
sprintf(CP_Comment, format, asctime(tmp_time));
fits_write_comment(ffp_FITS_Out,
                   CP_Comment,
                   &I_fits_return_status);
sprintf(CP_Comment, "%s", "RowDroop correction performed using ROWDRPCn for channel n");
fits_write_comment(ffp_FITS_Out,
                   CP_Comment,
                   &I_fits_return_status);

sprintf(CP_Keyname, "%s", "ROWDRPC1");
sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 1");
F_Num_In = Const->rowdroop[0];
fits_write_key(ffp_FITS_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
sprintf(CP_Keyname, "%s", "ROWDRPC2");
sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 2");
F_Num_In = Const->rowdroop[1];
fits_write_key(ffp_FITS_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
sprintf(CP_Keyname, "%s", "ROWDRPC3");
sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 3");
F_Num_In = Const->rowdroop[2];
fits_write_key(ffp_FITS_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
sprintf(CP_Keyname, "%s", "ROWDRPC4");
sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 4");
F_Num_In = Const->rowdroop[3];
fits_write_key(ffp_FITS_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);

fits_flush_file(ffp_FITS_Out, &I_fits_return_status);

fits_write_img( ffp_FITS_Out,
                TFLOAT,
                1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                FPP_Data_In,
                &I_fits_return_status);

fits_flush_file(ffp_FITS_Out,
                &I_fits_return_status);

fits_close_file(ffp_FITS_Out,
                &I_fits_return_status);

fits_close_file(ffp_FITS_In,
                &I_fits_return_status);

if(I_fits_return_status)
Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;

unc_out = 1;
/* write out the image uncertainty  data */
if (strcmp(Fnames->CP_Filename_FITS_Unc_Out, ""))
{
   fits_create_file( &ffp_FITS_Unc_Out,
                   Fnames->CP_Filename_FITS_Unc_Out,
                   &I_fits_return_status);
     
   if (strcmp(Fnames->CP_Filename_FITS_Unc_In, ""))
       fits_open_file( &ffp_FITS_Unc_In,
                       Fnames->CP_Filename_FITS_Unc_In,
                       READONLY, 
                       &I_fits_return_status);

   else
   {
       unc_out = 0;
       fits_open_file( &ffp_FITS_Unc_In,
                       Fnames->CP_Filename_FITS_In,
                       READONLY, 
                       &I_fits_return_status);
    } 
  

  
  fits_copy_header(ffp_FITS_Unc_In,
                   ffp_FITS_Unc_Out,
                   &I_fits_return_status);
  
  if (I_fits_return_status)
  {Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status; return;}

  /* Add module specific keywords */
  
  sprintf(CP_Comment, "%s", "Begin ROWDROOP module additions");
  fits_write_comment(ffp_FITS_Unc_Out,
                     CP_Comment,
                     &I_fits_return_status);
  if (!unc_out)
  {
       sprintf(CP_Comment, "%s", "This is the Uncertainty Image");
       fits_write_comment(ffp_FITS_Unc_Out,
                          CP_Comment,
                          &I_fits_return_status);
  }
  
  sprintf(CP_Comment, "%s", asctime(tmp_time));
  string_length = strlen(CP_Comment)-1;
  sprintf(format, "%%.%ds", string_length);
  sprintf(CP_Comment, format, asctime(tmp_time));
  fits_write_comment(ffp_FITS_Unc_Out,
                     CP_Comment,
                     &I_fits_return_status);
  sprintf(CP_Comment, "%s", "RowDroop correction performed using ROWDRPCn for channel n");
  fits_write_comment(ffp_FITS_Unc_Out,
                     CP_Comment,
                     &I_fits_return_status);
  
  sprintf(CP_Keyname, "%s", "ROWDRPC1");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 1");
  F_Num_In = Const->rowdroop[0];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC2");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 2");
  F_Num_In = Const->rowdroop[1];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC3");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 3");
  F_Num_In = Const->rowdroop[2];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC4");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 4");
  F_Num_In = Const->rowdroop[3];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  
  sprintf(CP_Keyname, "%s", "ROWDRPU1");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 1");
  F_Num_In = Const->rowdroop_unc[0];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU2");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 2");
  F_Num_In = Const->rowdroop_unc[1];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU3");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 3");
  F_Num_In = Const->rowdroop_unc[2];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU4");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 4");
  F_Num_In = Const->rowdroop_unc[3];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  
  fits_flush_file(ffp_FITS_Unc_Out, &I_fits_return_status);
  
  fits_write_img( ffp_FITS_Unc_Out,
                  TFLOAT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  FPP_Data_Unc_In,
                  &I_fits_return_status);
  
  fits_flush_file(ffp_FITS_Unc_Out,
                  &I_fits_return_status);
  
  fits_close_file(ffp_FITS_Unc_Out,
                  &I_fits_return_status);
  
  fits_close_file(ffp_FITS_Unc_In,
                  &I_fits_return_status);
  
  if(I_fits_return_status)
  Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;
}  

/* write out the Rowdroop uncertainty data */
if (strcmp(Fnames->CP_Filename_FITS_Rd_Unc_Out, ""))
{
   fits_create_file( &ffp_FITS_Unc_Out,
                   Fnames->CP_Filename_FITS_Rd_Unc_Out,
                   &I_fits_return_status);
     
   if (strcmp(Fnames->CP_Filename_FITS_Unc_In, ""))
       fits_open_file( &ffp_FITS_Unc_In,
                       Fnames->CP_Filename_FITS_Unc_In,
                       READONLY, 
                       &I_fits_return_status);

   else
   {
       unc_out = 0;
       fits_open_file( &ffp_FITS_Unc_In,
                       Fnames->CP_Filename_FITS_In,
                       READONLY, 
                       &I_fits_return_status);
    } 
  

  
  fits_copy_header(ffp_FITS_Unc_In,
                   ffp_FITS_Unc_Out,
                   &I_fits_return_status);
  
  if (I_fits_return_status)
  {Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status; return;}

  /* Add module specific keywords */
  
   sprintf(CP_Keyname, "%s", "NAXIS1");
   sprintf(CP_Comment, "%s", " Axis Length");
   I_Num_In = Dim_In->I_NY;  
   fits_update_key(ffp_FITS_Out, TINT,
                   CP_Keyname, &I_Num_In,
                   CP_Comment, &I_fits_return_status);

   sprintf(CP_Keyname, "%s", "NAXIS2");
   sprintf(CP_Comment, "%s", "Num of planes");
   I_Num_In = Dim_In->I_NZ ; 
   fits_update_key(ffp_FITS_Out, TINT,
                   CP_Keyname, &I_Num_In,
                   CP_Comment, &I_fits_return_status);

   sprintf(CP_Keyname, "%s", "NAXIS3");
   sprintf(CP_Comment, "%s", " Num of channels");
   I_Num_In = 4;
   fits_update_key(ffp_FITS_Out, TINT,
                   CP_Keyname, &I_Num_In,
                   CP_Comment, &I_fits_return_status);



   sprintf(CP_Comment, "%s", "This is the Rowdroop Uncertainty Image");
   fits_write_comment(ffp_FITS_Unc_Out,
                      CP_Comment,
                      &I_fits_return_status);
  
  sprintf(CP_Comment, "%s", asctime(tmp_time));
  string_length = strlen(CP_Comment)-1;
  sprintf(format, "%%.%ds", string_length);
  sprintf(CP_Comment, format, asctime(tmp_time));
  fits_write_comment(ffp_FITS_Unc_Out,
                     CP_Comment,
                     &I_fits_return_status);
  sprintf(CP_Comment, "%s", "RowDroop correction performed using ROWDRPCn for channel n");
  fits_write_comment(ffp_FITS_Unc_Out,
                     CP_Comment,
                     &I_fits_return_status);
  
  sprintf(CP_Keyname, "%s", "ROWDRPC1");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 1");
  F_Num_In = Const->rowdroop[0];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC2");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 2");
  F_Num_In = Const->rowdroop[1];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC3");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 3");
  F_Num_In = Const->rowdroop[2];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPC4");
  sprintf(CP_Comment, "%s", "Row Droop coupling constant for chan 4");
  F_Num_In = Const->rowdroop[3];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  
  sprintf(CP_Keyname, "%s", "ROWDRPU1");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 1");
  F_Num_In = Const->rowdroop_unc[0];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU2");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 2");
  F_Num_In = Const->rowdroop_unc[1];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU3");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 3");
  F_Num_In = Const->rowdroop_unc[2];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  sprintf(CP_Keyname, "%s", "ROWDRPU4");
  sprintf(CP_Comment, "%s", "Uncertainty in coupling constant for chan 4");
  F_Num_In = Const->rowdroop_unc[3];
  fits_write_key(ffp_FITS_Unc_Out, TFLOAT, CP_Keyname, &F_Num_In, CP_Comment, &I_fits_return_status);
  
  fits_flush_file(ffp_FITS_Unc_Out, &I_fits_return_status);
  
  fits_write_img( ffp_FITS_Unc_Out,
                  TFLOAT,
                  1, Dim_In->I_NY*Dim_In->I_NZ*4, 
                  FPP_Data_Rd_Unc,
                  &I_fits_return_status);
  
  fits_flush_file(ffp_FITS_Unc_Out,
                  &I_fits_return_status);
  
  fits_close_file(ffp_FITS_Unc_Out,
                  &I_fits_return_status);
  
  fits_close_file(ffp_FITS_Unc_In,
                  &I_fits_return_status);
  
  if(I_fits_return_status)
  Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;
}  
  if(Stat->I_Verbose)
  printf("rowdroop_output: Results written to file(s).\n");
  
  }
