

/*********************************************** 
rowdroop_read_data.c

Purpose

Read the image data from the FITS files and save into an array.
Note that 2 input files will be read
1 - A-D saturation corrected data file
2 - droop corrected data file

Overview

Open both input  FITS file, allocate memory, read the data into an array, 

Definitions of Variables

Input:
CP_Filename_FITS_In = In Data File Name.
CP_Filename_FITS_In_AD = In Data File Name with AD sturation correction.

Output:
fpp_FITS_in = Pointer to fits file handle.
fpp_FITS_Unc_in = Pointer to fits file handle.
fpp_FITS_in_AD = Pointer to fits file handle.
fpp_FITS_Unc_in_AD = Pointer to fits file handle.
FPP_Data_In = Pointer to image data in array form.
FPP_Data_Unc_In = Pointer to uncertainty image data in array form.
FPP_Data_In_AD = Pointer to image data in array form.
FPP_Data_Unc_In_AD = Pointer to uncertainty image data in array form.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "rowdroop.h"
#include "rowdroop_errcodes.h"
#include "rowdroop_defs.h"


rowdroop_read_data(	
            Filenames   *Fnames,
            Constants   *Const, 
            Dims        *Dim_In,
            float	**FPP_Data_In_AD,
            float	**FPP_Data_Unc_In_AD,
            float	**FPP_Data_In,
            float	**FPP_Data_Unc_In,
            float	**FPP_Data_Rd_Unc,
            Status	*Stat,
            short       **SPP_pmask,
            short       **SPP_dmask, 
            short       **SPP_bmask,
            short       **SPP_cmask) {
int	i, j, k;
int	I_fits_return_status=0, I_ndims_found, I_anynull;
int     I_naxis, I_bitpix;
long NX, NY, NZ;
long	LP_naxes[3];
float	D_nullval=0;

fitsfile 	*ffp_FITS_In_AD;
fitsfile 	*ffp_FITS_Unc_In_AD;
fitsfile 	*ffp_FITS_In;
fitsfile 	*ffp_FITS_Unc_In;

fitsfile 	*ffp_FITS_pmask;
fitsfile 	*ffp_FITS_dmask;
fitsfile 	*ffp_FITS_bmask;
fitsfile        *ffp_FITS_cmask;
char    fits_comment[80];

if(Stat->I_status) return;

/*  Open and get array data from input fits file */

fits_open_file(	&ffp_FITS_In,
		Fnames->CP_Filename_FITS_In,
		READONLY,
		&I_fits_return_status);

if(I_fits_return_status)
{
 printf("Error in opening file [%s]\n", Fnames->CP_Filename_FITS_In);
 Stat->I_status = READ_DATA|FOPEN_FAILED; 
 return;
}

fits_read_key( ffp_FITS_In,
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
if(I_bitpix != -32)
{Stat->I_status = READ_DATA|BIT_ERROR; return;}

fits_read_key( ffp_FITS_In,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_ndims_found == 2) LP_naxes[2] = 1;

Dim_In->I_NX = LP_naxes[0];
Dim_In->I_NY = LP_naxes[1];
Dim_In->I_NZ = LP_naxes[2];

/* Allocate memory for the data */
/* Don't try if problem with fits file */

if(I_fits_return_status)
{Stat->I_status = READ_DATA|KEYWORD_READ; return;}

*FPP_Data_In = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2], 
          sizeof(float));

/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In==NULL) I_fits_return_status = MALLOC_FAILED;

if (Const->Rd_Unc_Fits)
{
   *FPP_Data_Rd_Unc = (float *) calloc( LP_naxes[1]*LP_naxes[2]*4,
                        sizeof(float));
   if(*FPP_Data_Rd_Unc==NULL) I_fits_return_status = MALLOC_FAILED;
}
/* Read the data */

fits_read_img(ffp_FITS_In,
              TFLOAT,
              1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
              &D_nullval,
              *FPP_Data_In,
              &I_anynull,
              &I_fits_return_status);
if(I_fits_return_status)
{Stat->I_status = READ_DATA|COULDNOTREAD; return;}

fits_close_file(ffp_FITS_In,
		&I_fits_return_status);

if(Stat->I_Verbose)	
   printf("Input image file read successfully..continue\n");

/***********************************************************************/
/*  Open and read input uncertainty image file. */
/*  initialized if file does not exist */

*FPP_Data_Unc_In = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if(*FPP_Data_Unc_In==NULL) I_fits_return_status = MALLOC_FAILED;


I_fits_return_status = 0;
if (strcmp(Fnames->CP_Filename_FITS_Unc_In, "")) {
  fits_open_file(&ffp_FITS_Unc_In, 
		Fnames->CP_Filename_FITS_Unc_In,
		READONLY,
		&I_fits_return_status);

if(I_fits_return_status)
{
 printf("Error in opening file [%s]\n", Fnames->CP_Filename_FITS_Unc_In);
 Stat->I_status = READ_DATA|FOPEN_FAILED; 
 return;
}

  fits_read_key( ffp_FITS_Unc_In, 
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
  if(I_bitpix != -32)
   {Stat->I_status = READ_DATA|BIT_ERROR; return;}

  fits_read_key( ffp_FITS_Unc_In,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

   fits_read_keys_lng( ffp_FITS_Unc_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_ndims_found == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY || NZ != Dim_In->I_NZ)
      {Stat->I_status = READ_DATA|UNMATCHD_DIMS; return;}

/* Allocate memory for the data */
/* Don't try if problem with fits file */

   if(I_fits_return_status)
   {Stat->I_status = READ_DATA|KEYWORD_READ; return;}


/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
/* Read the data */

   fits_read_img(ffp_FITS_Unc_In, 
                 TFLOAT,
                 1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
                 &D_nullval,
                 *FPP_Data_Unc_In,
                 &I_anynull,
                 &I_fits_return_status);

   if(I_fits_return_status)
   {Stat->I_status = READ_DATA|COULDNOTREAD; return;}

   fits_close_file(ffp_FITS_Unc_In,
		   &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_DATA|KEYWORD_READ; return;}


}  
else
{
     for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
            *(*FPP_Data_Unc_In +i) = 0.0;
   }
}

if(Stat->I_Verbose)	
  printf("Input image Uncertainty file read successfully..continue\n");

/***********************************************************************/
/*  Open and read input image with A-to-D saturation correction.
    This will be used to compute the magnitude of rowdroop correction.
*/
I_fits_return_status = 0;
fits_open_file(	&ffp_FITS_In_AD,
		Fnames->CP_Filename_FITS_In_AD,
		READONLY,
		&I_fits_return_status);

if(I_fits_return_status)
{
 printf("Error in opening file [%s]\n", Fnames->CP_Filename_FITS_In_AD);
 Stat->I_status = READ_DATA|FOPEN_FAILED; 
 return;
}

fits_read_key( ffp_FITS_In_AD,
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
if(I_bitpix != -32)
   {Stat->I_status = READ_DATA|BIT_ERROR; return;}

fits_read_key( ffp_FITS_In_AD,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_In_AD,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_ndims_found == 2) LP_naxes[2] = 1;

NX = LP_naxes[0];
NY = LP_naxes[1];
NZ = LP_naxes[2];

if (NX != Dim_In->I_NX || NY != Dim_In->I_NY || NZ != Dim_In->I_NZ)
   {Stat->I_status = READ_DATA|UNMATCHD_DIMS; return;}


if(I_fits_return_status)
      {Stat->I_status = READ_DATA|KEYWORD_READ; return;}

*FPP_Data_In_AD = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2], 
          sizeof(float));

/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In_AD==NULL) I_fits_return_status = MALLOC_FAILED;
/* Read the data */

fits_read_img(ffp_FITS_In_AD,
              TFLOAT,
              1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
              &D_nullval,
              *FPP_Data_In_AD,
              &I_anynull,
              &I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|COULDNOTREAD; return;}

fits_close_file(ffp_FITS_In_AD,
		&I_fits_return_status);


if(Stat->I_Verbose)	
  printf("Input AD image file read successfully..continue\n");

/***********************************************************************/
/*  Open and read input uncertainty image file for AD. */
/*  initialized if file does not exist */

*FPP_Data_Unc_In_AD = (float *) calloc( Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
          sizeof(float));
if(*FPP_Data_Unc_In_AD==NULL) I_fits_return_status = MALLOC_FAILED;


I_fits_return_status = 0;
if (strcmp(Fnames->CP_Filename_FITS_Unc_In_AD, "")) {

  fits_open_file(&ffp_FITS_Unc_In_AD, 
		Fnames->CP_Filename_FITS_Unc_In_AD,
		READONLY,
		&I_fits_return_status);

if(I_fits_return_status)
{
 printf("Error in opening file [%s]\n", Fnames->CP_Filename_FITS_Unc_In_AD);
 Stat->I_status = READ_DATA|FOPEN_FAILED; 
 return;
}

  fits_read_key( ffp_FITS_Unc_In_AD, 
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
  if(I_bitpix != -32)
   {Stat->I_status = READ_DATA|BIT_ERROR; return;}

  fits_read_key( ffp_FITS_Unc_In_AD,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

   fits_read_keys_lng( ffp_FITS_Unc_In_AD,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if (I_naxis == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY || NZ != Dim_In->I_NZ)
      {Stat->I_status = READ_DATA|UNMATCHD_DIMS; return;}

/* Allocate memory for the data */
/* Don't try if problem with fits file */

   if(I_fits_return_status)
      {Stat->I_status = READ_DATA|KEYWORD_READ; return;}


/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
/* Read the data */

   fits_read_img(ffp_FITS_Unc_In_AD, 
                 TFLOAT,
                 1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
                 &D_nullval,
                 *FPP_Data_Unc_In_AD,
                 &I_anynull,
                 &I_fits_return_status);

   if(I_fits_return_status)
   {Stat->I_status = READ_DATA|COULDNOTREAD; return;}

   fits_close_file(ffp_FITS_Unc_In_AD,
		   &I_fits_return_status);



}  
else
{
     for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i) {
            *(*FPP_Data_Unc_In_AD +i) = 0.0;
   }
}

/***********************************************************************/
/*  Open and read pmask file.  Optional input. 
    If file can't be opened create dmask assuming all pixes are good.
    Otherwise read in from file.
*/

I_fits_return_status = 0;
*SPP_pmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Pmask, "")) {
   fits_open_file( &ffp_FITS_pmask,
                   Fnames->CP_Filename_Pmask,
                   READONLY,
                   &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|FOPEN_FAILED; return;}

   fits_read_keys_lng( ffp_FITS_pmask,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|KEYWORD_READ; return;}

   if (I_ndims_found == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY)
      {Stat->I_status = READ_MASK|UNMATCHD_DIMS; return;}


   fits_read_img( ffp_FITS_pmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_pmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_pmask, 
                   &I_fits_return_status);
}
else {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_pmask +i) = 0;
   }
}


if(Stat->I_Verbose)	
  printf("Pmask image file read successfully..continue\n");
/***********************************************************************/
/*  Open and read dmask file.  Optional input.  */

I_fits_return_status = 0;
*SPP_dmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Dmask, "")) {
   fits_open_file( &ffp_FITS_dmask,
                   Fnames->CP_Filename_Dmask,
                   READONLY,
                   &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|FOPEN_FAILED; return;}

   fits_read_keys_lng( ffp_FITS_dmask,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|KEYWORD_READ; return;}

   if (I_ndims_found == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY || NZ != Dim_In->I_NZ)
      {Stat->I_status = READ_MASK|UNMATCHD_DIMS; return;}


   fits_read_img( ffp_FITS_dmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  &D_nullval,
                  *SPP_dmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_dmask, 
                   &I_fits_return_status);
}
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i) {
       *(*SPP_dmask +i) = 0;
   }


if(Stat->I_Verbose)	
  printf("Dmask image file read successfully..continue\n");

/***********************************************************************/
/*  Open and read bmask file.  Optional input.  */

I_fits_return_status = 0;
*SPP_bmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));
if (strcmp(Fnames->CP_Filename_Bmask, "")) {
   fits_open_file( &ffp_FITS_bmask,
                   Fnames->CP_Filename_Bmask,
                   READONLY,
                   &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|FOPEN_FAILED; return;}

   fits_read_keys_lng( ffp_FITS_bmask,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|KEYWORD_READ; return;}

   if (I_ndims_found == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY)
      {Stat->I_status = READ_MASK|UNMATCHD_DIMS; return;}


   fits_read_img( ffp_FITS_bmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_bmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_bmask, 
                   &I_fits_return_status);
}
else {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_bmask +i) = 0;
   }
}


if(Stat->I_Verbose)	
  printf("Bmask image file read successfully..continue\n");
/*  Open and read cmask file.  Optional input.  */

I_fits_return_status = 0;
*SPP_cmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Cmask, "")) {
   fits_open_file( &ffp_FITS_cmask,
                   Fnames->CP_Filename_Cmask,
                   READONLY,
                   &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|FOPEN_FAILED; return;}

   fits_read_keys_lng( ffp_FITS_cmask,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

   if(I_fits_return_status)
      {Stat->I_status = READ_MASK|KEYWORD_READ; return;}

   if (I_ndims_found == 2) LP_naxes[2] = 1;

   NX = LP_naxes[0];
   NY = LP_naxes[1];
   NZ = LP_naxes[2];

   if (NX != Dim_In->I_NX || NY != Dim_In->I_NY || NZ != Dim_In->I_NZ)
      {Stat->I_status = READ_MASK|UNMATCHD_DIMS; return;}


   fits_read_img( ffp_FITS_cmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_cmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_cmask,
                   &I_fits_return_status);
}
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i) {
       *(*SPP_cmask +i) = 0;
   }


if(Stat->I_Verbose)	
  printf("Cmask image file read successfully..continue\n");

if(Stat->I_Verbose)	{
  printf("rowdroop_read_data: Data read from file:\n%s\n",
  Fnames->CP_Filename_FITS_In);
  printf("rowdroop_read_data: I_NX = %d\n", Dim_In->I_NX);
  printf("rowdroop_read_data: I_NY = %d\n", Dim_In->I_NY);
  printf("rowdroop_read_data: I_NZ = %d\n", Dim_In->I_NZ);
}

}
