#define PROGRAM "slope_finder"
/* This module does slope estimatation for IRS ramps. (previously done by "imagest") */

#define VERSION "1.4"
/* UPDATES ********************************************************************
1.4 080923 fixed bug which was including deltas from "initial delete" samples when computing stdev
1.3 080221 update BUNIT keyword card to "e-/sec"
           replaced 2 includes of subroutine call defs (im and param) with actual call defs
1.2 071002 fixed bug misprocessing radhit sample deleting CR-7908
           removed many DEBUG statements
1.1 070918 fixed bug where radhit after super-small subramp
1.0 070904 added optional debug outupt -info_oulier
           added option to skip initial samples -skip_n  -skip_minramp
           added slope threshold for deleting after radhit -radhit_thresh
           (removed radhit correction logic)
0.7y 070830 added optional debug outupt -info_radhit
0.7x 070828 *** radhit correction very preliminary ***
           new parameter: max samples deleted after radhit
           add option to apply slope correction factor after rasdhits (dfX bits)
           fixed bug in peakup saturation truncation
0.6 070815 took abs(varriance_correlated)
0.5 070730 fixed bug for p_sig2 when points are missing from segment
0.4 070312 add ramp-truncation for saturated peakups
           exclude "BAD" samples from bit propagation from DMASK to BMASK
0.3 070306 fixed uncertainty calculations to use constant p_unc
            removed unc generation (will put in SNEST_IRS)
            removed readnoise parameter (don't seem necessary)
            added "dual" (pre/post outlier) and "o2" (method2 mean deltas)
0.2 070221 made outlier rejection more efficient
0.1 070220 changed name to slope_finder (slopest sounded like "slow pest")
0.3 070116 initial prototype

Notes on function/terminology:
 "method1" is the the Karl Gordon method (correlated uncertainties)
 "method2" is the mean of deltas method

Notes on code itself:
 1. The MAIN routine is at bottom of page.
 2. Calls subroutine in: downlink/pipeline/include/tools/nl_and_params.c
******************************************************************************/

#define DMASK_BIT_SAT    8 /* (bit 3) in DMASK, bit for saturation */
                          /* was    8192 (bit 13) */ 
#define BMASK_BIT_PMASK 16384 /* (bit 14) in BMASK, bit for PMASK fatal */

/* image size info */
#define NAXIS1           128
#define NAXIS2           128
#define MAX_NAXIS3        16
#define NPIX_PLANE       (NAXIS1*NAXIS2)
#define MAXPIX_CUBE      (NPIX_PLANE*MAX_NAXIS3)

/* boundries of peakup areas (coordinates from 0 to 127) */
                /* subtracted 1 from cropfits.nl */
#define PEAKUP_R_X1  88
#define PEAKUP_R_X2 121
#define PEAKUP_R_Y1  68
#define PEAKUP_R_Y2 116
#define PEAKUP_B_X1  89
#define PEAKUP_B_X2 123
#define PEAKUP_B_Y1   6
#define PEAKUP_B_Y2  54
#define PEAKUP_R_NPIX ((PEAKUP_R_X2-PEAKUP_R_X1+1)*(PEAKUP_R_Y2-PEAKUP_R_Y1+1))
#define PEAKUP_B_NPIX ((PEAKUP_B_X2-PEAKUP_B_X1+1)*(PEAKUP_B_Y2-PEAKUP_B_Y1+1))

/* status values for input samples */
#define SAMPLE_OK    0
#define SAMPLE_JUMP 1 /* break in continuity from previous pixel */
#define SAMPLE_BAD   2 /* NaN or FATAL */
#define SAMPLE_BAD_JUMP   (SAMPLE_BAD | SAMPLE_JUMP)

#define MAXFILENAMESIZE  256

/* for -info_radhit file */
#define PLANE_JUMP   (0*NPIX_PLANE)
#define PLANE_BEFORE (1*NPIX_PLANE)
#define PLANE_AFTER  (2*NPIX_PLANE)
#define PLANE_COUNT  (3*NPIX_PLANE)

/* includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "nanvalue.h"

/* Added TYB 5/1/09 */
#include "fitsio.h"

/* definitions of calls to the "im" subroutine library */
int im_open(char* filename, char* mode);
int im_hdr(int fid, int* in_fids, int op);
int im_close(int fid);
int im_rkey_i(int fid, int* value, char* keyword);
int im_wkey_i(int fid, int value, char* keyword);
int im_wkey_t(int fid, char* card, char* operation);
int im_wkey_c(int fid, char* card, char* operation);
int im_rpix_d(int fid, double* pixel_array, int max_pixels);
int im_rpix_s(int fid, short*  pixel_array, int max_pixels);
int im_wpix_d(int fid, double* pixel_array, int n_pixels);
int im_wpix_s(int fid, short*  pixel_array, int n_pixels);

/* definitions of calls to the "param" subroutine library 070824 */
void param(void* param_value, char* info1, char* info2);
int nl_process(int argc, char** argv,
               char* program_name, char* program_version,
               char* namelist_start_code, char* namelist_end_code);
int param_begin(int argc, char** argv,
               char* program_name, char* program_version);
int param_exit(int status, char* program_name);
void param_print();

/* PARAMETERS (from command line & namelist) */
/* allocated static to scope within this source file */
static char filename_influx[MAXFILENAMESIZE];
static char filename_inunc[MAXFILENAMESIZE];
static char filename_indmask[MAXFILENAMESIZE];
static char filename_inpmask[MAXFILENAMESIZE];
static char filename_outbmask[MAXFILENAMESIZE];
static char filename_o_flux[MAXFILENAMESIZE];
static char filename_o_unc[MAXFILENAMESIZE];
static char filename_o_pre_flux[MAXFILENAMESIZE];
static char filename_o_pre_unc[MAXFILENAMESIZE];
static char filename_o_dual_flux[MAXFILENAMESIZE];
static char filename_o_dual_unc[MAXFILENAMESIZE];
static char filename_o2_flux[MAXFILENAMESIZE];
static char filename_o2_unc[MAXFILENAMESIZE];
static char filename_o2_pre_flux[MAXFILENAMESIZE];
static char filename_o2_pre_unc[MAXFILENAMESIZE];
static char filename_o2_dual_flux[MAXFILENAMESIZE];
static char filename_o2_dual_unc[MAXFILENAMESIZE];
static char filename_info_outlier[MAXFILENAMESIZE];
static char filename_info_radhit[MAXFILENAMESIZE];
static char bitpass_string[MAXFILENAMESIZE];
static int  pmask_fatal;
static int  dmask_fatal0;
static int  dmask_fatal1;
static int  dmask_radhit;
static int radhit_maxdel;
static double radhit_thresh; 
static int outlier_n_iter; /* N iterations of outlier rejection */
static double outlier_n_sigma; /* for outlier rejection by slope */
static int outlier_min_deltas; /* no outlier reject if less than this */
static int  min_subramp;
static int  init_skip_n;
static int  init_skip_minramp;
static double  saturation_pct;
static double  exposure_time;
static double gain;
static int  verbose;
static char filename_nl[MAXFILENAMESIZE];

/* GLOBAL data */
double normalization_factor;


/**************************************************************************/
/* UTILITY SUBROUTINE(S) */
/**************************************************************************/

void FATAL_char (char *message, char *param) {
/* prints ERROR message, then exits */
  /* prints error msg, with printf parameter; then exits with error status */
  char msg_nl[300];
  strcpy(msg_nl, "ERROR: ");
  strcat(msg_nl, message);
  strcat(msg_nl, "\n");
  printf(msg_nl, param);
  exit(64); /* abnormal termination */
}


/**************************************************************************/
void process_parameters( /* get parameters fomr command line and namelist */
/**************************************************************************/
 int argc, char** argv)
{
  int stat;

  param(filename_influx,  "-i IN_FLUX input flux file","infile");
  param(filename_inunc,   "-iu IN_UNC input uncertainty file","infile");
  param(filename_indmask, "-id IN_DMASK input DMASK file","infile - NONE");
  param(filename_inpmask, "-ip IN_PMASK input PMASK file","infile - NONE");

  param(filename_o_flux, "-o_flux OUT_FLUX output flux file", "outfile");
  param(filename_o_unc,  "-o_unc  OUT_UNC output uncertainty file", "outfile - NONE");
  param(filename_outbmask, "-ob  OUT_BMASK output BMASK file", "outfile - NONE");

  param(&pmask_fatal,    "-pf  PMASK_FATAL bits (pixel bad)", "int - 0");
  param(&dmask_fatal0,   "-df0  DMASK_FATAL0 bits (sample good, but delta from prev is bad)", "int - 0");
  param(&dmask_fatal1,   "-df1  DMASK_FATAL1 bits (sample bad)", "int - 0");
  param(&dmask_radhit,   "-dfX  DMASK_RADHIT bit (sample is a radhit)", "int - 0");
  param(&radhit_maxdel,  "-radhit_maxdel RADHIT_MAX_DEL max samples to delete after a radhit", "int - 1");
  param(&radhit_thresh,  "-radhit_thresh RADHIT_THRESH max slope for deleting samples following a radhit", "double - 0.0");
  param(&bitpass_string, "-m  BITPASS from DMASK to BMASK 3:4+12:2+1:7", "string - NONE");

  param(&exposure_time, "-t  EXPOSURE_TIME", "double - 1.0");
  param(&gain, "-gain  GAIN DN to elec/sec", "double - 4.6");

  param(&saturation_pct, "-sat_pct  PEAKUP_SAT_PCT Pct of pixels in either peakup area to reject entire plane (LL only)", "double - 200");

  param(&outlier_n_iter, "-rej_iter  OUTLIER_N_ITERATIONS max number of outlier rejection iterations", "int - 2");
  param(&outlier_n_sigma, "-rej_nsig  OUTLIER_N_SIGMA outlier rejection slope n_sigma ", "double - 2.0");
  param(&outlier_min_deltas, "-rej_min  OUTLIER_MIN_DELTAS outlier rejection min deltas", "int - 4");
  param(&min_subramp, "-mr  MIN_SUBRAMP minimum number of samples in a ramp or sub-ramp", "int - 2");
  param(&init_skip_n, "-skip_n  SKIP_N number of inintial samples to skip", "int - 0");
  param(&init_skip_minramp, "-skip_minramp  SKIP_MINRAMP minimum rampsize before skipping any inintial samples", "int - 5");

  param(&verbose, "-v VERBOSE level of printed messages", "int - 1");
  param(filename_nl,  "-n NAMELIST_FILE name of Namelist file", "infile - namelist.nl");

  param(filename_o_pre_flux,  "-o_pre_flux  OUT_PRE_FLUX output pre-outlier flux file", "outfile - NONE");
  param(filename_o_pre_unc,   "-o_pre_unc   OUT_PRE_UNC output pre-outlier rej. uncertainty file", "outfile - NONE");
  param(filename_o_dual_flux,  "-o_dual_flux  OUT_BOTH_FLUX output b2-plane flux file", "outfile - NONE");
  param(filename_o_dual_unc,  "-o_dual_unc  OUT_BOTH_UNC output b2-plane uncertainty file", "outfile - NONE");

  param(filename_o2_flux,  "-o2_flux  OUT2_PRE_FLUX method2 post-outlier flux file", "outfile - NONE");
  param(filename_o2_unc,   "-o2_unc   OUT2_PRE_UNC method2 post-outlier rej. uncertainty file", "outfile - NONE");
  param(filename_o2_pre_flux,  "-o2_pre_flux  OUT2_PRE_FLUX method2 pre-outlier flux file", "outfile - NONE");
  param(filename_o2_pre_unc,   "-o2_pre_unc   OUT2_PRE_UNC method2 pre-outlier rej. uncertainty file", "outfile - NONE");
  param(filename_o2_dual_flux,  "-o2_dual_flux  OUT2_BOTH_FLUX method2 b2-plane flux file", "outfile - NONE");
  param(filename_o2_dual_unc,  "-o2_dual_unc  OUT2_BOTH_UNC method2 b2-plane uncertainty file", "outfile - NONE");
  param(filename_info_outlier,  "-o_info_outlier  INFO_OUTLIER outlier information FITS file", "outfile - NONE");
  param(filename_info_radhit,  "-o_info_radhit  INFO_RADHIT radhit information FITS file", "outfile - NONE");

  /**ENDPARAMS **/
  
  /* parse namelist and params */
  stat = nl_process(argc, argv, PROGRAM, VERSION, "&SLOPE_FINDER", "&END");
  if (stat != 0) exit(64); /* abnormal termination */

  stat = param_begin(argc, argv, PROGRAM, VERSION);
  if (stat != 0) exit(64); /* abnormal termination */

  param_print();   /* Print the parameter settings to stdout. */
}

/**************************************************************************/
void read_fits_image( /* reads a fits image  */
/**************************************************************************/
/* IN */
 char* filename, /* name of file to read */
/* IN/OUT */
 int* _nsamp_cube, /* pointer to number of sample in image -- 
                  value returned if value pointed to is 0
                  validated otherwise */
/* OUT */
 double* array_r8, /* sample array of type double (0 if using short) */
 short*  array_i2, /* sample array of type short (0 if using double) */
 int* _fid) /* fid from IM open -- ignored if pointer is 0 */
{
  int fid;
  int i, naxis, npix;

  /* open the fie */
  fid = im_open(filename, "r");
  if (fid < 0) FATAL_char("Trouble opening file %s", filename);

  /* check dimensions */
  im_rkey_i(fid, &naxis, "NAXIS");
  im_rkey_i(fid, &i, "NAXIS1");
  if (i != NAXIS1) FATAL_char("File %s has incorrect NAXIS1", filename);
  im_rkey_i(fid, &i, "NAXIS2");
  if (i != NAXIS2) FATAL_char("File %s has incorrect NAXIS2", filename);

  if (_nsamp_cube==0) { /* 2D image */
    if (naxis != 2) FATAL_char("File %s should have NAXIS=2" , filename);
    npix = NPIX_PLANE;
  }
  else { /* 3D image */
    if (naxis != 3) FATAL_char("File %s should have NAXIS=3" , filename);
    im_rkey_i(fid, &i, "NAXIS3");
    if (*_nsamp_cube==0) { /* needs to be set */
      *_nsamp_cube = NPIX_PLANE * i;
      if (*_nsamp_cube > MAXPIX_CUBE)
           FATAL_char("File %s NAXIS3 is too large", filename);
    } else { /* already set */
      if (i != *_nsamp_cube/NPIX_PLANE) FATAL_char("File %s has incorrect NAXIS3", filename);
    }
    npix = *_nsamp_cube;
  }

  /* read the pixels in */
  if (array_r8) i = im_rpix_d(fid, array_r8, npix);
  else          i = im_rpix_s(fid, array_i2, npix);
  if (i != npix) FATAL_char("Trouble reading pixels from %s", filename);

  /* return FID if requested, otherwise close file */
  if (_fid) {
    *_fid = fid; /* return fid if pointer was given and leave open to allow copying header */
  }
  else im_close(fid);
}


/**************************************************************************/
void set_samp_status( /* set samp_status */
/**************************************************************************/
/* IN */
double* influx, double* inunc, short* inpmask, short* indmask, int nsamp_cube,
/* OUT */
 int* samp_status, int* samp_radhit)
{
  int i1, i2, i3, j;
  int count_truncate = 0;
  int count_pmask = 0;
  int count_nan = 0;
  int count_dmask_radhit = 0;
  int count_dmask_fatal = 0;
  int count_dmask_break = 0;
  int count_initial = 0; /* initial deleted pixels */
  int count_ok = 0;
  int peakup_sat_n, peakup_oversaturated;
  double peakup_sat_pct;

  /* compute normalization factor */
  normalization_factor = (nsamp_cube / NPIX_PLANE) / exposure_time;
  if (verbose >= 1) printf ("Normalization factor = %f\n",normalization_factor);

  /* initialize status */
  for(i3 = 0; i3<nsamp_cube; i3 ++) {
    samp_status[i3] = SAMPLE_OK;
    samp_radhit[i3] = 0;
  }

  /* throw out planes where peakup is saturated */
  if (saturation_pct <= 100.0) {
    peakup_oversaturated = 0; /* over-saturation not found yet */
    for (i1 = 0; i1<nsamp_cube; i1+=NPIX_PLANE) { /* do planes */

      /* check if RED peakup is over-saturated */
      peakup_sat_n = 0;
      for (i2 = i1+(PEAKUP_R_Y1)*NAXIS1;
            i2<=i1+(PEAKUP_R_Y2)*NAXIS1; i2+=NAXIS1) { /* do lines */
        for (i3 = i2+PEAKUP_R_X1;
              i3<=i2+PEAKUP_R_X2; i3++) { /* do samples */
           if (indmask[i3] & DMASK_BIT_SAT) peakup_sat_n ++;
        }
      }
      peakup_sat_pct = 100.0 * (double) peakup_sat_n / PEAKUP_R_NPIX;
      if (peakup_sat_pct >= saturation_pct) peakup_oversaturated = 1;

      /* check if BLUE peakup is over-saturated */
      peakup_sat_n = 0;
      for (i2 = i1+(PEAKUP_B_Y1)*NAXIS1;
            i2<=i1+(PEAKUP_B_Y2)*NAXIS1; i2+=NAXIS1) { /* do lines */
        for (i3 = i2+PEAKUP_B_X1;
              i3<=i2+PEAKUP_B_X2; i3++) { /* do samples */
          if (indmask[i3] & DMASK_BIT_SAT) peakup_sat_n ++;
        }
      }
      peakup_sat_pct = 100.0 * (double) peakup_sat_n / PEAKUP_B_NPIX;
      if (peakup_sat_pct >= saturation_pct) peakup_oversaturated = 1;

      /* if over-saturated detected, set all samples in plane to BAD */
      if (peakup_oversaturated != 0) {
        printf ("**** Peakup array saturation, plane %d ****\n", 1+ i1/NPIX_PLANE);
        for (i3 = i1; i3<i1+NPIX_PLANE; i3++) {
          samp_status[i3] = SAMPLE_BAD_JUMP;
          count_truncate++;
        }
      }

    } /* end for i1 */
  } /* end if want to check for peakup saturation */


  /* flag first N sampless */
  if ( (init_skip_n > 0) && (nsamp_cube/NPIX_PLANE >= init_skip_minramp) ) {
    for (i2=0; i2<NPIX_PLANE; i2++) {
      for(i3 = i2; i3<init_skip_n*NPIX_PLANE; i3 += NPIX_PLANE) {
        samp_status[i3] = SAMPLE_BAD_JUMP;
        count_initial++;
      }
    }
  }

  /* flag the "bad" samples */
  for (i2=0; i2<NPIX_PLANE; i2++) {
    for(i3 = i2; i3<nsamp_cube; i3 += NPIX_PLANE) {
      if (samp_status[i3] != SAMPLE_OK) continue; /* already flagged */
      if (inpmask[i2] & pmask_fatal) { /*pmask is 2D array */
        samp_status[i3] = SAMPLE_BAD_JUMP;
        count_pmask++;
      } else if (iznanorinfd(inunc[i3]))  {
        samp_status[i3] = SAMPLE_BAD_JUMP;
        count_nan++;
      } else if (indmask[i3] & dmask_radhit) {
        samp_radhit[i3] = 1;
        samp_status[i3] = SAMPLE_JUMP;
        count_dmask_radhit++;
      } else if (indmask[i3] & dmask_fatal1) {
        samp_status[i3] = SAMPLE_BAD;
        count_dmask_fatal++;
      } else if (indmask[i3] & dmask_fatal0) {
        samp_status[i3] = SAMPLE_JUMP;
        count_dmask_break++;
      }
    } /* end for i3 */
  } /* end for i2 */

  /* print out statistics */
  if(verbose >= 2) {
    count_ok = nsamp_cube - (count_truncate + count_pmask + count_nan
          + count_dmask_fatal + count_dmask_break + count_dmask_radhit
          + count_initial);
    printf("samples status: peakup_sat=%d pmask=%d NaN=%d\n",
                       count_truncate, count_pmask, count_nan);
    printf("         radhit=%d dmask_fatal=%d dmask_break=%d\n",
                      count_dmask_radhit, 
                      count_dmask_fatal, count_dmask_break);
    printf("              initial=%d, ok=%d\n",
                       count_initial,  count_ok);
  }

}


/**************************************************************************/
void compute_deltas(
/**************************************************************************/
/* IN */
 double influx[],
 int pix_needs_update[],
 int nsamp_cube,
/* UPDATED */
 int samp_status[],
/* OUT */
 double delta[], /* deltas (3D) */
 int    delta_n[], /* number of deltas (2D) */
 double delta_mean[], /* mean of deltas (2D) */
 double delta_stdev[], /* standard dev deltas (2D) */
 double outflux2[], 
 double outunc2[]) 

{
  double sum, sum_sqd;
  int n, i2, i3;
  int n_bads_between; /* number of break samples before this */
  double correction_factor;
  int count_delta_n[16];
  int distance;

  for (n = 0; n<16; n++) count_delta_n[n] = 0;
  correction_factor = (nsamp_cube / NPIX_PLANE) / exposure_time;

  /* compute deltas */
  for (i2=0; i2<NPIX_PLANE; i2++) {
    if (pix_needs_update[i2]==0) continue; /* already computed, unchanged */
    n_bads_between = 0;
    n       = 0;
    sum     = 0.0;
    sum_sqd = 0.0;
    /* skip initial bad pixels */
    for(i3 = i2           ; i3<nsamp_cube; i3 += NPIX_PLANE) {
      if ( (samp_status[i3] & SAMPLE_BAD) == 0) break;
    }
    /* find all the deltas */
    for(i3 = i3+NPIX_PLANE; i3<nsamp_cube; i3 += NPIX_PLANE) {
      if (samp_status[i3] & SAMPLE_JUMP) {
        n_bads_between = 0;
      } else if (samp_status[i3] & SAMPLE_BAD) {
        n_bads_between++;
      } else {
        distance = n_bads_between+1;
        delta[i3] = (influx[i3] - influx[i3- distance*NPIX_PLANE]) / distance;
        n_bads_between = 0;
        sum     += delta[i3];
        sum_sqd += delta[i3] * delta[i3];
        n ++;
      }
    }
    delta_n[i2]          = n;
    count_delta_n[n]++;
    if (n>0) delta_mean[i2]  = sum / n;
    else     delta_mean[i2]  = NANVALUE;
    if (n>1) delta_stdev[i2] = sqrt( (sum_sqd - sum*sum/n) / (n-1.0) );
    else     delta_stdev[i2] = NANVALUE;
    outflux2[i2] = delta_mean[i2]              * correction_factor;
    outunc2[i2]  = (delta_stdev[i2] / sqrt(n)) * correction_factor;
  }

  if (verbose >=3) {
    for (n = 0; n<16; n++) {
      printf ("%2d deltas = %d\n", n, count_delta_n[n]);
    }
  }
}


/**************************************************************************/
int flag_outliers(
/**************************************************************************/
/* IN */
 double delta[],
 int    delta_n[],
 double delta_mean[],
 double delta_stdev[],
 int pix_needs_update[],
 int nsamp_cube,
 short n_deltas_left[], /* for info_outlier output file */
 int iteration_n,
/* UPDATE */
 int samp_status[])
{
  double min_good, max_good; /* cutoffs for outliers */
  int i2, i3;
  int count_outliers = 0;

  for (i2=0; i2<NPIX_PLANE; i2++) {
    if (pix_needs_update[i2] == 0) continue; /* all already found */
    pix_needs_update[i2] = 0;
    if (delta_n[i2] < outlier_min_deltas) continue;
    if (iznanorinfd(delta_mean[i2]))  continue;
    if (iznanorinfd(delta_stdev[i2])) continue;
    min_good = delta_mean[i2] - delta_stdev[i2] * outlier_n_sigma;
    max_good = delta_mean[i2] + delta_stdev[i2] * outlier_n_sigma;

    for(i3 = i2+NPIX_PLANE; i3<nsamp_cube; i3 += NPIX_PLANE) {
      if (samp_status[i3] != SAMPLE_OK) continue;
      if (delta[i3] < min_good || delta[i3] > max_good) {
        if (i3+NPIX_PLANE < nsamp_cube
          && samp_status[i3+NPIX_PLANE] == SAMPLE_OK
          && delta[i3+NPIX_PLANE] >= min_good
          && delta[i3+NPIX_PLANE] <= max_good)
        { /* a "JUMP" (next samp relatively continuous) */
          samp_status[i3] = SAMPLE_JUMP;
        } else { /* single bad point (but next samp relatively continuous) */
          samp_status[i3] = SAMPLE_BAD;
        }
        count_outliers++;
        pix_needs_update[i2] = 1;
        n_deltas_left[i2+iteration_n*NPIX_PLANE]
         = n_deltas_left[i2+(iteration_n-1)*NPIX_PLANE] - 1;
        break; /* flag max of one pixel per iteration */
      }
    }
  } /* end for i2 */

  if (verbose >=2) {
    printf ("outliers = %d\n", count_outliers);
  }

  return count_outliers;

} /* end flag_outliers() */


/**************************************************************************/
void delete_after_radhit(
/**************************************************************************/
/* IN */
 double delta_mean[NPIX_PLANE],
 int samp_radhit[MAXPIX_CUBE],
 int nsamp_cube,
/* UPD */
 int samp_status[MAXPIX_CUBE])
{
  int i2, i3;
  int k, k_max;
  int count_radhit_del = 0; /* number of samples deleted after radhits */
  for (i2=0; i2<NPIX_PLANE; i2++) {
    if (iznanorinfd(delta_mean[i2]))  continue;
    if (delta_mean[i2] * normalization_factor > radhit_thresh) continue;
    for(i3 = i2+NPIX_PLANE; i3<nsamp_cube; i3 += NPIX_PLANE) {
      if (samp_radhit[i3] == 0) continue;
      k_max = i3 + (radhit_maxdel*NPIX_PLANE);
      if (k_max > nsamp_cube) k_max = nsamp_cube;
      for (k=i3; k<k_max; k += NPIX_PLANE) {
        samp_status[k] = SAMPLE_BAD_JUMP;
        count_radhit_del++;
      }
    }
  }

  if (verbose >= 1) printf("For radhits, %d samples deleted\n",
                    count_radhit_del);

} /* end delete_after_radhit() */


/**************************************************************************/
void fit_a_la_karl_gordon( /* compute slope of a segment */
/**************************************************************************/
/* IN */
 double x[], /* x-coordinate (~sample number) */
 double y[], /* flux */
 double y_sig[], /* sigma of flux */
 double gain, /* conversion factor from original DN */
 int n, /* number of elements in arrays (use 0 to n-1) */
/* OUT */
 double* _slope, double* _sigma) /* computed slope and sigma */
{
  /* major variables */
  double S, Sx, Sxx, Sy, Sxy; /* from first loop */
  double sum_w, sum_xw; /* from second loop */
  double D; /* capital Delta in the Karl Gordon paper */
  double variance_random, variance_correlated;
  double slope_final, sigma_final; /* final (returned) values */

  /* tmp vrbls */
  int i;
  double w; /* weight (1/sig**2) */
  double p_sig2_nominal; /* sigma**2 if delta_x is 1.0 */
  double delta_x;
  double p_sig2; /* sigma**2 for current sample */
  double tmp;

  /* compute sums and sums-of-sums and D */
  S = 0.0; Sx = 0.0; Sxx = 0.0; Sy = 0.0; Sxy = 0.0;
  for(i = n-1; i >= 0; i--) { /* sumUp from last to first+1 */
    w = 1.0/(y_sig[i]*y_sig[i]);
    S   += w;
    Sx  += x[i]*w;
    Sxx += x[i]*x[i]*w;
    Sy  += y[i]*w;
    Sxy += x[i]*y[i]*w;

/**** old (hard) way:
    if (i != 0) tmp = y[i] - y[i-1];
    else        tmp = y[i];
    Sy  += tmp * S;
    Sxy += tmp * Sx;
*****/
  } 

  /* compute D, slope_final, p_sig2_nominal */
  D = S*Sxx - Sx*Sx;
  slope_final = (S*Sxy - Sx*Sy)/D;
  p_sig2_nominal = slope_final * gain; /* from an email from Karl Gordon */

  /* compute variance_correlated */
  variance_correlated = 0.0;
  sum_w = 0.0;
  sum_xw = 0.0;
  for(i = n-1; i > 0; i--) {
    delta_x = x[i] - x[i-1];
    if (delta_x <= 0) delta_x = 1.0; /* shouldn't happen, but just in case */
    p_sig2 = p_sig2_nominal * delta_x;
    w = 1.0 / (y_sig[i]*y_sig[i]);
    sum_w += w;
    sum_xw += x[i]*w;
    tmp = S*sum_xw - Sx*sum_w;
    variance_correlated += ( p_sig2 / (D*D) ) * (tmp*tmp);
  }

  /* combine noise components */
  variance_random = S / D;

  if (variance_correlated < 0.0) variance_correlated = -variance_correlated; /* abs() kludge */

  sigma_final = sqrt(variance_random + variance_correlated);

  /* store return values */
  *_slope = slope_final;
  *_sigma = sigma_final;

} /* end fit_a_la_karl_gordon */


/**************************************************************************/
void compute_slopes( /* compute all the slopes (calls fit_a_la_karl_gordon) */
/**************************************************************************/
/* IN */
 double* influx, double* inunc, int* samp_status, int* samp_radhit,
 int nsamp_cube,
 short* info_outlier,
/* OUT */
 double* outflux, double* outunc, double* info_radhit)
{
int k; /* for debug */
  int i2, i3, i4, j;
  double x; /* sample number within input ramp ramp */
  int samp; /* sample number in data sent to fitting subroutine */
  double t_outflux;
  double t_outunc;
  double t_wgt;
  double total_wgt; /* runnung sum of all wgts of all segments */
  double total_flux_wgt; /* runnung sum of all flux*wgt of all segments */
  int n_segments; /* number of valid segments */
  double ramp_x  [MAX_NAXIS3];
  double ramp_val[MAX_NAXIS3];
  double ramp_unc[MAX_NAXIS3];
  double radhit_prev_flux;
  int radhit_prev_n;
  int n_radhit_adj;
  double delta;
  int j_max;

  for(i3 = 0; i3<4*NPIX_PLANE; i3 ++) { info_radhit[i3] = 0.0; }

  /* go through pixels, one by one */
  for (i2=0; i2<NPIX_PLANE; i2++) {
    total_wgt = 0.0;
    total_flux_wgt = 0.0;
    n_segments = 0;
    n_radhit_adj = 0;
    radhit_prev_flux = 0.0;
    radhit_prev_n = 0;

    /* find segments and call Karl Gordon's formula */
    x = 0;
    samp=0;
    for(i3 = i2; i3<nsamp_cube; i3 += NPIX_PLANE) {

      /* add this point to the ramp (unless "bad") */
      if ((samp_status[i3] & SAMPLE_BAD) ==0) {
        ramp_val[samp] = influx[i3];
        ramp_unc[samp] = inunc[i3];
        ramp_x  [samp] = x;

        samp++;
      }
      x += 1.0;

      /* compute slope if next pixel is a break or this pixel is last */
      if ( (i3>=nsamp_cube-NPIX_PLANE)
        || ((samp_status[i3+NPIX_PLANE] & SAMPLE_JUMP)!=0) ) { 

        if (samp >= min_subramp) {

          /* call Karl Gordon to compute slope of segment */
          fit_a_la_karl_gordon( ramp_x, ramp_val, ramp_unc, gain, samp,
                                &t_outflux, &t_outunc);

          /* add in the slope of this segment if it was found */
          if (t_outunc > 0.0) {
            t_wgt = 1.0 / (t_outunc * t_outunc);
            total_wgt += t_wgt;
            total_flux_wgt += t_outflux * t_wgt;
            n_segments ++;
          }

          if (radhit_prev_flux != 0.0) {
            info_radhit[i2+PLANE_COUNT] += 1.0;
            if (info_radhit[i2+PLANE_COUNT] > 1.0) continue;
            info_radhit[i2+PLANE_JUMP]  = delta;
            info_radhit[i2+PLANE_BEFORE]  = radhit_prev_flux;
            info_radhit[i2+PLANE_AFTER]  = t_outflux;
            if (verbose >= 4) {
              printf("RADHIT%d %f %d %f %f\n", n_radhit_adj, delta,
                     radhit_prev_n, radhit_prev_flux, t_outflux);
              if (verbose >= 5) {
              for (i4=i2; i4<nsamp_cube; i4+=NPIX_PLANE)
               printf("SAMP %2d %f %d %d\n",
                  i4/NPIX_PLANE, influx[i4], samp_status[i4], samp_radhit[i4]);
              }
            }
            radhit_prev_flux = 0.0;

            /* if followed by a radhit */
            if (i3<nsamp_cube-NPIX_PLANE && samp_radhit[i3+NPIX_PLANE]!=0 ) {

              /* save radhit info */
              if (samp_status[i3]==SAMPLE_OK) {
                delta = influx[i3+NPIX_PLANE] - influx[i3];
                n_radhit_adj ++;
                radhit_prev_flux = t_outflux;
                radhit_prev_n    = samp;
              }
            } /* end -- if followed by radhit */

          } /* if enough samps */

        }
        samp=0;
      } /* end if time to compute */
    } /* end for i3 */

    if (  n_segments >= 1) {
      outflux[i2] = (total_flux_wgt / total_wgt) * normalization_factor;
      outunc[i2]  = (1.0 / sqrt(total_wgt)     ) * normalization_factor;
    }
    else {
      outflux[i2] = NANVALUE;
      outunc[i2]  = NANVALUE;
    }

  } /* end for i2 */

} /* end compute_slopes */


/******************************/
void populate_outbmask( /* populate the outbmask array */
/**************************************************************************/
/* IN */
 short *inpmask, /* PMASK array (2D) */
 short *indmask, /* DMASK array (3D) */
 char* bitpass_string, /* control string from user */
 int* samp_status, /* status per sample */
 int nsamp_cube, /* number of samples in 3D image (DMASK) */
/* OUT */
 short *outbmask) /* BMASK array */
{
  int i2, i3, t, i;
  char c;
  int n_tokens = 0;
  char bp_symbol[33];
  int bp_value[33];

  int n_pass_masks;
  short bmask_pass[33];
  short dmask_pass[33];

  /* Added to avoid bit-shifting, may not be vital - TYB 5/1/09 */
  short power[16] = {1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,
		     16384,0x8000};

  /* first decode the string from user */
  bp_symbol[0] = '+';
  bp_value[0]  = -1;
  for (i=0; bitpass_string[i]; i++) {
    c = bitpass_string[i];
    if (c == '+' || c == ':') {
      n_tokens++;
      bp_symbol[n_tokens] = c;
      bp_value[n_tokens]  = -1;
      continue;
    }
    if (c >= '0' && c <= '9') {
      if (bp_value[n_tokens] != -1) FATAL_char("more than 2 digits in bit num of bitpass string %s", bitpass_string+i);
      bp_value[n_tokens] = c - '0';
      c = bitpass_string[i+1];
      if (c >= '0' && c <= '9') {
        i++;
        bp_value[n_tokens] = bp_value[n_tokens] * 10 + c - '0';
      }
      if (bp_value[n_tokens] > 15) FATAL_char("bitnum > 15 in bitpass string at %s", bitpass_string+i-2);
      continue;
    }
  }

  n_pass_masks = 0;
  for (t=0; t<=n_tokens; t+=2) { /* determine mask _pass bit patterns */

    /* Eliminate the bit-shifting - TYB 5/1/09 */
     dmask_pass[n_pass_masks] = power[bp_value[t]];
     bmask_pass[n_pass_masks] = power[bp_value[t+1]];

    //    dmask_pass[n_pass_masks] = 1 << bp_value[t];
    //    bmask_pass[n_pass_masks] = 1 << bp_value[t+1];

    n_pass_masks ++;
  }

  /* now actually pass all the bits from DMASK to BMASK */
  for (i2=0; i2<NPIX_PLANE; i2++) { /* for each out pixel */
    outbmask[i2] = 0;
    if (inpmask[i2] & pmask_fatal) outbmask[i2] |= BMASK_BIT_PMASK;
    for(i3 = i2; i3<nsamp_cube; i3 += NPIX_PLANE) { /* go thru all planes */
      if ( (samp_status[i3] & SAMPLE_BAD) != 0) continue;

      /* This appears to be an error - TYB 6/1/09 */
 
              for (t=0; t < n_pass_masks; t++) { /* go through all pass mask bits */
      //            for (t=0; t<=n_pass_masks; t++) { /* go through all pass mask bits */

        if (indmask[i3] & dmask_pass[t]) outbmask[i2] |= bmask_pass[t];
      }
    }
  }
}


/**************************************************************************/
void write_fits_image ( /* writes a FITS image */
/**************************************************************************/
  /* (hard coded to write NPIX_PLANE pixels) */
/* IN */
 char *filename, /* name of output file to be created/written */
 double* array_r8, /* double pixel array (=0 if doing short */
 short*  array_i2, /* short  pixel array (=0 if doing double) */
 int header_fid /* fid of file to use for generating output file header */
) {
  int i, fid;
  int header_fid_array[4];

/* Switch to CFITSIO routines for the real case - TYB 5/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int ikeyval = 0;
  int numofkeys;
  int first_history;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";

/* just return if NONE */
  if (strcmp(filename, "NONE") ==0) return;

  if(array_r8){
    strcat(myoutfile, "!");
    strcat(myoutfile, filename);
    fits_create_file(&ffp_FITS, myoutfile, &I_fits_return_status);
    fits_open_file(&ffp_FITS_In, filename_influx, READONLY, &I_fits_return_status);

    /* Copy the non-HISTORY headers */

    fits_get_hdrspace(ffp_FITS_In, &numofkeys, &morekeys, &I_fits_return_status);
    for(i=0; i<numofkeys; i++){
      fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
      if(strncmp(card,"HISTORY ", 8) != 0) {
	fits_write_record(ffp_FITS, card, &I_fits_return_status);
      }
    }
       
    /* Copy the HISTORY headers to the end */

    first_history = 1;
    for(i=0; i<numofkeys; i++){
      fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
      if(strncmp(card,"HISTORY ", 8) == 0) {
	if (first_history && card[8] != '*'){
	  fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
	  first_history = 0;
	}   
	fits_write_record(ffp_FITS, card, &I_fits_return_status);
      }
    }
    if (first_history){
      fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
    }    

    //    fits_copy_hdu (ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);
    fits_close_file(ffp_FITS_In, &I_fits_return_status);
    ikeyval = 2;
    fits_update_key(ffp_FITS, TINT, "NAXIS", &ikeyval, " ", &I_fits_return_status);
    fits_delete_key(ffp_FITS, "NAXIS3", &I_fits_return_status);
    fits_update_key(ffp_FITS, TSTRING, "BUNIT", "e-/sec", " ", &I_fits_return_status);
    fits_write_img(ffp_FITS, TDOUBLE, 1, NPIX_PLANE, array_r8, &I_fits_return_status);               
    fits_flush_file(ffp_FITS, &I_fits_return_status);
    fits_close_file(ffp_FITS, &I_fits_return_status);
  } else {

   /* Open file */
    fid = im_open(filename, "w");
    if (fid < 0) FATAL_char("Trouble opening output file %s", filename);

   /* Copy header from file influx file */
    header_fid_array[0] = header_fid;
    header_fid_array[1] = 0;
    im_hdr(fid, header_fid_array, 0);

   /* reduce NAXIS to 2 and delete NAXIS3 card*/
    im_rkey_i(fid, &i, "NAXIS3");
    im_wkey_t(fid,  0, "delete");
    im_wkey_i(fid,  2, "NAXIS");

   /* update BUNIT card */
    im_wkey_c(fid,  "e-/sec", "BUNIT");

   /* write out the pixels */
    im_wkey_i(fid,  16, "BITPIX"); /* reset FITS header for I2 data */
    i = im_wpix_s(fid, array_i2, NPIX_PLANE);

    if (i != NPIX_PLANE) FATAL_char("Trouble writing pixels to file %s", filename);

   /* close the file */
    im_close(fid);
  }
  //  /* just return if NONE */
  //  if (strcmp(filename, "NONE") ==0) return;

  //  /* Open file */
  //  fid = im_open(filename, "w");
  //  if (fid < 0) FATAL_char("Trouble opening output file %s", filename);

  //  /* Copy header from file influx file */
  //  header_fid_array[0] = header_fid;
  //  header_fid_array[1] = 0;
  //  im_hdr(fid, header_fid_array, 0);

  //  /* reduce NAXIS to 2 and delete NAXIS3 card*/
  //  im_rkey_i(fid, &i, "NAXIS3");
  //  im_wkey_t(fid,  0, "delete");
  //  im_wkey_i(fid,  2, "NAXIS");

  //  /* update BUNIT card */
  //  im_wkey_c(fid,  "e-/sec", "BUNIT");

  //  /* write out the pixels */
  //  if (array_r8) {
  //    i = im_wpix_d(fid, array_r8, NPIX_PLANE);
  //  } else {
  //    im_wkey_i(fid,  16, "BITPIX"); /* reset FITS header for I2 data */
  //    i = im_wpix_s(fid, array_i2, NPIX_PLANE);
  //  }
  //  if (i != NPIX_PLANE) FATAL_char("Trouble writing pixels to file %s", filename);

  //  /* close the file */
  //  im_close(fid);

}


/**************************************************************************/
void write_fits_dual ( /* writes a 2-plane FITS image */
/**************************************************************************/
/* IN */
 char *filename, /* name of output file to be created/written */
 double* array_plane1, /* double pixel array of first plane */
 double* array_plane2, /* double pixel array of second plane */
 int header_fid /* fid of file to use for generating output file header */
) {
  int i, fid;
  int header_fid_array[4];
  double array_copy[NPIX_PLANE]; /* need copy because write corrupts */

/* Switch to CFITSIO routines - TYB 5/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int ikeyval = 0;
  int numofkeys;
  int first_history;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";

/* just return if NONE */
  if (strcmp(filename, "NONE") ==0) return;

  strcat(myoutfile, "!");
  strcat(myoutfile, filename);
  fits_create_file(&ffp_FITS, myoutfile, &I_fits_return_status);
  fits_open_file(&ffp_FITS_In, filename_influx, READONLY, &I_fits_return_status);

  /* Copy the non-HISTORY headers */

  fits_get_hdrspace(ffp_FITS_In, &numofkeys, &morekeys, &I_fits_return_status);
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) != 0) {
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
       
  /* Copy the HISTORY headers to the end */

  first_history = 1;
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) == 0) {
      if (first_history && card[8] != '*'){
	fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
	first_history = 0;
      }   
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
  if (first_history){
    fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
  }    

    //    fits_copy_hdu (ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);
  fits_close_file(ffp_FITS_In, &I_fits_return_status);
  ikeyval = 3;
  fits_update_key(ffp_FITS, TINT, "NAXIS", &ikeyval, " ", &I_fits_return_status);
  ikeyval = 2;
  fits_update_key(ffp_FITS, TINT, "NAXIS3", &ikeyval, " ", &I_fits_return_status);
  fits_update_key(ffp_FITS, TSTRING, "BUNIT", "e-/sec", " ", &I_fits_return_status);
  /* write out the pixels */
  for (i=0; i<NPIX_PLANE; i++) array_copy[i] = array_plane1[i];
  fits_write_img(ffp_FITS, TDOUBLE, 1, NPIX_PLANE, array_copy, &I_fits_return_status);               
  for (i=0; i<NPIX_PLANE; i++) array_copy[i] = array_plane2[i];
  fits_write_img(ffp_FITS, TDOUBLE, 1+NPIX_PLANE, NPIX_PLANE, array_copy, &I_fits_return_status);               
  fits_flush_file(ffp_FITS, &I_fits_return_status);
  fits_close_file(ffp_FITS, &I_fits_return_status);

  //  /* just return if NONE */
  //  if (strcmp(filename, "NONE") ==0) return;

  //  /* Open file */
  //  fid = im_open(filename, "w");
  //  if (fid < 0) FATAL_char("Trouble opening output file %s", filename);

  //  /* Copy header from file influx file */
  //  header_fid_array[0] = header_fid;
  //  header_fid_array[1] = 0;
  //  im_hdr(fid, header_fid_array, 0);

  //  /* set NAXIS3 to 2 */
  //  im_wkey_i(fid,  3, "NAXIS");
  //  im_wkey_i(fid,  2, "NAXIS3");

  //  /* update BUNIT card */
  //  im_wkey_c(fid,  "e-/sec", "BUNIT");

  //  /* write out the pixels */
  //  for (i=0; i<NPIX_PLANE; i++) array_copy[i] = array_plane1[i];
  //  i = im_wpix_d(fid, array_copy, NPIX_PLANE);
  //  if (i != NPIX_PLANE) FATAL_char("Trouble writing pixels to %s (p1", filename);
  //  for (i=0; i<NPIX_PLANE; i++) array_copy[i] = array_plane2[i];
  //  i = im_wpix_d(fid, array_copy, NPIX_PLANE);
  //  if (i != NPIX_PLANE) FATAL_char("Trouble writing pixels to %s (p2", filename);

  //  /* close the file */
  //  im_close(fid);
}


/**************************************************************************/
void write_fits_4planes ( /* writes a double 4-plane FITS image */
/**************************************************************************/
/* IN */
 char *filename, /* name of output file to be created/written */
 double* array_planes, /* double pixel array 4 pixel planes */
 int header_fid /* fid of file to use for generating output file header */
) {
  int i, fid;
  int header_fid_array[4];
  double array_copy[4*NPIX_PLANE]; /* need copy because write corrupts */

/* Switch to CFITSIO routines - TYB 5/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int ikeyval = 0;
  int numofkeys;
  int first_history;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";
  char filenameold[MAXFILENAMESIZE] = "";
  char filenameint[MAXFILENAMESIZE] = "";
  char str_end[1];

/* just return if NONE */
  if (strcmp(filename, "NONE") ==0) return;

  /* Need to drop trailing comma */
  strcpy(filenameold, filename);
  strcpy(str_end, &filenameold[strlen(filenameold) - 1]);
  if (strncmp(str_end, ",", 1)==0){
    strncpy(filenameint, filenameold, strlen(filenameold)-1);
    strcpy(filename, filenameint);
  }

  strcat(myoutfile, "!");
  strcat(myoutfile, filename);
  fits_create_file(&ffp_FITS, myoutfile, &I_fits_return_status);
  fits_open_file(&ffp_FITS_In, filename_influx, READONLY, &I_fits_return_status);

  /* Copy the non-HISTORY headers */

  fits_get_hdrspace(ffp_FITS_In, &numofkeys, &morekeys, &I_fits_return_status);
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) != 0) {
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
       
  /* Copy the HISTORY headers to the end */

  first_history = 1;
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) == 0) {
      if (first_history && card[8] != '*'){
	fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
	first_history = 0;
      }   
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
  if (first_history){
    fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
  }    

    //    fits_copy_hdu (ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);
  fits_close_file(ffp_FITS_In, &I_fits_return_status);
  ikeyval = 3;
  fits_update_key(ffp_FITS, TINT, "NAXIS", &ikeyval, " ", &I_fits_return_status);
  ikeyval = 4;
  fits_update_key(ffp_FITS, TINT, "NAXIS3", &ikeyval, " ", &I_fits_return_status);
  //  fits_update_key(ffp_FITS, TSTRING, "BUNIT", "e-/sec", " ", &I_fits_return_status);
  /* write out the pixels */
  for (i=0; i<4*NPIX_PLANE; i++) array_copy[i] = array_planes[i];
  fits_write_img(ffp_FITS, TDOUBLE, 1, 4*NPIX_PLANE, array_copy, &I_fits_return_status);               
  fits_flush_file(ffp_FITS, &I_fits_return_status);
  fits_close_file(ffp_FITS, &I_fits_return_status);

  //  /* just return if NONE */
  //  if (strcmp(filename, "NONE") ==0) return;

  //  /* Open file */
  //  fid = im_open(filename, "w");
  //  if (fid < 0) FATAL_char("Trouble opening output file %s", filename);

  //  /* Copy header from file influx file */
  //  header_fid_array[0] = header_fid;
  //  header_fid_array[1] = 0;
  //  im_hdr(fid, header_fid_array, 0);

  //  /* set NAXIS3 to 4 */
  //  im_wkey_i(fid,  3, "NAXIS");
  //  im_wkey_i(fid,  4, "NAXIS3");

  //  /* write out the pixels */
  //  for (i=0; i<4*NPIX_PLANE; i++) array_copy[i] = array_planes[i];
  //  i = im_wpix_d(fid, array_copy, 4*NPIX_PLANE);
  //  if (i != 4*NPIX_PLANE) FATAL_char("Trouble writing pixels to %s (p1", filename);

  //  /* close the file */
  //  im_close(fid);
}


/**************************************************************************/
void write_fits_16planes ( /* writes a int 16-plane FITS image */
/**************************************************************************/
/* IN */
 char *filename, /* name of output file to be created/written */
 short* array_planes, /* int pixel array 16 pixel planes */
 int header_fid /* fid of file to use for generating output file header */
) {
  int i, fid;
  int header_fid_array[4];

  /* just return if NONE */
  if (strcmp(filename, "NONE") ==0) return;

  /* Open file */
  fid = im_open(filename, "w");
  if (fid < 0) FATAL_char("Trouble opening output file %s", filename);

  /* Copy header from file influx file */
  header_fid_array[0] = header_fid;
  header_fid_array[1] = 0;
  im_hdr(fid, header_fid_array, 0);

  /* reset NAXIS3 to 16 and BITPIX to 16 */
  im_wkey_i(fid,  16, "NAXIS3");
  im_wkey_i(fid,  16, "BITPIX");

  /* write out the pixels */
  i = im_wpix_s(fid, array_planes, MAXPIX_CUBE);
  if (i != MAXPIX_CUBE) FATAL_char("Trouble writing pixels to %s (p1", filename);

  /* close the file */
  im_close(fid);
}


/**************************************************************************/
/* MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN  */
/**************************************************************************/
main(int argc, char** argv)
{
  static double influx [MAXPIX_CUBE]; /* input flux array */
  static double inunc  [MAXPIX_CUBE]; /* input uncertainty array */
  static short  indmask[MAXPIX_CUBE]; /* input bmask array (3D) */
  static short  inpmask[NPIX_PLANE];  /* input pmask array (2D) */

  static int    samp_status[MAXPIX_CUBE]; /* sample status (3D) */
  static int    samp_radhit[MAXPIX_CUBE]; /* sample is a radhit (3D) */
  static int    pix_needs_update[NPIX_PLANE]; /* still need do delta,outliers */

  static double delta[MAXPIX_CUBE];       /* deltas plane-to-plane (3D) */
  static int    delta_n[NPIX_PLANE];  /* number of deltas (2D) */
  static double delta_mean[NPIX_PLANE];  /* mean of deltas (2D) */
  static double delta_stdev[NPIX_PLANE]; /* stdev of deltas (2D) */

  static double o_flux[NPIX_PLANE]; /* method1 post-outlier flux array */
  static double o_unc [NPIX_PLANE]; /* method1 post-outlier unc array */
  static double o_pre_flux[NPIX_PLANE]; /* method1 pre-outlier flux array */
  static double o_pre_unc [NPIX_PLANE]; /* method1 pre-outlier unc array */

  static double o2_flux[NPIX_PLANE]; /* method2 post-outlier flux array */
  static double o2_unc [NPIX_PLANE]; /* method2 post-outlier unc array */
  static double o2_pre_flux[NPIX_PLANE]; /* method2 pre-outlier flux array */
  static double o2_pre_unc [NPIX_PLANE]; /* method2 pre-outlier unc array */

  static short info_outlier [MAXPIX_CUBE]; /* counts after outlier rejection */
  static double info_radhit [4*NPIX_PLANE]; /* jump, before, after, N */

  static short  outbmask [NPIX_PLANE]; /* output mask array */
  int fid_influx; /* FID of influx file */
  int nsamp_cube; /* how many samples in 3D image */
  int iter_num; /* iteration number for outlier rejection */
  int i2, i3, n; /* tmp vrbls */

  /* process command line and namelist */
  process_parameters(argc, argv);

  /* read input files */
  nsamp_cube = 0; /* to force it to be set by first read */
  read_fits_image (filename_influx, &nsamp_cube, influx, 0,  &fid_influx);
  read_fits_image (filename_inunc, &nsamp_cube, inunc, 0, &fid_influx);
  read_fits_image (filename_indmask, &nsamp_cube, 0, indmask, 0);
  read_fits_image (filename_inpmask, 0 /* 2D file*/, 0, inpmask, 0);

/**** Ramp GENerator ****/
/*****
n=0;
for (i3=0; i3<nsamp_cube; i3+=NPIX_PLANE) {
 influx[i3]  = 100.0 + n * 10.0 + ((n%3)-1)*3.0;
 if (n>=5) influx[i3] += 2000.0;
 if (n==10) influx[i3] += 20000.0;
 inunc[i3]   = 50.0 - n;
 indmask[i3] = 0;
 if (n==13) indmask[i3] = 512;
 n++;
}
*****/

  /* set status per sample */
  set_samp_status(influx, inunc, inpmask, indmask, nsamp_cube,
       samp_status, samp_radhit);

  /* compute standard slopes before outlier rejection */
  if (strcmp(filename_o_pre_flux ,"NONE")!=0 
   || strcmp(filename_o_dual_flux,"NONE")!=0
   || strcmp(filename_o_pre_unc  ,"NONE")!=0
   || strcmp(filename_o_dual_unc ,"NONE")!=0)
  compute_slopes(influx, inunc, samp_status, samp_radhit, nsamp_cube,
info_outlier,
    o_pre_flux, o_pre_unc, info_radhit);

  /* outlier rejection, first iteration (saves in "pre" array) */
  for (i2=0; i2<NPIX_PLANE; i2++) {pix_needs_update[i2]=1;}
  compute_deltas (influx, pix_needs_update, nsamp_cube, samp_status, delta,
                   delta_n, delta_mean, delta_stdev, o2_pre_flux, o2_pre_unc);
  /* initialize info_outlier */
  for (i2=0; i2<NPIX_PLANE; i2++) {
    info_outlier[i2] = delta_n[i2];
    for (i3=i2+NPIX_PLANE; i3<MAXPIX_CUBE; i3+=NPIX_PLANE)
     info_outlier[i3] = -1;
  }
  n = flag_outliers (delta, delta_n, delta_mean, delta_stdev,
                    pix_needs_update, nsamp_cube,
                    info_outlier, 1,
                    samp_status);
  /* outlier rejection, second+ iteration */
  for (iter_num = 2; iter_num <= outlier_n_iter; iter_num++) {
    compute_deltas (influx, pix_needs_update, nsamp_cube, samp_status, delta,
                      delta_n, delta_mean, delta_stdev, o2_flux, o2_unc);
    n = flag_outliers (delta, delta_n, delta_mean, delta_stdev,
                      pix_needs_update, nsamp_cube,
                      info_outlier, iter_num,
                      samp_status);
    if (n == 0) break; /* no more outliers */
  }

  /* delete samples after radhits */
  delete_after_radhit(delta_mean, samp_radhit, nsamp_cube, samp_status);

  /* compute slopes using line-fitting formula for correlated uncertainties */
  compute_slopes(influx, inunc, samp_status, samp_radhit, nsamp_cube,
info_outlier,
    o_flux, o_unc, info_radhit);

  /* fill in BMASK values and write file */
  populate_outbmask(inpmask, indmask, bitpass_string, samp_status, nsamp_cube, outbmask);
  write_fits_image (filename_outbmask, 0, outbmask, fid_influx);

  /* write dual images (must do BEFORE singles - singles corrupt out_array) */
  write_fits_dual  (filename_o_dual_flux, o_pre_flux, o_flux, fid_influx);
  write_fits_dual  (filename_o_dual_unc , o_pre_unc , o_unc , fid_influx);
  write_fits_dual  (filename_o2_dual_flux, o2_pre_flux, o2_flux, fid_influx);
  write_fits_dual  (filename_o2_dual_unc , o2_pre_unc , o2_unc , fid_influx);

  /* write method1 images */
  write_fits_image (filename_o_flux, o_flux , 0, fid_influx);
  write_fits_image (filename_o_unc , o_unc  , 0, fid_influx);
  write_fits_image (filename_o_pre_flux, o_pre_flux , 0, fid_influx);
  write_fits_image (filename_o_pre_unc , o_pre_unc  , 0, fid_influx);

  /* write method2 images */
  write_fits_image (filename_o2_flux, o2_flux , 0, fid_influx);
  write_fits_image (filename_o2_unc , o2_unc  , 0, fid_influx);
  write_fits_image (filename_o2_pre_flux, o2_pre_flux , 0, fid_influx);
  write_fits_image (filename_o2_pre_unc , o2_pre_unc  , 0, fid_influx);

  /* write info images */
  write_fits_16planes(filename_info_outlier, info_outlier , fid_influx);
  write_fits_4planes (filename_info_radhit, info_radhit , fid_influx);

  param_exit(0, PROGRAM); /* normal exit */
}
