/* $Id: slremov_read_data.c,v 1.11 2009/02/02 17:51:00 bob Exp $ */


/*********************************************** 
slremov_read_data.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

Open the FITS file, allocate memory, read the data into an array, 

Definitions of Variables

Input:
CP_Filename_FITS_In = In Data File Name.

Output:
fpp_FITS_in = Pointer to fits file handle.
DPP_Data_In = Pointer to image data in array form.
 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.


***********************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitsio.h"
#include "slremov.h"
#include "slremov_defs.h"


void slremov_read_data(	
            Filenames   *Fnames,
	    Constants   *Const,
            Dims        *Dim_In,
            float	**FPP_Data_In,
            float       **FPP_Vars_In,
            Status	*Stat,
            short       **SPP_pmask,
            short       **SPP_dmask,
            short       **SPP_bmask,
            short       **SPP_cmask,
            short       **SPP_lmask)
{
char    line[MAX_FILENAME_LENGTH];
int	i, j, k, totalrow, nrow;
int	I_fits_return_status=0, I_ndims_found, I_anynull;
int     I_naxis, E_naxis, L_naxis, I_bitpix, E_bitpix;
long	LP_naxes[3];
long	LP_naxes1[3];
float   I_gain, I_tint;
float	D_nullval=0;
double   alpha, eta;
fitsfile 	*ffp_FITS_In;
fitsfile        *ffp_Vars_In;
fitsfile 	*ffp_FITS_pmask;
fitsfile 	*ffp_FITS_dmask;
fitsfile 	*ffp_FITS_bmask;
fitsfile        *ffp_FITS_cmask;
fitsfile 	*ffp_FITS_lmask;
char    fits_comment[FLEN_CARD];
FILE    *fp;

if(Stat->I_status) return;

/* Open and read decay constant file */

/* open order file */
if ((fp = fopen(Fnames->CP_Filename_Const_In, "r")) == NULL) {
     printf("Error opening order file %s\n", Fnames->CP_Filename_Const_In);
     Stat->I_status = READ_DATA|FOPEN_FAILED; 
     return;
}

totalrow=0;
while (fgets(line, MAX_FILENAME_LENGTH, fp) != NULL) {
   if (!(line[0] == 0x5c || line[0] == 0x7c || line[0] == 0x23)) { /* could be IPAC table */
      sscanf(line, "%d %lf %lf", &nrow, &alpha, &eta);
      Const->Decay_Alpha[nrow] = alpha;
      Const->Decay_Eta[nrow] = eta;
      totalrow++;  
      if (totalrow >= MAXROW) {
         printf("slremov_read_data: input decay constant file containing too many lines\n");
         Stat->I_status = READ_DATA|TOO_MANY_CONST;
      }
   }
}

fclose(fp);

/*  Open and get array data from input fits file */

fits_open_file(	&ffp_FITS_In,
		Fnames->CP_Filename_FITS_In,
		READWRITE,
		&I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

fits_read_key( ffp_FITS_In,
               TINT,
               "BITPIX",
               &I_bitpix,
               fits_comment,
               &I_fits_return_status);
if(I_bitpix != -32)
{Stat->I_status = READ_DATA|BIT_ERROR; return;}

fits_read_key( ffp_FITS_In,
               TINT,
               "NAXIS",
               &I_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_In,
                    "NAXIS",
                    1, I_naxis,
                    LP_naxes,
                    &I_ndims_found,
                    &I_fits_return_status);

if (I_naxis == 2) LP_naxes[2] = 1;

Dim_In->I_NX = LP_naxes[0];
Dim_In->I_NY = LP_naxes[1];
Dim_In->I_NZ = LP_naxes[2];

/* Allocate memory for the data */
/* Don't try if problem with fits file */

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

*FPP_Data_In = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2], 
          sizeof(float));
*FPP_Vars_In = (float *) calloc( LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
          sizeof(float));

/* Use I_fits_return_status to store status, 
   so if there was a problem with calloc
   subsequent fits functions can detect it */
if(*FPP_Data_In==NULL || *FPP_Vars_In == NULL) I_fits_return_status = MALLOC_FAILED;

/* Read the data */

fits_read_img(ffp_FITS_In,
              TFLOAT,
              1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
              &D_nullval,
              *FPP_Data_In,
              &I_anynull,
              &I_fits_return_status);

fits_close_file(ffp_FITS_In,
		&I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and get uncertainty data */

if (strcmp(Fnames->CP_Filename_Vars_In, "")) {

   fits_open_file( &ffp_Vars_In,
                   Fnames->CP_Filename_Vars_In,
                   READWRITE,
                   &I_fits_return_status);
   
   if(I_fits_return_status)
   {Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

   fits_read_key( ffp_Vars_In,
                  TINT,
                  "BITPIX",
                  &E_bitpix,
                  fits_comment,
                  &I_fits_return_status);
   if(E_bitpix != -32)
   {Stat->I_status = READ_DATA|BIT_ERROR; return;}

   fits_read_key( ffp_Vars_In,
                  TINT,
                  "NAXIS",
                  &E_naxis,
                  fits_comment,
                  &I_fits_return_status);

   fits_read_keys_lng( ffp_Vars_In,
                       "NAXIS",
                       1, E_naxis,
                       LP_naxes1,
                       &I_ndims_found,
                       &I_fits_return_status);

   if (E_naxis == 2) LP_naxes1[2] = 1;

   Dim_In->E_NX = LP_naxes1[0];
   Dim_In->E_NY = LP_naxes1[1];
   Dim_In->E_NZ = LP_naxes1[2];

   if (Dim_In->I_NX != Dim_In->E_NX || Dim_In->I_NY != Dim_In->E_NY)
      {
      printf("Error when reading input data:\n\n");
      printf("Data and uncertainty images have different pixels in at least one of the two dimensions\n");
      Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
      }

   if (Dim_In->I_NZ > Dim_In->E_NZ)
      {
      printf("Error when reading input data:\n\n");
      printf("Data cube has more layers than uncertainty cube.\n");
      Stat->I_status = READ_DATA|DIFF_CUBE_PLANE;
      }

   fits_read_img(  ffp_Vars_In,
                   TFLOAT,
                   1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                   &D_nullval,
                   *FPP_Vars_In,
                   &I_anynull,
                   &I_fits_return_status);

   fits_close_file(ffp_Vars_In,
                   &I_fits_return_status);

   if (I_fits_return_status)
      {
      Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; 
      return;
      }
} else {
   for (i=0; i<Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ; ++i)
       *(*FPP_Vars_In + i) = 1.0;
}

/*  Open and read lmask file.  Required input.
*/

*SPP_lmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY,
         sizeof(short));

fits_open_file( &ffp_FITS_lmask,
                Fnames->CP_Filename_Lmask,
                READWRITE,
                &I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

fits_read_key( ffp_FITS_lmask,
               TINT,
               "NAXIS",
               &L_naxis,
               fits_comment,
               &I_fits_return_status);

fits_read_keys_lng( ffp_FITS_lmask,
                    "NAXIS",
                    1, L_naxis,
                    LP_naxes1,
                    &I_ndims_found,
                    &I_fits_return_status);

if (L_naxis == 2) LP_naxes1[2] = 1;

Dim_In->L_NX = LP_naxes1[0];
Dim_In->L_NY = LP_naxes1[1];
Dim_In->L_NZ = LP_naxes1[2];

if (Dim_In->L_NZ != 1) {
   printf("Error when reading input data:\n\n");
   printf("Input Lmask has more than one planes\n");
   Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
}

if (Dim_In->I_NX != Dim_In->L_NX || Dim_In->I_NY != Dim_In->L_NY) {
   printf("Error when reading input data:\n\n");
   printf("Data and uncertainty images have different pixels in at least one of the two dimensions\n");
   Stat->I_status = READ_DATA|INCRT_ARRAY_DIM;
}

fits_read_img( ffp_FITS_lmask,
               TSHORT,
               1, Dim_In->I_NX*Dim_In->I_NY,
               &D_nullval,
               *SPP_lmask,
               &I_anynull,
               &I_fits_return_status);

fits_close_file(ffp_FITS_lmask,
                &I_fits_return_status);

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}


/*  Open and read pmask file.  Optional input. 
    If file can't be opened create dmask assuming all pixes are good.
    Otherwise read in from file.
*/

*SPP_pmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Pmask, ""))
   {
   fits_open_file( &ffp_FITS_pmask,
                   Fnames->CP_Filename_Pmask,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_pmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_pmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_pmask, 
                   &I_fits_return_status);
   }
else
   {
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i)
       {
       *(*SPP_pmask +i) = 0;
       }
   }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read dmask file.  Optional input.  */

*SPP_dmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Dmask, ""))
   {
   fits_open_file( &ffp_FITS_dmask,
                   Fnames->CP_Filename_Dmask,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_dmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ,
                  &D_nullval,
                  *SPP_dmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_dmask, 
                   &I_fits_return_status);
   }
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY*Dim_In->I_NZ-1; ++i)
       {
       *(*SPP_dmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read Bmask file.  Optional input.  */

*SPP_bmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY, 
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Bmask, ""))
   {
   fits_open_file( &ffp_FITS_bmask,
                   Fnames->CP_Filename_Bmask,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_bmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_bmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_bmask, 
                   &I_fits_return_status);
   }
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i)
       {
       *(*SPP_bmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

/*  Open and read Cmask file.  Optional input.  */

*SPP_cmask = (short *) calloc(Dim_In->I_NX*Dim_In->I_NY,
         sizeof(short));

if (strcmp(Fnames->CP_Filename_Cmask, ""))
   {
   fits_open_file( &ffp_FITS_cmask,
                   Fnames->CP_Filename_Cmask,
                   READWRITE,
                   &I_fits_return_status);

   fits_read_img( ffp_FITS_cmask,
                  TSHORT,
                  1, Dim_In->I_NX*Dim_In->I_NY,
                  &D_nullval,
                  *SPP_cmask,
                  &I_anynull,
                  &I_fits_return_status);

   fits_close_file(ffp_FITS_cmask,
                   &I_fits_return_status);
   }
else
   for (i=0; i<=Dim_In->I_NX*Dim_In->I_NY-1; ++i)
       {
       *(*SPP_cmask +i) = 0;
       }

if(I_fits_return_status)
{Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status; return;}

printf("fits %d\n",I_fits_return_status);

/* Return the fits status as slremov status, with mask to say so */
if(I_fits_return_status)
Stat->I_status = READ_DATA|FITS_MSG_MASK|I_fits_return_status;

if(Stat->I_Verbose)	{
printf("slremov_read_data: Data read from file:\n%s\n",
Fnames->CP_Filename_FITS_In);
printf("slremov_read_data: I_NX = %d\n", Dim_In->I_NX);
printf("slremov_read_data: I_NY = %d\n", Dim_In->I_NY);
printf("slremov_read_data: I_NZ = %d\n", Dim_In->I_NZ);
}
}
