#define PROGRAM "snest_irs"
#define VERSION "1.1"

/**
Reads 3-D IRS image (usually cvti2r4.fits) and estimates noise for each 
sample in the 3-D image.
**/


/** UPDATES UPDATES UPDATES UPDATES UPDATES UPDATES UPDATES UPDATES UPDATES
061004 1.1 fix bug in of im_hdr needing all 3 parameters set
060913 1.0 divide readnoise and  confusion by gain brfore combining
060522 0.2 Add namelist processing
           Add optional output of adjusted DCE
060511 0.1 Initial version
**/

/*=================================================================*/
/* DEFINEs DEFINEs DEFINEs DEFINEs DEFINEs DEFINEs DEFINEs DEFINEs */
/*=================================================================*/
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdio.h>
#include "tbl.h"

/* Added - TYB 5/1/09 */
#include "fitsio.h"

#define MAXFILENAMESIZE  256

#define OK_NAXIS1         128
#define OK_NAXIS2         128
#define MAX_NAXIS3        16
#define MAXPIXELS_DCE    (OK_NAXIS1*OK_NAXIS2*MAX_NAXIS3)

#define TERMINATE_NORMAL         0
#define TERMINATE_WARNING       32
#define TERMINATE_ERROR	        64
#ifndef FALSE
#define FALSE		         0
#define TRUE              (!FALSE)
#endif


/*===================================================================*/
/* PARAMETERS PARAMETERS PARAMETERS PARAMETERS PARAMETERS PARAMETERS */
/*===================================================================*/
char filename_indce[MAXFILENAMESIZE];
char filename_indmask[MAXFILENAMESIZE];
char filename_outunc[MAXFILENAMESIZE];
char filename_adjdce[MAXFILENAMESIZE];
char filename_namelist[MAXFILENAMESIZE];

double gain;
double readnoise;
double confusion;
double outlier_ndelta;

int dmask_fatal;

int  verbose;


/*=============================================================*/
/* DATA DATA DATA DATA DATA DATA DATA DATA DATA DATA DATA DATA */
/*=============================================================*/

/* IMAGE (PIXEL) ARRAYS */
double indce[MAXPIXELS_DCE]; /* from input file */
short  indmask[MAXPIXELS_DCE]; /* from input file */
double adjdce[MAXPIXELS_DCE]; /* adjusted dce */
double outunc[MAXPIXELS_DCE];

/* INFO ABOUT DCE FILE */
int naxis1;
int naxis2;
int naxis3;
int nramps; /* 2-D size of DCE file */
int npixels; /* number of pixels in DCE file */
int fid_indce; /* fid of input file (for passing header info) */

/* CONSTANTs */
double NaN_double;



/**************************************************************************/
FATAL_char (char *message, char *param) {
  /* print error msg, with printf parameter; then exit with error status */
  char msg_nl[500];
  strcpy(msg_nl, "FATAL ERROR ");
  strcat(msg_nl, message);
  strcat(msg_nl, "\n");
  printf(msg_nl, param);
  exit(TERMINATE_ERROR);
}



/**************************************************************************/
initialize(int argc, char **argv) {
/* get and check parameters */

#define NL_START     "&SNEST_IRS"
#define NL_END       "&END"

  param(filename_indce,   "-i IN_FLUX input DCE file","infile");
  param(filename_indmask, "-d IN_DMASK input DMASK file","infile - NONE");
  param(filename_outunc,  "-o  OUT_UNC output uncertainty file", "outfile");
  param(filename_adjdce,  "-a  OUT_ADJDCE output adjustde DCE file", "outfile - NONE");
  param(&dmask_fatal,   "-dmask_fatal DMASK_FATAL DMASK fatal bits", "int - 0");
  param(&outlier_ndelta,"-nd OUTLIER_NDELTA N-deltas for outlier", "double - 3.0");
  param(&gain,          "-gain GAIN gain", "double - 0");
  param(&readnoise,     "-readnoise READNOISE read noise", "double - 0");
  param(&confusion,   "-confusion CONFUSION confusion", "double - 0");
  param(&verbose,         "-v VERBOSE level of printed messages", "int - 1");
  param(filename_namelist,  "-n NAMELIST_FILE name of Namelist file", "infile - namelist.nl");


  /**ENDPARAMS **/
  
  /* parse namelist and params */
  if (nl_process(argc, argv, PROGRAM, VERSION, NL_START, NL_END))
          exit(TERMINATE_ERROR);
  if (param_begin(argc, argv, PROGRAM, VERSION) ) exit(TERMINATE_ERROR);
  param_print();   /* Print the parameter settings to stdout. */

  /* set up NaN_double */
  sscanf("NaN", "%lf",  &NaN_double);
}



/**************************************************************************/
read_indce ()
/* read in the DCE file */
{
  int i;

  /* open file */
  fid_indce = im_open(filename_indce, "r");
  if (fid_indce < 0) FATAL_char("opening DCE file %s", filename_indce);

  /* check dimensions */
  im_rkey_i(fid_indce, &i, "NAXIS");
  if (i != 3) FATAL_char("DCE file %s not 3 dimensional", filename_indce);
  im_rkey_i(fid_indce, &naxis1, "NAXIS1");
  if (naxis1 != OK_NAXIS1) FATAL_char("DCE %s wrong NAXIS1", filename_indce);
  im_rkey_i(fid_indce, &naxis2, "NAXIS2");
  if (naxis2 != OK_NAXIS2) FATAL_char("DCE %s wrong NAXIS2", filename_indce);
  im_rkey_i(fid_indce, &naxis3, "NAXIS3");
  if (naxis3 > MAX_NAXIS3) FATAL_char("DCE %s NAXIS3 too big", filename_indce);

  /* read pixels */
  nramps  = naxis1 * naxis2;
  npixels = nramps * naxis3;

  i = im_rpix_d(fid_indce, indce, npixels);
  if (i != npixels) FATAL_char("Trouble reading pixels in %s", filename_indce);
}



/**************************************************************************/
read_indmask()
/* read in the DMASK file if there is one, otherwise set indmask to 0 */
{
  int i;
  int fid;

  /* just set to 0 if NONE */
  if (strcmp(filename_indmask, "NONE")==0) {
    for (i=0; i<npixels; i++) indmask[i] = 0;
    return;
  }

  /* open file */
  fid = im_open(filename_indmask, "r");
  if (fid < 0) FATAL_char("opening in input file %s", filename_indmask);

  /* check dimensions */
  im_rkey_i(fid, &i, "NAXIS");
  if (i != 3) FATAL_char("Input file %s not 3 dimensional", filename_indmask);
  im_rkey_i(fid, &i, "NAXIS1");
  if (i != naxis1) FATAL_char("file %s incorrect NAXIS1", filename_indmask);
  im_rkey_i(fid, &i, "NAXIS2");
  if (i != naxis2) FATAL_char("file %s incorrect NAXIS2", filename_indmask);
  im_rkey_i(fid, &i, "NAXIS3");
  if (i != naxis3) FATAL_char("file %s incorrect NAXIS3", filename_indmask);

  /* read pixels */
  i = im_rpix_s(fid, indmask, npixels);
  if (i != npixels) FATAL_char("Trouble reading pixels in %s",filename_indmask);
}



/**************************************************************************/
zap_dmask_fatals ()
/* set indce values to NaN if any fatal DMASK bit is set */
{
  int i;

  for (i=0; i<npixels; i++) {
    if (indmask[i] & dmask_fatal) {
      indce[i] = NaN_double;
    }
  }

}


/**************************************************************************/
compute_adjdce ()
/* compute adjusted DCE array (move ramp above and away from zero) */
{
  int r;
  int i, j, n, n_last, n_candidate;
  double adjustment;
  double tmp;
  double delta[MAX_NAXIS3];
  double median_delta;
  double allowed_delta;
  double maxdcevalue;
  double max_value;
  double min_value;
  double big_value;
int DEBUG_flag;

  for (r=0; r<nramps; r++) { /* go through every ramp */

DEBUG_flag = 0;

    /* initialize adjdce[*] and find max_value */
    max_value = -9999999.0;
    for (i=r; i<npixels; i+=nramps) {
      adjdce[i] = indce[i];
      if (indce[i] > max_value) max_value = indce[i];
    }
    if (max_value == -9999999.0) return; /* all NaNs */

    /* find deltas of all points in ramp */ 
    n = 0;
    for (i=r; i<npixels-nramps; i+=nramps) {

      delta[n] = indce[i+nramps] - indce[i];

/*******
if (i==r && delta[n]>200.0) DEBUG_flag = 1;
if (DEBUG_flag)
printf("DEBUGdel %6d %2d %10lf %10lf %10lf\n",
 r, n, delta[n], indce[i+nramps], indce[i]);
*******/

      if (finite(delta[n])) {
        if (delta[n] < 0) delta[n] = -delta[n]; /* abs() */
        n++;
      }
    }
    if (n <= 0) return; /* no deltas - no need adjust */

    /* sort deltas (keep putting the highest one at the end) */ 
    for (n_last=n-1; n_last>0; n_last--) {
      for (n_candidate=0; n_candidate<n_last; n_candidate++) {
        if (delta[n_candidate] > delta[n_last]) {
          tmp = delta[n_last];
          delta[n_last] = delta[n_candidate];
          delta[n_candidate] = tmp;
        }
      }
    }

/*******
if (DEBUG_flag)
for (n_last=0; n_last<n; n_last++) {
printf("DEBUGsort %2d %10lf\n", n_last, delta[n_last]);
}
*******/

    /* pick median from (now sorted) deltas */
    median_delta = delta[(n+1)/2];

    /* set the outliers to big value */ 
    big_value = max_value * 3.0;
    for (i=r; i<npixels-nramps; i+=nramps) {
      allowed_delta = outlier_ndelta * median_delta;
      if (finite(indce[i])==0) continue; /* skip NaNs */
      for (j=i+nramps; j<npixels-nramps; j+=nramps) { /* scan for non-NaN */
        if (finite(indce[j])) break;
        allowed_delta += median_delta; /* increase reqt per NaN found */
      }
      if (j >= npixels-nramps) break; /* skipped over NaNs till end of ramp */
      if (indce[j] - indce[i] > allowed_delta) adjdce[i] = big_value;
      i = j - nramps; /* in case NaNs skipped over */
    }

    /* find minimum value (outliers excluded - have a high value now) */ 
    min_value = big_value;
    for (i=r; i<npixels; i+=nramps) {
      if (adjdce[i] < min_value) min_value = adjdce[i];
    }

    /* compute adjustment value (adjusts lowest value to delta) */ 
    adjustment = -min_value + median_delta;

/*******
if (DEBUG_flag)
printf("DEBUGajd min=%lf median_delta=%lf adj=%lf\n",
          min_value, median_delta, adjustment);
********/

    /* apply adjustment to ramp */ 
    for (i=r; i<npixels; i+=nramps) {
      if (finite(adjdce[i]) && adjdce[i]!=big_value)
       adjdce[i] += adjustment;
    }

  } /* end (for r) */
}



/**************************************************************************/
compute_outunc ()
/* compute uncertainty (noise) for each pixel */
{
  int i;
  double c2r2;

  c2r2 = (confusion * confusion + readnoise * readnoise) / (gain * gain);

  for (i=0; i<npixels; i++) {
    outunc[i] = sqrt( c2r2 + adjdce[i]/gain );
  }
}


/**************************************************************************/
write_adjdce ()
/* write out the adjusted DCE file */
{
  int i;
  int fid;
  int fid_array[4];

  /* Switch to CFITSIO routines - TYB 5/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int numofkeys = 0;
  int first_history;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";

  if (strcmp(filename_adjdce, "NONE")==0) return;

  strcat(myoutfile, "!");
  strcat(myoutfile, filename_adjdce);
  fits_create_file(&ffp_FITS,
		   myoutfile,
		   &I_fits_return_status);

  fits_open_file(&ffp_FITS_In,
		 filename_indce,READONLY,
		 &I_fits_return_status);

  /* Copy the non-HISTORY headers */

  fits_get_hdrspace(ffp_FITS_In, &numofkeys, &morekeys, &I_fits_return_status);
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) != 0) {
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
      
  /* Copy the HISTORY headers to the end */

  first_history = 1;
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) == 0) {
      if (first_history && card[8] != '*'){
        fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
        first_history = 0;
      }   
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
  if (first_history){
    fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
  }    

  //  fits_copy_hdu (ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);

  fits_close_file(ffp_FITS_In, &I_fits_return_status);

  fits_flush_file(ffp_FITS, &I_fits_return_status);

  fits_write_img(ffp_FITS,
		 TDOUBLE,1, npixels, outunc,
		 &I_fits_return_status);               

  fits_flush_file(ffp_FITS, &I_fits_return_status);

  fits_close_file(ffp_FITS, &I_fits_return_status);
  
  //  if (strcmp(filename_adjdce, "NONE")==0) return;

  //  /* Open file */
  //  fid = im_open(filename_adjdce, "w");
  //  if (fid < 0) FATAL_char("unable to open out unc file %s", filename_adjdce);

  //  /* Copy header from reference file */
  //  fid_array[0] = fid_indce;
  //  fid_array[1] = 0;
  //  i = im_hdr(fid, fid_array, 0);


  //  /* write pixels */
  //  i = im_wpix_d(fid, adjdce, npixels);
  //  if (i != npixels)
  //    FATAL_char("writing pixels to FLUX file %s", filename_adjdce);
  //  im_close(fid);
}


/**************************************************************************/
write_outunc ()
/* write out the uncertainty file */
{
  int i;
  int fid;
  int fid_array[4];

  /* Switch to CFITSIO routines - TYB 5/1/09 */

  int  I_fits_return_status=0;
  int morekeys = 0;
  int numofkeys = 0;
  int first_history;
  fitsfile *ffp_FITS;
  fitsfile *ffp_FITS_In;
  char card[82];
  char myoutfile[MAXFILENAMESIZE] = "";

  if (strcmp(filename_outunc, "NONE")==0) return;

  strcat(myoutfile, "!");
  strcat(myoutfile, filename_outunc);
  fits_create_file(&ffp_FITS,
		   myoutfile,
		   &I_fits_return_status);

  fits_open_file(&ffp_FITS_In,
		 filename_indce,READONLY,
		 &I_fits_return_status);

  /* Copy the non-HISTORY headers */

  fits_get_hdrspace(ffp_FITS_In, &numofkeys, &morekeys, &I_fits_return_status);
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) != 0) {
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
      
  /* Copy the HISTORY headers to the end */

  first_history = 1;
  for(i=0; i<numofkeys; i++){
    fits_read_record(ffp_FITS_In, i+1, card, &I_fits_return_status);
    if(strncmp(card,"HISTORY ", 8) == 0) {
      if (first_history && card[8] != '*'){
        fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
        first_history = 0;
      }   
      fits_write_record(ffp_FITS, card, &I_fits_return_status);
    }
  }
  if (first_history){
    fits_write_record(ffp_FITS, "HISTORY *0 (imported)", &I_fits_return_status);
  }    

  //  fits_copy_hdu (ffp_FITS_In, ffp_FITS, morekeys, &I_fits_return_status);

  fits_close_file(ffp_FITS_In, &I_fits_return_status);

  fits_flush_file(ffp_FITS, &I_fits_return_status);

  fits_write_img(ffp_FITS,
		 TDOUBLE,1, npixels, outunc,
		 &I_fits_return_status);               

  fits_flush_file(ffp_FITS, &I_fits_return_status);

  fits_close_file(ffp_FITS, &I_fits_return_status);
  
  //  if (strcmp(filename_outunc, "NONE")==0) return;

  //  /* Open file */
  //  fid = im_open(filename_outunc, "w");
  //  if (fid < 0) FATAL_char("unable to open out unc file %s", filename_outunc);

  //  /* Copy header from reference file */
  //  fid_array[0] = fid_indce;
  //  fid_array[1] = 0;
  //  i = im_hdr(fid, fid_array, 0);


  //  /* write pixels */
  //  i = im_wpix_d(fid, outunc, npixels);
  //  if (i != npixels)
  //    FATAL_char("writing pixels to FLUX file %s", filename_outunc);
  //  im_close(fid);
}




/**************************************************************************/
/* MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN  */
/**************************************************************************/
main(argc, argv)
int argc;
char **argv;
{
  initialize(argc, argv);

  read_indce();
  read_indmask();
  zap_dmask_fatals();
  compute_adjdce();
  compute_outunc();

  write_adjdce();
  write_outunc();

  param_exit(TERMINATE_NORMAL, PROGRAM);
}
