#define PROGRAM "velocity"
#define VERSION "1.0"

/*************************************************************************
 This module computes velocity correctnios for IRS BCDs

 NOTE:  uses program downlink/pipeline/include/tools/nl_and_params.c 
**************************************************************************/

/**UPDATES
060830 1.0 Initial version (Bob)
**/

#define MAXFILENAMESIZE       256
#define MAX_VEL_TBL_ELTS 10000
#define NOVAL -9999.00

#define TERMINATE_NORMAL        0
#define TERMINATE_WARNING      32
#define TERMINATE_ERROR	       64

#ifndef FALSE
#define FALSE		        0
#define TRUE             (!FALSE)
#endif

#define NL_START       "&VELOCITY"
#define NL_END             "&END"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "tbl.h"
#include "nanvalue.h"

/*****
From Dave Shupe in CR-7374:

1) Find the nearest "julian" value corresponding to MJD_OBS, or interpolate
   between the two nearest values, to obtain the Spitzer velocity components
   VX, VY and VZ. Remember that MJD_OBS is Julian date minus 2,400,000 days.

2) Compute the quantities X = cos(RA_FOV)*cos(DEC_FOV),
   Y = sin(RA_FOV)*cos(DEC_FOV), Z = sin(DEC_FOV), converting the arguments
   to radians of course if needed.

3) VHELCORR is the dot product of [X,Y,Z] and [VX,VY,VZ],
    i.e. VHELCORR = X*VX + Y*VY + Z*VZ.

4) VLSRCORR = VHELCORR + 0.145*X - 17.320*Y + 10.000*Z.

   If possible, it is desirable to limit the number of significant digits
   output to the header to one decimal place, to be in line with the
   uncertainty of the calculation.

******/


/*=====================================================*/
/*==== quit SUBROUTINE ===============================*/
/*=====================================================*/
void quit(char* message, char *submessage)
/* printing messag, then abort */
{
  char full_message[500];
  sprintf(full_message, message, submessage); /* embed submessage */
  printf("ERROR[velocity]: %s\n", full_message);
  exit(TERMINATE_ERROR);
}


/*=====================================================*/
/*==== tio_getind SUBROUTINE ==========================*/
/*=====================================================*/
int tio_getindex(int tfid, char* col_name, int *index)
/* get index of the named column */
/* aborts if not found */
{
  *index = tbl_column(tfid, col_name);
  if (*index < 0) quit("Table file missing column: %s\n", col_name);
  return TERMINATE_NORMAL;
}


/*=======================================================*/
/*==== velocity main SUBROUTINE =========================*/
/*=======================================================*/
velocity(argc, argv)
int argc;
char *argv[];
{
/********/
/* DATA */
/********/

  /* command-line parameters */
  char infitsFilename[MAXFILENAMESIZE];
  char outfitsFilename[MAXFILENAMESIZE];
  char velocityFilename[MAXFILENAMESIZE];
  int verbose;

  /*image file data*/
  int infits_fid, outfits_fid;
  int fid_in_array[3];
  double *ptr_pixels;
  int npixels, naxis1, naxis2;
  double mjd_obs, ra_rqst, dec_rqst, ra_fov, dec_fov; /* from header */
  
  /* table file data */
  int intbl_fid;
  struct tbl_filinfo  *rinfo;
  int tstat;

  /* data input from valocity table */
  double tbl_jd[MAX_VEL_TBL_ELTS]; /* day     2452897.500000000 */
  double tbl_vx[MAX_VEL_TBL_ELTS]; /* km/sec  4.402088796500650E+00 */
  double tbl_vy[MAX_VEL_TBL_ELTS]; /* km/sec  2.640841249016063E+01 */
  double tbl_vz[MAX_VEL_TBL_ELTS]; /* km/sec  1.208046218928024E+01 */
  int tbl_nrows;
  int ind_jd, ind_vx, ind_vy, ind_vz; /* column indicies */

  /* misc daa */
  int i, j, stat;
  char header_card[280];
  double frac;
  double x, y, z;
  double vx, vy, vz;
  double real_jd;
  double ra_used, dec_used; /* chosen from header kwd values */
  double ra_radians, dec_radians; /* ra, dec in radians */
  double vhelcorr, vlsrcorr;


/***************************/
/* get command line PARAMS */
/***************************/

  /* get the parameters and print them out */
  param(infitsFilename,  "-i INFILE name of input file",  "infile in.fits");
  param(outfitsFilename, "-o OUTFILE name of output file","outfile out.fits");
  param(velocityFilename,"-t VELOCITY_FILE name of velocity file", "infile velocity.tbl");
  param(&verbose,        "-v VERBOSE level of printed messages",   "int - 1");
  stat = param_begin(argc, argv, PROGRAM, VERSION);
  if (stat != 0) return stat;
  param_print();   /* Print the parameter settings to stdout. */


/***************************/
/* Read the velocity table */
/***************************/
  intbl_fid = tbl_open(velocityFilename, &rinfo); 
  stat   = tbl_fileinfo(intbl_fid, &rinfo);
  if (stat != 0)
  {
     printf("ERROR openning velocity file [%s]\n", velocityFilename);
     exit(64);
  }
  if (tio_getindex(intbl_fid, "julian", &ind_jd)) return TERMINATE_ERROR;
  if (tio_getindex(intbl_fid, "vx",     &ind_vx)) return TERMINATE_ERROR;
  if (tio_getindex(intbl_fid, "vy",     &ind_vy)) return TERMINATE_ERROR;
  if (tio_getindex(intbl_fid, "vz",     &ind_vz)) return TERMINATE_ERROR;

  tbl_nrows = 0;
  while (1)
  {
    stat = tbl_colval(intbl_fid, tbl_nrows, ind_jd, TBL_DOUBLE, &tbl_jd[tbl_nrows]); 
    //
    //  Don't know what this is - TYB 7/20/09
    //
    //    if (tstat == -4) break;
    //    else if (tstat < 0)
    //    {
    //       printf("Table I/O ERROR when reading file [%s]\n", velocityFilename);
    //       return TERMINATE_ERROR;
    //    }
    stat = tbl_colval(intbl_fid, tbl_nrows, ind_vx, TBL_DOUBLE, &tbl_vx[tbl_nrows]); 
    if (stat < 0) break;
    stat = tbl_colval(intbl_fid, tbl_nrows, ind_vy, TBL_DOUBLE, &tbl_vy[tbl_nrows]); 
    if (stat < 0) break;
    stat = tbl_colval(intbl_fid, tbl_nrows, ind_vz, TBL_DOUBLE, &tbl_vz[tbl_nrows]); 
    if (stat < 0) break;
    if (verbose >=4)
    {  
      if (tbl_nrows == 0) printf("\nVelocity table:\n");
      printf("row=%4d jd=%10.1f vx=%13.10f vy=%13.10f vz=%13.10f\n", 
        tbl_nrows, tbl_jd[tbl_nrows], tbl_vx[tbl_nrows], tbl_vy[tbl_nrows], tbl_vz[tbl_nrows]);
    }
    tbl_nrows++;
  }
  tbl_close(intbl_fid);


/************************/
/* Read the infits File */
/************************/

  /* open file */
  infits_fid = im_open(infitsFilename, "r");
  if (infits_fid == -1)
  {
    printf("ERROR: %s -- not a proper FITS file.\n", infitsFilename);
    return TERMINATE_ERROR;
  }
  if (infits_fid < 0)
  {
    printf("%s -- ERROR opening or reading file.\n", infitsFilename);
    return TERMINATE_ERROR;
  }

  /* read keywords */
  if (im_rkey_i(infits_fid, &naxis1,       "NAXIS1") == -1)
     printf("Keyword NAXIS1 is missing from FITS header.\n");
  if (im_rkey_i(infits_fid, &naxis2,       "NAXIS2") == -1)
     printf("Keyword NAXIS2 is missing from FITS header.\n");
  if (im_rkey_d(infits_fid, &mjd_obs,  "MJD_OBS") ==-1) mjd_obs = NOVAL;
  if (im_rkey_d(infits_fid, &ra_rqst,  "RA_RQST") ==-1) ra_rqst = NOVAL;
  if (im_rkey_d(infits_fid, &dec_rqst, "DEC_RQST")==-1) dec_rqst= NOVAL;
  if (im_rkey_d(infits_fid, &ra_fov,   "RA_FOV")  ==-1) ra_fov  = NOVAL;
  if (im_rkey_d(infits_fid, &dec_fov,  "DEC_FOV") ==-1) dec_fov = NOVAL;

/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
printf("DEBUG unset dec_fov\n");
  if (im_rkey_d(infits_fid, &dec_fov,  "DEC_SSX") ==-1) dec_fov = NOVAL;
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */
/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */

  if (verbose >= 3) {
    printf("mjd_obs=%f ra_rqst=%f dec_rqst=%f ra_fov=%f dec_fov=%f\n",
            mjd_obs, ra_rqst, dec_rqst, ra_fov, dec_fov);
  }

  /* check and process keywords */
  if (mjd_obs == NOVAL) quit("Input FITS file has no MJD_OBS keyword", "");
  ra_used = ra_rqst;
  if (ra_fov != NOVAL) ra_used = ra_fov;
  if (ra_used == NOVAL) quit("Input FITS file has no RA_RQST or RA_FOV keyword", "");
  dec_used = dec_rqst;
  if (dec_fov != NOVAL) dec_used = dec_fov;
  if (dec_used == NOVAL) quit("Input FITS file has no DEC_RQST or DEC_FOV keyword", "");

  /* read pixels */
  npixels = naxis1*naxis2;
  ptr_pixels = (double *) malloc(npixels * sizeof(double));
  if (ptr_pixels == NULL) quit("Can't allocate required memory", "");
  im_rpix_d(infits_fid, ptr_pixels, npixels);



/**********************/
/* Write outfits File */
/**********************/

  /* open output file */
  outfits_fid = im_open(outfitsFilename, "w");
  if (outfits_fid < 0) quit("Unable to open out file %s\n", outfitsFilename);

  /* copy header cards */
  fid_in_array[0] = infits_fid;
  fid_in_array[1] = 0;
  im_hdr(outfits_fid, fid_in_array, 0);

  /* compute vx, vy, vz from table */
  real_jd = mjd_obs + 2400000.5;
  for (i=0; i<tbl_nrows; i++) {
    if (real_jd <= tbl_jd[i]) break;
  }
  if (real_jd<tbl_jd[0]) {
    printf("\nWARNING[velocity]: Julian date before beginning of table\n", real_jd);
    vx = tbl_vx[0];
    vy = tbl_vy[0];
    vz = tbl_vz[0];
  }
  else if (i>=tbl_nrows) {
    printf("\nWARNING[velocity]: Julian date past end of table %f\n", real_jd);
    vx = tbl_vx[tbl_nrows-1];
    vy = tbl_vy[tbl_nrows-1];
    vz = tbl_vz[tbl_nrows-1];
  }
  else { /* interpolate */
    frac = (real_jd - tbl_jd[i-1]) / (tbl_jd[i] - tbl_jd[i-1]);
    vx = frac * tbl_vx[i] + (1.0-frac) * tbl_vx[i-1];
    vy = frac * tbl_vy[i] + (1.0-frac) * tbl_vy[i-1];
    vz = frac * tbl_vz[i] + (1.0-frac) * tbl_vz[i-1];
  }

  /* compute velocity correction */
  ra_radians  = ra_used  * (M_PI/180.0);
  dec_radians = dec_used * (M_PI/180.0);
  x = cos(ra_radians)*cos(dec_radians);
  y = sin(ra_radians)*cos(dec_radians);
  z = sin(dec_radians);
  vhelcorr = x*vx + y*vy + z*vz;
  vlsrcorr = vhelcorr + 0.145*x - 17.320*y + 10.000*z;
  if (verbose >=3) {
    printf("[velocity] jd=%.6f, tbl_jd=%.1f vx=%.9f,vy=%.9f, vz=%.9f frac=%5.2f\n",
            real_jd, tbl_jd[i], vx, vy, vz, frac*100.0);
    printf("[velocity] x=%.9f y=%.9f z=%.9f vhelcorr=%.9f vlsrcorr=%.9E\n",
            x, y, z, vhelcorr, vlsrcorr);
  }

  /* add header card to outfits */
  sprintf(header_card,"VHELCORR= %20.1f / Velocity correction (km/sec)", vhelcorr);
  if (verbose >=3) printf("%s\n", header_card);
  im_wkey_t(outfits_fid, header_card,"end");
  sprintf(header_card,"VLSRCORR= %20.1f / Velocity correction (km/sec)", vlsrcorr);
  if (verbose >=3) printf("%s\n", header_card);
  im_wkey_t(outfits_fid, header_card,"end");

  /* write pixels */
  im_wpix_d(outfits_fid, ptr_pixels, npixels);

  /* close files */
  im_close(infits_fid);
  im_close(outfits_fid);


/********/
/* EXIT */
/********/

  return TERMINATE_NORMAL;
}


/******************/
/******************/
/* MAIN MAIN MAIN */
/******************/
/******************/
main(argc, argv)
int argc;
char *argv[];
{
  int status;
  status = velocity(argc, argv);
  if (status < 0) return TERMINATE_ERROR;
  param_exit(status);
}
