# BCDACQ pipeline wrapper

#=======================================================================
# initialize
use strict;
require "$ENV{'CUPID'}/wrappers/wrapper_subs.pm"; # standard subroutines
my $band = initialize();
my $dir_include      = "$ENV{'CUPID'}/include"; # ancillary directory

#=======================================================================
# Get FITS keyword values from input file
my $kwd_SAMPTIME = get_kwd('SAMPTIME');
my $kwd_GAIN1    = get_kwd('GAIN1', '4.6');
my $kwd_APERTURE = get_kwd('APERTURE');
my $kwd_APERNAME = get_kwd('APERNAME');
my $kwd_TARGETNA = get_kwd('TARGETNA');
my $kwd_EXPTYPE  = get_kwd('EXPTYPE');
my $kwd_DCENUM   = get_kwd('DCENUM');

#=======================================================================
# skip this pipeline if DCENUM 3 (not correct type of data)
if ($kwd_DCENUM == 3) {
  message (1, "Skipping processing because DCENUM = 3");
  terminate();
}

#=======================================================================
# create files to update FITS header keywords
write_kwd_files(); # (code is at end of file)

#=======================================================================
#STEP CVTI2R4 convert pixels from integer to floating point
# (must remove dmask if exists because cvti2r4 will pick it up if it's there)
run_module("CVTI2R4:RM remove_file
IN  -   dmask.fits # Input DMASK datafile
");
run_module("CVTI2R4 cvti2r4
V   -t no             # two's complement conversion: yes no
V   -a 32768          # offset added to each pixel
IN  -i <INFILE1>      # Input raw FLUX datafile
V   -b -1             # value for missing pixel
V   -k 10             # set dmask bit 10 (1024) for saturated pixels
OUT -d bmask.fits     # Output BMASK datafile
OUT -o cvti2r4.fits <FLUX> # Output FLUX datafile
V   -s 65533          # saturation value in DN 32765 + 32768
");

#=======================================================================
#STEP ADDKWDS update FITS header keywords
# first delete keywords
run_module("ADDKWDS:DEL hdrupd8
UPD -  <FLUX>             # Input/Output FLUX datafile
V   -d .keywords_del.tmp  # delete keyword cards
V   -n -                  # don't add COMMENT card
");
# then add keywords
run_module("ADDKWDS hdrupd8
UPD -  <FLUX>            # Input/Output FLUX datafile
V   -a .keywords_add.tmp # append the tmp file keyword cards to add
V   -n -                 # don't add COMMENT card
");

#=======================================================================
#STEP DNTOFLUX convert from DN to electrons/sec
my $flux_factor = sprintf("%9.6f", $kwd_GAIN1 / (8.0 * $kwd_SAMPTIME) );
run_module("DNTOFLUX dntoflux
IN  -i  <FLUX> # Input FLUX datafile
OUT -o1 dntoflux.fits <FLUX> # Output FLUX datafile
V   -a  $dir_include            # location of ancillary directory
V   -s  e-/sec                  # units to put in FITS keyword
V   -co 1                       # operation: 1=readout-channel-dependent
V   -c1 $flux_factor            # for readout channel 1
V   -c2 $flux_factor            # for readout channel 2
V   -c3 $flux_factor            # for readout channel 3
V   -c4 $flux_factor            # for readout channel 4
");

#==================================================================
#STEP IMSUBTRACT Subtract the dark
run_module("IMSUBTRACT imsubtract
IN  -i1   <FLUX>              # Input datafile
CAL  -i2  dcsdark.fits        # 2-D Dark
OUT  -o   darksub.fits <FLUX> # Output dark subtracted datafile
");

#=======================================================================
#STEP DARKDRIFT remove "jail bars"
run_module("DARKDRIFT darkdrift
CDF -n  darkdrift.nl             # namelist file
IN  -i  <FLUX>                   # Input FLUX datafile
OUT -o1 darkdrift.fits <FLUX>    # Output FLUX datafile
OUT -o3 darkdrift2.tbl           # ascii-table-darkdrift
V   -a  $dir_include
V   -p  1                        # data plane
");

#=======================================================================
#STEP FLATAP apply FLAT
run_module("FLATAP flatap
IN  -i  <FLUX>               # Input FLUX datafile
CAL -f  dcsflat.fits         # FLATFIELD CAL file
OUT -o  flatap.fits <FLUX>   # Output FLUX datafile
V   -dm 28672                # BMASK fatal bits
# V   -pm 18176              # PMASK fatal bits
# V   -cm   444              # CMASK fatal bits
# CAL -p  pmask.fits         # PMASK CAL file
# CAL -c  flatfield_cmask.fits # FLATFIELD_CMASK CAL file
UPD -d  bmask.fits           # Input/Output BMASK datafile
# IN  -ui <UNC>              # Input UNC datafile
# OUT -uo func.fits <UNC>    # Output UNC datafile
");

#=======================================================================
#STEP FPGEN run 'final product generator' and add velocity keywords
# delete bcd file if it exists; FPGEN will crash if bcd file already exists
# add velocity header cards
run_module("FPGEN:VELOCITY velocity
IN  -i <FLUX>          # Input FLUX datafile
CDF -t velocity.tbl    # ASCII TABLE fluxcon CDF file
OUT -o <FLUX>          # Output FLUX datafile
");
run_module("FPGEN:RM remove_file
IN  -   acq_f.fits       # Input FLUX datafile
");
# Run FPGEN final product generator (standardize the FITS header keywords)
run_module("FPGEN fpgen
IN  -i <FLUX>          # Input FLUX datafile
OUT -o acq_f.fits <FLUX> # Output FLUX datafile
CDF -t fpg.tpl         # reformatting 'template' file
V   -h 1               # copy HISTORY lines too
V   -b 1               # remove blocks of blanks
V   -q 1               # 'quiet' mode
RESET_STATUS 32 0 # force status 32 (warning) change to 0 (OK) 
");

#=======================================================================
#STEP BLUE do BLUE cutput
run_module("BLUE:CROP1 cropfits2
CDF -n  cropfits2_b.nl # namelist file
IN  -i <FLUX> # Input FLUX datafile
OUT -o cropfits_b.fits # Output blue flux cutout
");
run_module("BLUE:CROP2 cropfits
CDF -n  cropfits_b.nl # namelist file
IN  -i bmask.fits # Input BMASK datafile
OUT -o mskb.fits # Output BLUE bmask cutout
V   -a  $dir_include # location of ancillary directory
");
run_module("BLUE dntoflux
CDF -n  dcs_dntoflux_b.nl # namelist file
IN  -i cropfits_b.fits # Input BLUE flux cutout
OUT -o1 acqb.fits # Output BLUE flux (calibrated) cutout
IN  -id mskb.fits # Input BLUE bmask cutout
V   -a  $dir_include # location of ancillary directory
IN  -iu cropfits_b.fits #  (bogus input so module will run)
V   -o2 dntof.fits #  (bogus output so module will run)
");

#=======================================================================
#STEP RED do RED cutput
run_module("RED:CROP1 cropfits2
CDF -n  cropfits2_r.nl # namelist file
IN  -i <FLUX> # Input FLUX datafile
OUT -o cropfits_r.fits # Output red flux cutout
");
run_module("RED:CROP2 cropfits
CDF -n  cropfits_r.nl # namelist file
IN  -i bmask.fits # Input BMASK datafile
OUT -o mskr.fits # Output RED bmask cutout
V   -a  $dir_include # location of ancillary directory
");
run_module("RED dntoflux
CDF -n  dcs_dntoflux_r.nl # namelist file
IN  -i cropfits_r.fits # Input RED flux cutout
OUT -o1 acqr.fits # Output RED flux (calibrated) cutout
IN  -id mskr.fits # Input RED bmask cutout
V   -a  $dir_include # location of ancillary directory
IN  -iu cropfits_r.fits #  (bogus input so module will run)
V   -o2 dntof.fits #  (bogus output so module will run)
");


#=======================================================================
# print END message, update STATUS file,  and exit the program
terminate();


#########################################################################
#########################################################################
sub write_kwd_files { # this code moved to avoid clutter near start of script

#-----------------------------------------------------------------------
# setenv SOS_REL (for use by FPGEN to make "CREATOR" kwd)
$ENV{'SOS_REL'} = $ENV{'CUPID_VERSION'};

#-----------------------------------------------------------------------
# create .keywords_del file (to delete unwanted FITS keywords)
open(HDRUPDATE, ">.keywords_del.tmp");
printf HDRUPDATE "CREATOR\n"; # will be replaced, below
printf HDRUPDATE "CAL_SET\n"; # will be replaced, below
printf HDRUPDATE "DPID\n"; # not needed
close(HDRUPDATE);

#-----------------------------------------------------------------------
# create .keywords_add file (to add new/updated FITS keywords)
open(HDRUPDATE, ">.keywords_add.tmp"); # open file to write keywords to
# generate CREATOR keyword
printf HDRUPDATE "CREATOR = '%-8s'           / version of CUPID used to create this file\n", $ENV{'CUPID_VERSION'};;
# generate CAL_SET keyword
my $calset_file = fetch_cal('cal_set.txt');
open(CALSET_FILE, "<$calset_file") || message(64, "Couldn't open ${calset_file}");
my @calset_words = split(' ', <CALSET_FILE>);
my $calset_id = $calset_words[0];
close(CALSET_FILE);
printf HDRUPDATE "CAL_SET = '%-8s'           / ID of set of CAL files used (by CUPID)\n", $calset_id;
# generate FOVID, FOVNAME, OBJECT keywords (for SPICE)
printf HDRUPDATE "FOVID   = %d / Aperture number\n", $kwd_APERTURE;
printf HDRUPDATE "FOVNAME = '%s' / Aperture name\n", $kwd_APERNAME;
printf HDRUPDATE "OBJECT  = '%s' / Target name\n", $kwd_TARGETNA;
# generate FLXCON keywords
my $fluxcon_file = fetch_cal('fluxcon.tbl');
my $inline;
open(FLUXCON, "<$fluxcon_file") || message(64, "Couldn't open ${fluxcon_file}");
while (defined ($inline = <FLUXCON>)) {
  chomp $inline;
  next unless ($inline gt " ");
  if ($inline !~ /^(\\|\|)/) {
    my ($junk, $order, $wave, $passband, $s_irs, $s_irs_sigma, $s_ref,
        $s_ref_sigma, $fluxconv, $fluxconv_sigma, $a5, $a5_sigma,
        $a4, $a4_sigma, $a3, $a3_sigma, $a2, $a2_sigma, $a1, $a1_sigma, $a0, $a0_sigma)
        = split /\s+/, $inline;
     printf HDRUPDATE "FLXCON%02d= %20.6f / [e/Jy] Flux conversion Order %d\n", $order, $fluxconv, $order;
     printf HDRUPDATE "FLXERR%02d= %20.6f / [e/Jy] Flux conversion uncertainty Order %d\n",
                       $order, $fluxconv_sigma, $order;
   }
}
close(HDRUPDATE);

}
