#######################################################################
# CUPID SHARED WRAPPER (PERL) SUBROUTINES
#######################################################################
{
use strict;


#==============================================
# global variables
#==============================================

# identifiers
our $PIPE_TYPE;
our $SUBSTEP; # STEPNAME:SUBSTEPNAME  -or- just STEPNAME if no substep
our $WHOLE_STEP; # STEPNAME
our @input_files; # primary input file(s)
our $prefix_datafile; # e.g. SPITZER_S0_20136960_0002_
our ($band, $sclk, $expcode); # for fetching CAL and CDF files
our %variables; # values of <VARIABLES> passed from step to step

# options
our %overrides; # "SET" $overrides{'STEPAME-flag'} = value
our %override_used; # "SET" $override_used{'STEPAME-flag'} = 1 when it has been used
our %skip_steps; # "SKIP" steps to skip  $skip_steps{'STEPNAME'} = 1;
our $start_step = ''; # "START" stepname to start processing at
our $stop_step = '';  # "STOP" stepname to stop processing after
our %do_after;    # "DO.AFTER"       $do_after{'STEP_AFTER'} = STEP_DO
our %need_run_c1; # "DO.AFTER" $need_run_c1{'STEP_AFTER'} = 1st need_run_coms
our %need_run_c2; # "DO.AFTER" $need_run_c2{'STEP_AFTER'} = last need_run_coms
our @need_run_coms; # "DO.AFTER" commands that need to be run

# directories
our $dir_cdf_archive;
our $dir_cal_archive;
our $dir_executables;

# logfile verbosity
our $show_run_actual; # actual system calls to modules
our $show_get_kwd   ; # keayords read from FITS files
our $show_fetch     ; # cal and cdf files fetched
our $show_get_env   ; # environment variables accessed (2 = even if default)
our $show_time_summ ; # summary at end of n seconds each step took

#==============================================
sub get_env
#==============================================
{ (my $env_name, my $deflt) = @_;
# get value of environment variable (from %ENV)
# FATAL if not defined and no default is given

  my $val = $ENV{$env_name};

  if (defined $val) {
    printf("ENV(%s) = %s\n", $env_name, $val) if ($show_get_env >= 1);
  } else {
    if (defined $deflt) {
      $val = $deflt;
      printf("ENV(%s) = %s\n", $env_name, $val) if ($show_get_env >= 2);
    } else {
      message (64, "A required environment variable is not defined: $env_name");
    }
  }
  return $val;
}

#==============================================
sub get_kwd
#==============================================
# get keyword value from FITS header
{ (my $kwd_name, my $deflt) = @_;
  our %kwd_hash;
  our $fits_file;


  if ($kwd_name ne 'GET_FITS_FILE') {
    my $val = $kwd_hash{$kwd_name};

    unless (defined $val) {
      if (defined $deflt) {
        $val = $deflt;
      } else {
        message (64, "File $fits_file must must have FITS kwd: $kwd_name");
      }
    }
    printf("KWD(%s) = %s\n", $kwd_name, $val) if ($show_get_kwd >= 1);
    return $val;
  }

  # read fits keywords into the hash
  $fits_file = $deflt;
  my ($card,$retval,$key,$val,$comment);
  message(0,"Reading FITS keywords from $fits_file") if ($show_get_kwd);
  sysopen (FITS, $fits_file, 0)
    or message(64, "Unable to open FITS file $fits_file to read keywords");
  while($retval = sysread(FITS, $card, 80)) {
    if ($retval < 80)
      {message(64, "Premature EOF on $fits_file while reading keywords");}
    if ($card =~ /[^ -~]/) # Non-printing character found
      {message(64, "Reading past header in $fits_file while reading keywords");}
    last if $card =~ /^END +$/; # End of header
    next if substr($card,8,1) ne "="; # Not a key=value pair
    ($key,$val,$comment) = $card =~ m/^(\S+)\s*=\s*(.*?)\s*(\/\s*(.*)?)?$/;
    if ((defined $key) && (defined $val)) {
       $val =~ s/^\'\s*(.*?)\s*\'$/$1/;   # Strip quotes and white space.
       $kwd_hash{$key} = $val;
    }
  }
  if (! (defined $retval)) 
     {message(64, "I/O error on $fits_file while reading keywords");}
  close FITS;
}


#==============================================
sub message
#==============================================
{ (my $level, my $msg_text) = @_;

# DOES:
#  - updates $wrapper_status
#  - prints appropriate "error level" message (INFO/WARN/FATAL)
#  - prints message itself
#  - exits if FATAL status
#
#  level:
#   64 to 255= FATAL ERROR
#   32 to 62 = WARNING
#          1 = skip (terminated)
#    2 to 31 = special message (generally something unusual but not serious)
#          0 = routine emphasized message (completely expected)
#
# special case: if msg_txt is GET_WRAPPER_STATUS, no print, just return status

  our $wrapper_status; $wrapper_status = 0 unless (defined $wrapper_status);

  if ($msg_text eq 'GET_WRAPPER_STATUS') {
    return $wrapper_status;
  }

  # update $wrapper_status
  $wrapper_status =  $level if ($level > $wrapper_status);

  # printout "error level" line
  if ($level >= 64) {
    printf "\n#########################################################\n";
    printf "FATAL ERROR status = %d (unable to continue)\n", $level;
  }
  elsif ($level >= 32) {
    printf "\n*********************************************************\n";
    printf "WARNING status = %d\n", $level;
  }
  elsif ($level >= 1) {
    printf "\n---------------------------------------------------------\n";
    printf "INFO status = %d\n", $level;
  }
  else {
    printf "\n---------------------------------------------------------\n";
  }

  # print the message itself
  printf "$msg_text\n";

  # print "end of message" line
  if ($level >= 32) {
    printf "*********************************************************\n\n";
  }
  else {
    printf "---------------------------------------------------------\n\n";
  }

  # exit if fatal
  terminate() if ($level >= 63);
  terminate() if ($level ==  1);
}

#==============================================
sub run_module
#==============================================
{my @lines = split("\n", $_[0]);
  use File::Copy;
  my (@flag, @value);
  my ($module, $params);
  our $step_num; $step_num = 0 unless (defined $step_num);
  my @reset_status_from=(); my @reset_status_to; my $reset_status_message='';

  for (my $i=0; $i<=$#lines; $i++) {
    my ($tokens, $comment) = split("#", $lines[$i]);
    my (@t_array) = split(' ', $tokens);
    if ($i==0) {
      if ($#t_array != 1)
        { message(200, "WRAPPER ERROR: run_module subroutine requires SUBSTEP_NAME and module_name"); }
      $SUBSTEP = $t_array[0];
      ($WHOLE_STEP, my @sstep) = split(':', $SUBSTEP);
      $module = $t_array[1];

      # check if this step is to be done.after; save command for later if so
      my $tgt_step = $do_after{$WHOLE_STEP};
      if (defined $tgt_step) {
        push @need_run_coms, $_[0]; # save entire command
        $need_run_c2{$tgt_step} = $#need_run_coms;
        $need_run_c1{$tgt_step} = $#need_run_coms unless (defined $need_run_c1{$tgt_step});
        delete $do_after{$WHOLE_STEP} if ($WHOLE_STEP eq $SUBSTEP);
        return;
      }

      # check $start_step  and $skp_steps
      if ($start_step eq $WHOLE_STEP) {
        message(2, "STARTing with STEP: $WHOLE_STEP");
        $start_step = ''; 
      }

      if (defined $skip_steps{$WHOLE_STEP}) {
        if ($SUBSTEP eq $WHOLE_STEP) {
          message(3, "Skipping STEP: $WHOLE_STEP");
          delete $skip_steps{$WHOLE_STEP}; # so won't get error message later
        }
        return; # don't run module
      }

      unless ($start_step gt '') {
        my $time_string = datetime_string($SUBSTEP);
        printf "\n=========================================================\n";
        if ($SUBSTEP eq $WHOLE_STEP) {
          $step_num += 1;
          printf "START_STEP %2d. %-13s  %s\n", $step_num, "$SUBSTEP", $time_string;
        } else {
          printf "START_SUBSTEP  %-13s  %s\n", "$SUBSTEP", $time_string;
        }
        printf "=========================================================\n\n";
      }

      next;


    }
    next if ($#t_array <= 0); # (no actual parameters, just ignore)
    if ($#t_array < 2) {
         message(64, "Calling $module, each parameter must have 3 parts: type, flag, value\n$tokens");}
    my ($toke_type, $one_flag, $one_value, $one_variable) = @t_array;
    if ($toke_type eq 'RESET_STATUS') {
      push (@reset_status_from, $one_flag);
      push (@reset_status_to,   $one_value);
      next;
    }
    next unless (defined $one_flag);
    $flag[$i] =  $one_flag;
    $value[$i] = $one_value;
    my $full_par_name = "$module$one_flag";

    if ($toke_type eq 'IN' || $toke_type eq 'OUT' || $toke_type eq 'UPD') {
      if (defined $overrides{$full_par_name}) {
        message(244, "May not override names of data fles (IN OUT UPD)");
      }
      if (substr($value[$i],0,1) ne '<') {
        $value[$i] = $prefix_datafile . $value[$i];
      } else {
        my $tmp = $variables{$value[$i]};
        unless (defined($tmp)) {
          my ($part1, $part2) = split(':', $value[$i]);
          if ($part2 eq 'ALT>') {
            $tmp = $variables{$part1 . '>'}; 
          }
        }
        unless (defined($tmp)) {
          message(251, "WRAPPER ERROR: undefined variable $value[$i]");
        }
        $value[$i] = $tmp;
      }
      if (defined $one_variable) { $variables{$one_variable} = $value[$i]; }
      next;
    }

    # process if overrides requested
    if (defined $overrides{$full_par_name}) {
      $value[$i] = $overrides{$full_par_name};
      print "OVERRIDE: $one_flag $value[$i]\n";
      $override_used{$full_par_name} = 1; # so can check at end if all used
      next;
    }

    if ($toke_type eq 'V') {
      next;
    }
    if ($toke_type eq 'CDF') {
      next if ($start_step gt ' '); # don't do fetches before start_step
      $value[$i] = fetch_cdf($value[$i]);
      next;
    }
    if ($toke_type eq 'CAL') {
      next if ($start_step gt ' '); # don't do fetches before start_step
      $value[$i] = fetch_cal($value[$i]);
      next;
    }
    message(250, "Unknown parameter type prefix: ${toke_type}_\n @lines");
  }

  return if ($start_step gt ' '); # don't run module if before start_step

  my $return_status;

  if ($module eq 'copy_file') {
    printf("PERL CALL: copy($value[1], $value[2])\n\n")
     if ($show_run_actual);
    $return_status = copy($value[1], $value[2]);
    if ($return_status ==1) { $return_status = 0; }
    else                    { $return_status = 64; }
  }

  elsif ($module eq 'remove_file') {
    printf("PERL CALL (remove file): unlink($value[1])\n\n")
     if ($show_run_actual);
    unlink($value[1]);
    $return_status = 0; # ignore actual return status
  }
 
  else { # call normal module
    my @command_array;
    push (@command_array, "$dir_executables/$module");

    for (my $j=1; $j<=$#lines; $j++) {
      if (defined $flag[$j] && $flag[$j] ne '-') {
        push (@command_array, $flag[$j]);
      }
      if (defined $value[$j] && $value[$j] ne '-') {
        push (@command_array, $value[$j]);
      }
    }

    # process overrides that were not mentioned in wrapper
    foreach my $key (keys %overrides) {
      next unless ( $module eq substr($key, 0, length($module)) );
      my $tflag = substr($key, length($module), 99);
      print "OVERRIDE: $tflag $overrides{$key}\n";
      push (@command_array, $tflag);
      push (@command_array, $overrides{$key});
      delete $overrides{$key}; # so can check at end if all used
    }

    if ($show_fetch) { printf "\n"; }

    print 'MODULE_CALL: ' . join(' ',@command_array) . "\n\n"
     if($show_run_actual);

    $return_status = system(@command_array);
    if ($return_status < 0 ) {
      printf "Command: %s\n", join(' ', @command_array);
      message(255, "Unable to run module");
    }

    $return_status = $return_status / 256; # discard rightmost 8 gits
    if ($return_status >= 64) {
      printf "Command: %s\n", join(' ', @command_array);
      message($return_status, "Module terminated abnormally");
    }

    # check if status should be reset
    for (my $k=0; $k<=$#reset_status_from; $k++) {
      if ($return_status == $reset_status_from[$k]) {
        $reset_status_message =
             "\n  (reset from status = $return_status)";
        $return_status = $reset_status_to[$k]
      }
    }
  }
  my $str;
  if ($SUBSTEP eq $WHOLE_STEP) {
    $str = sprintf "END_STEP %2d. %-13s status=%3d",
           $step_num, "$SUBSTEP", $return_status;
  } else {
    $str = sprintf "END_SUBSTEP  %-13s status=%3d",
            "$SUBSTEP", $return_status;
  }
  message($return_status, "$str $reset_status_message");
  if ($stop_step eq $SUBSTEP) {
    message (2, "STOPing after STEP: $WHOLE_STEP");
    terminate();
  }

  # see if there are any do.after (need_run) commands
  if (defined $need_run_c1{$SUBSTEP}) {
    my $c1=$need_run_c1{$SUBSTEP}; my $c2 = $need_run_c2{$SUBSTEP};
    delete $need_run_c1{$SUBSTEP};
    message(3, "Running following step out of normal order");
    for (my $c=$c1; $c<= $c2; $c++) {
      run_module($need_run_coms[$c]);
    }
  }
}


#==============================================
sub fetch_cdf  ## (sub-subroutine)
#==============================================
{my ($base_name) = @_;
  my $local_file = "../cdf/b${band}_$base_name";
  if (-e $local_file) {
    printf "FETCH $local_file (already there)\n" if ($show_fetch);
    return $local_file;
  }
  foreach my $file_name ('b'.$band.'_'.$base_name , $base_name) {
    my $archive_file = $dir_cdf_archive .'/'. $file_name;
    if (-e $archive_file) {
      print "FETCH $local_file <- $file_name\n" if ($show_fetch);
      copy ($archive_file, $local_file);
      return $local_file;
    }
  }
  message (64, "Unable to fetch $base_name
     from $dir_cdf_archive");
}


#==============================================
sub basename_if_qualified ## (sub-subroutine)
#==============================================
# returns: "" if not qualified, basename if qualified
# returns: (basename, delta_sclk) if it IS qualified
{my ($filename) = @_;
  my $sclk_irs9    =  772487104; # IRS campaign 9.5
  my $sclk_irs9999 = 9999999999; # end of mission

  my $c1 = substr($filename,0,1);
  return "" if ($c1 eq '.');
  return "" if ($c1 eq '_');
  return "" if ($c1 eq 'T');

  my @F = split('_', $filename);

  # process band
  if ($c1 eq 'b') {
    return "" if ($F[0] ne 'b' . $band);
    splice(@F, 0, 1);
  }

  # get expcode if normal dark
  if (defined $F[1] &&
    ( substr($F[1],0,4) eq 'dark' || substr($F[1],0,7) eq 'dcsdark') ) {
    unless (defined $expcode) {
      my $frametime   = get_kwd('HCFRAM_T', 0);
      my $framecount  = get_kwd('HCFRAM_C', 0);
      if (defined $frametime && defined $framecount) {
        $expcode = 'r6'   if ($frametime == 2 && $framecount ==  4);
        $expcode = 'r14'  if ($frametime == 2 && $framecount ==  8);
        $expcode = 'r30'  if ($frametime == 2 && $framecount == 16);
        $expcode = 'r60'  if ($frametime == 3 && $framecount == 16);
        $expcode = 'r120' if ($frametime == 4 && $framecount == 16);
        $expcode = 'r240' if ($frametime == 5 && $framecount == 16);
        $expcode = 'r480' if ($frametime == 6 && $framecount == 16);
        # below for DCS darks;  above for normal darks
        $expcode = 'r9'   if ($frametime == 2 && $framecount ==  2);
        $expcode = 'r18'  if ($frametime == 3 && $framecount ==  2);
        $expcode = 'r36'  if ($frametime == 4 && $framecount ==  2);
      }
      $expcode = 'rXXX' unless (defined $expcode);
    }
    return "" if ($F[0] ne $expcode);
    splice(@F, 0, 1);
  }

  my $part_filename = join('_', @F);
  my $fn1; my $sclk1; my $sclk2; my $fn2; my $w;
  while (1) { # find sclk coverage of file
    # if explicit SCLK range is in filename
    ($fn1, $sclk1, $sclk2, $fn2)
      = $part_filename =~ /(.+)_([0-9]+)_([0-9]{6,})(.+)/;
    last if (defined $fn1);

    # if "post_irs9" is in filename
    ($fn1, $w, $fn2) = $part_filename =~ /(.+)_(post_irs9)(.+)/;
    if (defined $fn1) { $sclk1 = $sclk_irs9; $sclk2=$sclk_irs9999; last; }

    # if "pre_irs9" is in filename
    ($fn1, $w, $fn2) = $part_filename =~ /(.+)_(pre_irs9)(.+)/;
    if (defined $fn1) { $sclk1 = 0; $sclk2 = $sclk_irs9; last; }

    # if "pre_IRS-9" is in filename
    ($fn1, $w, $fn2) = $part_filename =~ /(.+)_(pre_IRS-9)(.+)/;
    if (defined $fn1) { $sclk1 = 0; $sclk2 = $sclk_irs9; last; }

    # if no SCLK range in filename
    $sclk1 = 0; $sclk2 = $sclk_irs9999;
    $fn1 = $part_filename; $fn2 = '';
    last;
  }
  return "" if ($sclk < $sclk1);
  return "" if ($sclk > $sclk2);
  my $basename = $fn1 . $fn2;
  my $dsclk    = $sclk2 - $sclk1;
  return ($basename, $dsclk);
}


#==============================================
sub fetch_cal  ## (sub-subroutine)
#==============================================
{my ($base_name) = @_;
  our %cal_fullnames; # name of cal file in archive
  our %cal_delta_sclk; # sclk range (-1 for local files in local subdir)
  our $archive_scanned_flag; # set if the archive has been scanned
  my $cal_dir = '../cal';

  # get names of local cal files (if not done yet)
  unless (defined %cal_fullnames) {
    opendir (CALDIR, $cal_dir);
    my @filenames = readdir CALDIR;
    closedir CALDIR;
    for (my $i=0; $i<=$#filenames; $i++) {
      my ($shortname, $dsclk) = basename_if_qualified($filenames[$i]);
      next if ($shortname eq ""); # not a qualified cal file
      next if (defined $cal_fullnames{$shortname}); # already found one
      if (defined $cal_fullnames{$shortname}) {
        print "\n-----------------\n";
        print "   $cal_fullnames{$shortname}\n";
        print "   $filenames[$i]\n";
        message (64, "More than one $shortname in local cal sub-directory");
      }
      $cal_fullnames{$shortname} = $filenames[$i];
      $cal_delta_sclk{$shortname} = -1; # indicates from local directory
    }
  }

  # look up full filename
  my $fullname = $cal_fullnames{$base_name};

  # read filenames from cal archive if needed
  unless ( (defined $fullname) || (defined $archive_scanned_flag) ) {
    opendir (CALDIR, $dir_cal_archive);
    my @filenames = readdir CALDIR;
    closedir CALDIR;
    for (my $i=0; $i<=$#filenames; $i++) {
      my ($shortname, $dsclk) = basename_if_qualified($filenames[$i]);
      next if ($shortname eq ""); # not a qualified cal file
      if (defined $cal_fullnames{$shortname}) {
        next if ($dsclk > $cal_delta_sclk{$shortname});
      }
      $cal_fullnames{$shortname} = $filenames[$i];
      $cal_delta_sclk{$shortname} = $dsclk;
# print "cal arch found: $shortname $filenames[$i]\n";
    }
    $archive_scanned_flag = 1;
    $fullname = $cal_fullnames{$base_name};
  }

  # return filename if found, else ERROR out
  message (64, "Unable to fetch CAL $base_name") unless (defined $fullname);
  if ($cal_delta_sclk{$base_name} > 0) {
    print "FETCH $cal_dir/$fullname from archive\n" if ($show_fetch);
    copy ("$dir_cal_archive/$fullname", "$cal_dir/$fullname");
  } else {
    print "FETCH $cal_dir/$fullname (already there)\n" if ($show_fetch);
  }
  return $cal_dir .'/'. $fullname
}



#==============================================
sub datetime_string
#==============================================
{
  our @timing_step; our @timing_nsec; # for summary of seconds per run_module

  my ($par) = @_;

  if ($par eq 'TOTAL') { return $timing_nsec[$#timing_step]-$timing_nsec[0]; }

  unless ($par eq 'SUMMARY') {
    my $time = time();
    push (@timing_step, $par);
    push (@timing_nsec, $time);
    my ($sec, $min, $hr, $day, $mon, $yr);
    ($sec, $min, $hr, $day, $mon, $yr) = localtime($time);
    $mon += 1;
    $yr += 1900;
    my $text_time = sprintf "%04d/%02d/%02d %02d:%02d:%02d",
                      $yr, $mon,$day, $hr,$min,$sec;
    return $text_time;
  }

  # print summary
  return unless($show_time_summ);
  printf "-------------------------------\n";
  printf "  Module Timing Summary\n";
  printf "-------------------------------\n";
  printf "  SUBSTEP        NSEC.\n";
  for (my $i=1; $i<$#timing_step; $i++) {
    printf "%-14s %3d\n",
      "$timing_step[$i]", $timing_nsec[$i+1]-$timing_nsec[$i];
  }
}


#==============================================
sub initialize
#==============================================
{
  #------------------------------------------------------------------
  # set show_* variables for logfile verbosity
  $show_run_actual = 0; $show_get_kwd = 0; $show_fetch = 0;
  $show_get_env = 0; $show_time_summ = 0;
  my $env_CUPID_LOG   = get_env('CUPID_LOG', 2);
  $show_run_actual = 1 if ($env_CUPID_LOG >= 1);
  $show_get_kwd    = 1 if ($env_CUPID_LOG >= 2);
  $show_get_env    = 1 if ($env_CUPID_LOG >= 2);
  $show_fetch      = 1 if ($env_CUPID_LOG >= 2);
  $show_get_env    = 2 if ($env_CUPID_LOG >= 3);
  $show_time_summ  = 1 if ($env_CUPID_LOG >= 3);

  #------------------------------------------------------------------
  # get info from ARGV
  $PIPE_TYPE = $ARGV[0];
  @input_files = split(',', $ARGV[1]); # primary input  file
  $prefix_datafile = $ARGV[2]; # prefix for output data files

  #------------------------------------------------------------------
  # print first few lines of logfile
  printf "%s: %s  %s\n",
   $PIPE_TYPE, $prefix_datafile, datetime_string('STARTING');
  printf "OUTPUT DIRECTORY: $ENV{'PWD'}\n";
  printf "Pipeline log file --- %s\n", $ENV{'CUPID_VERSION'};
  print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";

  #------------------------------------------------------------------
  # get environment variables and set important paths
  my $dir_CUPID = $ENV{'CUPID'};
  $dir_cdf_archive = get_env('CUPID_CDF_ARCHIVE', "$dir_CUPID/cdf");
  $dir_cal_archive = get_env('CUPID_CAL_ARCHIVE', "$dir_CUPID/cal");
  $dir_executables = "$dir_CUPID/". get_env('CUPID_BIN'); # executables subdir

  #------------------------------------------------------------------
  # Process options
  printf ("OPTIONS:\n") if ($#ARGV >= 3);
  for (my $i=3; $i<=$#ARGV; $i++) {
    my ($option_type, @option_details) = split('\.', $ARGV[$i]);
    if ($option_type eq 'SET') {
      my $override_string = join ('.', @option_details);
      my ($override_module_param, $override_value) = split('=', $override_string);
      unless ($override_value gt ' ') {
        message (206, "Override $override_module_param has no '=' value");
      }
      # force module name to lower case
      my ($override_module, $override_param) = split('-', $override_module_param);
      $override_module_param = "\L$override_module" . "-$override_param";
      $overrides{"$override_module_param"} = $override_value;
      printf "   override: %s = %s\n", $override_module_param, $override_value;
    } else {
      my $option_stepname = "\U$option_details[0]";
      if      ($option_type eq 'SKIP') {
        $skip_steps{$option_stepname} = 1;
        printf "   skip: %s\n", $option_stepname;
      } elsif ($option_type eq 'START') {
        $start_step = $option_stepname;
        printf "   start at: %s\n", $option_stepname;
      } elsif ($option_type eq 'STOP') {
        $stop_step  = $option_stepname;
        printf "   stop after: %s\n", $option_stepname;
      } elsif ($option_type eq 'DO') {
        unless ("\U$option_details[1]" eq 'AFTER')
          { message (205, "'DO' option without 'AFTER'"); }
        $do_after{$option_stepname} = "\U$option_details[2]";
        printf "   do %s after %s\n", $option_stepname, "\U$option_details[2]";
      } else {
        message (207, "Unknown command line option: $option_type");
      }
    }
  }

  print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n";

  mkdir '../cdf'; mkdir '../cal';

  # Put @input_files  array into %variables
  for (my $k=0; $k<=$#input_files; $k++) {
    my $kwd = '<INFILE' . ($k+1) . '>';
    $variables{$kwd} = $input_files[$k];
  }

  # Get info from FITS header
  unless (defined $band) {
    get_kwd('GET_FITS_FILE', $input_files[0]);
    $band = get_kwd('CHNLNUM');
    $sclk = get_kwd('SCLK_OBS');
  }
  
  return $band;
}


#==============================================
sub terminate
#==============================================
{
  my $wrapper_status = message(0,'GET_WRAPPER_STATUS');
  my $datetime = datetime_string('ENDING');
  datetime_string('SUMMARY');
  my $nsec = datetime_string('TOTAL');

  # check that all overrides were used
  if ($wrapper_status < 64 && $wrapper_status != 1) {
    foreach my $key (keys %override_used) {delete $overrides{$key}; }
    if (scalar keys(%overrides) > 0) {
      print "\n\n*********************************************************\n";
      print "ERROR: Invalid module name for parameter override(s) requested:\n";
      print "       (must be lower case module name)\n";
      foreach my $key (keys %overrides) { print "  $key\n"; }
      print "*********************************************************\n";
      $wrapper_status = 244;
    }
  }

  my $status_text;
  if    ($wrapper_status >= 64) {$status_text = "==== ERROR ====";}
  elsif ($wrapper_status >= 32) {$status_text = "== WARNING ==";}
  elsif ($wrapper_status ==  1) {$status_text = "SKIPPED";}
  else                          {$status_text = "OK";}

  my $status_line = sprintf " %3d %-3s $PIPE_TYPE $prefix_datafile  $datetime  ($nsec sec.)", $wrapper_status, $status_text;

  printf "\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
  printf "$status_line\n";
  printf "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n";

  open STATUS, ">.STATUS"; printf STATUS "$status_line\n"; close STATUS;

  exit($wrapper_status);
}

return 1;
}
